// This unit is part of the GLScene Project, http://glscene.org
//
{ : VectorGeometry<p>

  Base classes and structures for GLScene.<p>

  Most common functions/procedures come in various flavours (using overloads),
  the naming convention is :<ul>
  <li>TypeOperation: functions returning a result, or accepting a "var" as last
  parameter to place result (VectorAdd, VectorCrossProduct...)
  <li>OperationType : procedures taking as first parameter a "var" that will be
  used as operand and result (AddVector, CombineVector...)
  </ul><p>
  As a general rule, procedures implementations (asm or not) are the fastest
  (up to 800% faster than function equivalents), due to reduced return value
  duplication overhead (the exception being the matrix operations).<p>

  For better performance, it is recommended <b>not</b> to use the "Math" unit
  that comes with Delphi, and only use functions/procedures from this unit
  (the single-based functions have been optimized and are up to 100% faster,
  than extended-based ones from "Math").<p>

  3DNow! SIMD instructions are automatically detected and used in *some* of the
  functions/procedures, typical gains (over FPU implementation) are approx a
  100% speed increase on K6-2/3, and 20-60% on K7, and sometimes more
  (f.i. 650% on 4x4 matrix multiplication for the K6, 300% for RSqrt on K7).<p>

  Cyrix, NexGen and other "exotic" CPUs may fault in the 3DNow! detection
  (initialization section), comment out or replace with your own detection
  routines if you want to support these. All AMD processors after K5, and
  all Intel processors after Pentium should be immune to this.<p>

  <b>History : </b><font size=-1><ul>

  <li>20/11/12 - PW - Added CPP compatibility using records instead arrays for vectors,
  moved IsPointInPolygon from GLGizmoEx;
  <li>05/06/12 - Maverick - Added SegmentPlaneIntersection routine
  <li>10/05/12 - Maverick - Added quad/disk intersection routines,
  c3PIdiv2 constant, some asm blocks
  <li>10/05/12 - Maverick - Added plane/triangle intersection routines,
  overloaded plane routines, linelinedistance routine
  <li>28/03/12 - Maverick - Added IsColinear test and Vector2d dot product
  <li>19/12/11 - Yar - Added VectorAdd for 2d vector (thanks microalexx)

  <li>10/06/11 - DaStr - Added some Vector2f routines
  Overloaded some procedures to accept both 3f and 4f vectors
  Marked some methods as inline
  Added SignStrict, MoveObjectAround(), GetSafeTurnAngle(),
  RectanglesIntersect(), RectangleContains(),
  AngleBetweenVectors(), ShiftObjectFromCenter()
  <li>11/05/11 - Yar - Added ClampInteger
  <li>25/11/10 - DaStr - Added InterpolateExp() and itExp mode
  <li>04/11/10 - DaStr - Removed duplicate standard type definitions
  <li>09/08/10 - Yar - Added CreateLookAtMatrix, CreateMatrixFromFrustum, CreatePerspectiveMatrix,
  CreateOrthoMatrix, CreatePickMatrix, Project, UnProject
  <li>14/06/10 - Yar - Added VectorArrayLerp for TTexPointArray
  <li>06/06/10 - Yar - Fixed warnings for FPC
  <li>15/04/10 - Yar - Bugfixed vector normalization on Intel processors
  (thanks mif) (BugTracker ID = 2987779)
  <li>02/04/10 - Yar - Added inline directive for small vector operations
  <li>12/03/09 - DanB - Added overloaded versions of IsVolumeClipped
  <li>09/10/08 - DanB - moved TRenderContextClippingInfo + IsVolumeClipped functions that
  use TRenderContextClippingInfo to GLRenderContextInfo.pas
  <li>21/02/07 - DaStr - Bugfixed InterpolatePower() to support negative Base
  and not round Exponent parameters
  <li>12/02/08 - Mrqzzz - Removed cPIdiv360, not needed anymore, by Pete,Dan Bartlett
  <li>12/02/08 - Mrqzzz - Dave Gravel added const cPIdiv360 to fix ResetAndPitchTurnRoll
  <li>18/11/07 - DaStr - Added MatrixInvert(), VectorDivide() functions
  <li>19/09/07 - DaStr - Added AddPoint() and PointAdd() procedures
  <li>18/09/07 - DaStr - Added two more overloaded versions of VectorScale()
  <li>29/08/07 - LC - Fixed BarycentricCoordinates to work with triangles in yz plane
  <li>27/08/07 - DaStr - Bugfixed VectorAffineFltToDbl and VectorFltToDbl
  (thanks Biot!) (BugTracker ID = 1782005)
  <li>23/08/07 - LC - Added BarycentricCoordinates
  <li>16/04/07 - DaStr - Removed TMatrix[3/4][w/e] types (declared in VectorTypes.pas)
  <li>25/03/07 - DaStr - Replaced Types with GLCrossPlatform for Delphi 5 support
  Added YZHmgVector and XZHmgVector
  <li>21/03/07 - DaStr - Bugfixed MatrixDecompose
  (thanks Burkhard Carstens) (BugTrackerID=1684438)
  <li>06/03/07 - DaStr - Added InterpolateXXX and MatrixLerp functions
  <li>03/03/07 - DaStr - Added [Vector/Matrix/Rect]Equals, Vector[2/3/4][i/f/s/b/d]Make
  Added Vector[More/Less](Equal)Then
  <li>15/02/07 - DaStr - Returned to old code formating style
  <li>29/01/07 - DaStr - Added IntersectSphereBox (Thanks to dikoe Kenguru)
  <li>23/01/07 - fig - Added TexpointEquals() function
  <li>13/01/07 - DaStr - Added RayCastBoxIntersect and IntersectTriangleBox (Thanx to dikoe Kenguru)
  Deleted types that were duplicated  (BugTrackerID = 1586318),
  some other types moved to VectorTypes unit
  <li>07/04/06 - DB - Fixed VectorArrayLerp_3DNow (affine) for n<=1 (dikoe Kenguru)
  <li>02/12/04 - MF - Added IsVolumeClipped overload that uses Frustum instead
  of rcci
  <li>02/08/04 - LR, YHC - BCB corrections: use record instead array
  move PAffineVectorArray, PVectorArray and PMatrixArray to VectorTypes Unit
  <li>08/07/04 - LR - Removed ../ from the GLScene.inc
  <li>26/10/03 - EG - Renamed from "Geometry.pas" to "VectorGeometry.pas"
  <li>17/10/03 - EG - Optimized Min/MaxInteger, some of the Min/MaxFloat
  <li>13/08/03 - SG - Added TQuaternionArray, PQuaternionArray and PQuaternion
  <li>21/07/03 - EG - Added RoundInt, faster Round/Round64, updated Power
  <li>04/07/03 - EG - New VectorCombine overload, some optimizations
  <li>18/06/03 - MF - Added PointSegmentClosestPoint, PointSegmentDistance,
  PointLineClosestPoint and PointLineDistance.
  <li>26/05/03 - EG - NO_ASM variant creation completed
  <li>22/05/03 - EG - All vSIMD asm tests should now be under GEOMETRY_NO_ASM control
  <li>20/05/03 - EG - Added MakeParallelProjectionMatrix
  <li>13/05/03 - EG - 3DNow! optimization for ClampValue
  <li>30/04/03 - EG - Hyperbolic trig functions (Aaron Hochwimmer)
  <li>14/02/03 - EG - Added ScaleAndRound
  <li>28/01/03 - EG - Affine matrix inversion and related functions (Dan Barlett)
  <li>29/10/02 - EG - New MinFloat overloads (Bob)
  <li>04/09/02 - EG - New Abs/Max functions, VectorTransform(affine, hmgMatrix)
  now considers the matrix as 4x3 (was 3x3)
  <li>21/08/02 - EG - Added Pack/UnPackRotationMatrix
  <li>13/08/02 - EG - Added Area functions
  <li>20/07/02 - EG - Fixed RayCastTriangleIntersect "backward" hits
  <li>05/07/02 - EG - Started adding non-asm variants (GEOMETRY_NO_ASM)
  <li>22/02/02 - EG - Temporary Quaternion fix for VectorAngleLerp
  <li>12/02/02 - EG - Added QuaternionFromEuler (Alex Grigny de Castro)
  <li>11/02/02 - EG - Non-spinned QuaternionSlerp (Alex Grigny de Castro)
  <li>07/02/02 - EG - Added AnglePreservingMatrixInvert
  <li>30/01/02 - EG - New Quaternion<->Matrix code (Alex Grigny de Castro)
  <li>29/01/02 - EG - Fixed AngleLerp, added DistanceBetweenAngles (Alex Grigny de Castro)
  <li>20/01/02 - EG - Added VectorArrayAdd, ScaleFloatArray, OffsetFloatArray
  <li>11/01/02 - EG - 3DNow Optim for VectorAdd (hmg)
  <li>10/01/02 - EG - Fixed VectorEquals ("True" wasn't Pascal compliant "1"),
  3DNow optims for vector mormalizations (affine),
  Added RSqrt
  <li>04/01/02 - EG - Updated/fixed RayCastTriangleIntersect
  <li>13/12/01 - EG - Fixed MakeReflectionMatrix
  <li>02/11/01 - EG - Faster mode for PrepareSinCosCache (by Nelson Chu)
  <li>22/08/01 - EG - Some new overloads
  <li>19/08/01 - EG - Added sphere raycasting functions
  <li>08/08/01 - EG - Added MaxFloat overloads
  <li>24/07/01 - EG - VectorAngle renamed to VectorAngleCosine to avoid confusions
  <li>06/07/01 - EG - Added NormalizeDegAngle
  <li>04/07/01 - EG - Now uses VectorTypes
  <li>18/03/01 - EG - Added AngleLerp and NormalizeAngle
  <li>15/03/01 - EG - Added Int, Ceil and Floor, faster "Frac"
  <li>06/03/01 - EG - Fix in PointInPolygon by Pavel Vassiliev
  <li>04/03/01 - EG - Added NormalizeVectorArray
  <li>03/03/01 - EG - Added MakeReflectionMatrix
  <li>02/03/01 - EG - New PointInPolygon code by Pavel Vassiliev
  <li>25/02/01 - EG - Fixed 'VectorSubstract', added VectorArrayLerp and a few minors
  <li>22/02/01 - EG - Added MinXYZ/MaxXYZ variants and Plane-Line intersection
  <li>21/02/01 - EG - Added Sign, MinFloat & MaxFloat
  <li>15/02/01 - EG - Faster Vector Transforms (3DNow! optimizations)
  <li>14/02/01 - EG - Faster Matrix multiplications (3DNow! & FPU optimizations),
  Added support for FPU-only sections
  <li>05/02/01 - EG - Faster VectorEquals
  <li>21/01/01 - EG - Fixed MakePoint/Vector affine variants (thx Jacques Tur)
  <li>17/01/00 - EG - VectoAdd return type fix (thx Jacques Tur),
  also added a few new overloads
  <li>05/11/00 - EG - Added RayCastPlaneIntersect
  <li>08/10/00 - EG - Added SetMatrix
  <li>13/08/00 - EG - Added Plane geometry support
  <li>06/08/00 - EG - Various minor additions
  <li>16/07/00 - EG - Added some new mixed vector/scalar funcs and new overloads
  <li>12/07/00 - EG - New overloads and replacements for Power, Trunc, Frac & Round
  <li>25/06/00 - EG - End of major update
  <li>13/06/00 - EG - Start of major update
  <li>09/06/00 - EG - Some additions and fixes in preparation for major changes
  <li>05/06/00 - EG - Added VectorLength overloads
  <li>26/05/00 - EG - [0..0] arrays changed to [0..cMaxArray]
  <li>23/05/00 - EG - Added intersection functions,
  Replaced some xxxAffinexxx funcs with overloads
  <li>22/03/00 - EG - Added MakeShadowMatrix (adapted from "OpenGL SuperBible" book)
  <li>21/03/00 - EG - Removed PWordArray (was a SysUtils's duplicate)
  <li>06/02/00 - EG - Added VectorEquals
  <li>05/02/00 - EG - Added some "const", more still needed,
  Added overloads for some of the MakeXXXVector funcs,
  Added homogeneous vector consts, VectorSpacing
  </ul>
}
unit VectorGeometry;

// This unit contains many needed types, functions and procedures for
// quaternion, vector and matrix arithmetics. It is specifically designed
// for geometric calculations within R3 (affine vector space)
// and R4 (homogeneous vector space).
//
// Note: The terms 'affine' or 'affine coordinates' are not really correct here
// because an 'affine transformation' describes generally a transformation which leads
// to a uniquely solvable system of equations and has nothing to do with the dimensionality
// of a vector. One could use 'projective coordinates' but this is also not really correct
// and since I haven't found a better name (or even any correct one), 'affine' is as good
// as any other one.
//
// Identifiers containing no dimensionality (like affine or homogeneous)
// and no datatype (integer..extended) are supposed as R4 representation
// with 'single' floating point type (examples are TVector, TMatrix,
// and TQuaternion). The default data type is 'single' ('GLFloat' for OpenGL)
// and used in all routines (except conversions and trigonometric functions).
//
// Routines with an open array as argument can either take Func([1,2,3,4,..]) or Func(Vect).
// The latter is prefered, since no extra stack operations is required.
// Note: Be careful while passing open array elements! If you pass more elements
// than there's room in the result the behaviour will be unpredictable.
//
// If not otherwise stated, all angles are given in radians
// (instead of degrees). Use RadToDeg or DegToRad to convert between them.
//
// Geometry.pas was assembled from different sources (like GraphicGems)
// and relevant books or based on self written code, respectivly.
//
// Note: Some aspects need to be considered when using Delphi and pure
// assembler code. Delphi esnures that the direction flag is always
// cleared while entering a function and expects it cleared on return.
// This is in particular important in routines with (CPU) string commands (MOVSD etc.)
// The registers EDI, ESI and EBX (as well as the stack management
// registers EBP and ESP) must not be changed! EAX, ECX and EDX are
// freely available and mostly used for parameter.
//
// Version 2.5
// last change : 04. January 2000
//
// (c) Copyright 1999, Dipl. Ing. Mike Lischke (public@lischke-online.de)

interface

// define for turning off assembly routines in this unit
// *experimental* and incomplete
{ .$DEFINE GEOMETRY_NO_ASM }

// Activate Inline
{$DEFINE GLS_INLINE}

// Activate support X Window System
{.$DEFINE GLS_X11_SUPPORT}


{$IFNDEF GEOMETRY_NO_ASM}
{$IFDEF GLS_INLINE}
{$DEFINE GLS_INLINE_VICE_ASM}
{$ENDIF}
{$ENDIF}

{$IFNDEF MSWINDOWS}
{$IFDEF WIN32}
// MSWINDOWS wasn't defined in Delphi 5, but WIN32 was
// so need to define MSWINDOWS here to provide Delphi 5 support
{$DEFINE MSWINDOWS}
{$ENDIF}
{$ENDIF}

{$IFDEF LINUX}
{$DEFINE GEOMETRY_NO_ASM}
{$DEFINE UNIX}
{$ENDIF}

{$IFDEF FPC}
{$UNDEF GLS_DELPHI_OR_CPPB}
{ .$DEFINE GLS_GENERIC_PREFIX }
{$DEFINE GEOMETRY_NO_ASM}
{$DEFINE GLS_NO_ASM}
{$TYPEINFO ON}
{$INLINE ON}
{$H+} // Enable long strings.
{$IFDEF FPC_HAS_TYPE_EXTENDED}
{$DEFINE GLS_PLATFORM_HAS_EXTENDED}
{$ENDIF}
{$ENDIF}

{$IFDEF CONDITIONALEXPRESSIONS}
{$IF CompilerVersion>=23}
{$DEFINE XE2_UP}
{$IFEND}
{$ENDIF}
uses
{$IFDEF MSWINDOWS}
  Windows,
{$ENDIF}
{$IFDEF UNIX}
  Unix, BaseUnix,
{$ENDIF}
{$IFDEF GLS_X11_SUPPORT}
  xlib,
{$ENDIF}

{$IFDEF FPC}
  Classes,
  SysUtils,
  StrUtils,
  Types,
  LCLVersion,
  LCLType,
  FileUtil,
  //Consts,
  Graphics,
  Controls,
  Forms,
  Dialogs
{$ELSE}

{$IFDEF XE2_UP}
  System.Classes,
  System.SysUtils,
  System.StrUtils,
  System.Types,
  VCL.Consts,
  VCL.Graphics,
  VCL.Controls,
  VCL.Forms,
  VCL.Dialogs
{$ELSE}
  Classes,
  SysUtils,
  StrUtils,
  Types,
  Consts,
  Graphics,
  Controls,
  Forms,
  Dialogs
{$ENDIF}
{$ENDIF}
    ;

// GLCrossPlatform
type
  THalfFloat = type Word;
  PHalfFloat = ^THalfFloat;

{$IFDEF FPC}
  DWORD  = System.DWORD;
  TPoint = Types.TPoint;
  PPoint = ^TPoint;
  TRect  = Types.TRect;
  PRect  = ^TRect;
{$ELSE}
  {$IFDEF XE2_UP}
  DWORD  = System.Types.DWORD; {$NODEFINE DWORD}
  TPoint = System.Types.TPoint; {$NODEFINE TPoint}
  PPoint = System.Types.PPoint; {$NODEFINE PPoint}
  TRect  = System.Types.TRect; {$NODEFINE TRect}
  PRect  = System.Types.PRect; {$NODEFINE PRect}
  {$ELSE}
  DWORD  = Types.DWORD; {$NODEFINE DWORD}
  TPoint = Types.TPoint; {$NODEFINE TPoint}
  PPoint = Types.PPoint; {$NODEFINE PPoint}
  TRect  = Types.TRect; {$NODEFINE TRect}
  PRect  = Types.PRect; {$NODEFINE PRect}
  {$ENDIF}
{$ENDIF}
{.$ENDIF}
  // GLCrossPlatform

  // VectorTypes
type
  // 2
  TVector2d = record
    case Integer of
      0:
        (V: array [0 .. 1] of double);
      1:
        (X: double;
          Y: double);
  end;

  TVector2f = record
    case Integer of
      0:
        (V: array [0 .. 1] of single);
      1:
        (X, Y: single);
  end;

  TVector2h = record
    case Integer of
      0:
        (V: array [0 .. 1] of THalfFloat);
      1:
        (X, Y: THalfFloat);
  end;

  TVector2i = record
    case Integer of
      0:
        (V: array [0 .. 1] of longint);
      1:
        (X, Y: longint);
  end;

  TVector2ui = record
    case Integer of
      0:
        (V: array [0 .. 1] of longword);
      1:
        (X, Y: longword);
  end;

  TVector2s = record
    case Integer of
      0:
        (V: array [0 .. 1] of smallint);
      1:
        (X, Y: smallint);
  end;

  TVector2b = record
    case Integer of
      0:
        (V: array [0 .. 1] of byte);
      1:
        (X, Y: byte);
  end;

  TVector2sb = record
    case Integer of
      0:
        (V: array [0 .. 1] of ShortInt);
      1:
        (X, Y: ShortInt);
  end;

  TVector2e = record
    case Integer of
      0:
        (V: array [0 .. 1] of Extended);
      1:
        (X, Y: Extended);
  end;

  TVector2w = record
    case Integer of
      0:
        (V: array [0 .. 1] of Word);
      1:
        (X, Y: Word);
  end;

  TVector2p = record
    case Integer of
      0:
        (V: array [0 .. 1] of Pointer);
      1:
        (X, Y: Pointer);
  end;

  // 3
  TVector3d = record
    case Integer of
      0:
        (V: array [0 .. 2] of double);
      1:
        (X, Y, Z: double);
  end;

  TVector3f = record
    case Integer of
      0:
        (V: array [0 .. 2] of single);
      1:
        (X, Y, Z: single);
  end;

  TVector3h = record
    case Integer of
      0:
        (V: array [0 .. 2] of THalfFloat);
      1:
        (X, Y, Z: THalfFloat);
  end;

  TVector3i = record
    case Integer of
      0:
        (V: array [0 .. 2] of longint);
      1:
        (X, Y, Z: longint);
  end;

  TVector3ui = record
    case Integer of
      0:
        (V: array [0 .. 2] of longword);
      1:
        (X, Y, Z: longword);
  end;

  TVector3s = record
    case Integer of
      0:
        (V: array [0 .. 2] of smallint);
      1:
        (X, Y, Z: smallint);
  end;

  TVector3b = record
    case Integer of
      0:
        (V: array [0 .. 2] of byte);
      1:
        (X, Y, Z: byte);
  end;

  TVector3sb = record
    case Integer of
      0:
        (V: array [0 .. 2] of ShortInt);
      1:
        (X, Y, Z: ShortInt);
  end;

  TVector3e = record
    case Integer of
      0:
        (V: array [0 .. 2] of Extended);
      1:
        (X, Y, Z: Extended);
  end;

  TVector3w = record
    case Integer of
      0:
        (V: array [0 .. 2] of Word);
      1:
        (X, Y, Z: Word);
  end;

  TVector3p = record
    case Integer of
      0:
        (V: array [0 .. 2] of Pointer);
      1:
        (X, Y, Z: Pointer);
  end;

  // 4
  TVector4d = record
    case Integer of
      0:
        (V: array [0 .. 3] of double);
      1:
        (X, Y, Z, W: double);
  end;

  TVector4f = record
    case Integer of
      0:
        (V: array [0 .. 3] of single);
      1:
        (X, Y, Z, W: single);
  end;

  TVector4h = record
    case Integer of
      0:
        (V: array [0 .. 3] of THalfFloat);
      1:
        (X, Y, Z, W: THalfFloat);
  end;

  TVector4i = record
    case Integer of
      0:
        (V: array [0 .. 3] of longint);
      1:
        (X, Y, Z, W: longint);
  end;

  TVector4ui = record
    case Integer of
      0:
        (V: array [0 .. 3] of longword);
      1:
        (X, Y, Z, W: longword);
  end;

  TVector4s = record
    case Integer of
      0:
        (V: array [0 .. 3] of smallint);
      1:
        (X, Y, Z, W: smallint);
  end;

  TVector4b = record
    case Integer of
      0:
        (V: array [0 .. 3] of byte);
      1:
        (X, Y, Z, W: byte);
  end;

  TVector4sb = record
    case Integer of
      0:
        (V: array [0 .. 3] of ShortInt);
      1:
        (X, Y, Z, W: ShortInt);
  end;

  TVector4e = record
    case Integer of
      0:
        (V: array [0 .. 3] of Extended);
      1:
        (X, Y, Z, W: Extended);
  end;

  TVector4w = record
    case Integer of
      0:
        (V: array [0 .. 3] of Word);
      1:
        (X, Y, Z, W: Word);
  end;

  TVector4p = record
    case Integer of
      0:
        (V: array [0 .. 3] of Pointer);
      1:
        (X, Y, Z, W: Pointer);
  end;

  TMatrix2d = record
    case Integer of
      0:
        (V: array [0 .. 1] of TVector2d);
      1:
        (X, Y: TVector2d);
  end;

  TMatrix2f = record
    case Integer of
      0:
        (V: array [0 .. 1] of TVector2f);
      1:
        (X, Y: TVector2f);
  end;

  TMatrix2i = record
    case Integer of
      0:
        (V: array [0 .. 1] of TVector2i);
      1:
        (X, Y: TVector2i);
  end;

  TMatrix2s = record
    case Integer of
      0:
        (V: array [0 .. 1] of TVector2s);
      1:
        (X, Y: TVector2s);
  end;

  TMatrix2b = record
    case Integer of
      0:
        (V: array [0 .. 1] of TVector2b);
      1:
        (X, Y: TVector2b);
  end;

  TMatrix2e = record
    case Integer of
      0:
        (V: array [0 .. 1] of TVector2e);
      1:
        (X, Y: TVector2e);
  end;

  TMatrix2w = record
    case Integer of
      0:
        (V: array [0 .. 1] of TVector2w);
      1:
        (X, Y: TVector2w);
  end;

  TMatrix2p = record
    case Integer of
      0:
        (V: array [0 .. 1] of TVector2p);
      1:
        (X, Y: TVector2p);
  end;

  TMatrix3d = record
    case Integer of
      0:
        (V: array [0 .. 2] of TVector3d);
      1:
        (X, Y, Z: TVector3d);
  end;

  TMatrix3f = record
    case Integer of
      0:
        (V: array [0 .. 2] of TVector3f);
      1:
        (X, Y, Z: TVector3f);
  end;

  TMatrix3i = record
    case Integer of
      0:
        (V: array [0 .. 2] of TVector3i);
      1:
        (X, Y, Z: TVector3i);
  end;

  TMatrix3s = record
    case Integer of
      0:
        (V: array [0 .. 2] of TVector3s);
      1:
        (X, Y, Z: TVector3s);
  end;

  TMatrix3b = record
    case Integer of
      0:
        (V: array [0 .. 2] of TVector3b);
      1:
        (X, Y, Z: TVector3b);
  end;

  TMatrix3e = record
    case Integer of
      0:
        (V: array [0 .. 2] of TVector3e);
      1:
        (X, Y, Z: TVector3e);
  end;

  TMatrix3w = record
    case Integer of
      0:
        (V: array [0 .. 2] of TVector3w);
      1:
        (X, Y, Z: TVector3w);
  end;

  TMatrix3p = record
    case Integer of
      0:
        (V: array [0 .. 2] of TVector3p);
      1:
        (X, Y, Z: TVector3p);
  end;

  TMatrix4d = record
    case Integer of
      0:
        (V: array [0 .. 3] of TVector4d);
      1:
        (X, Y, Z, W: TVector4d);
  end;

  TMatrix4f = record
    case Integer of
      0:
        (V: array [0 .. 3] of TVector4f);
      1:
        (X, Y, Z, W: TVector4f);
  end;

  TMatrix4i = record
    case Integer of
      0:
        (V: array [0 .. 3] of TVector4i);
      1:
        (X, Y, Z, W: TVector4i);
  end;

  TMatrix4s = record
    case Integer of
      0:
        (V: array [0 .. 3] of TVector4s);
      1:
        (X, Y, Z, W: TVector4s);
  end;

  TMatrix4b = record
    case Integer of
      0:
        (V: array [0 .. 3] of TVector4b);
      1:
        (X, Y, Z, W: TVector4b);
  end;

  TMatrix4e = record
    case Integer of
      0:
        (V: array [0 .. 3] of TVector4e);
      1:
        (X, Y, Z, W: TVector4e);
  end;

  TMatrix4w = record
    case Integer of
      0:
        (V: array [0 .. 3] of TVector4w);
      1:
        (X, Y, Z, W: TVector4w);
  end;

  TMatrix4p = record
    case Integer of
      0:
        (V: array [0 .. 3] of TVector4p);
      1:
        (X, Y, Z, W: TVector4p);
  end;

  TD3DVector = packed record
    case Integer of
      0:
        (X: single;
          Y: single;
          Z: single);
      1:
        (V: TVector3f);
  end;

  TD3DMatrix = packed record
    case Integer of
      0:
        (_11, _12, _13, _14: single;
          _21, _22, _23, _24: single;
          _31, _32, _33, _34: single;
          _41, _42, _43, _44: single);
      1:
        (M: TMatrix4f);
  end;

  // VectorTypes

const
  cMaxArray     = (MaxInt shr 4);
  cColinearBias = 1E-8;

type
  // data types needed for 3D graphics calculation,
  // included are 'C like' aliases for each type (to be
  // conformal with OpenGL types)
  PFloat = PSingle;

  PTexPoint = ^TTexPoint;

  TTexPoint = packed record
    S, T: single;
  end;

  // types to specify continous streams of a specific type
  // switch off range checking to access values beyond the limits
  PByteVector = ^TByteVector;
  PByteArray  = PByteVector;
  TByteVector = array [0 .. cMaxArray] of byte;

  PWordVector = ^TWordVector;
  TWordVector = array [0 .. cMaxArray] of Word;

  PIntegerVector = ^TIntegerVector;
  PIntegerArray  = PIntegerVector;
  TIntegerVector = array [0 .. cMaxArray] of Integer;

  PFloatVector = ^TFloatVector;
  PFloatArray  = PFloatVector;
  PSingleArray = PFloatArray;
  TFloatVector = array [0 .. cMaxArray] of single;
  TSingleArray = array of single;

  PDoubleVector = ^TDoubleVector;
  PDoubleArray  = PDoubleVector;
  TDoubleVector = array [0 .. cMaxArray] of double;

  PExtendedVector = ^TExtendedVector;
  PExtendedArray  = PExtendedVector;
  TExtendedVector = array [0 .. cMaxArray] of Extended;

  PPointerVector = ^TPointerVector;
  PPointerArray  = PPointerVector;
  TPointerVector = array [0 .. cMaxArray] of Pointer;

  PCardinalVector = ^TCardinalVector;
  PCardinalArray  = PCardinalVector;
  TCardinalVector = array [0 .. cMaxArray] of Cardinal;

  PLongWordVector = ^TLongWordVector;
  PLongWordArray  = PLongWordVector;
  TLongWordVector = array [0 .. cMaxArray] of longword;

  // common vector and matrix types with predefined limits
  // indices correspond like: x -> 0
  // y -> 1
  // z -> 2
  // w -> 3

  PHomogeneousByteVector = ^THomogeneousByteVector;
  THomogeneousByteVector = TVector4b;

  PHomogeneousWordVector = ^THomogeneousWordVector;
  THomogeneousWordVector = TVector4w;

  PHomogeneousIntVector = ^THomogeneousIntVector;
  THomogeneousIntVector = TVector4i;

  PHomogeneousFltVector = ^THomogeneousFltVector;
  THomogeneousFltVector = TVector4f;

  PHomogeneousDblVector = ^THomogeneousDblVector;
  THomogeneousDblVector = TVector4d;

  PHomogeneousExtVector = ^THomogeneousExtVector;
  THomogeneousExtVector = TVector4e;

  PHomogeneousPtrVector = ^THomogeneousPtrVector;
  THomogeneousPtrVector = TVector4p;

  PAffineByteVector = ^TAffineByteVector;
  TAffineByteVector = TVector3b;

  PAffineWordVector = ^TAffineWordVector;
  TAffineWordVector = TVector3w;

  PAffineIntVector = ^TAffineIntVector;
  TAffineIntVector = TVector3i;

  PAffineFltVector = ^TAffineFltVector;
  TAffineFltVector = TVector3f;

  PAffineDblVector = ^TAffineDblVector;
  TAffineDblVector = TVector3d;

  PAffineExtVector = ^TAffineExtVector;
  TAffineExtVector = TVector3e;

  PAffinePtrVector = ^TAffinePtrVector;
  TAffinePtrVector = TVector3p;

  PVector2f = ^TVector2f;

  // some simplified names
  PVector = ^TVector;
  TVector = THomogeneousFltVector;

  PHomogeneousVector = ^THomogeneousVector;
  THomogeneousVector = THomogeneousFltVector;

  PAffineVector = ^TAffineVector;
  TAffineVector = TVector3f;

  PVertex = ^TVertex;
  TVertex = TAffineVector;

  // arrays of vectors
  PAffineVectorArray = ^TAffineVectorArray;
  TAffineVectorArray = array [0 .. MaxInt shr 4] of TAffineVector;

  PVectorArray = ^TVectorArray;
  TVectorArray = array [0 .. MaxInt shr 5] of TVector;

  PTexPointArray = ^TTexPointArray;
  TTexPointArray = array [0 .. MaxInt shr 4] of TTexPoint;

  // matrices
  THomogeneousByteMatrix = TMatrix4b;

  THomogeneousWordMatrix = array [0 .. 3] of THomogeneousWordVector;

  THomogeneousIntMatrix = TMatrix4i;

  THomogeneousFltMatrix = TMatrix4f;

  THomogeneousDblMatrix = TMatrix4d;

  THomogeneousExtMatrix = array [0 .. 3] of THomogeneousExtVector;

  TAffineByteMatrix = TMatrix3b;

  TAffineWordMatrix = array [0 .. 2] of TAffineWordVector;

  TAffineIntMatrix = TMatrix3i;

  TAffineFltMatrix = TMatrix3f;

  TAffineDblMatrix = TMatrix3d;

  TAffineExtMatrix = array [0 .. 2] of TAffineExtVector;

  // some simplified names
  PMatrix = ^TMatrix;
  TMatrix = THomogeneousFltMatrix;

  TMatrixArray = array [0 .. MaxInt shr 7] of TMatrix;
  PMatrixArray = ^TMatrixArray;

  PHomogeneousMatrix = ^THomogeneousMatrix;
  THomogeneousMatrix = THomogeneousFltMatrix;

  PAffineMatrix = ^TAffineMatrix;
  TAffineMatrix = TAffineFltMatrix;

  { : A plane equation.<p>
    Defined by its equation A.x+B.y+C.z+D<p>, a plane can be mapped to the
    homogeneous space coordinates, and this is what we are doing here.<br>
    The typename is just here for easing up data manipulation. }
  THmgPlane       = TVector;
  TDoubleHmgPlane = THomogeneousDblVector;

  // q = ([x, y, z], w)
  PQuaternion = ^TQuaternion;

  TQuaternion = record
    ImagPart: TAffineVector;
    RealPart: single;
  end;

  PQuaternionArray = ^TQuaternionArray;
  TQuaternionArray = array [0 .. MaxInt shr 5] of TQuaternion;

  TRectangle = record
    Left, Top, Width, Height: Integer;
  end;

  TFrustum = record
    pLeft, pTop, pRight, pBottom, pNear, pFar: THmgPlane;
  end;

  TTransType = (ttScaleX, ttScaleY, ttScaleZ, ttShearXY, ttShearXZ, ttShearYZ, ttRotateX, ttRotateY,
    ttRotateZ, ttTranslateX, ttTranslateY, ttTranslateZ, ttPerspectiveX, ttPerspectiveY,
    ttPerspectiveZ, ttPerspectiveW);

  // used to describe a sequence of transformations in following order:
  // [Sx][Sy][Sz][ShearXY][ShearXZ][ShearZY][Rx][Ry][Rz][Tx][Ty][Tz][P(x,y,z,w)]
  // constants are declared for easier access (see MatrixDecompose below)
  TTransformations = array [TTransType] of single;

  TPackedRotationMatrix = array [0 .. 2] of smallint;

const
  // useful constants

  // TexPoints (2D space)
  XTexPoint: TTexPoint    = (S: 1; T: 0);
  YTexPoint: TTexPoint    = (S: 0; T: 1);
  XYTexPoint: TTexPoint   = (S: 1; T: 1);
  NullTexPoint: TTexPoint = (S: 0; T: 0);
  MidTexPoint: TTexPoint  = (S: 0.5; T: 0.5);

  // standard vectors
  XVector: TAffineVector      = (X: 1; Y: 0; Z: 0);
  YVector: TAffineVector      = (X: 0; Y: 1; Z: 0);
  ZVector: TAffineVector      = (X: 0; Y: 0; Z: 1);
  XYVector: TAffineVector     = (X: 1; Y: 1; Z: 0);
  XZVector: TAffineVector     = (X: 1; Y: 0; Z: 1);
  YZVector: TAffineVector     = (X: 0; Y: 1; Z: 1);
  XYZVector: TAffineVector    = (X: 1; Y: 1; Z: 1);
  NullVector: TAffineVector   = (X: 0; Y: 0; Z: 0);
  MinusXVector: TAffineVector = (X: - 1; Y: 0; Z: 0);
  MinusYVector: TAffineVector = (X: 0; Y: - 1; Z: 0);
  MinusZVector: TAffineVector = (X: 0; Y: 0; Z: - 1);
  // standard homogeneous vectors
  XHmgVector: THomogeneousVector    = (X: 1; Y: 0; Z: 0; W: 0);
  YHmgVector: THomogeneousVector    = (X: 0; Y: 1; Z: 0; W: 0);
  ZHmgVector: THomogeneousVector    = (X: 0; Y: 0; Z: 1; W: 0);
  WHmgVector: THomogeneousVector    = (X: 0; Y: 0; Z: 0; W: 1);
  XYHmgVector: THomogeneousVector   = (X: 1; Y: 1; Z: 0; W: 0);
  YZHmgVector: THomogeneousVector   = (X: 0; Y: 1; Z: 1; W: 0);
  XZHmgVector: THomogeneousVector   = (X: 1; Y: 0; Z: 1; W: 0);
  XYZHmgVector: THomogeneousVector  = (X: 1; Y: 1; Z: 1; W: 0);
  XYZWHmgVector: THomogeneousVector = (X: 1; Y: 1; Z: 1; W: 1);
  NullHmgVector: THomogeneousVector = (X: 0; Y: 0; Z: 0; W: 0);
  // standard homogeneous points
  XHmgPoint: THomogeneousVector    = (X: 1; Y: 0; Z: 0; W: 1);
  YHmgPoint: THomogeneousVector    = (X: 0; Y: 1; Z: 0; W: 1);
  ZHmgPoint: THomogeneousVector    = (X: 0; Y: 0; Z: 1; W: 1);
  WHmgPoint: THomogeneousVector    = (X: 0; Y: 0; Z: 0; W: 1);
  NullHmgPoint: THomogeneousVector = (X: 0; Y: 0; Z: 0; W: 1);

  IdentityMatrix: TAffineMatrix = (V: ((X: 1; Y: 0; Z: 0), (X: 0; Y: 1; Z: 0), (X: 0; Y: 0; Z: 1)));
  IdentityHmgMatrix: TMatrix = (V: ((X: 1; Y: 0; Z: 0; W: 0), (X: 0; Y: 1; Z: 0; W: 0), (X: 0; Y: 0;
    Z: 1; W: 0), (X: 0; Y: 0; Z: 0; W: 1)));
  IdentityHmgDblMatrix: THomogeneousDblMatrix = (V: ((X: 1; Y: 0; Z: 0; W: 0), (X: 0; Y: 1; Z: 0;
    W: 0), (X: 0; Y: 0; Z: 1; W: 0), (X: 0; Y: 0; Z: 0; W: 1)));
  EmptyMatrix: TAffineMatrix = (V: ((X: 0; Y: 0; Z: 0), (X: 0; Y: 0; Z: 0), (X: 0; Y: 0; Z: 0)));
  EmptyHmgMatrix: TMatrix = (V: ((X: 0; Y: 0; Z: 0; W: 0), (X: 0; Y: 0; Z: 0; W: 0), (X: 0; Y: 0;
    Z: 0; W: 0), (X: 0; Y: 0; Z: 0; W: 0)));


  // Quaternions

  IdentityQuaternion: TQuaternion = (ImagPart: (X: 0; Y: 0; Z: 0); RealPart: 1);

  // some very small numbers
  EPSILON: single  = 1E-40;
  EPSILON2: single = 1E-30;

  // ------------------------------------------------------------------------------
  // Vector functions
  // ------------------------------------------------------------------------------

function TexPointMake(const S, T: single): TTexPoint; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function AffineVectorMake(const X, Y, Z: single): TAffineVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function AffineVectorMake(const V: TVector): TAffineVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure SetAffineVector(out V: TAffineVector; const X, Y, Z: single); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure SetVector(out V: TAffineVector; const X, Y, Z: single); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure SetVector(out V: TAffineVector; const vSrc: TVector); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure SetVector(out V: TAffineVector; const vSrc: TAffineVector); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure SetVector(out V: TAffineDblVector; const vSrc: TAffineVector); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure SetVector(out V: TAffineDblVector; const vSrc: TVector); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorMake(const V: TAffineVector; W: single = 0): TVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorMake(const X, Y, Z: single; W: single = 0): TVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function PointMake(const X, Y, Z: single): TVector; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline;
{$ENDIF}
function PointMake(const V: TAffineVector): TVector; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline;
{$ENDIF}
function PointMake(const V: TVector): TVector; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline;
{$ENDIF}
procedure SetVector(out V: TVector; const X, Y, Z: single; W: single = 0); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure SetVector(out V: TVector; const av: TAffineVector; W: single = 0); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure SetVector(out V: TVector; const vSrc: TVector); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure MakePoint(out V: TVector; const X, Y, Z: single); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure MakePoint(out V: TVector; const av: TAffineVector); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure MakePoint(out V: TVector; const av: TVector); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure MakeVector(out V: TAffineVector; const X, Y, Z: single); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure MakeVector(out V: TVector; const X, Y, Z: single); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure MakeVector(out V: TVector; const av: TAffineVector); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure MakeVector(out V: TVector; const av: TVector); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure RstVector(var V: TAffineVector); overload; {$IFDEF GLS_INLINE_VICE_ASM_VICE_ASM}inline;
{$ENDIF}
procedure RstVector(var V: TVector); overload; {$IFDEF GLS_INLINE_VICE_ASM_VICE_ASM}inline; {$ENDIF}
// 2
function VectorEquals(const Vector1, Vector2: TVector2f): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM_VICE_ASM}inline; {$ENDIF}
function VectorEquals(const Vector1, Vector2: TVector2i): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorEquals(const V1, V2: TVector2d): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorEquals(const V1, V2: TVector2s): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorEquals(const V1, V2: TVector2b): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// 3
// function VectorEquals(const V1, V2: TVector3f): Boolean; overload; //declared further
function VectorEquals(const V1, V2: TVector3i): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorEquals(const V1, V2: TVector3d): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorEquals(const V1, V2: TVector3s): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorEquals(const V1, V2: TVector3b): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// 4
// function VectorEquals(const V1, V2: TVector4f): Boolean; overload; //declared further
function VectorEquals(const V1, V2: TVector4i): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorEquals(const V1, V2: TVector4d): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorEquals(const V1, V2: TVector4s): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorEquals(const V1, V2: TVector4b): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// 3x3
function MatrixEquals(const Matrix1, Matrix2: TMatrix3f): Boolean; overload;
function MatrixEquals(const Matrix1, Matrix2: TMatrix3i): Boolean; overload;
function MatrixEquals(const Matrix1, Matrix2: TMatrix3d): Boolean; overload;
function MatrixEquals(const Matrix1, Matrix2: TMatrix3s): Boolean; overload;
function MatrixEquals(const Matrix1, Matrix2: TMatrix3b): Boolean; overload;

// 4x4
function MatrixEquals(const Matrix1, Matrix2: TMatrix4f): Boolean; overload;
function MatrixEquals(const Matrix1, Matrix2: TMatrix4i): Boolean; overload;
function MatrixEquals(const Matrix1, Matrix2: TMatrix4d): Boolean; overload;
function MatrixEquals(const Matrix1, Matrix2: TMatrix4s): Boolean; overload;
function MatrixEquals(const Matrix1, Matrix2: TMatrix4b): Boolean; overload;

// 2x
function Vector2fMake(const X, Y: single): TVector2f; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline;
{$ENDIF}
function Vector2iMake(const X, Y: longint): TVector2i; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline;
{$ENDIF}
function Vector2sMake(const X, Y: smallint): TVector2s; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector2dMake(const X, Y: double): TVector2d; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline;
{$ENDIF}
function Vector2bMake(const X, Y: byte): TVector2b; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline;
{$ENDIF}
function Vector2fMake(const Vector: TVector3f): TVector2f; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector2iMake(const Vector: TVector3i): TVector2i; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector2sMake(const Vector: TVector3s): TVector2s; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector2dMake(const Vector: TVector3d): TVector2d; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector2bMake(const Vector: TVector3b): TVector2b; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector2fMake(const Vector: TVector4f): TVector2f; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector2iMake(const Vector: TVector4i): TVector2i; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector2sMake(const Vector: TVector4s): TVector2s; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector2dMake(const Vector: TVector4d): TVector2d; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector2bMake(const Vector: TVector4b): TVector2b; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// 3x
function Vector3fMake(const X: single; const Y: single = 0; const Z: single = 0): TVector3f;
  overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3iMake(const X: longint; const Y: longint = 0; const Z: longint = 0): TVector3i;
  overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3sMake(const X: smallint; const Y: smallint = 0; const Z: smallint = 0): TVector3s;
  overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3dMake(const X: double; const Y: double = 0; const Z: double = 0): TVector3d;
  overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3bMake(const X: byte; const Y: byte = 0; const Z: byte = 0): TVector3b; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3fMake(const Vector: TVector2f; const Z: single = 0): TVector3f; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3iMake(const Vector: TVector2i; const Z: longint = 0): TVector3i; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3sMake(const Vector: TVector2s; const Z: smallint = 0): TVector3s; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3dMake(const Vector: TVector2d; const Z: double = 0): TVector3d; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3bMake(const Vector: TVector2b; const Z: byte = 0): TVector3b; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3fMake(const Vector: TVector4f): TVector3f; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3iMake(const Vector: TVector4i): TVector3i; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3sMake(const Vector: TVector4s): TVector3s; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3dMake(const Vector: TVector4d): TVector3d; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector3bMake(const Vector: TVector4b): TVector3b; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// 4x
function Vector4fMake(const X: single; const Y: single = 0; const Z: single = 0;
  const W: single = 0): TVector4f; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4iMake(const X: longint; const Y: longint = 0; const Z: longint = 0;
  const W: longint = 0): TVector4i; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4sMake(const X: smallint; const Y: smallint = 0; const Z: smallint = 0;
  const W: smallint = 0): TVector4s; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4dMake(const X: double; const Y: double = 0; const Z: double = 0;
  const W: double = 0): TVector4d; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4bMake(const X: byte; const Y: byte = 0; const Z: byte = 0; const W: byte = 0)
  : TVector4b; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4fMake(const Vector: TVector3f; const W: single = 0): TVector4f; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4iMake(const Vector: TVector3i; const W: longint = 0): TVector4i; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4sMake(const Vector: TVector3s; const W: smallint = 0): TVector4s; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4dMake(const Vector: TVector3d; const W: double = 0): TVector4d; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4bMake(const Vector: TVector3b; const W: byte = 0): TVector4b; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4fMake(const Vector: TVector2f; const Z: single = 0; const W: single = 0): TVector4f;
  overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4iMake(const Vector: TVector2i; const Z: longint = 0; const W: longint = 0)
  : TVector4i; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4sMake(const Vector: TVector2s; const Z: smallint = 0; const W: smallint = 0)
  : TVector4s; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4dMake(const Vector: TVector2d; const Z: double = 0; const W: double = 0): TVector4d;
  overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function Vector4bMake(const Vector: TVector2b; const Z: byte = 0; const W: byte = 0): TVector4b;
  overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Vector comparison functions:
// ComparedVector
// 3f
function VectorMoreThen(const SourceVector, ComparedVector: TVector3f): Boolean; overload;
function VectorMoreEqualThen(const SourceVector, ComparedVector: TVector3f): Boolean; overload;

function VectorLessThen(const SourceVector, ComparedVector: TVector3f): Boolean; overload;
function VectorLessEqualThen(const SourceVector, ComparedVector: TVector3f): Boolean; overload;
// 4f
function VectorMoreThen(const SourceVector, ComparedVector: TVector4f): Boolean; overload;
function VectorMoreEqualThen(const SourceVector, ComparedVector: TVector4f): Boolean; overload;

function VectorLessThen(const SourceVector, ComparedVector: TVector4f): Boolean; overload;
function VectorLessEqualThen(const SourceVector, ComparedVector: TVector4f): Boolean; overload;
// 3i
function VectorMoreThen(const SourceVector, ComparedVector: TVector3i): Boolean; overload;
function VectorMoreEqualThen(const SourceVector, ComparedVector: TVector3i): Boolean; overload;

function VectorLessThen(const SourceVector, ComparedVector: TVector3i): Boolean; overload;
function VectorLessEqualThen(const SourceVector, ComparedVector: TVector3i): Boolean; overload;
// 4i
function VectorMoreThen(const SourceVector, ComparedVector: TVector4i): Boolean; overload;
function VectorMoreEqualThen(const SourceVector, ComparedVector: TVector4i): Boolean; overload;

function VectorLessThen(const SourceVector, ComparedVector: TVector4i): Boolean; overload;
function VectorLessEqualThen(const SourceVector, ComparedVector: TVector4i): Boolean; overload;

// 3s
function VectorMoreThen(const SourceVector, ComparedVector: TVector3s): Boolean; overload;
function VectorMoreEqualThen(const SourceVector, ComparedVector: TVector3s): Boolean; overload;

function VectorLessThen(const SourceVector, ComparedVector: TVector3s): Boolean; overload;
function VectorLessEqualThen(const SourceVector, ComparedVector: TVector3s): Boolean; overload;
// 4s
function VectorMoreThen(const SourceVector, ComparedVector: TVector4s): Boolean; overload;
function VectorMoreEqualThen(const SourceVector, ComparedVector: TVector4s): Boolean; overload;

function VectorLessThen(const SourceVector, ComparedVector: TVector4s): Boolean; overload;
function VectorLessEqualThen(const SourceVector, ComparedVector: TVector4s): Boolean; overload;

// ComparedNumber
// 3f
function VectorMoreThen(const SourceVector: TVector3f; const ComparedNumber: single)
  : Boolean; overload;
function VectorMoreEqualThen(const SourceVector: TVector3f; const ComparedNumber: single)
  : Boolean; overload;

function VectorLessThen(const SourceVector: TVector3f; const ComparedNumber: single)
  : Boolean; overload;
function VectorLessEqualThen(const SourceVector: TVector3f; const ComparedNumber: single)
  : Boolean; overload;
// 4f
function VectorMoreThen(const SourceVector: TVector4f; const ComparedNumber: single)
  : Boolean; overload;
function VectorMoreEqualThen(const SourceVector: TVector4f; const ComparedNumber: single)
  : Boolean; overload;

function VectorLessThen(const SourceVector: TVector4f; const ComparedNumber: single)
  : Boolean; overload;
function VectorLessEqualThen(const SourceVector: TVector4f; const ComparedNumber: single)
  : Boolean; overload;
// 3i
function VectorMoreThen(const SourceVector: TVector3i; const ComparedNumber: single)
  : Boolean; overload;
function VectorMoreEqualThen(const SourceVector: TVector3i; const ComparedNumber: single)
  : Boolean; overload;

function VectorLessThen(const SourceVector: TVector3i; const ComparedNumber: single)
  : Boolean; overload;
function VectorLessEqualThen(const SourceVector: TVector3i; const ComparedNumber: single)
  : Boolean; overload;
// 4i
function VectorMoreThen(const SourceVector: TVector4i; const ComparedNumber: single)
  : Boolean; overload;
function VectorMoreEqualThen(const SourceVector: TVector4i; const ComparedNumber: single)
  : Boolean; overload;

function VectorLessThen(const SourceVector: TVector4i; const ComparedNumber: single)
  : Boolean; overload;
function VectorLessEqualThen(const SourceVector: TVector4i; const ComparedNumber: single)
  : Boolean; overload;
// 3s
function VectorMoreThen(const SourceVector: TVector3s; const ComparedNumber: single)
  : Boolean; overload;
function VectorMoreEqualThen(const SourceVector: TVector3s; const ComparedNumber: single)
  : Boolean; overload;

function VectorLessThen(const SourceVector: TVector3s; const ComparedNumber: single)
  : Boolean; overload;
function VectorLessEqualThen(const SourceVector: TVector3s; const ComparedNumber: single)
  : Boolean; overload;
// 4s
function VectorMoreThen(const SourceVector: TVector4s; const ComparedNumber: single)
  : Boolean; overload;
function VectorMoreEqualThen(const SourceVector: TVector4s; const ComparedNumber: single)
  : Boolean; overload;

function VectorLessThen(const SourceVector: TVector4s; const ComparedNumber: single)
  : Boolean; overload;
function VectorLessEqualThen(const SourceVector: TVector4s; const ComparedNumber: single)
  : Boolean; overload;

function VectorAdd(const V1, V2: TVector2f): TVector2f; overload;
// : Returns the sum of two affine vectors
function VectorAdd(const V1, V2: TAffineVector): TAffineVector; overload;
// : Adds two vectors and places result in vr
procedure VectorAdd(const V1, V2: TAffineVector; var vr: TAffineVector); overload;
procedure VectorAdd(const V1, V2: TAffineVector; vr: PAffineVector); overload;
// : Returns the sum of two homogeneous vectors
function VectorAdd(const V1, V2: TVector): TVector; overload;
procedure VectorAdd(const V1, V2: TVector; var vr: TVector); overload;
// : Sums up f to each component of the vector
function VectorAdd(const V: TAffineVector; const f: single): TAffineVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Sums up f to each component of the vector
function VectorAdd(const V: TVector; const f: single): TVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Adds V2 to V1, result is placed in V1
procedure AddVector(var V1: TAffineVector; const V2: TAffineVector); overload;
// : Adds V2 to V1, result is placed in V1
procedure AddVector(var V1: TAffineVector; const V2: TVector); overload;
// : Adds V2 to V1, result is placed in V1
procedure AddVector(var V1: TVector; const V2: TVector); overload;
// : Sums up f to each component of the vector
procedure AddVector(var V: TAffineVector; const f: single); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Sums up f to each component of the vector
procedure AddVector(var V: TVector; const f: single); overload; {$IFDEF GLS_INLINE_VICE_ASM}inline;
{$ENDIF}
// : Adds V2 to V1, result is placed in V1. W coordinate is always 1.
procedure AddPoint(var V1: TVector; const V2: TVector); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Returns the sum of two homogeneous vectors. W coordinate is always 1.
function PointAdd(var V1: TVector; const V2: TVector): TVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Adds delta to nb texpoints in src and places result in dest
procedure TexPointArrayAdd(const src: PTexPointArray; const delta: TTexPoint; const nb: Integer;
  dest: PTexPointArray); overload;
procedure TexPointArrayScaleAndAdd(const src: PTexPointArray; const delta: TTexPoint;
  const nb: Integer; const scale: TTexPoint; dest: PTexPointArray); overload;
// : Adds delta to nb vectors in src and places result in dest
procedure VectorArrayAdd(const src: PAffineVectorArray; const delta: TAffineVector;
  const nb: Integer; dest: PAffineVectorArray); overload;

// : Returns V1-V2
function VectorSubtract(const V1, V2: TVector2f): TVector2f; overload;
// : Subtracts V2 from V1, result is placed in V1
procedure SubtractVector(var V1: TVector2f; const V2: TVector2f); overload;

// : Returns V1-V2
function VectorSubtract(const V1, V2: TAffineVector): TAffineVector; overload;
// : Subtracts V2 from V1 and return value in result
procedure VectorSubtract(const V1, V2: TAffineVector; var result: TAffineVector); overload;
// : Subtracts V2 from V1 and return value in result
procedure VectorSubtract(const V1, V2: TAffineVector; var result: TVector); overload;
// : Subtracts V2 from V1 and return value in result
procedure VectorSubtract(const V1: TVector; V2: TAffineVector; var result: TVector); overload;
// : Returns V1-V2
function VectorSubtract(const V1, V2: TVector): TVector; overload;
// : Subtracts V2 from V1 and return value in result
procedure VectorSubtract(const V1, V2: TVector; var result: TVector); overload;
// : Subtracts V2 from V1 and return value in result
procedure VectorSubtract(const V1, V2: TVector; var result: TAffineVector); overload;
function VectorSubtract(const V1: TAffineVector; delta: single): TAffineVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorSubtract(const V1: TVector; delta: single): TVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Subtracts V2 from V1, result is placed in V1
procedure SubtractVector(var V1: TAffineVector; const V2: TAffineVector); overload;
// : Subtracts V2 from V1, result is placed in V1
procedure SubtractVector(var V1: TVector; const V2: TVector); overload;

// : Combine the first vector with the second : vr:=vr+v*f
procedure CombineVector(var vr: TAffineVector; const V: TAffineVector; var f: single); overload;
procedure CombineVector(var vr: TAffineVector; const V: TAffineVector; pf: PFloat); overload;
// : Makes a linear combination of two texpoints
function TexPointCombine(const t1, t2: TTexPoint; f1, f2: single): TTexPoint;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Makes a linear combination of two vectors and return the result
function VectorCombine(const V1, V2: TAffineVector; const f1, f2: single): TAffineVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Makes a linear combination of three vectors and return the result
function VectorCombine3(const V1, V2, V3: TAffineVector; const f1, f2, F3: single): TAffineVector;
  overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure VectorCombine3(const V1, V2, V3: TAffineVector; const f1, f2, F3: single;
  var vr: TAffineVector); overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Combine the first vector with the second : vr:=vr+v*f
procedure CombineVector(var vr: TVector; const V: TVector; var f: single); overload;
// : Combine the first vector with the second : vr:=vr+v*f
procedure CombineVector(var vr: TVector; const V: TAffineVector; var f: single); overload;
// : Makes a linear combination of two vectors and return the result
function VectorCombine(const V1, V2: TVector; const f1, f2: single): TVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Makes a linear combination of two vectors and return the result
function VectorCombine(const V1: TVector; const V2: TAffineVector; const f1, f2: single): TVector;
  overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Makes a linear combination of two vectors and place result in vr
procedure VectorCombine(const V1: TVector; const V2: TAffineVector; const f1, f2: single;
  var vr: TVector); overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Makes a linear combination of two vectors and place result in vr
procedure VectorCombine(const V1, V2: TVector; const f1, f2: single; var vr: TVector); overload;
// : Makes a linear combination of two vectors and place result in vr, F1=1.0
procedure VectorCombine(const V1, V2: TVector; const f2: single; var vr: TVector); overload;
// : Makes a linear combination of three vectors and return the result
function VectorCombine3(const V1, V2, V3: TVector; const f1, f2, F3: single): TVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Makes a linear combination of three vectors and return the result
procedure VectorCombine3(const V1, V2, V3: TVector; const f1, f2, F3: single;
  var vr: TVector); overload;

{ : Calculates the dot product between V1 and V2.<p>
  Result:=V1[X] * V2[X] + V1[Y] * V2[Y] }
function VectorDotProduct(const V1, V2: TVector2f): single; overload;
{ : Calculates the dot product between V1 and V2.<p>
  Result:=V1[X] * V2[X] + V1[Y] * V2[Y] + V1[Z] * V2[Z] }
function VectorDotProduct(const V1, V2: TAffineVector): single; overload;
{ : Calculates the dot product between V1 and V2.<p>
  Result:=V1[X] * V2[X] + V1[Y] * V2[Y] + V1[Z] * V2[Z] }
function VectorDotProduct(const V1, V2: TVector): single; overload;
{ : Calculates the dot product between V1 and V2.<p>
  Result:=V1[X] * V2[X] + V1[Y] * V2[Y] + V1[Z] * V2[Z] }
function VectorDotProduct(const V1: TVector; const V2: TAffineVector): single; overload;

{ : Projects p on the line defined by o and direction.<p>
  Performs VectorDotProduct(VectorSubtract(p, origin), direction), which,
  if direction is normalized, computes the distance between origin and the
  projection of p on the (origin, direction) line. }
function PointProject(const p, origin, direction: TAffineVector): single; overload;
function PointProject(const p, origin, direction: TVector): single; overload;

// : Calculates the cross product between vector 1 and 2
function VectorCrossProduct(const V1, V2: TAffineVector): TAffineVector; overload;
// : Calculates the cross product between vector 1 and 2
function VectorCrossProduct(const V1, V2: TVector): TVector; overload;
// : Calculates the cross product between vector 1 and 2, place result in vr
procedure VectorCrossProduct(const V1, V2: TVector; var vr: TVector); overload;
// : Calculates the cross product between vector 1 and 2, place result in vr
procedure VectorCrossProduct(const V1, V2: TAffineVector; var vr: TVector); overload;
// : Calculates the cross product between vector 1 and 2, place result in vr
procedure VectorCrossProduct(const V1, V2: TVector; var vr: TAffineVector); overload;
// : Calculates the cross product between vector 1 and 2, place result in vr
procedure VectorCrossProduct(const V1, V2: TAffineVector; var vr: TAffineVector); overload;

// : Calculates linear interpolation between start and stop at point t
function Lerp(const start, stop, T: single): single; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Calculates angular interpolation between start and stop at point t
function AngleLerp(start, stop, T: single): single; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
{ : This is used for interpolating between 2 matrices. The result
  is used to reposition the model parts each frame. }
function MatrixLerp(const m1, m2: TMatrix; const delta: single): TMatrix;

{ : Calculates the angular distance between two angles in radians.<p>
  Result is in the [0; PI] range. }
function DistanceBetweenAngles(angle1, angle2: single): single;

// : Calculates linear interpolation between texpoint1 and texpoint2 at point t
function TexPointLerp(const t1, t2: TTexPoint; T: single): TTexPoint; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Calculates linear interpolation between vector1 and vector2 at point t
function VectorLerp(const V1, V2: TAffineVector; T: single): TAffineVector; overload;
// : Calculates linear interpolation between vector1 and vector2 at point t, places result in vr
procedure VectorLerp(const V1, V2: TAffineVector; T: single; var vr: TAffineVector); overload;
// : Calculates linear interpolation between vector1 and vector2 at point t
function VectorLerp(const V1, V2: TVector; T: single): TVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Calculates linear interpolation between vector1 and vector2 at point t, places result in vr
procedure VectorLerp(const V1, V2: TVector; T: single; var vr: TVector); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorAngleLerp(const V1, V2: TAffineVector; T: single): TAffineVector; overload;
function VectorAngleCombine(const V1, V2: TAffineVector; f: single): TAffineVector; overload;

// : Calculates linear interpolation between vector arrays
procedure VectorArrayLerp(const src1, src2: PVectorArray; T: single; n: Integer;
  dest: PVectorArray); overload;
procedure VectorArrayLerp(const src1, src2: PAffineVectorArray; T: single; n: Integer;
  dest: PAffineVectorArray); overload;
procedure VectorArrayLerp(const src1, src2: PTexPointArray; T: single; n: Integer;
  dest: PTexPointArray); overload;

type
  TGLInterpolationType = (itLinear, itPower, itSin, itSinAlt, itTan, itLn, itExp);

  { : There functions that do the same as "Lerp", but add some distortions. }
function InterpolatePower(const start, stop, delta: single; const DistortionDegree: single): single;
function InterpolateLn(const start, stop, delta: single; const DistortionDegree: single): single;
function InterpolateExp(const start, stop, delta: single; const DistortionDegree: single): single;

{ : Only valid where Delta belongs to [0..1] }
function InterpolateSin(const start, stop, delta: single): single;
function InterpolateTan(const start, stop, delta: single): single;

{ : "Alt" functions are valid everywhere }
function InterpolateSinAlt(const start, stop, delta: single): single;

function InterpolateCombinedFastPower(const OriginalStart, OriginalStop, OriginalCurrent: single;
  const TargetStart, TargetStop: single; const DistortionDegree: single): single;
function InterpolateCombinedSafe(const OriginalStart, OriginalStop, OriginalCurrent: single;
  const TargetStart, TargetStop: single; const DistortionDegree: single;
  const InterpolationType: TGLInterpolationType): single;
function InterpolateCombinedFast(const OriginalStart, OriginalStop, OriginalCurrent: single;
  const TargetStart, TargetStop: single; const DistortionDegree: single;
  const InterpolationType: TGLInterpolationType): single;
function InterpolateCombined(const start, stop, delta: single; const DistortionDegree: single;
  const InterpolationType: TGLInterpolationType): single;

{ : Calculates the length of a vector following the equation sqrt(x*x+y*y). }
function VectorLength(const X, Y: single): single; overload;
{ : Calculates the length of a vector following the equation sqrt(x*x+y*y+z*z). }
function VectorLength(const X, Y, Z: single): single; overload;
// : Calculates the length of a vector following the equation sqrt(x*x+y*y).
function VectorLength(const V: TVector2f): single; overload;
// : Calculates the length of a vector following the equation sqrt(x*x+y*y+z*z).
function VectorLength(const V: TAffineVector): single; overload;
// : Calculates the length of a vector following the equation sqrt(x*x+y*y+z*z+w*w).
function VectorLength(const V: TVector): single; overload;
{ : Calculates the length of a vector following the equation: sqrt(x*x+y*y+...).<p>
  Note: The parameter of this function is declared as open array. Thus
  there's no restriction about the number of the components of the vector. }
function VectorLength(const V: array of single): single; overload;

{ : Calculates norm of a vector which is defined as norm = x * x + y * y<p>
  Also known as "Norm 2" in the math world, this is sqr(VectorLength). }
function VectorNorm(const X, Y: single): single; overload;
{ : Calculates norm of a vector which is defined as norm = x*x + y*y + z*z<p>
  Also known as "Norm 2" in the math world, this is sqr(VectorLength). }
function VectorNorm(const V: TAffineVector): single; overload;
{ : Calculates norm of a vector which is defined as norm = x*x + y*y + z*z<p>
  Also known as "Norm 2" in the math world, this is sqr(VectorLength). }
function VectorNorm(const V: TVector): single; overload;
{ : Calculates norm of a vector which is defined as norm = v.V[0]*v.V[0] + ...<p>
  Also known as "Norm 2" in the math world, this is sqr(VectorLength). }
function VectorNorm(var V: array of single): single; overload;

// : Transforms a vector to unit length
procedure NormalizeVector(var V: TVector2f); overload;
// : Returns the vector transformed to unit length
// : Transforms a vector to unit length
procedure NormalizeVector(var V: TAffineVector); overload;
// : Transforms a vector to unit length
procedure NormalizeVector(var V: TVector); overload;
// : Returns the vector transformed to unit length
function VectorNormalize(const V: TVector2f): TVector2f; overload;
// : Returns the vector transformed to unit length
function VectorNormalize(const V: TAffineVector): TAffineVector; overload;
// : Returns the vector transformed to unit length (w component dropped)
function VectorNormalize(const V: TVector): TVector; overload;

// : Transforms vectors to unit length
procedure NormalizeVectorArray(list: PAffineVectorArray; n: Integer); overload;

{ : Calculates the cosine of the angle between Vector1 and Vector2.<p>
  Result = DotProduct(V1, V2) / (Length(V1) * Length(V2)) }
function VectorAngleCosine(const V1, V2: TAffineVector): single; overload;

{ : Calculates the cosine of the angle between Vector1 and Vector2.<p>
  Result = DotProduct(V1, V2) / (Length(V1) * Length(V2)) }
function VectorAngleCosine(const V1, V2: TVector): single; overload;

// : Negates the vector
function VectorNegate(const Vector: TAffineVector): TAffineVector; overload;
function VectorNegate(const Vector: TVector): TVector; overload;

// : Negates the vector
procedure NegateVector(var V: TAffineVector); overload;
// : Negates the vector
procedure NegateVector(var V: TVector); overload;
// : Negates the vector
procedure NegateVector(var V: array of single); overload;

// : Scales given vector by a factor
procedure ScaleVector(var V: TVector2f; factor: single); overload;
// : Scales given vector by a factor
procedure ScaleVector(var V: TAffineVector; factor: single); overload;
{ : Scales given vector by another vector.<p>
  v[x]:=v[x]*factor[x], v[y]:=v[y]*factor[y] etc. }
procedure ScaleVector(var V: TAffineVector; const factor: TAffineVector); overload;
// : Scales given vector by a factor
procedure ScaleVector(var V: TVector; factor: single); overload;
{ : Scales given vector by another vector.<p>
  v[x]:=v[x]*factor[x], v[y]:=v[y]*factor[y] etc. }
procedure ScaleVector(var V: TVector; const factor: TVector); overload;

// : Returns a vector scaled by a factor
function VectorScale(const V: TVector2f; factor: single): TVector2f; overload;
// : Returns a vector scaled by a factor
function VectorScale(const V: TAffineVector; factor: single): TAffineVector; overload;
// : Scales a vector by a factor and places result in vr
procedure VectorScale(const V: TAffineVector; factor: single; var vr: TAffineVector); overload;
// : Returns a vector scaled by a factor
function VectorScale(const V: TVector; factor: single): TVector; overload;
// : Scales a vector by a factor and places result in vr
procedure VectorScale(const V: TVector; factor: single; var vr: TVector); overload;
// : Scales a vector by a factor and places result in vr
procedure VectorScale(const V: TVector; factor: single; var vr: TAffineVector); overload;
// : Scales given vector by another vector
function VectorScale(const V: TAffineVector; const factor: TAffineVector): TAffineVector; overload;
// : RScales given vector by another vector
function VectorScale(const V: TVector; const factor: TVector): TVector; overload;

{ : Divides given vector by another vector.<p>
  v[x]:=v[x]/divider[x], v[y]:=v[y]/divider[y] etc. }
procedure DivideVector(var V: TVector; const divider: TVector); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
procedure DivideVector(var V: TAffineVector; const divider: TAffineVector); overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorDivide(const V: TVector; const divider: TVector): TVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function VectorDivide(const V: TAffineVector; const divider: TAffineVector): TAffineVector;
  overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : True if all components are equal.
function TexpointEquals(const p1, p2: TTexPoint): Boolean; {$IFDEF GLS_INLINE_VICE_ASM}inline;
{$ENDIF}
// : True if all components are equal.
function RectEquals(const Rect1, Rect2: TRect): Boolean; {$IFDEF GLS_INLINE_VICE_ASM}inline;
{$ENDIF}
// : True if all components are equal.
function VectorEquals(const V1, V2: TVector): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM_VICE_ASM}inline; {$ENDIF}
// : True if all components are equal.
function VectorEquals(const V1, V2: TAffineVector): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM_VICE_ASM}inline; {$ENDIF}
// : True if X, Y and Z components are equal.
function AffineVectorEquals(const V1, V2: TVector): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM_VICE_ASM}inline; {$ENDIF}
// : True if x=y=z=0, w ignored
function VectorIsNull(const V: TVector): Boolean; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline;
{$ENDIF}
// : True if x=y=z=0, w ignored
function VectorIsNull(const V: TAffineVector): Boolean; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
{ : Calculates Abs(v1[x]-v2[x])+Abs(v1[y]-v2[y]), also know as "Norm1".<p> }
function VectorSpacing(const V1, V2: TTexPoint): single; overload;
{ : Calculates Abs(v1[x]-v2[x])+Abs(v1[y]-v2[y])+..., also know as "Norm1".<p> }
function VectorSpacing(const V1, V2: TAffineVector): single; overload;
{ : Calculates Abs(v1[x]-v2[x])+Abs(v1[y]-v2[y])+..., also know as "Norm1".<p> }
function VectorSpacing(const V1, V2: TVector): single; overload;

{ : Calculates distance between two vectors.<p>
  ie. sqrt(sqr(v1[x]-v2[x])+...) }
function VectorDistance(const V1, V2: TAffineVector): single; overload;
{ : Calculates distance between two vectors.<p>
  ie. sqrt(sqr(v1[x]-v2[x])+...) (w component ignored) }
function VectorDistance(const V1, V2: TVector): single; overload;

{ : Calculates the "Norm 2" between two vectors.<p>
  ie. sqr(v1[x]-v2[x])+... }
function VectorDistance2(const V1, V2: TAffineVector): single; overload;
{ : Calculates the "Norm 2" between two vectors.<p>
  ie. sqr(v1[x]-v2[x])+... (w component ignored) }
function VectorDistance2(const V1, V2: TVector): single; overload;

{ : Calculates a vector perpendicular to N.<p>
  N is assumed to be of unit length, subtract out any component parallel to N }
function VectorPerpendicular(const V, n: TAffineVector): TAffineVector;
// : Reflects vector V against N (assumes N is normalized)
function VectorReflect(const V, n: TAffineVector): TAffineVector;
// : Rotates Vector about Axis with Angle radians
procedure RotateVector(var Vector: TVector; const axis: TAffineVector; angle: single); overload;
// : Rotates Vector about Axis with Angle radians
procedure RotateVector(var Vector: TVector; const axis: TVector; angle: single); overload;

// : Rotate given vector around the Y axis (alpha is in rad)
procedure RotateVectorAroundY(var V: TAffineVector; alpha: single);
// : Returns given vector rotated around the X axis (alpha is in rad)
function VectorRotateAroundX(const V: TAffineVector; alpha: single): TAffineVector; overload;
// : Returns given vector rotated around the Y axis (alpha is in rad)
function VectorRotateAroundY(const V: TAffineVector; alpha: single): TAffineVector; overload;
// : Returns given vector rotated around the Y axis in vr (alpha is in rad)
procedure VectorRotateAroundY(const V: TAffineVector; alpha: single;
  var vr: TAffineVector); overload;
// : Returns given vector rotated around the Z axis (alpha is in rad)
function VectorRotateAroundZ(const V: TAffineVector; alpha: single): TAffineVector; overload;

// : Vector components are replaced by their Abs() value. }
procedure AbsVector(var V: TVector); overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Vector components are replaced by their Abs() value. }
procedure AbsVector(var V: TAffineVector); overload; {$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Returns a vector with components replaced by their Abs value. }
function VectorAbs(const V: TVector): TVector; overload; {$IFDEF GLS_INLINE_VICE_ASM}inline;
{$ENDIF}
// : Returns a vector with components replaced by their Abs value. }
function VectorAbs(const V: TAffineVector): TAffineVector; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
// : Returns true if both vector are colinear
function IsColinear(const V1, V2: TVector2f): Boolean; overload;
// : Returns true if both vector are colinear
function IsColinear(const V1, V2: TAffineVector): Boolean; overload;
// : Returns true if both vector are colinear
function IsColinear(const V1, V2: TVector): Boolean; overload;

// ------------------------------------------------------------------------------
// Matrix functions
// ------------------------------------------------------------------------------

procedure SetMatrix(var dest: THomogeneousDblMatrix; const src: TMatrix); overload;
procedure SetMatrix(var dest: TAffineMatrix; const src: TMatrix); overload;
procedure SetMatrix(var dest: TMatrix; const src: TAffineMatrix); overload;

procedure SetMatrixRow(var dest: TMatrix; rowNb: Integer; const aRow: TVector); overload;

// : Creates scale matrix
function CreateScaleMatrix(const V: TAffineVector): TMatrix; overload;
// : Creates scale matrix
function CreateScaleMatrix(const V: TVector): TMatrix; overload;
// : Creates translation matrix
function CreateTranslationMatrix(const V: TAffineVector): TMatrix; overload;
// : Creates translation matrix
function CreateTranslationMatrix(const V: TVector): TMatrix; overload;
{ : Creates a scale+translation matrix.<p>
  Scale is applied BEFORE applying offset }
function CreateScaleAndTranslationMatrix(const scale, offset: TVector): TMatrix; overload;
// : Creates matrix for rotation about x-axis (angle in rad)
function CreateRotationMatrixX(const sine, cosine: single): TMatrix; overload;
function CreateRotationMatrixX(const angle: single): TMatrix; overload;
// : Creates matrix for rotation about y-axis (angle in rad)
function CreateRotationMatrixY(const sine, cosine: single): TMatrix; overload;
function CreateRotationMatrixY(const angle: single): TMatrix; overload;
// : Creates matrix for rotation about z-axis (angle in rad)
function CreateRotationMatrixZ(const sine, cosine: single): TMatrix; overload;
function CreateRotationMatrixZ(const angle: single): TMatrix; overload;
// : Creates a rotation matrix along the given Axis by the given Angle in radians.
function CreateRotationMatrix(const anAxis: TAffineVector; angle: single): TMatrix; overload;
function CreateRotationMatrix(const anAxis: TVector; angle: single): TMatrix; overload;
// : Creates a rotation matrix along the given Axis by the given Angle in radians.
function CreateAffineRotationMatrix(const anAxis: TAffineVector; angle: single): TAffineMatrix;

// : Multiplies two 3x3 matrices
function MatrixMultiply(const m1, m2: TAffineMatrix): TAffineMatrix; overload;
// : Multiplies two 4x4 matrices
function MatrixMultiply(const m1, m2: TMatrix): TMatrix; overload;
// : Multiplies M1 by M2 and places result in MResult
procedure MatrixMultiply(const m1, m2: TMatrix; var MResult: TMatrix); overload;

// : Transforms a homogeneous vector by multiplying it with a matrix
function VectorTransform(const V: TVector; const M: TMatrix): TVector; overload;
// : Transforms a homogeneous vector by multiplying it with a matrix
function VectorTransform(const V: TVector; const M: TAffineMatrix): TVector; overload;
// : Transforms an affine vector by multiplying it with a matrix
function VectorTransform(const V: TAffineVector; const M: TMatrix): TAffineVector; overload;
// : Transforms an affine vector by multiplying it with a matrix
function VectorTransform(const V: TAffineVector; const M: TAffineMatrix): TAffineVector; overload;

// : Determinant of a 3x3 matrix
function MatrixDeterminant(const M: TAffineMatrix): single; overload;
// : Determinant of a 4x4 matrix
function MatrixDeterminant(const M: TMatrix): single; overload;

{ : Adjoint of a 4x4 matrix.<p>
  used in the computation of the inverse of a 4x4 matrix }
procedure AdjointMatrix(var M: TMatrix); overload;
{ : Adjoint of a 3x3 matrix.<p>
  used in the computation of the inverse of a 3x3 matrix }
procedure AdjointMatrix(var M: TAffineMatrix); overload;

// : Multiplies all elements of a 3x3 matrix with a factor
procedure ScaleMatrix(var M: TAffineMatrix; const factor: single); overload;
// : Multiplies all elements of a 4x4 matrix with a factor
procedure ScaleMatrix(var M: TMatrix; const factor: single); overload;

// : Adds the translation vector into the matrix
procedure TranslateMatrix(var M: TMatrix; const V: TAffineVector); overload;
procedure TranslateMatrix(var M: TMatrix; const V: TVector); overload;

{ : Normalize the matrix and remove the translation component.<p>
  The resulting matrix is an orthonormal matrix (Y direction preserved, then Z) }
procedure NormalizeMatrix(var M: TMatrix);

// : Computes transpose of 3x3 matrix
procedure TransposeMatrix(var M: TAffineMatrix); overload;
// : Computes transpose of 4x4 matrix
procedure TransposeMatrix(var M: TMatrix); overload;

// : Finds the inverse of a 4x4 matrix
procedure InvertMatrix(var M: TMatrix); overload;
function MatrixInvert(const M: TMatrix): TMatrix; overload;

// : Finds the inverse of a 3x3 matrix;
procedure InvertMatrix(var M: TAffineMatrix); overload;
function MatrixInvert(const M: TAffineMatrix): TAffineMatrix; overload;

{ : Finds the inverse of an angle preserving matrix.<p>
  Angle preserving matrices can combine translation, rotation and isotropic
  scaling, other matrices won't be properly inverted by this function. }
function AnglePreservingMatrixInvert(const mat: TMatrix): TMatrix;

{ : Decompose a non-degenerated 4x4 transformation matrix into the sequence of transformations that produced it.<p>
  Modified by ml then eg, original Author: Spencer W. Thomas, University of Michigan<p>
  The coefficient of each transformation is returned in the corresponding
  element of the vector Tran.<p>
  Returns true upon success, false if the matrix is singular. }
function MatrixDecompose(const M: TMatrix; var Tran: TTransformations): Boolean;

function CreateLookAtMatrix(const eye, center, normUp: TVector): TMatrix;
function CreateMatrixFromFrustum(Left, Right, Bottom, Top, ZNear, ZFar: single): TMatrix;
function CreatePerspectiveMatrix(FOV, Aspect, ZNear, ZFar: single): TMatrix;
function CreateOrthoMatrix(Left, Right, Bottom, Top, ZNear, ZFar: single): TMatrix;
function CreatePickMatrix(X, Y, deltax, deltay: single; const viewport: TVector4i): TMatrix;
function Project(objectVector: TVector; const ViewProjMatrix: TMatrix; const viewport: TVector4i;
  out WindowVector: TVector): Boolean;
function UnProject(WindowVector: TVector; ViewProjMatrix: TMatrix; const viewport: TVector4i;
  out objectVector: TVector): Boolean;
// ------------------------------------------------------------------------------
// Plane functions
// ------------------------------------------------------------------------------

// : Computes the parameters of a plane defined by three points.
function PlaneMake(const p1, p2, p3: TAffineVector): THmgPlane; overload;
function PlaneMake(const p1, p2, p3: TVector): THmgPlane; overload;
// : Computes the parameters of a plane defined by a point and a normal.
function PlaneMake(const point, normal: TAffineVector): THmgPlane; overload;
function PlaneMake(const point, normal: TVector): THmgPlane; overload;
// : Converts from single to double representation
procedure SetPlane(var dest: TDoubleHmgPlane; const src: THmgPlane);

// : Normalize a plane so that point evaluation = plane distance. }
procedure NormalizePlane(var plane: THmgPlane);

{ : Calculates the cross-product between the plane normal and plane to point vector.<p>
  This functions gives an hint as to were the point is, if the point is in the
  half-space pointed by the vector, result is positive.<p>
  This function performs an homogeneous space dot-product. }
function PlaneEvaluatePoint(const plane: THmgPlane; const point: TAffineVector): single; overload;
function PlaneEvaluatePoint(const plane: THmgPlane; const point: TVector): single; overload;

{ : Calculate the normal of a plane defined by three points. }
function CalcPlaneNormal(const p1, p2, p3: TAffineVector): TAffineVector; overload;
procedure CalcPlaneNormal(const p1, p2, p3: TAffineVector; var vr: TAffineVector); overload;
procedure CalcPlaneNormal(const p1, p2, p3: TVector; var vr: TAffineVector); overload;

{ : Returns true if point is in the half-space defined by a plane with normal.<p>
  The plane itself is not considered to be in the tested halfspace. }
function PointIsInHalfSpace(const point, planePoint, planeNormal: TVector): Boolean; overload;
function PointIsInHalfSpace(const point, planePoint, planeNormal: TAffineVector): Boolean; overload;
function PointIsInHalfSpace(const point: TAffineVector; plane: THmgPlane): Boolean; overload;

{ : Computes algebraic distance between point and plane.<p>
  Value will be positive if the point is in the halfspace pointed by the normal,
  negative on the other side. }
function PointPlaneDistance(const point, planePoint, planeNormal: TVector): single; overload;
function PointPlaneDistance(const point, planePoint, planeNormal: TAffineVector): single; overload;
function PointPlaneDistance(const point: TAffineVector; plane: THmgPlane): single; overload;

{ : Computes point to plane projection. Plane and direction have to be normalized }
function PointPlaneOrthoProjection(const point: TAffineVector; const plane: THmgPlane;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;
function PointPlaneProjection(const point, direction: TAffineVector; const plane: THmgPlane;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;

{ : Computes segment / plane intersection return false if there isn't an intersection }
function SegmentPlaneIntersection(const ptA, ptB: TAffineVector; const plane: THmgPlane;
  var inter: TAffineVector): Boolean;

{ : Computes point to triangle projection. Direction has to be normalized }
function PointTriangleOrthoProjection(const point, ptA, ptB, ptC: TAffineVector;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;
function PointTriangleProjection(const point, direction, ptA, ptB, ptC: TAffineVector;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;

{ : Returns true if line intersect ABC triangle. }
function IsLineIntersectTriangle(const point, direction, ptA, ptB, ptC: TAffineVector): Boolean;

{ : Computes point to Quad projection. Direction has to be normalized. Quad have to be flat and convex }
function PointQuadOrthoProjection(const point, ptA, ptB, ptC, ptD: TAffineVector;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;
function PointQuadProjection(const point, direction, ptA, ptB, ptC, ptD: TAffineVector;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;

{ : Returns true if line intersect ABCD quad. Quad have to be flat and convex }
function IsLineIntersectQuad(const point, direction, ptA, ptB, ptC, ptD: TAffineVector): Boolean;

{ : Computes point to disk projection. Direction has to be normalized }
function PointDiskOrthoProjection(const point, center, up: TAffineVector; const radius: single;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;
function PointDiskProjection(const point, direction, center, up: TAffineVector;
  const radius: single; var inter: TAffineVector; bothface: Boolean = True): Boolean;

{ : Computes closest point on a segment (a segment is a limited line). }
function PointSegmentClosestPoint(const point, segmentStart, segmentStop: TAffineVector)
  : TAffineVector; overload;
function PointSegmentClosestPoint(const point, segmentStart, segmentStop: TVector)
  : TVector; overload;

{ : Computes algebraic distance between segment and line (a segment is a limited line). }
function PointSegmentDistance(const point, segmentStart, segmentStop: TAffineVector): single;

{ : Computes closest point on a line. }
function PointLineClosestPoint(const point, linePoint, lineDirection: TAffineVector): TAffineVector;

{ : Computes algebraic distance between point and line. }
function PointLineDistance(const point, linePoint, lineDirection: TAffineVector): single;

{ : Computes the closest points (2) given two segments. }
procedure SegmentSegmentClosestPoint(const S0Start, S0Stop, S1Start, S1Stop: TAffineVector;
  var Segment0Closest, Segment1Closest: TAffineVector);

{ : Computes the closest distance between two segments. }
function SegmentSegmentDistance(const S0Start, S0Stop, S1Start, S1Stop: TAffineVector): single;

{ : Computes the closest distance between two lines. }
function LineLineDistance(const linePt0, lineDir0, linePt1, lineDir1: TAffineVector): single;

// ------------------------------------------------------------------------------
// Quaternion functions
// ------------------------------------------------------------------------------

type
  TEulerOrder = (eulXYZ, eulXZY, eulYXZ, eulYZX, eulZXY, eulZYX);

  // : Creates a quaternion from the given values
function QuaternionMake(const Imag: array of single; Real: single): TQuaternion;
// : Returns the conjugate of a quaternion
function QuaternionConjugate(const Q: TQuaternion): TQuaternion;
// : Returns the magnitude of the quaternion
function QuaternionMagnitude(const Q: TQuaternion): single;
// : Normalizes the given quaternion
procedure NormalizeQuaternion(var Q: TQuaternion);

// : Constructs a unit quaternion from two points on unit sphere
function QuaternionFromPoints(const V1, V2: TAffineVector): TQuaternion;
// : Converts a unit quaternion into two points on a unit sphere
procedure QuaternionToPoints(const Q: TQuaternion; var ArcFrom, ArcTo: TAffineVector);
// : Constructs a unit quaternion from a rotation matrix
function QuaternionFromMatrix(const mat: TMatrix): TQuaternion;
{ : Constructs a rotation matrix from (possibly non-unit) quaternion.<p>
  Assumes matrix is used to multiply column vector on the left:<br>
  vnew = mat vold.<p>
  Works correctly for right-handed coordinate system and right-handed rotations. }
function QuaternionToMatrix(quat: TQuaternion): TMatrix;
{ : Constructs an affine rotation matrix from (possibly non-unit) quaternion.<p> }
function QuaternionToAffineMatrix(quat: TQuaternion): TAffineMatrix;
// : Constructs quaternion from angle (in deg) and axis
function QuaternionFromAngleAxis(const angle: single; const axis: TAffineVector): TQuaternion;
// : Constructs quaternion from Euler angles
function QuaternionFromRollPitchYaw(const r, p, Y: single): TQuaternion;
// : Constructs quaternion from Euler angles in arbitrary order (angles in degrees)
function QuaternionFromEuler(const X, Y, Z: single; eulerOrder: TEulerOrder): TQuaternion;

{ : Returns quaternion product qL * qR.<p>
  Note: order is important!<p>
  To combine rotations, use the product QuaternionMuliply(qSecond, qFirst),
  which gives the effect of rotating by qFirst then qSecond. }
function QuaternionMultiply(const qL, qR: TQuaternion): TQuaternion;

{ : Spherical linear interpolation of unit quaternions with spins.<p>
  QStart, QEnd - start and end unit quaternions<br>
  t            - interpolation parameter (0 to 1)<br>
  Spin         - number of extra spin rotations to involve<br> }
function QuaternionSlerp(const QStart, QEnd: TQuaternion; Spin: Integer; T: single)
  : TQuaternion; overload;
function QuaternionSlerp(const source, dest: TQuaternion; const T: single): TQuaternion; overload;

// ------------------------------------------------------------------------------
// Logarithmic and exponential functions
// ------------------------------------------------------------------------------

{ : Return ln(1 + X),  accurate for X near 0. }
function LnXP1(X: Extended): Extended;
{ : Log base 10 of X }
function Log10(X: Extended): Extended;
{ : Log base 2 of X }
function Log2(X: Extended): Extended; overload;
{ : Log base 2 of X }
function Log2(X: single): single; overload;
{ : Log base N of X }
function LogN(Base, X: Extended): Extended;
{ : Raise base to an integer. }
function IntPower(Base: Extended; Exponent: Integer): Extended;
{ : Raise base to any power.<p>
  For fractional exponents, or |exponents| > MaxInt, base must be > 0. }
function Power(const Base, Exponent: single): single; overload;
{ : Raise base to an integer. }
function Power(Base: single; Exponent: Integer): single; overload;
function Power(Base: single; Exponent: Int64): single; overload;

// ------------------------------------------------------------------------------
// Trigonometric functions
// ------------------------------------------------------------------------------

function DegToRad(const Degrees: Extended): Extended; overload;
function DegToRad(const Degrees: single): single; overload;
function RadToDeg(const Radians: Extended): Extended; overload;
function RadToDeg(const Radians: single): single; overload;

// : Normalize to an angle in the [-PI; +PI] range
function NormalizeAngle(angle: single): single;
// : Normalize to an angle in the [-180; 180] range
function NormalizeDegAngle(angle: single): single;

// : Calculates sine and cosine from the given angle Theta
{$IFDEF GLS_PLATFORM_HAS_EXTENDED}
procedure SinCos(const Theta: Extended; out Sin, Cos: Extended); overload;
{$ENDIF}
// : Calculates sine and cosine from the given angle Theta
procedure SinCos(const Theta: double; out Sin, Cos: double); overload;
// : Calculates sine and cosine from the given angle Theta
procedure SinCos(const Theta: single; out Sin, Cos: single); overload;
{ : Calculates sine and cosine from the given angle Theta and Radius.<p>
  sin and cos values calculated from theta are multiplicated by radius. }
{$IFDEF GLS_PLATFORM_HAS_EXTENDED}
procedure SinCos(const Theta, radius: double; out Sin, Cos: Extended); overload;
{$ENDIF}
{ : Calculates sine and cosine from the given angle Theta and Radius.<p>
  sin and cos values calculated from theta are multiplicated by radius. }
procedure SinCos(const Theta, radius: double; out Sin, Cos: double); overload;
{ : Calculates sine and cosine from the given angle Theta and Radius.<p>
  sin and cos values calculated from theta are multiplicated by radius. }
procedure SinCos(const Theta, radius: single; out Sin, Cos: single); overload;

{ : Fills up the two given dynamic arrays with sin cos values.<p>
  start and stop angles must be given in degrees, the number of steps is
  determined by the length of the given arrays. }
procedure PrepareSinCosCache(var S, c: array of single; startAngle, stopAngle: single);

function ArcCos(const X: Extended): Extended; overload;
function ArcCos(const X: single): single; overload;
function ArcSin(const X: Extended): Extended; overload;
function ArcSin(const X: single): single; overload;
function ArcTan2(const Y, X: Extended): Extended; overload;
function ArcTan2(const Y, X: single): single; overload;
{ : Fast ArcTan2 approximation, about 0.07 rads accuracy. }
function FastArcTan2(Y, X: single): single;
function Tan(const X: Extended): Extended; overload;
function Tan(const X: single): single; overload;
function CoTan(const X: Extended): Extended; overload;
function CoTan(const X: single): single; overload;

// ------------------------------------------------------------------------------
// Hyperbolic Trigonometric functions
// ------------------------------------------------------------------------------

function Sinh(const X: single): single; overload;
function Sinh(const X: double): double; overload;
function Cosh(const X: single): single; overload;
function Cosh(const X: double): double; overload;

// ------------------------------------------------------------------------------
// Miscellanious math functions
// ------------------------------------------------------------------------------

{ : Computes 1/Sqrt(v).<p> }
function RSqrt(V: single): single;
{ : Computes 1/Sqrt(Sqr(x)+Sqr(y)). }
function RLength(X, Y: single): single;
{ : Computes an integer sqrt approximation.<p> }
function ISqrt(i: Integer): Integer;
{ : Computes an integer length Result:=Sqrt(x*x+y*y). }
function ILength(X, Y: Integer): Integer; overload;
function ILength(X, Y, Z: Integer): Integer; overload;

{$IFNDEF GEOMETRY_NO_ASM}
{ : Computes Exp(ST(0)) and leaves result on ST(0) }
procedure RegisterBasedExp;
{$ENDIF}
{ : Generates a random point on the unit sphere.<p>
  Point repartition is correctly isotropic with no privilegied direction. }
procedure RandomPointOnSphere(var p: TAffineVector);

{ : Rounds the floating point value to the closest integer.<p>
  Behaves like Round but returns a floating point value like Int. }
function RoundInt(V: single): single; overload;
function RoundInt(V: Extended): Extended; overload;

{$IFNDEF GEOMETRY_NO_ASM}
function Trunc(V: single): Integer; overload;
function Trunc64(V: Extended): Int64; overload;
function Int(V: single): single; overload;
function Int(V: Extended): Extended; overload;
function Frac(V: single): single; overload;
function Frac(V: Extended): Extended; overload;
function Round(V: single): Integer; overload;
function Round64(V: single): Int64; overload;
function Round64(V: Extended): Int64; overload;
{$ELSE}
function Trunc(X: Extended): Int64;
function Round(X: Extended): Int64;
function Frac(X: Extended): Extended;
{$ENDIF}
function Ceil(V: single): Integer; overload;
function Ceil64(V: Extended): Int64; overload;
function Floor(V: single): Integer; overload;
function Floor64(V: Extended): Int64; overload;

{ : Multiples i by s and returns the rounded result.<p> }
function ScaleAndRound(i: Integer; var S: single): Integer;

{ : Returns the sign of the x value using the (-1, 0, +1) convention }
function Sign(X: single): Integer;
function SignStrict(X: single): Integer;

{ : Returns True if x is in [a; b] }
function IsInRange(const X, a, b: single): Boolean; overload;
function IsInRange(const X, a, b: double): Boolean; overload;

{ : Returns True if p is in the cube defined by d. }
function IsInCube(const p, d: TAffineVector): Boolean; overload;
function IsInCube(const p, d: TVector): Boolean; overload;

{ : Returns the minimum value of the array. }
function MinFloat(values: PSingleArray; nbItems: Integer): single; overload;
function MinFloat(values: PDoubleArray; nbItems: Integer): double; overload;
function MinFloat(values: PExtendedArray; nbItems: Integer): Extended; overload;
{ : Returns the minimum of given values. }
function MinFloat(const V1, V2: single): single; overload;
function MinFloat(const V: array of single): single; overload;
function MinFloat(const V1, V2: double): double; overload;
{$IFDEF GLS_PLATFORM_HAS_EXTENDED}
function MinFloat(const V1, V2: Extended): Extended; overload;
{$ENDIF}
function MinFloat(const V1, V2, V3: single): single; overload;
function MinFloat(const V1, V2, V3: double): double; overload;
{$IFDEF GLS_PLATFORM_HAS_EXTENDED}
function MinFloat(const V1, V2, V3: Extended): Extended; overload;
{$ENDIF}
{ : Returns the maximum value of the array. }
function MaxFloat(values: PSingleArray; nbItems: Integer): single; overload;
function MaxFloat(values: PDoubleArray; nbItems: Integer): double; overload;
function MaxFloat(values: PExtendedArray; nbItems: Integer): Extended; overload;
function MaxFloat(const V: array of single): single; overload;
{ : Returns the maximum of given values. }
function MaxFloat(const V1, V2: single): single; overload;
function MaxFloat(const V1, V2: double): double; overload;
{$IFDEF GLS_PLATFORM_HAS_EXTENDED}
function MaxFloat(const V1, V2: Extended): Extended; overload;
{$ENDIF GLS_PLATFORM_HAS_EXTENDED}
function MaxFloat(const V1, V2, V3: single): single; overload;
function MaxFloat(const V1, V2, V3: double): double; overload;
{$IFDEF GLS_PLATFORM_HAS_EXTENDED}
function MaxFloat(const V1, V2, V3: Extended): Extended; overload;
{$ENDIF GLS_PLATFORM_HAS_EXTENDED}
function MinInteger(const V1, V2: Integer): Integer; overload;
function MinInteger(const V1, V2: Cardinal): Cardinal; overload;
function MinInteger(const V1, V2, V3: Integer): Integer; overload;
function MinInteger(const V1, V2, V3: Cardinal): Cardinal; overload;

function MaxInteger(const V1, V2: Integer): Integer; overload;
function MaxInteger(const V1, V2: Cardinal): Cardinal; overload;
function MaxInteger(const V1, V2, V3: Integer): Integer; overload;
function MaxInteger(const V1, V2, V3: Cardinal): Cardinal; overload;

function ClampInteger(const value, min, max: Integer): Integer; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
function ClampInteger(const value, min, max: Cardinal): Cardinal; overload;
{$IFDEF GLS_INLINE_VICE_ASM}inline; {$ENDIF}
{ : Computes the triangle's area. }
function TriangleArea(const p1, p2, p3: TAffineVector): single; overload;
{ : Computes the polygons's area.<p>
  Points must be coplanar. Polygon needs not be convex. }
function PolygonArea(const p: PAffineVectorArray; nSides: Integer): single; overload;
{ : Computes a 2D triangle's signed area.<p>
  Only X and Y coordinates are used, Z is ignored. }
function TriangleSignedArea(const p1, p2, p3: TAffineVector): single; overload;
{ : Computes a 2D polygon's signed area.<p>
  Only X and Y coordinates are used, Z is ignored. Polygon needs not be convex. }
function PolygonSignedArea(const p: PAffineVectorArray; nSides: Integer): single; overload;

{ : Multiplies values in the array by factor.<p>
  This function is especially efficient for large arrays, it is not recommended
  for arrays that have less than 10 items.<br>
  Expected performance is 4 to 5 times that of a Deliph-compiled loop on AMD
  CPUs, and 2 to 3 when 3DNow! isn't available. }
procedure ScaleFloatArray(values: PSingleArray; nb: Integer; var factor: single); overload;
procedure ScaleFloatArray(var values: TSingleArray; factor: single); overload;

{ : Adds delta to values in the array.<p>
  Array size must be a multiple of four. }
procedure OffsetFloatArray(values: PSingleArray; nb: Integer; var delta: single); overload;
procedure OffsetFloatArray(var values: array of single; delta: single); overload;
procedure OffsetFloatArray(valuesDest, valuesDelta: PSingleArray; nb: Integer); overload;

{ : Returns the max of the X, Y and Z components of a vector (W is ignored). }
function MaxXYZComponent(const V: TVector): single; overload;
function MaxXYZComponent(const V: TAffineVector): single; overload;
{ : Returns the min of the X, Y and Z components of a vector (W is ignored). }
function MinXYZComponent(const V: TVector): single; overload;
function MinXYZComponent(const V: TAffineVector): single; overload;
{ : Returns the max of the Abs(X), Abs(Y) and Abs(Z) components of a vector (W is ignored). }
function MaxAbsXYZComponent(V: TVector): single;
{ : Returns the min of the Abs(X), Abs(Y) and Abs(Z) components of a vector (W is ignored). }
function MinAbsXYZComponent(V: TVector): single;
{ : Replace components of v with the max of v or v1 component.<p>
  Maximum is computed per component. }
procedure MaxVector(var V: TVector; const V1: TVector); overload;
procedure MaxVector(var V: TAffineVector; const V1: TAffineVector); overload;
{ : Replace components of v with the min of v or v1 component.<p>
  Minimum is computed per component. }
procedure MinVector(var V: TVector; const V1: TVector); overload;
procedure MinVector(var V: TAffineVector; const V1: TAffineVector); overload;

{ : Sorts given array in ascending order.<p>
  NOTE : current implementation is a slow bubble sort... }
procedure SortArrayAscending(var a: array of Extended);

{ : Clamps aValue in the aMin-aMax interval.<p> }
function ClampValue(const aValue, aMin, aMax: single): single; overload;
{ : Clamps aValue in the aMin-INF interval.<p> }
function ClampValue(const aValue, aMin: single): single; overload;

{ : Returns the detected optimization mode.<p>
  Returned values is either 'FPU', '3DNow!' or 'SSE'. }
function GeometryOptimizationMode: String;

{ : Begins a FPU-only section.<p>
  You can use a FPU-only section to force use of FPU versions of the math
  functions, though typically slower than their SIMD counterparts, they have
  a higher precision (80 bits internally) that may be required in some cases.<p>
  Each BeginFPUOnlySection call must be balanced by a EndFPUOnlySection (calls
  can be nested). }
procedure BeginFPUOnlySection;
{ : Ends a FPU-only section.<p>
  See BeginFPUOnlySection. }
procedure EndFPUOnlySection;

// --------------------- Unstandardized functions after these lines
// --------------------- Unstandardized functions after these lines
// --------------------- Unstandardized functions after these lines
// --------------------- Unstandardized functions after these lines
// --------------------- Unstandardized functions after these lines

// mixed functions

{ : Turn a triplet of rotations about x, y, and z (in that order) into an equivalent rotation around a single axis (all in radians).<p> }
function ConvertRotation(const Angles: TAffineVector): TVector;

// miscellaneous functions

function MakeAffineDblVector(var V: array of double): TAffineDblVector;
function MakeDblVector(var V: array of double): THomogeneousDblVector;
function VectorAffineDblToFlt(const V: TAffineDblVector): TAffineVector;
function VectorDblToFlt(const V: THomogeneousDblVector): THomogeneousVector;
function VectorAffineFltToDbl(const V: TAffineVector): TAffineDblVector;
function VectorFltToDbl(const V: TVector): THomogeneousDblVector;

function PointInPolygon(var xp, yp: array of single; X, Y: single): Boolean;
function IsPointInPolygon(Polygon: array of TPoint; p: TPoint): Boolean;

procedure DivMod(Dividend: Integer; Divisor: Word; var result, Remainder: Word);

// coordinate system manipulation functions

// : Rotates the given coordinate system (represented by the matrix) around its Y-axis
function Turn(const Matrix: TMatrix; angle: single): TMatrix; overload;
// : Rotates the given coordinate system (represented by the matrix) around MasterUp
function Turn(const Matrix: TMatrix; const MasterUp: TAffineVector; angle: single)
  : TMatrix; overload;
// : Rotates the given coordinate system (represented by the matrix) around its X-axis
function Pitch(const Matrix: TMatrix; angle: single): TMatrix; overload;
// : Rotates the given coordinate system (represented by the matrix) around MasterRight
function Pitch(const Matrix: TMatrix; const MasterRight: TAffineVector; angle: single)
  : TMatrix; overload;
// : Rotates the given coordinate system (represented by the matrix) around its Z-axis
function Roll(const Matrix: TMatrix; angle: single): TMatrix; overload;
// : Rotates the given coordinate system (represented by the matrix) around MasterDirection
function Roll(const Matrix: TMatrix; const MasterDirection: TAffineVector; angle: single)
  : TMatrix; overload;

// intersection functions

{ : Compute the intersection point "res" of a line with a plane.<p>
  Return value:<ul>
  <li>0 : no intersection, line parallel to plane
  <li>1 : res is valid
  <li>-1 : line is inside plane
  </ul><br>
  Adapted from:<br>
  E.Hartmann, Computeruntersttzte Darstellende Geometrie, B.G. Teubner Stuttgart 1988 }
function IntersectLinePlane(const point, direction: TVector; const plane: THmgPlane;
  intersectPoint: PVector = nil): Integer; overload;

{ : Compute intersection between a triangle and a box.<p>
  Returns True if an intersection was found. }
function IntersectTriangleBox(const p1, p2, p3, aMinExtent, aMaxExtent: TAffineVector): Boolean;

{ : Compute intersection between a Sphere and a box.<p>
  Up, Direction and Right must be normalized!
  Use CubDepht, CubeHeight and CubeWidth to scale TGLCube. }
function IntersectSphereBox(const SpherePos: TVector; const SphereRadius: single;
  const BoxMatrix: TMatrix; const BoxScale: TAffineVector; intersectPoint: PAffineVector = nil;
  normal: PAffineVector = nil; depth: PSingle = nil): Boolean;

{ : Compute intersection between a ray and a plane.<p>
  Returns True if an intersection was found, the intersection point is placed
  in intersectPoint is the reference is not nil. }
function RayCastPlaneIntersect(const rayStart, rayVector: TVector;
  const planePoint, planeNormal: TVector; intersectPoint: PVector = nil): Boolean; overload;
function RayCastPlaneXZIntersect(const rayStart, rayVector: TVector; const planeY: single;
  intersectPoint: PVector = nil): Boolean; overload;

{ : Compute intersection between a ray and a triangle. }
function RayCastTriangleIntersect(const rayStart, rayVector: TVector;
  const p1, p2, p3: TAffineVector; intersectPoint: PVector = nil; intersectNormal: PVector = nil)
  : Boolean; overload;
{ : Compute the min distance a ray will pass to a point.<p> }
function RayCastMinDistToPoint(const rayStart, rayVector: TVector; const point: TVector): single;
{ : Determines if a ray will intersect with a given sphere.<p> }
function RayCastIntersectsSphere(const rayStart, rayVector: TVector; const sphereCenter: TVector;
  const SphereRadius: single): Boolean; overload;
{ : Calculates the intersections between a sphere and a ray.<p>
  Returns 0 if no intersection is found (i1 and i2 untouched), 1 if one
  intersection was found (i1 defined, i2 untouched), and 2 is two intersections
  were found (i1 and i2 defined). }
function RayCastSphereIntersect(const rayStart, rayVector: TVector; const sphereCenter: TVector;
  const SphereRadius: single; var i1, i2: TVector): Integer; overload;
{ : Compute intersection between a ray and a box.<p>
  Returns True if an intersection was found, the intersection point is
  placed in intersectPoint if the reference is not nil. }
function RayCastBoxIntersect(const rayStart, rayVector, aMinExtent, aMaxExtent: TAffineVector;
  intersectPoint: PAffineVector = nil): Boolean;

// Some 2d intersection functions.

{ : Determine if 2 rectanges intersect. }
function RectanglesIntersect(const ACenterOfRect1, ACenterOfRect2, ASizeOfRect1,
  ASizeOfRect2: TVector2f): Boolean;

{ : Determine if BigRect completely contains SmallRect. }
function RectangleContains(const ACenterOfBigRect1, ACenterOfSmallRect2, ASizeOfBigRect1,
  ASizeOfSmallRect2: TVector2f; const AEps: single = 0.0): Boolean;

{ : Computes the visible radius of a sphere in a perspective projection.<p>
  This radius can be used for occlusion culling (cone extrusion) or 2D
  intersection testing. }
function SphereVisibleRadius(distance, radius: single): single;

{ : Extracts a TFrustum for combined modelview and projection matrices. }
function ExtractFrustumFromModelViewProjection(const modelViewProj: TMatrix): TFrustum;

// : Determines if volume is clipped or not
function IsVolumeClipped(const objPos: TAffineVector; const objRadius: single;
  const Frustum: TFrustum): Boolean; overload;
function IsVolumeClipped(const objPos: TVector; const objRadius: single; const Frustum: TFrustum)
  : Boolean; overload;
function IsVolumeClipped(const min, max: TAffineVector; const Frustum: TFrustum): Boolean; overload;

// misc funcs

{ : Creates a parallel projection matrix.<p>
  Transformed points will projected on the plane along the specified direction. }
function MakeParallelProjectionMatrix(const plane: THmgPlane; const dir: TVector): TMatrix;

{ : Creates a shadow projection matrix.<p>
  Shadows will be projected onto the plane defined by planePoint and planeNormal,
  from lightPos. }
function MakeShadowMatrix(const planePoint, planeNormal, lightPos: TVector): TMatrix;

{ : Builds a reflection matrix for the given plane.<p>
  Reflection matrix allow implementing planar reflectors in OpenGL (mirrors). }
function MakeReflectionMatrix(const planePoint, planeNormal: TAffineVector): TMatrix;

{ : Packs an homogeneous rotation matrix to 6 bytes.<p>
  The 6:64 (or 6:36) compression ratio is achieved by computing the quaternion
  associated to the matrix and storing its Imaginary components at 16 bits
  precision each.<br>
  Deviation is typically below 0.01% and around 0.1% in worst case situations.<p>
  Note: quaternion conversion is faster and more robust than an angle decomposition. }
function PackRotationMatrix(const mat: TMatrix): TPackedRotationMatrix;
{ : Restores a packed rotation matrix.<p>
  See PackRotationMatrix. }
function UnPackRotationMatrix(const packedMatrix: TPackedRotationMatrix): TMatrix;

{ : Calculates the barycentric coordinates for the point p on the triangle
  defined by the vertices v1, v2 and v3. That is, solves
  p = u * v1 + v * v2 + (1-u-v) * v3
  for u,v.
  Returns true if the point is inside the triangle, false otherwise.<p>
  NOTE: This function assumes that the point lies on the plane defined by the triangle.
  If this is not the case, the function will not work correctly! }
function BarycentricCoordinates(const V1, V2, V3, p: TAffineVector; var u, V: single): Boolean;

{ : Calculates angles for the Camera.MoveAroundTarget(pitch, turn) procedure.
  Initially from then GLCameraColtroller unit, requires AOriginalUpVector to contain only -1, 0 or 1.
  Result contains pitch and turn angles. }
function GetSafeTurnAngle(const AOriginalPosition, AOriginalUpVector, ATargetPosition,
  AMoveAroundTargetCenter: TVector): TVector2f; overload;
function GetSafeTurnAngle(const AOriginalPosition, AOriginalUpVector, ATargetPosition,
  AMoveAroundTargetCenter: TAffineVector): TVector2f; overload;

{ : Extracted from Camera.MoveAroundTarget(pitch, turn). }
function MoveObjectAround(const AMovingObjectPosition, AMovingObjectUp, ATargetPosition: TVector;
  pitchDelta, turnDelta: single): TVector;

{ : Calcualtes Angle between 2 Vectors: (A-CenterPoint) and (B-CenterPoint). In radians. }
function AngleBetweenVectors(const a, b, ACenterPoint: TVector): single; overload;
function AngleBetweenVectors(const a, b, ACenterPoint: TAffineVector): single; overload;

{ : AOriginalPosition - Object initial position.
  ACenter - some point, from which is should be distanced.

  ADistance + AFromCenterSpot - distance, which object should keep from ACenter
  or
  ADistance + not AFromCenterSpot - distance, which object should shift from his current position away from center.
}
function ShiftObjectFromCenter(const AOriginalPosition: TVector; const ACenter: TVector;
  const ADistance: single; const AFromCenterSpot: Boolean): TVector; overload;
function ShiftObjectFromCenter(const AOriginalPosition: TAffineVector; const ACenter: TAffineVector;
  const ADistance: single; const AFromCenterSpot: Boolean): TAffineVector; overload;

const
  cPI: single       = 3.141592654;
  cPIdiv180: single = 0.017453292;
  c180divPI: single = 57.29577951;
  c2PI: single      = 6.283185307;
  cPIdiv2: single   = 1.570796326;
  cPIdiv4: single   = 0.785398163;
  c3PIdiv2: single  = 4.71238898;
  c3PIdiv4: single  = 2.35619449;
  cInv2PI: single   = 1 / 6.283185307;
  cInv360: single   = 1 / 360;
  c180: single      = 180;
  c360: single      = 360;
  cOneHalf: single  = 0.5;
  cLn10: single     = 2.302585093;

  // Ranges of the IEEE floating point types, including denormals
  // with Math.pas compatible name
  MinSingle   = 1.5E-45;
  MaxSingle   = 3.4E+38;
  MinDouble   = 5.0E-324;
  MaxDouble   = 1.7E+308;
  MinExtended = 3.4E-4932;
  MaxExtended = 1.1E+4932;
  MinComp     = -9.223372036854775807E+18;
  MaxComp     = 9.223372036854775807E+18;

var
  // this var is adjusted during "initialization", current values are
  // + 0 : use standard optimized FPU code
  // + 1 : use 3DNow! optimized code (requires K6-2/3 CPU)
  // + 2 : use Intel SSE code (Pentium III, NOT IMPLEMENTED YET !)
  vSIMD: byte = 0;

  // --------------------------------------------------------------
  // --------------------------------------------------------------
  // --------------------------------------------------------------
implementation

// --------------------------------------------------------------
// --------------------------------------------------------------
// --------------------------------------------------------------

uses Math;

const
{$IFNDEF GEOMETRY_NO_ASM}
  // FPU status flags (high order byte)
  cwChop: Word = $1F3F;
{$ENDIF}
  // to be used as descriptive indices
  X = 0;
  Y = 1;
  Z = 2;
  W = 3;

  cZero: single       = 0.0;
  cOne: single        = 1.0;
  cOneDotFive: single = 0.5;

  // OptimizationMode
  //
function GeometryOptimizationMode: String;
begin
  case vSIMD of
    0:
      result := 'FPU';
    1:
      result := '3DNow!';
    2:
      result := 'SSE';
  else
    result := '*ERR*';
  end;
end;

// BeginFPUOnlySection
//
var
  vOldSIMD              : byte;
  vFPUOnlySectionCounter: Integer;

procedure BeginFPUOnlySection;
begin
  if vFPUOnlySectionCounter = 0 then
    vOldSIMD := vSIMD;
  Inc(vFPUOnlySectionCounter);
  vSIMD := 0;
end;

// EndFPUOnlySection
//
procedure EndFPUOnlySection;
begin
  Dec(vFPUOnlySectionCounter);
  Assert(vFPUOnlySectionCounter >= 0);
  if vFPUOnlySectionCounter = 0 then
    vSIMD := vOldSIMD;
end;

// ------------------------------------------------------------------------------
// ----------------- vector functions -------------------------------------------
// ------------------------------------------------------------------------------

// TexPointMake
//
function TexPointMake(const S, T: single): TTexPoint;
begin
  result.S := S;
  result.T := T;
end;

// AffineVectorMake
//
function AffineVectorMake(const X, Y, Z: single): TAffineVector; overload;
begin
  result.V[0] := X;
  result.V[1] := Y;
  result.V[2] := Z;
end;

// AffineVectorMake
//
function AffineVectorMake(const V: TVector): TAffineVector;
begin
  result.V[0] := V.V[0];
  result.V[1] := V.V[1];
  result.V[2] := V.V[2];
end;

// SetAffineVector
//
procedure SetAffineVector(out V: TAffineVector; const X, Y, Z: single); overload;
begin
  V.V[0] := X;
  V.V[1] := Y;
  V.V[2] := Z;
end;

// SetVector (affine)
//
procedure SetVector(out V: TAffineVector; const X, Y, Z: single);
begin
  V.V[0] := X;
  V.V[1] := Y;
  V.V[2] := Z;
end;

// SetVector (affine-hmg)
//
procedure SetVector(out V: TAffineVector; const vSrc: TVector);
begin
  V.V[0] := vSrc.V[0];
  V.V[1] := vSrc.V[1];
  V.V[2] := vSrc.V[2];
end;

// SetVector (affine-affine)
//
procedure SetVector(out V: TAffineVector; const vSrc: TAffineVector);
begin
  V.V[0] := vSrc.V[0];
  V.V[1] := vSrc.V[1];
  V.V[2] := vSrc.V[2];
end;

// SetVector (affine double - affine single)
//
procedure SetVector(out V: TAffineDblVector; const vSrc: TAffineVector);
begin
  V.V[0] := vSrc.V[0];
  V.V[1] := vSrc.V[1];
  V.V[2] := vSrc.V[2];
end;

// SetVector (affine double - hmg single)
//
procedure SetVector(out V: TAffineDblVector; const vSrc: TVector);
begin
  V.V[0] := vSrc.V[0];
  V.V[1] := vSrc.V[1];
  V.V[2] := vSrc.V[2];
end;

// VectorMake
//
function VectorMake(const V: TAffineVector; W: single = 0): TVector;
begin
  result.V[0] := V.V[0];
  result.V[1] := V.V[1];
  result.V[2] := V.V[2];
  result.V[3] := W;
end;

// VectorMake
//
function VectorMake(const X, Y, Z: single; W: single = 0): TVector;
begin
  result.V[0] := X;
  result.V[1] := Y;
  result.V[2] := Z;
  result.V[3] := W;
end;

// PointMake (xyz)
//
function PointMake(const X, Y, Z: single): TVector; overload;
begin
  result.V[0] := X;
  result.V[1] := Y;
  result.V[2] := Z;
  result.V[3] := 1;
end;

// PointMake (affine)
//
function PointMake(const V: TAffineVector): TVector; overload;
begin
  result.V[0] := V.V[0];
  result.V[1] := V.V[1];
  result.V[2] := V.V[2];
  result.V[3] := 1;
end;

// PointMake (hmg)
//
function PointMake(const V: TVector): TVector; overload;
begin
  result.V[0] := V.V[0];
  result.V[1] := V.V[1];
  result.V[2] := V.V[2];
  result.V[3] := 1;
end;

// SetVector
//
procedure SetVector(out V: TVector; const X, Y, Z: single; W: single = 0);
begin
  V.V[0] := X;
  V.V[1] := Y;
  V.V[2] := Z;
  V.V[3] := W;
end;

// SetVector
//
procedure SetVector(out V: TVector; const av: TAffineVector; W: single = 0);
begin
  V.V[0] := av.V[0];
  V.V[1] := av.V[1];
  V.V[2] := av.V[2];
  V.V[3] := W;
end;

// SetVector
//
procedure SetVector(out V: TVector; const vSrc: TVector);
begin
  // faster than memcpy, move or ':=' on the TVector...
  V.V[0] := vSrc.V[0];
  V.V[1] := vSrc.V[1];
  V.V[2] := vSrc.V[2];
  V.V[3] := vSrc.V[3];
end;

// MakePoint
//
procedure MakePoint(out V: TVector; const X, Y, Z: single);
begin
  V.V[0] := X;
  V.V[1] := Y;
  V.V[2] := Z;
  V.V[3] := 1.0;
end;

// MakePoint
//
procedure MakePoint(out V: TVector; const av: TAffineVector);
begin
  V.V[0] := av.V[0];
  V.V[1] := av.V[1];
  V.V[2] := av.V[2];
  V.V[3] := 1.0; // cOne
end;

// MakePoint
//
procedure MakePoint(out V: TVector; const av: TVector);
begin
  V.V[0] := av.V[0];
  V.V[1] := av.V[1];
  V.V[2] := av.V[2];
  V.V[3] := 1.0; // cOne
end;

// MakeVector
//
procedure MakeVector(out V: TAffineVector; const X, Y, Z: single); overload;
begin
  V.V[0] := X;
  V.V[1] := Y;
  V.V[2] := Z;
end;

// MakeVector
//
procedure MakeVector(out V: TVector; const X, Y, Z: single);
begin
  V.V[0] := X;
  V.V[1] := Y;
  V.V[2] := Z;
  V.V[3] := 0.0 // cZero;
end;

// MakeVector
//
procedure MakeVector(out V: TVector; const av: TAffineVector);
begin
  V.V[0] := av.V[0];
  V.V[1] := av.V[1];
  V.V[2] := av.V[2];
  V.V[3] := 0.0 // cZero;
end;

// MakeVector
//
procedure MakeVector(out V: TVector; const av: TVector);
begin
  V.V[0] := av.V[0];
  V.V[1] := av.V[1];
  V.V[2] := av.V[2];
  V.V[3] := 0.0; // cZero;
end;

// RstVector (affine)
//
procedure RstVector(var V: TAffineVector);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  xor   edx, edx
  mov   [eax], edx
  mov   [eax+4], edx
  mov   [eax+8], edx
  {$ELSE}
begin
  V.V[0] := 0;
  V.V[1] := 0;
  V.V[2] := 0;
{$ENDIF}
end;

// RstVector (hmg)
//
procedure RstVector(var V: TVector);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  xor   edx, edx
  mov   [eax], edx
  mov   [eax+4], edx
  mov   [eax+8], edx
  mov   [eax+12], edx
  {$ELSE}
begin
  V.V[0] := 0;
  V.V[1] := 0;
  V.V[2] := 0;
  V.V[3] := 0;
{$ENDIF}
end;

// VectorAdd (func)
//
function VectorAdd(const V1, V2: TVector2f): TVector2f;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FADD DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FADD DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  {$ELSE}
begin
  result.V[0] := V1.V[0] + V2.V[0];
  result.V[1] := V1.V[1] + V2.V[1];
{$ENDIF}
end;

// VectorAdd (func, affine)
//
function VectorAdd(const V1, V2: TAffineVector): TAffineVector;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FADD DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FADD DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  FLD  DWORD PTR [EAX+8]
  FADD DWORD PTR [EDX+8]
  FSTP DWORD PTR [ECX+8]
  {$ELSE}
begin
  result.V[0] := V1.V[0] + V2.V[0];
  result.V[1] := V1.V[1] + V2.V[1];
  result.V[2] := V1.V[2] + V2.V[2];
{$ENDIF}
end;

// VectorAdd (proc, affine)
//
procedure VectorAdd(const V1, V2: TAffineVector; var vr: TAffineVector); overload;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FADD DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FADD DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  FLD  DWORD PTR [EAX+8]
  FADD DWORD PTR [EDX+8]
  FSTP DWORD PTR [ECX+8]
  {$ELSE}
begin
  vr.V[0] := V1.V[0] + V2.V[0];
  vr.V[1] := V1.V[1] + V2.V[1];
  vr.V[2] := V1.V[2] + V2.V[2];
{$ENDIF}
end;

// VectorAdd (proc, affine)
//
procedure VectorAdd(const V1, V2: TAffineVector; vr: PAffineVector); overload;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FADD DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FADD DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  FLD  DWORD PTR [EAX+8]
  FADD DWORD PTR [EDX+8]
  FSTP DWORD PTR [ECX+8]
  {$ELSE}
begin
  vr^.V[0] := V1.V[0] + V2.V[0];
  vr^.V[1] := V1.V[1] + V2.V[1];
  vr^.V[2] := V1.V[2] + V2.V[2];
{$ENDIF}
end;

// VectorAdd (hmg)
//
function VectorAdd(const V1, V2: TVector): TVector;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:
  db $0F,$6F,$00           /// movq  mm0, [eax]
  db $0F,$0F,$02,$9E       /// pfadd mm0, [edx]
  db $0F,$7F,$01           /// movq  [ecx], mm0
  db $0F,$6F,$48,$08       /// movq  mm1, [eax+8]
  db $0F,$0F,$4A,$08,$9E   /// pfadd mm1, [edx+8]
  db $0F,$7F,$49,$08       /// movq  [ecx+8], mm1
  db $0F,$0E               /// femms
  ret

@@FPU:
  FLD  DWORD PTR [EAX]
  FADD DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FADD DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  FLD  DWORD PTR [EAX+8]
  FADD DWORD PTR [EDX+8]
  FSTP DWORD PTR [ECX+8]
  FLD  DWORD PTR [EAX+12]
  FADD DWORD PTR [EDX+12]
  FSTP DWORD PTR [ECX+12]
  {$ELSE}
begin
  result.V[0] := V1.V[0] + V2.V[0];
  result.V[1] := V1.V[1] + V2.V[1];
  result.V[2] := V1.V[2] + V2.V[2];
  result.V[3] := V1.V[3] + V2.V[3];
{$ENDIF}
end;

// VectorAdd (hmg, proc)
//
procedure VectorAdd(const V1, V2: TVector; var vr: TVector);
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:
  db $0F,$6F,$00           /// movq  mm0, [eax]
  db $0F,$0F,$02,$9E       /// pfadd mm0, [edx]
  db $0F,$7F,$01           /// movq  [ecx], mm0
  db $0F,$6F,$48,$08       /// movq  mm1, [eax+8]
  db $0F,$0F,$4A,$08,$9E   /// pfadd mm1, [edx+8]
  db $0F,$7F,$49,$08       /// movq  [ecx+8], mm1
  db $0F,$0E               /// femms
  ret

@@FPU:
  FLD  DWORD PTR [EAX]
  FADD DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FADD DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  FLD  DWORD PTR [EAX+8]
  FADD DWORD PTR [EDX+8]
  FSTP DWORD PTR [ECX+8]
  FLD  DWORD PTR [EAX+12]
  FADD DWORD PTR [EDX+12]
  FSTP DWORD PTR [ECX+12]
  {$ELSE}
begin
  vr.V[0] := V1.V[0] + V2.V[0];
  vr.V[1] := V1.V[1] + V2.V[1];
  vr.V[2] := V1.V[2] + V2.V[2];
  vr.V[3] := V1.V[3] + V2.V[3];
{$ENDIF}
end;

// VectorAdd (affine, single)
//
function VectorAdd(const V: TAffineVector; const f: single): TAffineVector;
begin
  result.V[0] := V.V[0] + f;
  result.V[1] := V.V[1] + f;
  result.V[2] := V.V[2] + f;
end;

// VectorAdd (hmg, single)
//
function VectorAdd(const V: TVector; const f: single): TVector;
begin
  result.V[0] := V.V[0] + f;
  result.V[1] := V.V[1] + f;
  result.V[2] := V.V[2] + f;
  result.V[3] := V.V[3] + f;
end;

// PointAdd (hmg, W = 1)
//
function PointAdd(var V1: TVector; const V2: TVector): TVector;
begin
  result.V[0] := V1.V[0] + V2.V[0];
  result.V[1] := V1.V[1] + V2.V[1];
  result.V[2] := V1.V[2] + V2.V[2];
  result.V[3] := 1;
end;

// AddVector (affine)
//
procedure AddVector(var V1: TAffineVector; const V2: TAffineVector);
// EAX contains address of V1
// EDX contains address of V2
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FADD DWORD PTR [EDX]
  FSTP DWORD PTR [EAX]
  FLD  DWORD PTR [EAX+4]
  FADD DWORD PTR [EDX+4]
  FSTP DWORD PTR [EAX+4]
  FLD  DWORD PTR [EAX+8]
  FADD DWORD PTR [EDX+8]
  FSTP DWORD PTR [EAX+8]
  {$ELSE}
begin
  V1.V[0] := V1.V[0] + V2.V[0];
  V1.V[1] := V1.V[1] + V2.V[1];
  V1.V[2] := V1.V[2] + V2.V[2];
{$ENDIF}
end;

// AddVector (affine)
//
procedure AddVector(var V1: TAffineVector; const V2: TVector);
// EAX contains address of V1
// EDX contains address of V2
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FADD DWORD PTR [EDX]
  FSTP DWORD PTR [EAX]
  FLD  DWORD PTR [EAX+4]
  FADD DWORD PTR [EDX+4]
  FSTP DWORD PTR [EAX+4]
  FLD  DWORD PTR [EAX+8]
  FADD DWORD PTR [EDX+8]
  FSTP DWORD PTR [EAX+8]
  {$ELSE}
begin
  V1.V[0] := V1.V[0] + V2.V[0];
  V1.V[1] := V1.V[1] + V2.V[1];
  V1.V[2] := V1.V[2] + V2.V[2];
{$ENDIF}
end;

// AddVector (hmg)
//
procedure AddVector(var V1: TVector; const V2: TVector);
// EAX contains address of V1
// EDX contains address of V2
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:
  db $0F,$6F,$00           /// MOVQ  MM0, [EAX]
  db $0F,$0F,$02,$9E       /// PFADD MM0, [EDX]
  db $0F,$7F,$00           /// MOVQ  [EAX], MM0
  db $0F,$6F,$48,$08       /// MOVQ  MM1, [EAX+8]
  db $0F,$0F,$4A,$08,$9E   /// PFADD MM1, [EDX+8]
  db $0F,$7F,$48,$08       /// MOVQ  [EAX+8], MM1
  db $0F,$0E               /// FEMMS
  ret
@@FPU:
  FLD  DWORD PTR [EAX]
  FADD DWORD PTR [EDX]
  FSTP DWORD PTR [EAX]
  FLD  DWORD PTR [EAX+4]
  FADD DWORD PTR [EDX+4]
  FSTP DWORD PTR [EAX+4]
  FLD  DWORD PTR [EAX+8]
  FADD DWORD PTR [EDX+8]
  FSTP DWORD PTR [EAX+8]
  FLD  DWORD PTR [EAX+12]
  FADD DWORD PTR [EDX+12]
  FSTP DWORD PTR [EAX+12]
  {$ELSE}
begin
  V1.V[0] := V1.V[0] + V2.V[0];
  V1.V[1] := V1.V[1] + V2.V[1];
  V1.V[2] := V1.V[2] + V2.V[2];
  V1.V[3] := V1.V[3] + V2.V[3];
{$ENDIF}
end;

// AddVector (affine)
//
procedure AddVector(var V: TAffineVector; const f: single);
begin
  V.V[0] := V.V[0] + f;
  V.V[1] := V.V[1] + f;
  V.V[2] := V.V[2] + f;
end;

// AddVector (hmg)
//
procedure AddVector(var V: TVector; const f: single);
begin
  V.V[0] := V.V[0] + f;
  V.V[1] := V.V[1] + f;
  V.V[2] := V.V[2] + f;
  V.V[3] := V.V[3] + f;
end;

// AddPoint (hmg, W = 1)
//
procedure AddPoint(var V1: TVector; const V2: TVector);
begin
  V1.V[0] := V1.V[0] + V2.V[0];
  V1.V[1] := V1.V[1] + V2.V[1];
  V1.V[2] := V1.V[2] + V2.V[2];
  V1.V[3] := 1;
end;

// TexPointArrayAdd
//
procedure TexPointArrayAdd(const src: PTexPointArray; const delta: TTexPoint; const nb: Integer;
  dest: PTexPointArray); overload;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  or    ecx, ecx
  jz    @@End

  test  vSIMD, 1
  jnz   @@3DNow

  push edi
  mov   edi, dest

@@FPULoop:
  fld   dword ptr [eax]
  fadd  dword ptr [edx]
  fstp  dword ptr [edi]
  fld   dword ptr [eax+4]
  fadd  dword ptr [edx+4]
  fstp  dword ptr [edi+4]

  add   eax, 8
  add   edi, 8
  dec   ecx
  jnz   @@FPULoop

  pop edi
  jmp   @@End

@@3DNow:
  db $0F,$6F,$02           /// movq  mm0, [edx]
  mov   edx, dest

@@3DNowLoop:
  db $0F,$6F,$10           /// movq  mm2, [eax]
  db $0F,$0F,$D0,$9E       /// pfadd mm2, mm0
  db $0F,$7F,$12           /// movq  [edx], mm2

  add   eax, 8
  add   edx, 8
  dec   ecx
  jnz   @@3DNowLoop

  db $0F,$0E               /// femms

@@End:
  {$ELSE}
var
  i: Integer;
begin
  for i := 0 to nb - 1 do
  begin
    dest^[i].S := src^[i].S + delta.S;
    dest^[i].T := src^[i].T + delta.T;
  end;
{$ENDIF}
end;

// TexPointArrayScaleAndAdd
//
procedure TexPointArrayScaleAndAdd(const src: PTexPointArray; const delta: TTexPoint;
  const nb: Integer; const scale: TTexPoint; dest: PTexPointArray); overload;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  or    ecx, ecx
  jz    @@End

  test  vSIMD, 1
  jnz   @@3DNow

  push  edi
  push  esi
  mov   edi, dest
  mov   esi, scale

@@FPULoop:
  fld   dword ptr [eax]
  fmul  dword ptr [esi]
  fadd  dword ptr [edx]
  fstp  dword ptr [edi]
  fld   dword ptr [eax+4]
  fmul  dword ptr [esi+4]
  fadd  dword ptr [edx+4]
  fstp  dword ptr [edi+4]

  add   eax, 8
  add   edi, 8
  dec   ecx
  jnz   @@FPULoop

  pop   esi
  pop   edi
  jmp   @@End

@@3DNow:
  db $0F,$6F,$02           /// movq  mm0, [edx]
  mov   edx, scale
  db $0F,$6F,$0A           /// movq  mm1, [edx]
  mov   edx, dest

@@3DNowLoop:
  db $0F,$6F,$10           /// movq  mm2, [eax]
  db $0F,$0F,$D1,$B4       /// pfmul mm2, mm1
  db $0F,$0F,$D0,$9E       /// pfadd mm2, mm0
  db $0F,$7F,$12           /// movq  [edx], mm2

  add   eax, 8
  add   edx, 8
  dec   ecx
  jnz   @@3DNowLoop

  db $0F,$0E               /// femms
@@End:
  {$ELSE}
var
  i: Integer;
begin
  for i := 0 to nb - 1 do
  begin
    dest^[i].S := src^[i].S * scale.S + delta.S;
    dest^[i].T := src^[i].T * scale.T + delta.T;
  end;
{$ENDIF}
end;

// VectorArrayAdd
//
procedure VectorArrayAdd(const src: PAffineVectorArray; const delta: TAffineVector;
  const nb: Integer; dest: PAffineVectorArray);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  or    ecx, ecx
  jz    @@End

  test  vSIMD, 1
  jnz   @@3DNow

  push edi
  mov   edi, dest

@@FPULoop:
  fld   dword ptr [eax]
  fadd  dword ptr [edx]
  fstp  dword ptr [edi]
  fld   dword ptr [eax+4]
  fadd  dword ptr [edx+4]
  fstp  dword ptr [edi+4]
  fld   dword ptr [eax+8]
  fadd  dword ptr [edx+8]
  fstp  dword ptr [edi+8]

  add   eax, 12
  add   edi, 12
  dec   ecx
  jnz   @@FPULoop

  pop edi
  jmp   @@End

@@3DNow:
  db $0F,$6F,$02           /// movq  mm0, [edx]
  db $0F,$6E,$4A,$08       /// movd  mm1, [edx+8]
  mov   edx, dest

@@3DNowLoop:
  db $0F,$6F,$10           /// movq  mm2, [eax]
  db $0F,$6E,$58,$08       /// movd  mm3, [eax+8]
  db $0F,$0F,$D0,$9E       /// pfadd mm2, mm0
  db $0F,$0F,$D9,$9E       /// pfadd mm3, mm1
  db $0F,$7F,$12           /// movq  [edx], mm2
  db $0F,$7E,$5A,$08       /// movd  [edx+8], mm3

  add   eax, 12
  add   edx, 12
  dec   ecx
  jnz   @@3DNowLoop

  db $0F,$0E               /// femms

@@End:
  {$ELSE}
var
  i: Integer;
begin
  for i := 0 to nb - 1 do
  begin
    dest^[i].V[0] := src^[i].V[0] + delta.V[0];
    dest^[i].V[1] := src^[i].V[1] + delta.V[1];
    dest^[i].V[2] := src^[i].V[2] + delta.V[2];
  end;
{$ENDIF}
end;

// VectorSubtract (func, affine)
//
function VectorSubtract(const V1, V2: TAffineVector): TAffineVector;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FSTP DWORD PTR [ECX+8]
  {$ELSE}
begin
  result.V[0] := V1.V[0] - V2.V[0];
  result.V[1] := V1.V[1] - V2.V[1];
  result.V[2] := V1.V[2] - V2.V[2];
{$ENDIF}
end;

// VectorSubtract (func, 2f)
//
function VectorSubtract(const V1, V2: TVector2f): TVector2f;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  {$ELSE}
begin
  result.V[0] := V1.V[0] - V2.V[0];
  result.V[1] := V1.V[1] - V2.V[1];
{$ENDIF}
end;

// VectorSubtract (proc, affine)
//
procedure VectorSubtract(const V1, V2: TAffineVector; var result: TAffineVector);
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FSTP DWORD PTR [ECX+8]
  {$ELSE}
begin
  result.V[0] := V1.V[0] - V2.V[0];
  result.V[1] := V1.V[1] - V2.V[1];
  result.V[2] := V1.V[2] - V2.V[2];
{$ENDIF}
end;

// VectorSubtract (proc, affine-hmg)
//
procedure VectorSubtract(const V1, V2: TAffineVector; var result: TVector);
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FSTP DWORD PTR [ECX+8]
  xor   eax, eax
  mov   [ECX+12], eax
  {$ELSE}
begin
  result.V[0] := V1.V[0] - V2.V[0];
  result.V[1] := V1.V[1] - V2.V[1];
  result.V[2] := V1.V[2] - V2.V[2];
  result.V[3] := 0;
{$ENDIF}
end;

// VectorSubtract
//
procedure VectorSubtract(const V1: TVector; V2: TAffineVector; var result: TVector);
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FSTP DWORD PTR [ECX+8]
  mov   edx, [eax+12]
  mov   [ECX+12], edx
  {$ELSE}
begin
  result.V[0] := V1.V[0] - V2.V[0];
  result.V[1] := V1.V[1] - V2.V[1];
  result.V[2] := V1.V[2] - V2.V[2];
  result.V[3] := V1.V[0];
{$ENDIF}
end;

// VectorSubtract (hmg)
//
function VectorSubtract(const V1, V2: TVector): TVector;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:
  db $0F,$6F,$00           /// MOVQ  MM0, [EAX]
  db $0F,$0F,$02,$9A       /// PFSUB MM0, [EDX]
  db $0F,$7F,$01           /// MOVQ  [ECX], MM0
  db $0F,$6F,$48,$08       /// MOVQ  MM1, [EAX+8]
  db $0F,$0F,$4A,$08,$9A   /// PFSUB MM1, [EDX+8]
  db $0F,$7F,$49,$08       /// MOVQ  [ECX+8], MM1
  db $0F,$0E               /// FEMMS
  ret
@@FPU:
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FSTP DWORD PTR [ECX+8]
  FLD  DWORD PTR [EAX+12]
  FSUB DWORD PTR [EDX+12]
  FSTP DWORD PTR [ECX+12]
  {$ELSE}
begin
  result.V[0] := V1.V[0] - V2.V[0];
  result.V[1] := V1.V[1] - V2.V[1];
  result.V[2] := V1.V[2] - V2.V[2];
  result.V[3] := V1.V[3] - V2.V[3];
{$ENDIF}
end;

// VectorSubtract (proc, hmg)
//
procedure VectorSubtract(const V1, V2: TVector; var result: TVector);
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:
  db $0F,$6F,$00           /// MOVQ  MM0, [EAX]
  db $0F,$0F,$02,$9A       /// PFSUB MM0, [EDX]
  db $0F,$7F,$01           /// MOVQ  [ECX], MM0
  db $0F,$6F,$48,$08       /// MOVQ  MM1, [EAX+8]
  db $0F,$0F,$4A,$08,$9A   /// PFSUB MM1, [EDX+8]
  db $0F,$7F,$49,$08       /// MOVQ  [ECX+8], MM1
  db $0F,$0E               /// FEMMS
  ret
@@FPU:
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FSTP DWORD PTR [ECX+8]
  FLD  DWORD PTR [EAX+12]
  FSUB DWORD PTR [EDX+12]
  FSTP DWORD PTR [ECX+12]
  {$ELSE}
begin
  result.V[0] := V1.V[0] - V2.V[0];
  result.V[1] := V1.V[1] - V2.V[1];
  result.V[2] := V1.V[2] - V2.V[2];
  result.V[3] := V1.V[3] - V2.V[3];
{$ENDIF}
end;

// VectorSubtract (proc, affine)
//
procedure VectorSubtract(const V1, V2: TVector; var result: TAffineVector); overload;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FSTP DWORD PTR [ECX]
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FSTP DWORD PTR [ECX+4]
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FSTP DWORD PTR [ECX+8]
  {$ELSE}
begin
  result.V[0] := V1.V[0] - V2.V[0];
  result.V[1] := V1.V[1] - V2.V[1];
  result.V[2] := V1.V[2] - V2.V[2];
{$ENDIF}
end;

// VectorSubtract (affine, single)
//
function VectorSubtract(const V1: TAffineVector; delta: single): TAffineVector;
begin
  result.V[0] := V1.V[0] - delta;
  result.V[1] := V1.V[1] - delta;
  result.V[2] := V1.V[2] - delta;
end;

// VectorSubtract (hmg, single)
//
function VectorSubtract(const V1: TVector; delta: single): TVector;
begin
  result.V[0] := V1.V[0] - delta;
  result.V[1] := V1.V[1] - delta;
  result.V[2] := V1.V[2] - delta;
  result.V[3] := V1.V[3] - delta;
end;

// SubtractVector (affine)
//
procedure SubtractVector(var V1: TAffineVector; const V2: TAffineVector);
// EAX contains address of V1
// EDX contains address of V2
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FSTP DWORD PTR [EAX]
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FSTP DWORD PTR [EAX+4]
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FSTP DWORD PTR [EAX+8]
  {$ELSE}
begin
  V1.V[0] := V1.V[0] - V2.V[0];
  V1.V[1] := V1.V[1] - V2.V[1];
  V1.V[2] := V1.V[2] - V2.V[2];
{$ENDIF}
end;

// SubtractVector (2f)
//
procedure SubtractVector(var V1: TVector2f; const V2: TVector2f);
// EAX contains address of V1
// EDX contains address of V2
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FSTP DWORD PTR [EAX]
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FSTP DWORD PTR [EAX+4]
  {$ELSE}
begin
  V1.V[0] := V1.V[0] - V2.V[0];
  V1.V[1] := V1.V[1] - V2.V[1];
{$ENDIF}
end;

// SubtractVector (hmg)
//
procedure SubtractVector(var V1: TVector; const V2: TVector);
// EAX contains address of V1
// EDX contains address of V2
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:
  db $0F,$6F,$00           /// MOVQ  MM0, [EAX]
  db $0F,$0F,$02,$9A       /// PFSUB MM0, [EDX]
  db $0F,$7F,$00           /// MOVQ  [EAX], MM0
  db $0F,$6F,$48,$08       /// MOVQ  MM1, [EAX+8]
  db $0F,$0F,$4A,$08,$9A   /// PFSUB MM1, [EDX+8]
  db $0F,$7F,$48,$08       /// MOVQ  [EAX+8], MM1
  db $0F,$0E               /// FEMMS
  ret
@@FPU:
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FSTP DWORD PTR [EAX]
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FSTP DWORD PTR [EAX+4]
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FSTP DWORD PTR [EAX+8]
  FLD  DWORD PTR [EAX+12]
  FSUB DWORD PTR [EDX+12]
  FSTP DWORD PTR [EAX+12]
  {$ELSE}
begin
  V1.V[0] := V1.V[0] - V2.V[0];
  V1.V[1] := V1.V[1] - V2.V[1];
  V1.V[2] := V1.V[2] - V2.V[2];
  V1.V[3] := V1.V[3] - V2.V[3];
{$ENDIF}
end;

// CombineVector (var)
//
procedure CombineVector(var vr: TAffineVector; const V: TAffineVector; var f: single);
// EAX contains address of vr
// EDX contains address of v
// ECX contains address of f
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EDX]
  FMUL DWORD PTR [ECX]
  FADD DWORD PTR [EAX]
  FSTP DWORD PTR [EAX]
  FLD  DWORD PTR [EDX+4]
  FMUL DWORD PTR [ECX]
  FADD DWORD PTR [EAX+4]
  FSTP DWORD PTR [EAX+4]
  FLD  DWORD PTR [EDX+8]
  FMUL DWORD PTR [ECX]
  FADD DWORD PTR [EAX+8]
  FSTP DWORD PTR [EAX+8]
  {$ELSE}
begin
  vr.V[0] := vr.V[0] + V.V[0] * f;
  vr.V[1] := vr.V[1] + V.V[1] * f;
  vr.V[2] := vr.V[2] + V.V[2] * f;
{$ENDIF}
end;

// CombineVector (pointer)
//
procedure CombineVector(var vr: TAffineVector; const V: TAffineVector; pf: PFloat);
// EAX contains address of vr
// EDX contains address of v
// ECX contains address of f
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EDX]
  FMUL DWORD PTR [ECX]
  FADD DWORD PTR [EAX]
  FSTP DWORD PTR [EAX]
  FLD  DWORD PTR [EDX+4]
  FMUL DWORD PTR [ECX]
  FADD DWORD PTR [EAX+4]
  FSTP DWORD PTR [EAX+4]
  FLD  DWORD PTR [EDX+8]
  FMUL DWORD PTR [ECX]
  FADD DWORD PTR [EAX+8]
  FSTP DWORD PTR [EAX+8]
  {$ELSE}
begin
  vr.V[0] := vr.V[0] + V.V[0] * pf^;
  vr.V[1] := vr.V[1] + V.V[1] * pf^;
  vr.V[2] := vr.V[2] + V.V[2] * pf^;
{$ENDIF}
end;

// TexPointCombine
//
function TexPointCombine(const t1, t2: TTexPoint; f1, f2: single): TTexPoint;
begin
  result.S := (f1 * t1.S) + (f2 * t2.S);
  result.T := (f1 * t1.T) + (f2 * t2.T);
end;

// VectorCombine
//
function VectorCombine(const V1, V2: TAffineVector; const f1, f2: single): TAffineVector;
begin
  result.V[X] := (f1 * V1.V[X]) + (f2 * V2.V[X]);
  result.V[Y] := (f1 * V1.V[Y]) + (f2 * V2.V[Y]);
  result.V[Z] := (f1 * V1.V[Z]) + (f2 * V2.V[Z]);
end;

// VectorCombine3 (func)
//
function VectorCombine3(const V1, V2, V3: TAffineVector; const f1, f2, F3: single): TAffineVector;
begin
  result.V[X] := (f1 * V1.V[X]) + (f2 * V2.V[X]) + (F3 * V3.V[X]);
  result.V[Y] := (f1 * V1.V[Y]) + (f2 * V2.V[Y]) + (F3 * V3.V[Y]);
  result.V[Z] := (f1 * V1.V[Z]) + (f2 * V2.V[Z]) + (F3 * V3.V[Z]);
end;

// VectorCombine3 (vector)
//
procedure VectorCombine3(const V1, V2, V3: TAffineVector; const f1, f2, F3: single;
  var vr: TAffineVector);
begin
  vr.V[X] := (f1 * V1.V[X]) + (f2 * V2.V[X]) + (F3 * V3.V[X]);
  vr.V[Y] := (f1 * V1.V[Y]) + (f2 * V2.V[Y]) + (F3 * V3.V[Y]);
  vr.V[Z] := (f1 * V1.V[Z]) + (f2 * V2.V[Z]) + (F3 * V3.V[Z]);
end;

// CombineVector
//
procedure CombineVector(var vr: TVector; const V: TVector; var f: single); overload;
// EAX contains address of vr
// EDX contains address of v
// ECX contains address of f
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:
  db $0F,$6E,$11           /// MOVD  MM2, [ECX]
  db $0F,$62,$D2           /// PUNPCKLDQ MM2, MM2
  db $0F,$6F,$02           /// MOVQ  MM0, [EDX]
  db $0F,$0F,$C2,$B4       /// PFMUL MM0, MM2
  db $0F,$0F,$00,$9E       /// PFADD MM0, [EAX]
  db $0F,$7F,$00           /// MOVQ  [EAX], MM0
  db $0F,$6F,$4A,$08       /// MOVQ  MM1, [EDX+8]
  db $0F,$0F,$CA,$B4       /// PFMUL MM1, MM2
  db $0F,$0F,$48,$08,$9E   /// PFADD MM1, [EAX+8]
  db $0F,$7F,$48,$08       /// MOVQ  [EAX+8], MM1
  db $0F,$0E               /// FEMMS
  ret
@@FPU:
  FLD  DWORD PTR [EDX]
  FMUL DWORD PTR [ECX]
  FADD DWORD PTR [EAX]
  FSTP DWORD PTR [EAX]
  FLD  DWORD PTR [EDX+4]
  FMUL DWORD PTR [ECX]
  FADD DWORD PTR [EAX+4]
  FSTP DWORD PTR [EAX+4]
  FLD  DWORD PTR [EDX+8]
  FMUL DWORD PTR [ECX]
  FADD DWORD PTR [EAX+8]
  FSTP DWORD PTR [EAX+8]
  FLD  DWORD PTR [EDX+12]
  FMUL DWORD PTR [ECX]
  FADD DWORD PTR [EAX+12]
  FSTP DWORD PTR [EAX+12]
  {$ELSE}
begin
  vr.V[0] := vr.V[0] + V.V[0] * f;
  vr.V[1] := vr.V[1] + V.V[1] * f;
  vr.V[2] := vr.V[2] + V.V[2] * f;
  vr.V[3] := vr.V[3] + V.V[3] * f;
{$ENDIF}
end;

// CombineVector
//
procedure CombineVector(var vr: TVector; const V: TAffineVector; var f: single); overload;
// EAX contains address of vr
// EDX contains address of v
// ECX contains address of f
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EDX]
  FMUL DWORD PTR [ECX]
  FADD DWORD PTR [EAX]
  FSTP DWORD PTR [EAX]
  FLD  DWORD PTR [EDX+4]
  FMUL DWORD PTR [ECX]
  FADD DWORD PTR [EAX+4]
  FSTP DWORD PTR [EAX+4]
  FLD  DWORD PTR [EDX+8]
  FMUL DWORD PTR [ECX]
  FADD DWORD PTR [EAX+8]
  FSTP DWORD PTR [EAX+8]
  {$ELSE}
begin
  vr.V[0] := vr.V[0] + V.V[0] * f;
  vr.V[1] := vr.V[1] + V.V[1] * f;
  vr.V[2] := vr.V[2] + V.V[2] * f;
{$ENDIF}
end;

// VectorCombine
//
function VectorCombine(const V1, V2: TVector; const f1, f2: single): TVector;
begin
  result.V[X] := (f1 * V1.V[X]) + (f2 * V2.V[X]);
  result.V[Y] := (f1 * V1.V[Y]) + (f2 * V2.V[Y]);
  result.V[Z] := (f1 * V1.V[Z]) + (f2 * V2.V[Z]);
  result.V[W] := (f1 * V1.V[W]) + (f2 * V2.V[W]);
end;

// VectorCombine
//
function VectorCombine(const V1: TVector; const V2: TAffineVector; const f1, f2: single)
  : TVector; overload;
begin
  result.V[X] := (f1 * V1.V[X]) + (f2 * V2.V[X]);
  result.V[Y] := (f1 * V1.V[Y]) + (f2 * V2.V[Y]);
  result.V[Z] := (f1 * V1.V[Z]) + (f2 * V2.V[Z]);
  result.V[W] := f1 * V1.V[W];
end;

// VectorCombine
//
procedure VectorCombine(const V1, V2: TVector; const f1, f2: single; var vr: TVector); overload;
// EAX contains address of v1
// EDX contains address of v2
// ECX contains address of vr
// ebp+$c points to f1
// ebp+$8 points to f2
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:                   // 246354
  db $0F,$6E,$4D,$0C       /// MOVD  MM1, [EBP+$0C]
  db $0F,$62,$C9           /// PUNPCKLDQ MM1, MM1
  db $0F,$6E,$55,$08       /// MOVD  MM2, [EBP+$08]
  db $0F,$62,$D2           /// PUNPCKLDQ MM2, MM2

  db $0F,$6F,$18           /// MOVQ  MM3, [EAX]
  db $0F,$0F,$D9,$B4       /// PFMUL MM3, MM1
  db $0F,$6F,$22           /// MOVQ  MM4, [EDX]
  db $0F,$0F,$E2,$B4       /// PFMUL MM4, MM2
  db $0F,$0F,$DC,$9E       /// PFADD MM3, MM4
  db $0F,$7F,$19           /// MOVQ  [ECX], MM3

  db $0F,$6F,$68,$08       /// MOVQ  MM5, [EAX+8]
  db $0F,$0F,$E9,$B4       /// PFMUL MM5, MM1
  db $0F,$6F,$72,$08       /// MOVQ  MM6, [EDX+8]
  db $0F,$0F,$F2,$B4       /// PFMUL MM6, MM2
  db $0F,$0F,$EE,$9E       /// PFADD MM5, MM6
  db $0F,$7F,$69,$08       /// MOVQ  [ECX+8], MM5

  db $0F,$0E               /// FEMMS
  pop ebp
  ret $08

@@FPU:      // 327363
  FLD  DWORD PTR [EAX]
  FMUL DWORD PTR [EBP+$0C]
  FLD  DWORD PTR [EDX]
  FMUL DWORD PTR [EBP+$08]
  FADD
  FSTP DWORD PTR [ECX]

  FLD  DWORD PTR [EAX+4]
  FMUL DWORD PTR [EBP+$0C]
  FLD  DWORD PTR [EDX+4]
  FMUL DWORD PTR [EBP+$08]
  FADD
  FSTP DWORD PTR [ECX+4]

  FLD  DWORD PTR [EAX+8]
  FMUL DWORD PTR [EBP+$0C]
  FLD  DWORD PTR [EDX+8]
  FMUL DWORD PTR [EBP+$08]
  FADD
  FSTP DWORD PTR [ECX+8]

  FLD  DWORD PTR [EAX+12]
  FMUL DWORD PTR [EBP+$0C]
  FLD  DWORD PTR [EDX+12]
  FMUL DWORD PTR [EBP+$08]
  FADD
  FSTP DWORD PTR [ECX+12]
  {$ELSE}
begin
  vr.V[0] := (f1 * V1.V[0]) + (f2 * V2.V[0]);
  vr.V[1] := (f1 * V1.V[1]) + (f2 * V2.V[1]);
  vr.V[2] := (f1 * V1.V[2]) + (f2 * V2.V[2]);
  vr.V[3] := (f1 * V1.V[3]) + (f2 * V2.V[3]);
{$ENDIF}
end;

// VectorCombine (F1=1.0)
//
procedure VectorCombine(const V1, V2: TVector; const f2: single; var vr: TVector); overload;
// EAX contains address of v1
// EDX contains address of v2
// ECX contains address of vr
// ebp+$8 points to f2
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:                   // 121559
  db $0F,$6E,$55,$08       /// MOVD  MM2, [EBP+$08]
  db $0F,$62,$D2           /// PUNPCKLDQ MM2, MM2

  db $0F,$6F,$22           /// MOVQ  MM4, [EDX]
  db $0F,$6F,$72,$08       /// MOVQ  MM6, [EDX+8]

  db $0F,$0F,$E2,$B4       /// PFMUL MM4, MM2
  db $0F,$0F,$F2,$B4       /// PFMUL MM6, MM2

  db $0F,$0F,$20,$9E       /// PFADD MM4, [EAX]
  db $0F,$0F,$70,$08,$9E   /// PFADD MM6, [EAX+8]

  db $0F,$7F,$21           /// MOVQ  [ECX], MM4
  db $0F,$7F,$71,$08       /// MOVQ  [ECX+8], MM6

  db $0F,$0E               /// FEMMS
  pop ebp
  ret $04

@@FPU:      // 171379
  FLD  DWORD PTR [EBP+$08]

  FLD  DWORD PTR [EDX]
  FMUL ST, ST(1)
  FADD DWORD PTR [EAX]
  FSTP DWORD PTR [ECX]

  FLD  DWORD PTR [EDX+4]
  FMUL ST, ST(1)
  FADD DWORD PTR [EAX+4]
  FSTP DWORD PTR [ECX+4]

  FLD  DWORD PTR [EDX+8]
  FMUL ST, ST(1)
  FADD DWORD PTR [EAX+8]
  FSTP DWORD PTR [ECX+8]

  FLD  DWORD PTR [EDX+12]
  FMULP
  FADD DWORD PTR [EAX+12]
  FSTP DWORD PTR [ECX+12]
  {$ELSE}
begin // 201283
  vr.V[0] := V1.V[0] + (f2 * V2.V[0]);
  vr.V[1] := V1.V[1] + (f2 * V2.V[1]);
  vr.V[2] := V1.V[2] + (f2 * V2.V[2]);
  vr.V[3] := V1.V[3] + (f2 * V2.V[3]);
{$ENDIF}
end;

// VectorCombine
//
procedure VectorCombine(const V1: TVector; const V2: TAffineVector; const f1, f2: single;
  var vr: TVector);
begin
  vr.V[X] := (f1 * V1.V[X]) + (f2 * V2.V[X]);
  vr.V[Y] := (f1 * V1.V[Y]) + (f2 * V2.V[Y]);
  vr.V[Z] := (f1 * V1.V[Z]) + (f2 * V2.V[Z]);
  vr.V[W] := f1 * V1.V[W];
end;

// VectorCombine3
//
function VectorCombine3(const V1, V2, V3: TVector; const f1, f2, F3: single): TVector;
begin
  result.V[X] := (f1 * V1.V[X]) + (f2 * V2.V[X]) + (F3 * V3.V[X]);
  result.V[Y] := (f1 * V1.V[Y]) + (f2 * V2.V[Y]) + (F3 * V3.V[Y]);
  result.V[Z] := (f1 * V1.V[Z]) + (f2 * V2.V[Z]) + (F3 * V3.V[Z]);
  result.V[W] := (f1 * V1.V[W]) + (f2 * V2.V[W]) + (F3 * V3.V[W]);
end;

// VectorCombine3
//
procedure VectorCombine3(const V1, V2, V3: TVector; const f1, f2, F3: single; var vr: TVector);
// EAX contains address of v1
// EDX contains address of v2
// ECX contains address of v3
// EBX contains address of vr
// ebp+$14 points to f1
// ebp+$10 points to f2
// ebp+$0c points to f3
begin
{$IFNDEF GEOMETRY_NO_ASM}
  asm
    test vSIMD, 1
    jz @@FPU
  @@3DNow:                   // 197
    db $0F,$6E,$4D,$14       /// MOVD  MM1, [EBP+$14]
    db $0F,$62,$C9           /// PUNPCKLDQ MM1, MM1
    db $0F,$6E,$55,$10       /// MOVD  MM2, [EBP+$10]
    db $0F,$62,$D2           /// PUNPCKLDQ MM2, MM2
    db $0F,$6E,$5D,$0C       /// MOVD  MM3, [EBP+$0C]
    db $0F,$62,$DB           /// PUNPCKLDQ MM3, MM3

    db $0F,$6F,$20           /// MOVQ  MM4, [EAX]
    db $0F,$0F,$E1,$B4       /// PFMUL MM4, MM1
    db $0F,$6F,$2A           /// MOVQ  MM5, [EDX]
    db $0F,$0F,$EA,$B4       /// PFMUL MM5, MM2
    db $0F,$0F,$E5,$9E       /// PFADD MM4, MM5
    db $0F,$6F,$31           /// MOVQ  MM6, [ECX]
    db $0F,$0F,$F3,$B4       /// PFMUL MM6, MM3
    db $0F,$0F,$E6,$9E       /// PFADD MM4, MM6
    db $0F,$7F,$23           /// MOVQ  [EBX], MM4

    db $0F,$6F,$78,$08       /// MOVQ  MM7, [EAX+8]
    db $0F,$0F,$F9,$B4       /// PFMUL MM7, MM1
    db $0F,$6F,$42,$08       /// MOVQ  MM0, [EDX+8]
    db $0F,$0F,$C2,$B4       /// PFMUL MM0, MM2
    db $0F,$0F,$F8,$9E       /// PFADD MM7, MM0
    db $0F,$6F,$69,$08       /// MOVQ  MM5, [ECX+8]
    db $0F,$0F,$EB,$B4       /// PFMUL MM5, MM3
    db $0F,$0F,$FD,$9E       /// PFADD MM7, MM5
    db $0F,$7F,$7B,$08       /// MOVQ  [EBX+8], MM7

    db $0F,$0E               /// FEMMS
    pop ebx
    pop ebp
    ret $10
  @@FPU:      // 263
  end;
{$ENDIF}
  vr.V[X] := (f1 * V1.V[X]) + (f2 * V2.V[X]) + (F3 * V3.V[X]);
  vr.V[Y] := (f1 * V1.V[Y]) + (f2 * V2.V[Y]) + (F3 * V3.V[Y]);
  vr.V[Z] := (f1 * V1.V[Z]) + (f2 * V2.V[Z]) + (F3 * V3.V[Z]);
  vr.V[W] := (f1 * V1.V[W]) + (f2 * V2.V[W]) + (F3 * V3.V[W]);
end;

// VectorDotProduct (2f)
//
function VectorDotProduct(const V1, V2: TVector2f): single;
// EAX contains address of V1
// EDX contains address of V2
// result is stored in ST(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [eax]
  FMUL DWORD PTR [edx]
  FLD DWORD PTR [eax+4]
  FMUL DWORD PTR [edx+4]
  faddp
end;
{$ELSE}
begin
  result := V1.V[0] * V2.V[0] + V1.V[1] * V2.V[1];
end;
{$ENDIF}

// VectorDotProduct (affine)
//
function VectorDotProduct(const V1, V2: TAffineVector): single;
// EAX contains address of V1
// EDX contains address of V2
// result is stored in ST(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [eax]
  FMUL DWORD PTR [edx]
  FLD DWORD PTR [eax+4]
  FMUL DWORD PTR [edx+4]
  faddp
  FLD DWORD PTR [eax+8]
  FMUL DWORD PTR [edx+8]
  faddp
end;
{$ELSE}
begin
  result := V1.V[0] * V2.V[0] + V1.V[1] * V2.V[1] + V1.V[2] * V2.V[2];
end;
{$ENDIF}

// VectorDotProduct (hmg)
//
function VectorDotProduct(const V1, V2: TVector): single;
// EAX contains address of V1
// EDX contains address of V2
// result is stored in ST(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [EAX]
  FMUL DWORD PTR [EDX]
  FLD DWORD PTR [EAX + 4]
  FMUL DWORD PTR [EDX + 4]
  FADDP
  FLD DWORD PTR [EAX + 8]
  FMUL DWORD PTR [EDX + 8]
  FADDP
  FLD DWORD PTR [EAX + 12]
  FMUL DWORD PTR [EDX + 12]
  FADDP
  {$ELSE}
begin
  result := V1.V[0] * V2.V[0] + V1.V[1] * V2.V[1] + V1.V[2] * V2.V[2] + V1.V[3] * V2.V[3];
{$ENDIF}
end;

// VectorDotProduct
//
function VectorDotProduct(const V1: TVector; const V2: TAffineVector): single;
// EAX contains address of V1
// EDX contains address of V2
// result is stored in ST(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [EAX]
  FMUL DWORD PTR [EDX]
  FLD DWORD PTR [EAX + 4]
  FMUL DWORD PTR [EDX + 4]
  FADDP
  FLD DWORD PTR [EAX + 8]
  FMUL DWORD PTR [EDX + 8]
  FADDP
  {$ELSE}
begin
  result := V1.V[0] * V2.V[0] + V1.V[1] * V2.V[1] + V1.V[2] * V2.V[2];
{$ENDIF}
end;

// PointProject (affine)
//
function PointProject(const p, origin, direction: TAffineVector): single;
// EAX -> p, EDX -> origin, ECX -> direction
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld   dword ptr [eax]
  fsub  dword ptr [edx]
  fmul  dword ptr [ecx]
  fld   dword ptr [eax+4]
  fsub  dword ptr [edx+4]
  fmul  dword ptr [ecx+4]
  fadd
  fld   dword ptr [eax+8]
  fsub  dword ptr [edx+8]
  fmul  dword ptr [ecx+8]
  fadd
  {$ELSE}
begin
  result := direction.V[0] * (p.V[0] - origin.V[0]) + direction.V[1] * (p.V[1] - origin.V[1]) +
    direction.V[2] * (p.V[2] - origin.V[2]);
{$ENDIF}
end;

// PointProject (vector)
//
function PointProject(const p, origin, direction: TVector): single;
// EAX -> p, EDX -> origin, ECX -> direction
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld   dword ptr [eax]
  fsub  dword ptr [edx]
  fmul  dword ptr [ecx]
  fld   dword ptr [eax+4]
  fsub  dword ptr [edx+4]
  fmul  dword ptr [ecx+4]
  fadd
  fld   dword ptr [eax+8]
  fsub  dword ptr [edx+8]
  fmul  dword ptr [ecx+8]
  fadd
  {$ELSE}
begin
  result := direction.V[0] * (p.V[0] - origin.V[0]) + direction.V[1] * (p.V[1] - origin.V[1]) +
    direction.V[2] * (p.V[2] - origin.V[2]);
{$ENDIF}
end;

// VectorCrossProduct
//
function VectorCrossProduct(const V1, V2: TAffineVector): TAffineVector;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld   dword ptr [eax+$4]
  fmul  dword ptr [edx+$8]
  fld   dword ptr [eax+$8]
  fmul  dword ptr [edx+$4]
  fsubp
  fstp  dword ptr [ecx]

  fld   dword ptr [eax+$8]
  fmul  dword ptr [edx]
  fld   dword ptr [eax]
  fmul  dword ptr [edx+$8]
  fsubp
  fstp  dword ptr [ecx+$4]

  fld   dword ptr [eax]
  fmul  dword ptr [edx+$4]
  fld   dword ptr [eax+$4]
  fmul  dword ptr [edx]
  fsubp
  fstp  dword ptr [ecx+$8]
  {$ELSE}
begin
  result.V[X] := V1.V[Y] * V2.V[Z] - V1.V[Z] * V2.V[Y];
  result.V[Y] := V1.V[Z] * V2.V[X] - V1.V[X] * V2.V[Z];
  result.V[Z] := V1.V[X] * V2.V[Y] - V1.V[Y] * V2.V[X];
{$ENDIF}
end;

// VectorCrossProduct
//
function VectorCrossProduct(const V1, V2: TVector): TVector;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld   dword ptr [eax+$4]
  fmul  dword ptr [edx+$8]
  fld   dword ptr [eax+$8]
  fmul  dword ptr [edx+$4]
  fsubp
  fstp  dword ptr [ecx]

  fld   dword ptr [eax+$8]
  fmul  dword ptr [edx]
  fld   dword ptr [eax]
  fmul  dword ptr [edx+$8]
  fsubp
  fstp  dword ptr [ecx+$4]

  fld   dword ptr [eax]
  fmul  dword ptr [edx+$4]
  fld   dword ptr [eax+$4]
  fmul  dword ptr [edx]
  fsubp
  fstp  dword ptr [ecx+$8]

  xor   eax, eax
  mov   [ecx+$c], eax
  {$ELSE}
begin
  result.V[X] := V1.V[Y] * V2.V[Z] - V1.V[Z] * V2.V[Y];
  result.V[Y] := V1.V[Z] * V2.V[X] - V1.V[X] * V2.V[Z];
  result.V[Z] := V1.V[X] * V2.V[Y] - V1.V[Y] * V2.V[X];
  result.V[W] := 0;
{$ENDIF}
end;

// VectorCrossProduct
//
procedure VectorCrossProduct(const V1, V2: TVector; var vr: TVector);
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld   dword ptr [eax+$4]
  fmul  dword ptr [edx+$8]
  fld   dword ptr [eax+$8]
  fmul  dword ptr [edx+$4]
  fsubp
  fstp  dword ptr [ecx]

  fld   dword ptr [eax+$8]
  fmul  dword ptr [edx]
  fld   dword ptr [eax]
  fmul  dword ptr [edx+$8]
  fsubp
  fstp  dword ptr [ecx+$4]

  fld   dword ptr [eax]
  fmul  dword ptr [edx+$4]
  fld   dword ptr [eax+$4]
  fmul  dword ptr [edx]
  fsubp
  fstp  dword ptr [ecx+$8]

  xor   eax, eax
  mov   [ecx+$c], eax
  {$ELSE}
begin
  vr.V[X] := V1.V[Y] * V2.V[Z] - V1.V[Z] * V2.V[Y];
  vr.V[Y] := V1.V[Z] * V2.V[X] - V1.V[X] * V2.V[Z];
  vr.V[Z] := V1.V[X] * V2.V[Y] - V1.V[Y] * V2.V[X];
  vr.V[W] := 0;
{$ENDIF}
end;

// VectorCrossProduct
//
procedure VectorCrossProduct(const V1, V2: TAffineVector; var vr: TVector); overload;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld   dword ptr [eax+$4]
  fmul  dword ptr [edx+$8]
  fld   dword ptr [eax+$8]
  fmul  dword ptr [edx+$4]
  fsubp
  fstp  dword ptr [ecx]

  fld   dword ptr [eax+$8]
  fmul  dword ptr [edx]
  fld   dword ptr [eax]
  fmul  dword ptr [edx+$8]
  fsubp
  fstp  dword ptr [ecx+$4]

  fld   dword ptr [eax]
  fmul  dword ptr [edx+$4]
  fld   dword ptr [eax+$4]
  fmul  dword ptr [edx]
  fsubp
  fstp  dword ptr [ecx+$8]

  xor   eax, eax
  mov   [ecx+$c], eax
  {$ELSE}
begin
  vr.V[X] := V1.V[Y] * V2.V[Z] - V1.V[Z] * V2.V[Y];
  vr.V[Y] := V1.V[Z] * V2.V[X] - V1.V[X] * V2.V[Z];
  vr.V[Z] := V1.V[X] * V2.V[Y] - V1.V[Y] * V2.V[X];
  vr.V[W] := 0;
{$ENDIF}
end;

// VectorCrossProduct
//
procedure VectorCrossProduct(const V1, V2: TVector; var vr: TAffineVector); overload;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld   dword ptr [eax+$4]
  fmul  dword ptr [edx+$8]
  fld   dword ptr [eax+$8]
  fmul  dword ptr [edx+$4]
  fsubp
  fstp  dword ptr [ecx]

  fld   dword ptr [eax+$8]
  fmul  dword ptr [edx]
  fld   dword ptr [eax]
  fmul  dword ptr [edx+$8]
  fsubp
  fstp  dword ptr [ecx+$4]

  fld   dword ptr [eax]
  fmul  dword ptr [edx+$4]
  fld   dword ptr [eax+$4]
  fmul  dword ptr [edx]
  fsubp
  fstp  dword ptr [ecx+$8]
  {$ELSE}
begin
  vr.V[X] := V1.V[Y] * V2.V[Z] - V1.V[Z] * V2.V[Y];
  vr.V[Y] := V1.V[Z] * V2.V[X] - V1.V[X] * V2.V[Z];
  vr.V[Z] := V1.V[X] * V2.V[Y] - V1.V[Y] * V2.V[X];
{$ENDIF}
end;

// VectorCrossProduct
//
procedure VectorCrossProduct(const V1, V2: TAffineVector; var vr: TAffineVector); overload;
// EAX contains address of V1
// EDX contains address of V2
// ECX contains the result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld   dword ptr [eax+$4]
  fmul  dword ptr [edx+$8]
  fld   dword ptr [eax+$8]
  fmul  dword ptr [edx+$4]
  fsubp
  fstp  dword ptr [ecx]

  fld   dword ptr [eax+$8]
  fmul  dword ptr [edx]
  fld   dword ptr [eax]
  fmul  dword ptr [edx+$8]
  fsubp
  fstp  dword ptr [ecx+$4]

  fld   dword ptr [eax]
  fmul  dword ptr [edx+$4]
  fld   dword ptr [eax+$4]
  fmul  dword ptr [edx]
  fsubp
  fstp  dword ptr [ecx+$8]
  {$ELSE}
begin
  vr.V[X] := V1.V[Y] * V2.V[Z] - V1.V[Z] * V2.V[Y];
  vr.V[Y] := V1.V[Z] * V2.V[X] - V1.V[X] * V2.V[Z];
  vr.V[Z] := V1.V[X] * V2.V[Y] - V1.V[Y] * V2.V[X];
{$ENDIF}
end;

// Lerp
//
function Lerp(const start, stop, T: single): single;
begin
  result := start + (stop - start) * T;
end;

// Angle Lerp
//
function AngleLerp(start, stop, T: single): single;
var
  d: single;
begin
  start := NormalizeAngle(start);
  stop  := NormalizeAngle(stop);
  d     := stop - start;
  if d > PI then
  begin
    // positive d, angle on opposite side, becomes negative i.e. changes direction
    d := -d - c2PI;
  end
  else if d < -PI then
  begin
    // negative d, angle on opposite side, becomes positive i.e. changes direction
    d := d + c2PI;
  end;
  result := start + d * T;
end;

// DistanceBetweenAngles
//
function DistanceBetweenAngles(angle1, angle2: single): single;
begin
  angle1 := NormalizeAngle(angle1);
  angle2 := NormalizeAngle(angle2);
  result := Abs(angle2 - angle1);
  if result > PI then
    result := c2PI - result;
end;

// TexPointLerp
//
function TexPointLerp(const t1, t2: TTexPoint; T: single): TTexPoint; overload;
begin
  result.S := t1.S + (t2.S - t1.S) * T;
  result.T := t1.T + (t2.T - t1.T) * T;
end;

// VectorAffineLerp
//
function VectorLerp(const V1, V2: TAffineVector; T: single): TAffineVector;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld   t

  fld   dword ptr [eax+0]
  fld   dword ptr [edx+0]
  fsub  st(0), st(1)
  fmul  st(0), st(2)
  faddp
  fstp  dword ptr [ecx+0]

  fld   dword ptr [eax+4]
  fld   dword ptr [edx+4]
  fsub  st(0), st(1)
  fmul  st(0), st(2)
  faddp
  fstp  dword ptr [ecx+4]

  fld   dword ptr [eax+8]
  fld   dword ptr [edx+8]
  fsub  st(0), st(1)
  fmul  st(0), st(2)
  faddp
  fstp  dword ptr [ecx+8]

  ffree st(0)
  {$ELSE}
begin
  result.V[X] := V1.V[X] + (V2.V[X] - V1.V[X]) * T;
  result.V[Y] := V1.V[Y] + (V2.V[Y] - V1.V[Y]) * T;
  result.V[Z] := V1.V[Z] + (V2.V[Z] - V1.V[Z]) * T;
{$ENDIF}
end;

// VectorLerp
//
procedure VectorLerp(const V1, V2: TAffineVector; T: single; var vr: TAffineVector);
// EAX contains address of v1
// EDX contains address of v2
// EBX contains address of t
// ECX contains address of vr
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld   t

  fld   dword ptr [eax+0]
  fld   dword ptr [edx+0]
  fsub  st(0), st(1)
  fmul  st(0), st(2)
  faddp
  fstp  dword ptr [ecx+0]

  fld   dword ptr [eax+4]
  fld   dword ptr [edx+4]
  fsub  st(0), st(1)
  fmul  st(0), st(2)
  faddp
  fstp  dword ptr [ecx+4]

  fld   dword ptr [eax+8]
  fld   dword ptr [edx+8]
  fsub  st(0), st(1)
  fmul  st(0), st(2)
  faddp
  fstp  dword ptr [ecx+8]

  ffree st(0)
  {$ELSE}
begin
  vr.V[X] := V1.V[X] + (V2.V[X] - V1.V[X]) * T;
  vr.V[Y] := V1.V[Y] + (V2.V[Y] - V1.V[Y]) * T;
  vr.V[Z] := V1.V[Z] + (V2.V[Z] - V1.V[Z]) * T;
{$ENDIF}
end;

// VectorLerp
//
function VectorLerp(const V1, V2: TVector; T: single): TVector;
begin
  result.V[X] := V1.V[X] + (V2.V[X] - V1.V[X]) * T;
  result.V[Y] := V1.V[Y] + (V2.V[Y] - V1.V[Y]) * T;
  result.V[Z] := V1.V[Z] + (V2.V[Z] - V1.V[Z]) * T;
  result.V[W] := V1.V[W] + (V2.V[W] - V1.V[W]) * T;
end;

// VectorLerp
//
procedure VectorLerp(const V1, V2: TVector; T: single; var vr: TVector);
begin
  vr.V[X] := V1.V[X] + (V2.V[X] - V1.V[X]) * T;
  vr.V[Y] := V1.V[Y] + (V2.V[Y] - V1.V[Y]) * T;
  vr.V[Z] := V1.V[Z] + (V2.V[Z] - V1.V[Z]) * T;
  vr.V[W] := V1.V[W] + (V2.V[W] - V1.V[W]) * T;
end;

// VectorAngleLerp
//
function VectorAngleLerp(const V1, V2: TAffineVector; T: single): TAffineVector;
var
  q1, q2, qR: TQuaternion;
  M         : TMatrix;
  Tran      : TTransformations;
begin
  if VectorEquals(V1, V2) then
  begin
    result := V1;
  end
  else
  begin
    q1 := QuaternionFromEuler(VectorGeometry.RadToDeg(V1.V[0]), VectorGeometry.RadToDeg(V1.V[1]),
      VectorGeometry.RadToDeg(V1.V[2]), eulZYX);
    q2 := QuaternionFromEuler(VectorGeometry.RadToDeg(V2.V[0]), VectorGeometry.RadToDeg(V2.V[1]),
      VectorGeometry.RadToDeg(V2.V[2]), eulZYX);
    qR := QuaternionSlerp(q1, q2, T);
    M  := QuaternionToMatrix(qR);
    MatrixDecompose(M, Tran);
    result.V[0] := Tran[ttRotateX];
    result.V[1] := Tran[ttRotateY];
    result.V[2] := Tran[ttRotateZ];
  end;
end;

// VectorAngleCombine
//
function VectorAngleCombine(const V1, V2: TAffineVector; f: single): TAffineVector;
begin
  result := VectorCombine(V1, V2, 1, f);
end;

// VectorArrayLerp_3DNow (hmg)
//
{$IFNDEF GEOMETRY_NO_ASM}

procedure VectorArrayLerp_3DNow(const src1, src2: PVectorArray; T: single; n: Integer;
  dest: PVectorArray); stdcall; overload;
var
  pt: ^single;
begin
  pt := @T;
  asm
    push ebx
    push edi

    mov   eax, src1
    mov   edx, src2
    mov   ecx, n
    mov   ebx, dest
    mov   edi, pt

    db $0F,$0E               /// femms

    db $0F,$6E,$3F           /// movd     mm7, [edi]
    db $0F,$62,$FF           /// punpckldq mm7, mm7

  @@Loop:
    db $0F,$6F,$00           /// movq     mm0, [eax]
    db $0F,$6F,$50,$08       /// movq     mm2, [eax+8]
    db $0F,$6F,$C8           /// movq     mm1, mm0
    db $0F,$6F,$DA           /// movq     mm3, mm2
    db $0F,$0F,$02,$AA       /// pfsubr   mm0, [edx]
    db $0F,$0F,$52,$08,$AA   /// pfsubr   mm2, [edx+8]
    db $0F,$0D,$4B,$20       /// prefetchw [ebx+32]
    db $0F,$0F,$C7,$B4       /// pfmul    mm0, mm7
    db $0F,$0F,$D7,$B4       /// pfmul    mm2, mm7
    add   eax, 16
    add   edx, 16
    db $0F,$0D,$40,$20       /// prefetch [eax+32]
    db $0F,$0F,$C1,$9E       /// pfadd    mm0, mm1
    db $0F,$0F,$D3,$9E       /// pfadd    mm2, mm3
    db $0F,$0D,$42,$20       /// prefetch [edx+32]
    db $0F,$7F,$03           /// movq     [ebx], mm0
    db $0F,$7F,$53,$08       /// movq     [ebx+8], mm2

    add   ebx, 16

    dec   ecx
    jnz @@Loop

    db $0F,$0E               /// femms

    pop edi
    pop ebx
  end;
end;
{$ENDIF}

// VectorArrayLerp (hmg)
//
procedure VectorArrayLerp(const src1, src2: PVectorArray; T: single; n: Integer;
  dest: PVectorArray);
var
  i: Integer;
begin
{$IFNDEF GEOMETRY_NO_ASM}
  if vSIMD = 1 then
    VectorArrayLerp_3DNow(src1, src2, T, n, dest)
  else {$ENDIF} begin
    for i := 0 to n - 1 do
    begin
      dest^[i].V[0] := src1^[i].V[0] + (src2^[i].V[0] - src1^[i].V[0]) * T;
      dest^[i].V[1] := src1^[i].V[1] + (src2^[i].V[1] - src1^[i].V[1]) * T;
      dest^[i].V[2] := src1^[i].V[2] + (src2^[i].V[2] - src1^[i].V[2]) * T;
      dest^[i].V[3] := src1^[i].V[3] + (src2^[i].V[3] - src1^[i].V[3]) * T;
    end;
  end;
end;

// VectorArrayLerp_3DNow (affine)
//
{$IFNDEF GEOMETRY_NO_ASM}

procedure VectorArrayLerp_3DNow(const src1, src2: PAffineVectorArray; T: single; n: Integer;
  dest: PAffineVectorArray); stdcall; overload;
var
  pt: ^single;
begin
  pt := @T;
  asm
    push ebx
    push edi

    mov   eax, src1
    mov   edx, src2
    mov   ecx, n

    cmp   ecx, 1
    jbe   @@End

    shr   ecx, 1
    mov   ebx, dest
    mov   edi, pt

    db $0F,$0E               /// femms

    db $0F,$6E,$3F           /// movd     mm7, [edi]
    db $0F,$62,$FF           /// punpckldq mm7, mm7

  @@Loop:
    db $0F,$6F,$00           /// movq     mm0, [eax]
    db $0F,$6F,$50,$08       /// movq     mm2, [eax+8]
    db $0F,$6F,$60,$10       /// movq     mm4, [eax+16]
    db $0F,$6F,$C8           /// movq     mm1, mm0
    db $0F,$6F,$DA           /// movq     mm3, mm2
    db $0F,$6F,$EC           /// movq     mm5, mm4
    db $0F,$0F,$02,$AA       /// pfsubr   mm0, [edx]
    db $0F,$0F,$52,$08,$AA   /// pfsubr   mm2, [edx+8]
    db $0F,$0F,$62,$10,$AA   /// pfsubr   mm4, [edx+16]
    db $0F,$0D,$4B,$40       /// prefetchw [ebx+64]
    db $0F,$0F,$C7,$B4       /// pfmul    mm0, mm7
    db $0F,$0F,$D7,$B4       /// pfmul    mm2, mm7
    db $0F,$0F,$E7,$B4       /// pfmul    mm4, mm7
    db $0F,$0D,$40,$40       /// prefetch [eax+64]
    add   eax, 24
    add   edx, 24
    db $0F,$0F,$C1,$9E       /// pfadd    mm0, mm1
    db $0F,$0F,$D3,$9E       /// pfadd    mm2, mm3
    db $0F,$0F,$E5,$9E       /// pfadd    mm4, mm5
    db $0F,$0D,$42,$40       /// prefetch [edx+64]
    db $0F,$7F,$03           /// movq     [ebx], mm0
    db $0F,$7F,$53,$08       /// movq     [ebx+8], mm2
    db $0F,$7F,$63,$10       /// movq     [ebx+16], mm4

    add   ebx, 24

    dec   ecx
    jnz @@Loop

    db $0F,$0E               /// femms

  @@End:
    pop edi
    pop ebx
  end;
  if (n and 1) = 1 then
    VectorLerp(src1[n - 1], src2[n - 1], T, dest[n - 1]);
end;
{$ENDIF}

// VectorArrayLerp (affine)
//
procedure VectorArrayLerp(const src1, src2: PAffineVectorArray; T: single; n: Integer;
  dest: PAffineVectorArray);
var
  i: Integer;
begin
{$IFNDEF GEOMETRY_NO_ASM}
  if vSIMD = 1 then
    VectorArrayLerp_3DNow(src1, src2, T, n, dest)
  else {$ENDIF} begin
    for i := 0 to n - 1 do
    begin
      dest^[i].V[0] := src1^[i].V[0] + (src2^[i].V[0] - src1^[i].V[0]) * T;
      dest^[i].V[1] := src1^[i].V[1] + (src2^[i].V[1] - src1^[i].V[1]) * T;
      dest^[i].V[2] := src1^[i].V[2] + (src2^[i].V[2] - src1^[i].V[2]) * T;
    end;
  end;
end;

procedure VectorArrayLerp(const src1, src2: PTexPointArray; T: single; n: Integer;
  dest: PTexPointArray);
var
  i: Integer;
begin
  for i := 0 to n - 1 do
  begin
    dest^[i].S := src1^[i].S + (src2^[i].S - src1^[i].S) * T;
    dest^[i].T := src1^[i].T + (src2^[i].T - src1^[i].T) * T;
  end;
end;

// InterpolateCombined
//
function InterpolateCombined(const start, stop, delta: single; const DistortionDegree: single;
  const InterpolationType: TGLInterpolationType): single;
begin
  case InterpolationType of
    itLinear:
      result := Lerp(start, stop, delta);
    itPower:
      result := InterpolatePower(start, stop, delta, DistortionDegree);
    itSin:
      result := InterpolateSin(start, stop, delta);
    itSinAlt:
      result := InterpolateSinAlt(start, stop, delta);
    itTan:
      result := InterpolateTan(start, stop, delta);
    itLn:
      result := InterpolateLn(start, stop, delta, DistortionDegree);
    itExp:
      result := InterpolateExp(start, stop, delta, DistortionDegree);
  else
    begin
      result := -1;
      Assert(False);
    end;
  end;
end;

// InterpolateCombinedFastPower
//
function InterpolateCombinedFastPower(const OriginalStart, OriginalStop, OriginalCurrent: single;
  const TargetStart, TargetStop: single; const DistortionDegree: single): single;
begin
  result := InterpolatePower(TargetStart, TargetStop, (OriginalCurrent - OriginalStart) /
    (OriginalStop - OriginalStart), DistortionDegree);
end;

// InterpolateCombinedSafe
//
function InterpolateCombinedSafe(const OriginalStart, OriginalStop, OriginalCurrent: single;
  const TargetStart, TargetStop: single; const DistortionDegree: single;
  const InterpolationType: TGLInterpolationType): single;
var
  ChangeDelta: single;
begin
  if OriginalStop = OriginalStart then
    result := TargetStart
  else
  begin
    ChangeDelta := (OriginalCurrent - OriginalStart) / (OriginalStop - OriginalStart);
    result      := InterpolateCombined(TargetStart, TargetStop, ChangeDelta, DistortionDegree,
      InterpolationType);
  end;
end;

// InterpolateCombinedFast
//
function InterpolateCombinedFast(const OriginalStart, OriginalStop, OriginalCurrent: single;
  const TargetStart, TargetStop: single; const DistortionDegree: single;
  const InterpolationType: TGLInterpolationType): single;
var
  ChangeDelta: single;
begin
  ChangeDelta := (OriginalCurrent - OriginalStart) / (OriginalStop - OriginalStart);
  result      := InterpolateCombined(TargetStart, TargetStop, ChangeDelta, DistortionDegree,
    InterpolationType);
end;

// InterpolateLn
//
function InterpolateLn(const start, stop, delta: single; const DistortionDegree: single): single;
begin
  result := (stop - start) * Ln(1 + delta * DistortionDegree) / Ln(1 + DistortionDegree) + start;
end;

// InterpolateExp
//
function InterpolateExp(const start, stop, delta: single; const DistortionDegree: single): single;
begin
  result := (stop - start) * Exp(-DistortionDegree * (1 - delta)) + start;
end;

// InterpolateSinAlt
//
function InterpolateSinAlt(const start, stop, delta: single): single;
begin
  result := (stop - start) * delta * Sin(delta * PI / 2) + start;
end;

// InterpolateSin
//
function InterpolateSin(const start, stop, delta: single): single;
begin
  result := (stop - start) * Sin(delta * PI / 2) + start;
end;

// InterpolateTan
//
function InterpolateTan(const start, stop, delta: single): single;
begin
  result := (stop - start) * VectorGeometry.Tan(delta * PI / 4) + start;
end;

// InterpolatePower
//
function InterpolatePower(const start, stop, delta: single; const DistortionDegree: single): single;
var
  i: Integer;
begin
  if (Round(DistortionDegree) <> DistortionDegree) and (delta < 0) then
  begin
    i      := Round(DistortionDegree);
    result := (stop - start) * VectorGeometry.Power(delta, i) + start;
  end
  else
    result := (stop - start) * VectorGeometry.Power(delta, DistortionDegree) + start;
end;

// MatrixLerp
//
function MatrixLerp(const m1, m2: TMatrix; const delta: single): TMatrix;
var
  i, J: Integer;
begin
  for J                := 0 to 3 do
    for i              := 0 to 3 do
      result.V[i].V[J] := m1.V[i].V[J] + (m2.V[i].V[J] - m1.V[i].V[J]) * delta;
end;

// VectorLength (array)
//
function VectorLength(const V: array of single): single;
// EAX contains address of V
// EDX contains the highest index of V
// the result is returned in ST(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLDZ                           // initialize sum
@@Loop:
  FLD  DWORD PTR [EAX  +  4 * EDX] // load a component
  FMUL ST, ST
  FADDP
  SUB  EDX, 1
  JNL  @@Loop
  FSQRT
  {$ELSE}
var
  i: Integer;
begin
  result   := 0;
  for i    := Low(V) to High(V) do
    result := result + Sqr(V[i]);
  result   := Sqrt(result);
{$ENDIF}
end;

// VectorLength  (x, y)
//
function VectorLength(const X, Y: single): single;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD X
  FMUL ST, ST
  FLD Y
  FMUL ST, ST
  FADD
  FSQRT
  {$ELSE}
begin
  result := Sqrt(X * X + Y * Y);
{$ENDIF}
end;

// VectorLength (x, y, z)
//
function VectorLength(const X, Y, Z: single): single;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD X
  FMUL ST, ST
  FLD Y
  FMUL ST, ST
  FADD
  FLD Z
  FMUL ST, ST
  FADD
  FSQRT
  {$ELSE}
begin
  result := Sqrt(X * X + Y * Y + Z * Z);
{$ENDIF}
end;

// VectorLength
//
function VectorLength(const V: TVector2f): single;
// EAX contains address of V
// result is passed in ST(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FMUL ST, ST
  FLD  DWORD PTR [EAX+4]
  FMUL ST, ST
  FADDP
  FSQRT
  {$ELSE}
begin
  result := Sqrt(VectorNorm(V.V[0], V.V[1]));
{$ENDIF}
end;

// VectorLength
//
function VectorLength(const V: TAffineVector): single;
// EAX contains address of V
// result is passed in ST(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FMUL ST, ST
  FLD  DWORD PTR [EAX+4]
  FMUL ST, ST
  FADDP
  FLD  DWORD PTR [EAX+8]
  FMUL ST, ST
  FADDP
  FSQRT
  {$ELSE}
begin
  result := Sqrt(VectorNorm(V));
{$ENDIF}
end;

// VectorLength
//
function VectorLength(const V: TVector): single;
// EAX contains address of V
// result is passed in ST(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FMUL ST, ST
  FLD  DWORD PTR [EAX+4]
  FMUL ST, ST
  FADDP
  FLD  DWORD PTR [EAX+8]
  FMUL ST, ST
  FADDP
  FSQRT
  {$ELSE}
begin
  result := Sqrt(VectorNorm(V));
{$ENDIF}
end;

// VectorNorm
//
function VectorNorm(const X, Y: single): single;
begin
  result := Sqr(X) + Sqr(Y);
end;

// VectorNorm (affine)
//
function VectorNorm(const V: TAffineVector): single;
// EAX contains address of V
// result is passed in ST(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [EAX];
  FMUL ST, ST
  FLD DWORD PTR [EAX+4];
  FMUL ST, ST
  FADD
  FLD DWORD PTR [EAX+8];
  FMUL ST, ST
  FADD
  {$ELSE}
begin
  result := V.V[0] * V.V[0] + V.V[1] * V.V[1] + V.V[2] * V.V[2];
{$ENDIF}
end;

// VectorNorm (hmg)
//
function VectorNorm(const V: TVector): single;
// EAX contains address of V
// result is passed in ST(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [EAX];
  FMUL ST, ST
  FLD DWORD PTR [EAX+4];
  FMUL ST, ST
  FADD
  FLD DWORD PTR [EAX+8];
  FMUL ST, ST
  FADD
  {$ELSE}
begin
  result := V.V[0] * V.V[0] + V.V[1] * V.V[1] + V.V[2] * V.V[2];
{$ENDIF}
end;

// VectorNorm
//
function VectorNorm(var V: array of single): single;
// EAX contains address of V
// EDX contains highest index in V
// result is passed in ST(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLDZ                           // initialize sum
@@Loop:
  FLD  DWORD PTR [EAX + 4 * EDX] // load a component
  FMUL ST, ST                    // make square
  FADDP                          // add previous calculated sum
  SUB  EDX, 1
  JNL  @@Loop
  {$ELSE}
var
  i: Integer;
begin
  result   := 0;
  for i    := Low(V) to High(V) do
    result := result + V[i] * V[i];
{$ENDIF}
end;

// NormalizeVector (2f)
//
procedure NormalizeVector(var V: TVector2f);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
@@FPU:
  mov   ecx, eax
  FLD  DWORD PTR [ECX]
  FMUL ST, ST
  FLD  DWORD PTR [ECX+4]
  FMUL ST, ST
  FADD
  FLDZ
  FCOMP
  FNSTSW AX
  sahf
  jz @@result
  FSQRT
  FLD1
  FDIVR
@@result:
  FLD  ST
  FMUL DWORD PTR [ECX]
  FSTP DWORD PTR [ECX]
  FLD  ST
  FMUL DWORD PTR [ECX+4]
  FSTP DWORD PTR [ECX+4]
  {$ELSE}
var
  invLen: single;
  vn    : single;
begin
  vn := VectorNorm(V.V);
  if vn > 0 then
  begin
    invLen := RSqrt(vn);
    V.V[0] := V.V[0] * invLen;
    V.V[1] := V.V[1] * invLen;
  end;
{$ENDIF}
end;

// NormalizeVector (affine)
//
procedure NormalizeVector(var V: TAffineVector);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:
  db $0F,$6F,$00           /// movq        mm0,[eax]
  db $0F,$6E,$48,$08       /// movd        mm1,[eax+8]
  db $0F,$6F,$E0           /// movq        mm4,mm0
  db $0F,$6F,$D9           /// movq        mm3,mm1
  db $0F,$0F,$C0,$B4       /// pfmul       mm0,mm0
  db $0F,$0F,$C9,$B4       /// pfmul       mm1,mm1
  db $0F,$0F,$C0,$AE       /// pfacc       mm0,mm0
  db $0F,$0F,$C1,$9E       /// pfadd       mm0,mm1
  db $0F,$0F,$C8,$97       /// pfrsqrt     mm1,mm0
  db $0F,$6F,$D1           /// movq        mm2,mm1

  db $0F,$0F,$C9,$B4       /// pfmul       mm1,mm1
  db $0F,$0F,$C8,$A7       /// pfrsqit1    mm1,mm0
  db $0F,$0F,$CA,$B6       /// pfrcpit2    mm1,mm2
  db $0F,$62,$C9           /// punpckldq   mm1,mm1
  db $0F,$0F,$D9,$B4       /// pfmul       mm3,mm1
  db $0F,$0F,$E1,$B4       /// pfmul       mm4,mm1
  db $0F,$7E,$58,$08       /// movd        [eax+8],mm3
  db $0F,$7F,$20           /// movq        [eax],mm4
@@norm_end:
  db $0F,$0E               /// femms
  ret

@@FPU:
  mov   ecx, eax
  FLD  DWORD PTR [ECX]
  FMUL ST, ST
  FLD  DWORD PTR [ECX+4]
  FMUL ST, ST
  FADD
  FLD  DWORD PTR [ECX+8]
  FMUL ST, ST
  FADD
  FLDZ
  FCOMP
  FNSTSW AX
  sahf
  jz @@result
  FSQRT
  FLD1
  FDIVR
@@result:
  FLD  ST
  FMUL DWORD PTR [ECX]
  FSTP DWORD PTR [ECX]
  FLD  ST
  FMUL DWORD PTR [ECX+4]
  FSTP DWORD PTR [ECX+4]
  FMUL DWORD PTR [ECX+8]
  FSTP DWORD PTR [ECX+8]
  {$ELSE}
var
  invLen: single;
  vn    : single;
begin
  vn := VectorNorm(V);
  if vn > 0 then
  begin
    invLen := RSqrt(vn);
    V.V[0] := V.V[0] * invLen;
    V.V[1] := V.V[1] * invLen;
    V.V[2] := V.V[2] * invLen;
  end;
{$ENDIF}
end;

// VectorNormalize
//
function VectorNormalize(const V: TVector2f): TVector2f;
var
  invLen: single;
  vn    : single;
begin
  vn := VectorNorm(V.V[0], V.V[1]);
  if vn = 0 then
    result := V
  else
  begin
    invLen      := RSqrt(vn);
    result.V[0] := V.V[0] * invLen;
    result.V[1] := V.V[1] * invLen;
  end;
end;

// VectorNormalize
//
function VectorNormalize(const V: TAffineVector): TAffineVector;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:
  db $0F,$6F,$00           /// movq        mm0,[eax]
  db $0F,$6E,$48,$08       /// movd        mm1,[eax+8]
  db $0F,$6F,$E0           /// movq        mm4,mm0
  db $0F,$6F,$D9           /// movq        mm3,mm1
  db $0F,$0F,$C0,$B4       /// pfmul       mm0,mm0
  db $0F,$0F,$C9,$B4       /// pfmul       mm1,mm1
  db $0F,$0F,$C0,$AE       /// pfacc       mm0,mm0
  db $0F,$0F,$C1,$9E       /// pfadd       mm0,mm1
  db $0F,$0F,$C8,$97       /// pfrsqrt     mm1,mm0
  db $0F,$6F,$D1           /// movq        mm2,mm1

  db $0F,$0F,$C9,$B4       /// pfmul       mm1,mm1
  db $0F,$0F,$C8,$A7       /// pfrsqit1    mm1,mm0
  db $0F,$0F,$CA,$B6       /// pfrcpit2    mm1,mm2
  db $0F,$62,$C9           /// punpckldq   mm1,mm1
  db $0F,$0F,$D9,$B4       /// pfmul       mm3,mm1
  db $0F,$0F,$E1,$B4       /// pfmul       mm4,mm1
  db $0F,$7E,$5A,$08       /// movd        [edx+8],mm3
  db $0F,$7F,$22           /// movq        [edx],mm4
@@norm_end:
  db $0F,$0E               /// femms
  ret

@@FPU:
  mov   ecx, eax
  FLD  DWORD PTR [ECX]
  FMUL ST, ST
  FLD  DWORD PTR [ECX+4]
  FMUL ST, ST
  FADD
  FLD  DWORD PTR [ECX+8]
  FMUL ST, ST
  FADD
  FLDZ
  FCOMP
  FNSTSW AX
  sahf
  jz @@result
  FSQRT
  FLD1
  FDIVR
@@result:
  FLD  ST
  FMUL DWORD PTR [ECX]
  FSTP DWORD PTR [EDX]
  FLD  ST
  FMUL DWORD PTR [ECX+4]
  FSTP DWORD PTR [EDX+4]
  FMUL DWORD PTR [ECX+8]
  FSTP DWORD PTR [EDX+8]
  {$ELSE}
var
  invLen: single;
  vn    : single;
begin
  vn := VectorNorm(V);
  if vn = 0 then
    SetVector(result, V)
  else
  begin
    invLen      := RSqrt(vn);
    result.V[0] := V.V[0] * invLen;
    result.V[1] := V.V[1] * invLen;
    result.V[2] := V.V[2] * invLen;
  end;
{$ENDIF}
end;

// NormalizeVectorArray
//
procedure NormalizeVectorArray(list: PAffineVectorArray; n: Integer);
// EAX contains list
// EDX contains n
{$IFNDEF GEOMETRY_NO_ASM}
asm
  OR    EDX, EDX
  JZ    @@End
  test vSIMD, 1
  jz @@FPU
@@3DNowLoop:
  db $0F,$6F,$00           /// movq        mm0,[eax]
  db $0F,$6E,$48,$08       /// movd        mm1,[eax+8]
  db $0F,$6F,$E0           /// movq        mm4,mm0
  db $0F,$6F,$D9           /// movq        mm3,mm1
  db $0F,$0F,$C0,$B4       /// pfmul       mm0,mm0
  db $0F,$0F,$C9,$B4       /// pfmul       mm1,mm1
  db $0F,$0F,$C0,$AE       /// pfacc       mm0,mm0
  db $0F,$0F,$C1,$9E       /// pfadd       mm0,mm1
  db $0F,$0F,$C8,$97       /// pfrsqrt     mm1,mm0
  db $0F,$6F,$D1           /// movq        mm2,mm1

  db $0F,$0F,$C9,$B4       /// pfmul       mm1,mm1
  db $0F,$0F,$C8,$A7       /// pfrsqit1    mm1,mm0
  db $0F,$0F,$CA,$B6       /// pfrcpit2    mm1,mm2
  db $0F,$62,$C9           /// punpckldq   mm1,mm1
  db $0F,$0F,$D9,$B4       /// pfmul       mm3,mm1
  db $0F,$0F,$E1,$B4       /// pfmul       mm4,mm1
  db $0F,$7E,$58,$08       /// movd        [eax+8],mm3
  db $0F,$7F,$20           /// movq        [eax],mm4
@@norm_end:
  db $0F,$0E               /// femms
  add   eax, 12
  db $0F,$0D,$40,$60       /// PREFETCH    [EAX+96]
  dec   edx
  jnz   @@3DNowLOOP
  ret

@@FPU:
  mov   ecx, eax
@@FPULoop:
  FLD   DWORD PTR [ECX]
  FMUL  ST, ST
  FLD   DWORD PTR [ECX+4]
  FMUL  ST, ST
  FADD
  FLD   DWORD PTR [ECX+8]
  FMUL  ST, ST
  FADD
  FLDZ
  FCOMP
  FNSTSW AX
  sahf
  jz @@result
  FSQRT
  FLD1
  FDIVR
@@result:
  FLD   ST
  FMUL  DWORD PTR [ECX]
  FSTP  DWORD PTR [ECX]
  FLD   ST
  FMUL  DWORD PTR [ECX+4]
  FSTP  DWORD PTR [ECX+4]
  FMUL  DWORD PTR [ECX+8]
  FSTP  DWORD PTR [ECX+8]
  ADD   ECX, 12
  DEC   EDX
  JNZ   @@FPULOOP
@@End:
  {$ELSE}
var
  i: Integer;
begin
  for i := 0 to n - 1 do
    NormalizeVector(list^[i]);
{$ENDIF}
end;

// NormalizeVector (hmg)
//
procedure NormalizeVector(var V: TVector);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:
  db $0F,$6F,$00           /// movq        mm0,[eax]
  db $0F,$6E,$48,$08       /// movd        mm1,[eax+8]
  db $0F,$6F,$E0           /// movq        mm4,mm0
  db $0F,$6F,$D9           /// movq        mm3,mm1
  db $0F,$0F,$C0,$B4       /// pfmul       mm0,mm0
  db $0F,$0F,$C9,$B4       /// pfmul       mm1,mm1
  db $0F,$0F,$C0,$AE       /// pfacc       mm0,mm0
  db $0F,$0F,$C1,$9E       /// pfadd       mm0,mm1
  db $0F,$0F,$C8,$97       /// pfrsqrt     mm1,mm0
  db $0F,$6F,$D1           /// movq        mm2,mm1

  db $0F,$0F,$C9,$B4       /// pfmul       mm1,mm1
  db $0F,$0F,$C8,$A7       /// pfrsqit1    mm1,mm0
  db $0F,$0F,$CA,$B6       /// pfrcpit2    mm1,mm2
  db $0F,$62,$C9           /// punpckldq   mm1,mm1
  db $0F,$0F,$D9,$B4       /// pfmul       mm3,mm1
  db $0F,$0F,$E1,$B4       /// pfmul       mm4,mm1
  db $0F,$7E,$58,$08       /// movd        [eax+8],mm3
  db $0F,$7F,$20           /// movq        [eax],mm4
@@norm_end:
  db $0F,$0E               /// femms
  xor   edx, edx
  mov   [eax+12], edx
  ret

@@FPU:
  mov   ecx, eax
  FLD  DWORD PTR [ECX]
  FMUL ST, ST
  FLD  DWORD PTR [ECX+4]
  FMUL ST, ST
  FADD
  FLD  DWORD PTR [ECX+8]
  FMUL ST, ST
  FADD
  FLDZ
  FCOMP
  FNSTSW AX
  sahf
  jz @@result
  FSQRT
  FLD1
  FDIVR
@@result:
  FLD  ST
  FMUL DWORD PTR [ECX]
  FSTP DWORD PTR [ECX]
  FLD  ST
  FMUL DWORD PTR [ECX+4]
  FSTP DWORD PTR [ECX+4]
  FMUL DWORD PTR [ECX+8]
  FSTP DWORD PTR [ECX+8]
  xor   edx, edx
  mov   [ecx+12], edx
  {$ELSE}
var
  invLen: single;
  vn    : single;
begin
  vn := VectorNorm(V);
  if vn > 0 then
  begin
    invLen := RSqrt(vn);
    V.V[0] := V.V[0] * invLen;
    V.V[1] := V.V[1] * invLen;
    V.V[2] := V.V[2] * invLen;
  end;
  V.V[3] := 0;
{$ENDIF}
end;

// VectorNormalize (hmg, func)
//
function VectorNormalize(const V: TVector): TVector;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:
  db $0F,$6F,$00           /// movq        mm0,[eax]
  db $0F,$6E,$48,$08       /// movd        mm1,[eax+8]
  db $0F,$6F,$E0           /// movq        mm4,mm0
  db $0F,$6F,$D9           /// movq        mm3,mm1
  db $0F,$0F,$C0,$B4       /// pfmul       mm0,mm0
  db $0F,$0F,$C9,$B4       /// pfmul       mm1,mm1
  db $0F,$0F,$C0,$AE       /// pfacc       mm0,mm0
  db $0F,$0F,$C1,$9E       /// pfadd       mm0,mm1
  db $0F,$0F,$C8,$97       /// pfrsqrt     mm1,mm0
  db $0F,$6F,$D1           /// movq        mm2,mm1

  db $0F,$0F,$C9,$B4       /// pfmul       mm1,mm1
  db $0F,$0F,$C8,$A7       /// pfrsqit1    mm1,mm0
  db $0F,$0F,$CA,$B6       /// pfrcpit2    mm1,mm2
  db $0F,$62,$C9           /// punpckldq   mm1,mm1
  db $0F,$0F,$D9,$B4       /// pfmul       mm3,mm1
  db $0F,$0F,$E1,$B4       /// pfmul       mm4,mm1
  db $0F,$7E,$5A,$08       /// movd        [edx+8],mm3
  db $0F,$7F,$22           /// movq        [edx],mm4
@@norm_end:
  db $0F,$0E               /// femms
  xor   eax, eax
  mov   [edx+12], eax
  ret

@@FPU:
  mov	ecx, eax
  FLD  DWORD PTR [ECX]
  FMUL ST, ST
  FLD  DWORD PTR [ECX+4]
  FMUL ST, ST
  FADD
  FLD  DWORD PTR [ECX+8]
  FMUL ST, ST
  FADD
  FLDZ
  FCOMP
  FNSTSW AX
  sahf
  jz @@result
  FSQRT
  FLD1
  FDIVR
@@result:
  FLD  ST
  FMUL DWORD PTR [ECX]
  FSTP DWORD PTR [EDX]
  FLD  ST
  FMUL DWORD PTR [ECX+4]
  FSTP DWORD PTR [EDX+4]
  FMUL DWORD PTR [ECX+8]
  FSTP DWORD PTR [EDX+8]
  xor   ecx, ecx
  mov   [edx+12], ecx
  {$ELSE}
var
  invLen: single;
  vn    : single;
begin
  vn := VectorNorm(V);
  if vn = 0 then
    SetVector(result, V)
  else
  begin
    invLen      := RSqrt(vn);
    result.V[0] := V.V[0] * invLen;
    result.V[1] := V.V[1] * invLen;
    result.V[2] := V.V[2] * invLen;
  end;
  result.V[3] := 0;
{$ENDIF}
end;

// VectorAngleCosine
//
function VectorAngleCosine(const V1, V2: TAffineVector): single;
// EAX contains address of Vector1
// EDX contains address of Vector2
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [EAX]           // V1[0]
  FLD ST                        // double V1[0]
  FMUL ST, ST                   // V1[0]^2 (prep. for divisor)
  FLD DWORD PTR [EDX]           // V2[0]
  FMUL ST(2), ST                // ST(2):=V1[0] * V2[0]
  FMUL ST, ST                   // V2[0]^2 (prep. for divisor)
  FLD DWORD PTR [EAX + 4]       // V1[1]
  FLD ST                        // double V1[1]
  FMUL ST, ST                   // ST(0):=V1[1]^2
  FADDP ST(3), ST               // ST(2):=V1[0]^2 + V1[1] *  * 2
  FLD DWORD PTR [EDX + 4]       // V2[1]
  FMUL ST(1), ST                // ST(1):=V1[1] * V2[1]
  FMUL ST, ST                   // ST(0):=V2[1]^2
  FADDP ST(2), ST               // ST(1):=V2[0]^2 + V2[1]^2
  FADDP ST(3), ST               // ST(2):=V1[0] * V2[0] + V1[1] * V2[1]
  FLD DWORD PTR [EAX + 8]       // load V2[1]
  FLD ST                        // same calcs go here
  FMUL ST, ST                   // (compare above)
  FADDP ST(3), ST
  FLD DWORD PTR [EDX + 8]
  FMUL ST(1), ST
  FMUL ST, ST
  FADDP ST(2), ST
  FADDP ST(3), ST
  FMULP                         // ST(0):=(V1[0]^2 + V1[1]^2 + V1[2]) *
  // (V2[0]^2 + V2[1]^2 + V2[2])
  FSQRT                         // sqrt(ST(0))
  FDIVP                         // ST(0):=Result:=ST(1) / ST(0)
  // the result is expected in ST(0), if it's invalid, an error is raised
  {$ELSE}
begin
  result := VectorDotProduct(V1, V2) / (VectorLength(V1) * VectorLength(V2));
{$ENDIF}
end;

// VectorAngleCosine
//
function VectorAngleCosine(const V1, V2: TVector): single;
// EAX contains address of Vector1
// EDX contains address of Vector2
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [EAX]           // V1[0]
  FLD ST                        // double V1[0]
  FMUL ST, ST                   // V1[0]^2 (prep. for divisor)
  FLD DWORD PTR [EDX]           // V2[0]
  FMUL ST(2), ST                // ST(2):=V1[0] * V2[0]
  FMUL ST, ST                   // V2[0]^2 (prep. for divisor)
  FLD DWORD PTR [EAX + 4]       // V1[1]
  FLD ST                        // double V1[1]
  FMUL ST, ST                   // ST(0):=V1[1]^2
  FADDP ST(3), ST               // ST(2):=V1[0]^2 + V1[1] *  * 2
  FLD DWORD PTR [EDX + 4]       // V2[1]
  FMUL ST(1), ST                // ST(1):=V1[1] * V2[1]
  FMUL ST, ST                   // ST(0):=V2[1]^2
  FADDP ST(2), ST               // ST(1):=V2[0]^2 + V2[1]^2
  FADDP ST(3), ST               // ST(2):=V1[0] * V2[0] + V1[1] * V2[1]
  FLD DWORD PTR [EAX + 8]       // load V2[1]
  FLD ST                        // same calcs go here
  FMUL ST, ST                   // (compare above)
  FADDP ST(3), ST
  FLD DWORD PTR [EDX + 8]
  FMUL ST(1), ST
  FMUL ST, ST
  FADDP ST(2), ST
  FADDP ST(3), ST
  FMULP                         // ST(0):=(V1[0]^2 + V1[1]^2 + V1[2]) *
  // (V2[0]^2 + V2[1]^2 + V2[2])
  FSQRT                         // sqrt(ST(0))
  FDIVP                         // ST(0):=Result:=ST(1) / ST(0)
  // the result is expected in ST(0), if it's invalid, an error is raised
  {$ELSE}
begin
  result := VectorDotProduct(V1, V2) / (VectorLength(V1) * VectorLength(V2));
{$ENDIF}
end;

// VectorNegate (affine)
//
function VectorNegate(const Vector: TAffineVector): TAffineVector;
// EAX contains address of v
// EDX contains address of Result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [EAX]
  FCHS
  FSTP DWORD PTR [EDX]
  FLD DWORD PTR [EAX+4]
  FCHS
  FSTP DWORD PTR [EDX+4]
  FLD DWORD PTR [EAX+8]
  FCHS
  FSTP DWORD PTR [EDX+8]
  {$ELSE}
begin
  result.V[0] := -Vector.V[0];
  result.V[1] := -Vector.V[1];
  result.V[2] := -Vector.V[2];
{$ENDIF}
end;

// VectorNegate (hmg)
//
function VectorNegate(const Vector: TVector): TVector;
// EAX contains address of v
// EDX contains address of Result
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [EAX]
  FCHS
  FSTP DWORD PTR [EDX]
  FLD DWORD PTR [EAX+4]
  FCHS
  FSTP DWORD PTR [EDX+4]
  FLD DWORD PTR [EAX+8]
  FCHS
  FSTP DWORD PTR [EDX+8]
  FLD DWORD PTR [EAX+12]
  FCHS
  FSTP DWORD PTR [EDX+12]
  {$ELSE}
begin
  result.V[0] := -Vector.V[0];
  result.V[1] := -Vector.V[1];
  result.V[2] := -Vector.V[2];
  result.V[3] := -Vector.V[3];
{$ENDIF}
end;

// NegateVector
//
procedure NegateVector(var V: TAffineVector);
// EAX contains address of v
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [EAX]
  FCHS
  FSTP DWORD PTR [EAX]
  FLD DWORD PTR [EAX+4]
  FCHS
  FSTP DWORD PTR [EAX+4]
  FLD DWORD PTR [EAX+8]
  FCHS
  FSTP DWORD PTR [EAX+8]
  {$ELSE}
begin
  V.V[0] := -V.V[0];
  V.V[1] := -V.V[1];
  V.V[2] := -V.V[2];
{$ENDIF}
end;

// NegateVector
//
procedure NegateVector(var V: TVector);
// EAX contains address of v
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [EAX]
  FCHS
  FSTP DWORD PTR [EAX]
  FLD DWORD PTR [EAX+4]
  FCHS
  FSTP DWORD PTR [EAX+4]
  FLD DWORD PTR [EAX+8]
  FCHS
  FSTP DWORD PTR [EAX+8]
  FLD DWORD PTR [EAX+12]
  FCHS
  FSTP DWORD PTR [EAX+12]
  {$ELSE}
begin
  V.V[0] := -V.V[0];
  V.V[1] := -V.V[1];
  V.V[2] := -V.V[2];
  V.V[3] := -V.V[3];
{$ENDIF}
end;

// NegateVector
//
procedure NegateVector(var V: array of single);
// EAX contains address of V
// EDX contains highest index in V
{$IFNDEF GEOMETRY_NO_ASM}
asm
@@Loop:
  FLD DWORD PTR [EAX + 4 * EDX]
  FCHS
  WAIT
  FSTP DWORD PTR [EAX + 4 * EDX]
  DEC EDX
  JNS @@Loop
  {$ELSE}
var
  i: Integer;
begin
  for i  := Low(V) to High(V) do
    V[i] := -V[i];
{$ENDIF}
end;

// ScaleVector (2f)
//
procedure ScaleVector(var V: TVector2f; factor: single);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EAX]
  FLD  DWORD PTR [EAX+4]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EAX+4]
  {$ELSE}
begin
  V.V[0] := V.V[0] * factor;
  V.V[1] := V.V[1] * factor;
{$ENDIF}
end;

// ScaleVector (affine)
//
procedure ScaleVector(var V: TAffineVector; factor: single);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EAX]
  FLD  DWORD PTR [EAX+4]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EAX+4]
  FLD  DWORD PTR [EAX+8]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EAX+8]
  {$ELSE}
begin
  V.V[0] := V.V[0] * factor;
  V.V[1] := V.V[1] * factor;
  V.V[2] := V.V[2] * factor;
{$ENDIF}
end;

// ScaleVector (hmg)
//
procedure ScaleVector(var V: TVector; factor: single);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test     vSIMD, 1
  jz @@FPU

@@3DNow:      // 121824

  db $0F,$6E,$4D,$08       /// movd        mm1, [ebp+8]
  db $0F,$62,$C9           /// punpckldq   mm1, mm1

  db $0F,$6F,$00           /// movq        mm0, [eax]
  db $0F,$6F,$50,$08       /// movq        mm2, [eax+8]
  db $0F,$0F,$C1,$B4       /// pfmul       mm0, mm1
  db $0F,$0F,$D1,$B4       /// pfmul       mm2, mm1
  db $0F,$7F,$00           /// movq        [eax], mm0
  db $0F,$7F,$50,$08       /// movq        [eax+8], mm2

  db $0F,$0E               /// femms

  pop   ebp
  ret   $04

@@FPU:        // 155843
  FLD  DWORD PTR [EBP+8]

  FLD  DWORD PTR [EAX]
  FMUL ST, ST(1)
  FSTP DWORD PTR [EAX]
  FLD  DWORD PTR [EAX+4]
  FMUL ST, ST(1)
  FSTP DWORD PTR [EAX+4]
  FLD  DWORD PTR [EAX+8]
  FMUL ST, ST(1)
  FSTP DWORD PTR [EAX+8]
  FLD  DWORD PTR [EAX+12]
  FMULP
  FSTP DWORD PTR [EAX+12]
  {$ELSE}
begin
  V.V[0] := V.V[0] * factor;
  V.V[1] := V.V[1] * factor;
  V.V[2] := V.V[2] * factor;
  V.V[3] := V.V[3] * factor;
{$ENDIF}
end;

// ScaleVector (affine vector)
//
procedure ScaleVector(var V: TAffineVector; const factor: TAffineVector);
begin
  V.V[0] := V.V[0] * factor.V[0];
  V.V[1] := V.V[1] * factor.V[1];
  V.V[2] := V.V[2] * factor.V[2];
end;

// ScaleVector (hmg vector)
//
procedure ScaleVector(var V: TVector; const factor: TVector);
begin
  V.V[0] := V.V[0] * factor.V[0];
  V.V[1] := V.V[1] * factor.V[1];
  V.V[2] := V.V[2] * factor.V[2];
  V.V[3] := V.V[3] * factor.V[3];
end;

// VectorScale (2f)
//
function VectorScale(const V: TVector2f; factor: single): TVector2f;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX]
  FLD  DWORD PTR [EAX+4]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX+4]
  {$ELSE}
begin
  result.V[0] := V.V[0] * factor;
  result.V[1] := V.V[1] * factor;
{$ENDIF}
end;

// VectorScale (affine)
//
function VectorScale(const V: TAffineVector; factor: single): TAffineVector;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX]
  FLD  DWORD PTR [EAX+4]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX+4]
  FLD  DWORD PTR [EAX+8]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX+8]
  {$ELSE}
begin
  result.V[0] := V.V[0] * factor;
  result.V[1] := V.V[1] * factor;
  result.V[2] := V.V[2] * factor;
{$ENDIF}
end;

// VectorScale (proc, affine)
//
procedure VectorScale(const V: TAffineVector; factor: single; var vr: TAffineVector);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX]
  FLD  DWORD PTR [EAX+4]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX+4]
  FLD  DWORD PTR [EAX+8]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX+8]
  {$ELSE}
begin
  vr.V[0] := V.V[0] * factor;
  vr.V[1] := V.V[1] * factor;
  vr.V[2] := V.V[2] * factor;
{$ENDIF}
end;

// VectorScale (hmg)
//
function VectorScale(const V: TVector; factor: single): TVector;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX]
  FLD  DWORD PTR [EAX+4]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX+4]
  FLD  DWORD PTR [EAX+8]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX+8]
  FLD  DWORD PTR [EAX+12]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX+12]
  {$ELSE}
begin
  result.V[0] := V.V[0] * factor;
  result.V[1] := V.V[1] * factor;
  result.V[2] := V.V[2] * factor;
  result.V[3] := V.V[3] * factor;
{$ENDIF}
end;

// VectorScale (proc, hmg)
//
procedure VectorScale(const V: TVector; factor: single; var vr: TVector);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX]
  FLD  DWORD PTR [EAX+4]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX+4]
  FLD  DWORD PTR [EAX+8]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX+8]
  FLD  DWORD PTR [EAX+12]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX+12]
  {$ELSE}
begin
  vr.V[0] := V.V[0] * factor;
  vr.V[1] := V.V[1] * factor;
  vr.V[2] := V.V[2] * factor;
  vr.V[3] := V.V[3] * factor;
{$ENDIF}
end;

// VectorScale (proc, hmg-affine)
//
procedure VectorScale(const V: TVector; factor: single; var vr: TAffineVector);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX]
  FLD  DWORD PTR [EAX+4]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX+4]
  FLD  DWORD PTR [EAX+8]
  FMUL DWORD PTR [EBP+8]
  FSTP DWORD PTR [EDX+8]
  {$ELSE}
begin
  vr.V[0] := V.V[0] * factor;
  vr.V[1] := V.V[1] * factor;
  vr.V[2] := V.V[2] * factor;
{$ENDIF}
end;

// VectorScale (func, affine)
//
function VectorScale(const V: TAffineVector; const factor: TAffineVector): TAffineVector;
begin
  result.V[0] := V.V[0] * factor.V[0];
  result.V[1] := V.V[1] * factor.V[1];
  result.V[2] := V.V[2] * factor.V[2];
end;

// VectorScale (func, hmg)
//
function VectorScale(const V: TVector; const factor: TVector): TVector;
begin
  result.V[0] := V.V[0] * factor.V[0];
  result.V[1] := V.V[1] * factor.V[1];
  result.V[2] := V.V[2] * factor.V[2];
  result.V[3] := V.V[3] * factor.V[3];
end;

// DivideVector
//
procedure DivideVector(var V: TVector; const divider: TVector);
begin
  V.V[0] := V.V[0] / divider.V[0];
  V.V[1] := V.V[1] / divider.V[1];
  V.V[2] := V.V[2] / divider.V[2];
  V.V[3] := V.V[3] / divider.V[3];
end;

// DivideVector
//
procedure DivideVector(var V: TAffineVector; const divider: TAffineVector); overload;
begin
  V.V[0] := V.V[0] / divider.V[0];
  V.V[1] := V.V[1] / divider.V[1];
  V.V[2] := V.V[2] / divider.V[2];
end;

// VectorDivide
//
function VectorDivide(const V: TVector; const divider: TVector): TVector; overload;
begin
  result.V[0] := V.V[0] / divider.V[0];
  result.V[1] := V.V[1] / divider.V[1];
  result.V[2] := V.V[2] / divider.V[2];
  result.V[3] := V.V[3] / divider.V[3];
end;

// VectorDivide
//
function VectorDivide(const V: TAffineVector; const divider: TAffineVector): TAffineVector;
  overload;
begin
  result.V[0] := V.V[0] / divider.V[0];
  result.V[1] := V.V[1] / divider.V[1];
  result.V[2] := V.V[2] / divider.V[2];
end;

// TexpointEquals
//
function TexpointEquals(const p1, p2: TTexPoint): Boolean;
begin
  result := (p1.S = p2.S) and (p1.T = p2.T);
end;

// RectEquals
//
function RectEquals(const Rect1, Rect2: TRect): Boolean;
begin
  result := (Rect1.Left = Rect2.Left) and (Rect1.Right = Rect2.Right) and (Rect1.Top = Rect2.Top)
    and (Rect1.Bottom = Rect2.Left);
end;

// VectorEquals (hmg vector)
//
function VectorEquals(const V1, V2: TVector): Boolean;
// EAX contains address of v1
// EDX contains highest of v2
{$IFNDEF GEOMETRY_NO_ASM}
asm
  mov ecx, [edx]
  cmp ecx, [eax]
  jne @@Diff
  mov ecx, [edx+$4]
  cmp ecx, [eax+$4]
  jne @@Diff
  mov ecx, [edx+$8]
  cmp ecx, [eax+$8]
  jne @@Diff
  mov ecx, [edx+$C]
  cmp ecx, [eax+$C]
  jne @@Diff
@@Equal:
  mov eax, 1
  ret
@@Diff:
  xor eax, eax
  {$ELSE}
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]) and (V1.V[2] = V2.V[2]) and
    (V1.V[3] = V2.V[3]);
{$ENDIF}
end;

// VectorEquals (affine vector)
//
function VectorEquals(const V1, V2: TAffineVector): Boolean;
// EAX contains address of v1
// EDX contains highest of v2
{$IFNDEF GEOMETRY_NO_ASM}
asm
  mov ecx, [edx]
  cmp ecx, [eax]
  jne @@Diff
  mov ecx, [edx+$4]
  cmp ecx, [eax+$4]
  jne @@Diff
  mov ecx, [edx+$8]
  cmp ecx, [eax+$8]
  jne @@Diff
@@Equal:
  mov al, 1
  ret
@@Diff:
  xor eax, eax
@@End:
  {$ELSE}
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]) and (V1.V[2] = V2.V[2]);
{$ENDIF}
end;

// AffineVectorEquals (hmg vector)
//
function AffineVectorEquals(const V1, V2: TVector): Boolean;
// EAX contains address of v1
// EDX contains highest of v2
{$IFNDEF GEOMETRY_NO_ASM}
asm
  mov ecx, [edx]
  cmp ecx, [eax]
  jne @@Diff
  mov ecx, [edx+$4]
  cmp ecx, [eax+$4]
  jne @@Diff
  mov ecx, [edx+$8]
  cmp ecx, [eax+$8]
  jne @@Diff
@@Equal:
  mov eax, 1
  ret
@@Diff:
  xor eax, eax
  {$ELSE}
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]) and (V1.V[2] = V2.V[2]);
{$ENDIF}
end;

// VectorIsNull (hmg)
//
function VectorIsNull(const V: TVector): Boolean;
begin
  result := ((V.V[0] = 0) and (V.V[1] = 0) and (V.V[2] = 0));
end;

// VectorIsNull (affine)
//
function VectorIsNull(const V: TAffineVector): Boolean; overload;
begin
  result := ((V.V[0] = 0) and (V.V[1] = 0) and (V.V[2] = 0));
end;

// VectorSpacing (texpoint)
//
function VectorSpacing(const V1, V2: TTexPoint): single; overload;
// EAX contains address of v1
// EDX contains highest of v2
// Result  is passed on the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FABS
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FABS
  FADD
  {$ELSE}
begin
  result := Abs(V2.S - V1.S) + Abs(V2.T - V1.T);
{$ENDIF}
end;

// VectorSpacing (affine)
//
function VectorSpacing(const V1, V2: TAffineVector): single;
// EAX contains address of v1
// EDX contains highest of v2
// Result  is passed on the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FABS
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FABS
  FADD
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FABS
  FADD
  {$ELSE}
begin
  result := Abs(V2.V[0] - V1.V[0]) + Abs(V2.V[1] - V1.V[1]) + Abs(V2.V[2] - V1.V[2]);
{$ENDIF}
end;

// VectorSpacing (Hmg)
//
function VectorSpacing(const V1, V2: TVector): single;
// EAX contains address of v1
// EDX contains highest of v2
// Result  is passed on the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FABS
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FABS
  FADD
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FABS
  FADD
  FLD  DWORD PTR [EAX+12]
  FSUB DWORD PTR [EDX+12]
  FABS
  FADD
  {$ELSE}
begin
  result := Abs(V2.V[0] - V1.V[0]) + Abs(V2.V[1] - V1.V[1]) + Abs(V2.V[2] - V1.V[2]) +
    Abs(V2.V[3] - V1.V[3]);
{$ENDIF}
end;

// VectorDistance (affine)
//
function VectorDistance(const V1, V2: TAffineVector): single;
// EAX contains address of v1
// EDX contains highest of v2
// Result  is passed on the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FMUL ST, ST
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FMUL ST, ST
  FADD
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FMUL ST, ST
  FADD
  FSQRT
  {$ELSE}
begin
  result := Sqrt(Sqr(V2.V[0] - V1.V[0]) + Sqr(V2.V[1] - V1.V[1]) + Sqr(V2.V[2] - V1.V[2]));
{$ENDIF}
end;

// VectorDistance (hmg)
//
function VectorDistance(const V1, V2: TVector): single;
// EAX contains address of v1
// EDX contains highest of v2
// Result  is passed on the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FMUL ST, ST
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FMUL ST, ST
  FADD
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FMUL ST, ST
  FADD
  FSQRT
  {$ELSE}
begin
  result := Sqrt(Sqr(V2.V[0] - V1.V[0]) + Sqr(V2.V[1] - V1.V[1]) + Sqr(V2.V[2] - V1.V[2]));
{$ENDIF}
end;

// VectorDistance2 (affine)
//
function VectorDistance2(const V1, V2: TAffineVector): single;
// EAX contains address of v1
// EDX contains highest of v2
// Result is passed on the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FMUL ST, ST
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FMUL ST, ST
  FADD
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FMUL ST, ST
  FADD
  {$ELSE}
begin
  result := Sqr(V2.V[0] - V1.V[0]) + Sqr(V2.V[1] - V1.V[1]) + Sqr(V2.V[2] - V1.V[2]);
{$ENDIF}
end;

// VectorDistance2 (hmg)
//
function VectorDistance2(const V1, V2: TVector): single;
// EAX contains address of v1
// EDX contains highest of v2
// Result is passed on the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSUB DWORD PTR [EDX]
  FMUL ST, ST
  FLD  DWORD PTR [EAX+4]
  FSUB DWORD PTR [EDX+4]
  FMUL ST, ST
  FADD
  FLD  DWORD PTR [EAX+8]
  FSUB DWORD PTR [EDX+8]
  FMUL ST, ST
  FADD
  {$ELSE}
begin
  result := Sqr(V2.V[0] - V1.V[0]) + Sqr(V2.V[1] - V1.V[1]) + Sqr(V2.V[2] - V1.V[2]);
{$ENDIF}
end;

// VectorPerpendicular
//
function VectorPerpendicular(const V, n: TAffineVector): TAffineVector;
var
  dot: single;
begin
  dot         := VectorDotProduct(V, n);
  result.V[X] := V.V[X] - dot * n.V[X];
  result.V[Y] := V.V[Y] - dot * n.V[Y];
  result.V[Z] := V.V[Z] - dot * n.V[Z];
end;

// VectorReflect
//
function VectorReflect(const V, n: TAffineVector): TAffineVector;
begin
  result := VectorCombine(V, n, 1, -2 * VectorDotProduct(V, n));
end;

// RotateVector
//
procedure RotateVector(var Vector: TVector; const axis: TAffineVector; angle: single);
var
  rotMatrix: TMatrix4f;
begin
  rotMatrix := CreateRotationMatrix(axis, angle);
  Vector    := VectorTransform(Vector, rotMatrix);
end;

// RotateVector
//
procedure RotateVector(var Vector: TVector; const axis: TVector; angle: single); overload;
var
  rotMatrix: TMatrix4f;
begin
  rotMatrix := CreateRotationMatrix(PAffineVector(@axis)^, angle);
  Vector    := VectorTransform(Vector, rotMatrix);
end;

// RotateVectorAroundY
//
procedure RotateVectorAroundY(var V: TAffineVector; alpha: single);
var
  c, S, v0: single;
begin
  VectorGeometry.SinCos(alpha, S, c);
  v0     := V.V[0];
  V.V[0] := c * v0 + S * V.V[2];
  V.V[2] := c * V.V[2] - S * v0;
end;

// VectorRotateAroundX (func)
//
function VectorRotateAroundX(const V: TAffineVector; alpha: single): TAffineVector;
var
  c, S: single;
begin
  VectorGeometry.SinCos(alpha, S, c);
  result.V[0] := V.V[0];
  result.V[1] := c * V.V[1] + S * V.V[2];
  result.V[2] := c * V.V[2] - S * V.V[1];
end;

// VectorRotateAroundY (func)
//
function VectorRotateAroundY(const V: TAffineVector; alpha: single): TAffineVector;
var
  c, S: single;
begin
  VectorGeometry.SinCos(alpha, S, c);
  result.V[1] := V.V[1];
  result.V[0] := c * V.V[0] + S * V.V[2];
  result.V[2] := c * V.V[2] - S * V.V[0];
end;

// VectorRotateAroundY (proc)
//
procedure VectorRotateAroundY(const V: TAffineVector; alpha: single; var vr: TAffineVector);
var
  c, S: single;
begin
  VectorGeometry.SinCos(alpha, S, c);
  vr.V[1] := V.V[1];
  vr.V[0] := c * V.V[0] + S * V.V[2];
  vr.V[2] := c * V.V[2] - S * V.V[0];
end;

// VectorRotateAroundZ (func)
//
function VectorRotateAroundZ(const V: TAffineVector; alpha: single): TAffineVector;
var
  c, S: single;
begin
  VectorGeometry.SinCos(alpha, S, c);
  result.V[0] := c * V.V[0] + S * V.V[1];
  result.V[1] := c * V.V[1] - S * V.V[0];
  result.V[2] := V.V[2];
end;

// AbsVector (hmg)
//
procedure AbsVector(var V: TVector);
begin
  V.V[0] := Abs(V.V[0]);
  V.V[1] := Abs(V.V[1]);
  V.V[2] := Abs(V.V[2]);
  V.V[3] := Abs(V.V[3]);
end;

// AbsVector (affine)
//
procedure AbsVector(var V: TAffineVector);
begin
  V.V[0] := Abs(V.V[0]);
  V.V[1] := Abs(V.V[1]);
  V.V[2] := Abs(V.V[2]);
end;

// VectorAbs (hmg)
//
function VectorAbs(const V: TVector): TVector;
begin
  result.V[0] := Abs(V.V[0]);
  result.V[1] := Abs(V.V[1]);
  result.V[2] := Abs(V.V[2]);
  result.V[3] := Abs(V.V[3]);
end;

// VectorAbs (affine)
//
function VectorAbs(const V: TAffineVector): TAffineVector;
begin
  result.V[0] := Abs(V.V[0]);
  result.V[1] := Abs(V.V[1]);
  result.V[2] := Abs(V.V[2]);
end;

// IsColinear (2f)
//
function IsColinear(const V1, V2: TVector2f): Boolean; overload;
var
  a, b, c: single;
begin
  a      := VectorDotProduct(V1, V1);
  b      := VectorDotProduct(V1, V2);
  c      := VectorDotProduct(V2, V2);
  result := (a * c - b * b) < cColinearBias;
end;

// IsColinear (affine)
//
function IsColinear(const V1, V2: TAffineVector): Boolean; overload;
var
  a, b, c: single;
begin
  a      := VectorDotProduct(V1, V1);
  b      := VectorDotProduct(V1, V2);
  c      := VectorDotProduct(V2, V2);
  result := (a * c - b * b) < cColinearBias;
end;

// IsColinear (hmg)
//
function IsColinear(const V1, V2: TVector): Boolean; overload;
var
  a, b, c: single;
begin
  a      := VectorDotProduct(V1, V1);
  b      := VectorDotProduct(V1, V2);
  c      := VectorDotProduct(V2, V2);
  result := (a * c - b * b) < cColinearBias;
end;

// SetMatrix (single->double)
//
procedure SetMatrix(var dest: THomogeneousDblMatrix; const src: TMatrix);
var
  i: Integer;
begin
  for i := X to W do
  begin
    dest.V[i].X := src.V[i].X;
    dest.V[i].Y := src.V[i].Y;
    dest.V[i].Z := src.V[i].Z;
    dest.V[i].W := src.V[i].W;
  end;
end;

// SetMatrix (hmg->affine)
//
procedure SetMatrix(var dest: TAffineMatrix; const src: TMatrix);
begin
  dest.V[0].X := src.V[0].X;
  dest.V[0].Y := src.V[0].Y;
  dest.V[0].Z := src.V[0].Z;
  dest.V[1].X := src.V[1].X;
  dest.V[1].Y := src.V[1].Y;
  dest.V[1].Z := src.V[1].Z;
  dest.V[2].X := src.V[2].X;
  dest.V[2].Y := src.V[2].Y;
  dest.V[2].Z := src.V[2].Z;
end;

// SetMatrix (affine->hmg)
//
procedure SetMatrix(var dest: TMatrix; const src: TAffineMatrix);
begin
  dest.V[0].X := src.V[0].X;
  dest.V[0].Y := src.V[0].Y;
  dest.V[0].Z := src.V[0].Z;
  dest.V[0].W := 0;
  dest.V[1].X := src.V[1].X;
  dest.V[1].Y := src.V[1].Y;
  dest.V[1].Z := src.V[1].Z;
  dest.V[1].W := 0;
  dest.V[2].X := src.V[2].X;
  dest.V[2].Y := src.V[2].Y;
  dest.V[2].Z := src.V[2].Z;
  dest.V[2].W := 0;
  dest.V[3].X := 0;
  dest.V[3].Y := 0;
  dest.V[3].Z := 0;
  dest.V[3].W := 1;
end;

// SetMatrixRow
//
procedure SetMatrixRow(var dest: TMatrix; rowNb: Integer; const aRow: TVector);
begin
  dest.V[0].V[rowNb] := aRow.V[0];
  dest.V[1].V[rowNb] := aRow.V[1];
  dest.V[2].V[rowNb] := aRow.V[2];
  dest.V[3].V[rowNb] := aRow.V[3];
end;

// CreateScaleMatrix (affine)
//
function CreateScaleMatrix(const V: TAffineVector): TMatrix;
begin
  result     := IdentityHmgMatrix;
  result.X.X := V.V[X];
  result.Y.Y := V.V[Y];
  result.Z.Z := V.V[Z];
end;

// CreateScaleMatrix (Hmg)
//
function CreateScaleMatrix(const V: TVector): TMatrix;
begin
  result     := IdentityHmgMatrix;
  result.X.X := V.V[X];
  result.Y.Y := V.V[Y];
  result.Z.Z := V.V[Z];
end;

// CreateTranslationMatrix (affine)
//
function CreateTranslationMatrix(const V: TAffineVector): TMatrix;
begin
  result     := IdentityHmgMatrix;
  result.W.X := V.V[X];
  result.W.Y := V.V[Y];
  result.W.Z := V.V[Z];
end;

// CreateTranslationMatrix (hmg)
//
function CreateTranslationMatrix(const V: TVector): TMatrix;
begin
  result     := IdentityHmgMatrix;
  result.W.X := V.V[X];
  result.W.Y := V.V[Y];
  result.W.Z := V.V[Z];
end;

// CreateScaleAndTranslationMatrix
//
function CreateScaleAndTranslationMatrix(const scale, offset: TVector): TMatrix;
begin
  result     := IdentityHmgMatrix;
  result.X.X := scale.V[X];
  result.W.X := offset.V[X];
  result.Y.Y := scale.V[Y];
  result.W.Y := offset.V[Y];
  result.Z.Z := scale.V[Z];
  result.W.Z := offset.V[Z];
end;

// CreateRotationMatrixX
//
function CreateRotationMatrixX(const sine, cosine: single): TMatrix;
begin
  result     := EmptyHmgMatrix;
  result.X.X := 1;
  result.Y.Y := cosine;
  result.Y.Z := sine;
  result.Z.Y := -sine;
  result.Z.Z := cosine;
  result.W.W := 1;
end;

// CreateRotationMatrixX
//
function CreateRotationMatrixX(const angle: single): TMatrix;
var
  S, c: single;
begin
  VectorGeometry.SinCos(angle, S, c);
  result := CreateRotationMatrixX(S, c);
end;

// CreateRotationMatrixY
//
function CreateRotationMatrixY(const sine, cosine: single): TMatrix;
begin
  result     := EmptyHmgMatrix;
  result.X.X := cosine;
  result.X.Z := -sine;
  result.Y.Y := 1;
  result.Z.X := sine;
  result.Z.Z := cosine;
  result.W.W := 1;
end;

// CreateRotationMatrixY
//
function CreateRotationMatrixY(const angle: single): TMatrix;
var
  S, c: single;
begin
  VectorGeometry.SinCos(angle, S, c);
  result := CreateRotationMatrixY(S, c);
end;

// CreateRotationMatrixZ
//
function CreateRotationMatrixZ(const sine, cosine: single): TMatrix;
begin
  result     := EmptyHmgMatrix;
  result.X.X := cosine;
  result.X.Y := sine;
  result.Y.X := -sine;
  result.Y.Y := cosine;
  result.Z.Z := 1;
  result.W.W := 1;
end;

// CreateRotationMatrixZ
//
function CreateRotationMatrixZ(const angle: single): TMatrix;
var
  S, c: single;
begin
  VectorGeometry.SinCos(angle, S, c);
  result := CreateRotationMatrixZ(S, c);
end;

// CreateRotationMatrix (affine)
//
function CreateRotationMatrix(const anAxis: TAffineVector; angle: single): TMatrix;
var
  axis                          : TAffineVector;
  cosine, sine, one_minus_cosine: single;
begin
  VectorGeometry.SinCos(angle, sine, cosine);
  one_minus_cosine := 1 - cosine;
  axis             := VectorNormalize(anAxis);

  result.X.X := (one_minus_cosine * axis.V[0] * axis.V[0]) + cosine;
  result.X.Y := (one_minus_cosine * axis.V[0] * axis.V[1]) - (axis.V[2] * sine);
  result.X.Z := (one_minus_cosine * axis.V[2] * axis.V[0]) + (axis.V[1] * sine);
  result.X.W := 0;

  result.Y.X := (one_minus_cosine * axis.V[0] * axis.V[1]) + (axis.V[2] * sine);
  result.Y.Y := (one_minus_cosine * axis.V[1] * axis.V[1]) + cosine;
  result.Y.Z := (one_minus_cosine * axis.V[1] * axis.V[2]) - (axis.V[0] * sine);
  result.Y.W := 0;

  result.Z.X := (one_minus_cosine * axis.V[2] * axis.V[0]) + (axis.V[1] * sine);
  result.Z.Y := (one_minus_cosine * axis.V[1] * axis.V[2]) - (axis.V[0] * sine);
  result.Z.Z := (one_minus_cosine * axis.V[2] * axis.V[2]) + cosine;
  result.Z.W := 0;

  result.W.X := 0;
  result.W.Y := 0;
  result.W.Z := 0;
  result.W.W := 1;
end;

// CreateRotationMatrix (hmg)
//
function CreateRotationMatrix(const anAxis: TVector; angle: single): TMatrix;
begin
  result := CreateRotationMatrix(PAffineVector(@anAxis)^, angle);
end;

// CreateAffineRotationMatrix
//
function CreateAffineRotationMatrix(const anAxis: TAffineVector; angle: single): TAffineMatrix;
var
  axis                          : TAffineVector;
  cosine, sine, one_minus_cosine: single;
begin
  VectorGeometry.SinCos(angle, sine, cosine);
  one_minus_cosine := 1 - cosine;
  axis             := VectorNormalize(anAxis);

  result.X.X := (one_minus_cosine * Sqr(axis.V[0])) + cosine;
  result.X.Y := (one_minus_cosine * axis.V[0] * axis.V[1]) - (axis.V[2] * sine);
  result.X.Z := (one_minus_cosine * axis.V[2] * axis.V[0]) + (axis.V[1] * sine);

  result.Y.X := (one_minus_cosine * axis.V[0] * axis.V[1]) + (axis.V[2] * sine);
  result.Y.Y := (one_minus_cosine * Sqr(axis.V[1])) + cosine;
  result.Y.Z := (one_minus_cosine * axis.V[1] * axis.V[2]) - (axis.V[0] * sine);

  result.Z.X := (one_minus_cosine * axis.V[2] * axis.V[0]) - (axis.V[1] * sine);
  result.Z.Y := (one_minus_cosine * axis.V[1] * axis.V[2]) + (axis.V[0] * sine);
  result.Z.Z := (one_minus_cosine * Sqr(axis.V[2])) + cosine;
end;

// MatrixMultiply (3x3 func)
//
function MatrixMultiply(const m1, m2: TAffineMatrix): TAffineMatrix;
begin
{$IFNDEF GEOMETRY_NO_ASM}
  if vSIMD = 1 then
  begin
    asm
      db $0F,$0E               /// femms
      xchg eax, ecx

      db $0F,$6E,$7A,$08       /// movd        mm7,[edx+8]
      db $0F,$6E,$72,$20       /// movd        mm6,[edx+32]
      db $0F,$62,$7A,$14       /// punpckldq   mm7,[edx+20]
      db $0F,$6F,$01           /// movq        mm0,[ecx]
      db $0F,$6E,$59,$08       /// movd        mm3,[ecx+8]
      db $0F,$6F,$C8           /// movq        mm1,mm0
      db $0F,$0F,$C7,$B4       /// pfmul       mm0,mm7
      db $0F,$6F,$D1           /// movq        mm2,mm1
      db $0F,$62,$C9           /// punpckldq   mm1,mm1
      db $0F,$0F,$0A,$B4       /// pfmul       mm1,[edx]
      db $0F,$6A,$D2           /// punpckhdq   mm2,mm2
      db $0F,$0F,$52,$0C,$B4   /// pfmul       mm2,[edx+12]
      db $0F,$0F,$C0,$AE       /// pfacc       mm0,mm0
      db $0F,$6F,$E3           /// movq        mm4,mm3
      db $0F,$62,$DB           /// punpckldq   mm3,mm3
      db $0F,$0F,$5A,$18,$B4   /// pfmul       mm3,[edx+24]
      db $0F,$0F,$D1,$9E       /// pfadd       mm2,mm1
      db $0F,$0F,$E6,$B4       /// pfmul       mm4,mm6
      db $0F,$6F,$69,$0C       /// movq        mm5,[ecx+12]
      db $0F,$0F,$D3,$9E       /// pfadd       mm2,mm3
      db $0F,$6E,$59,$14       /// movd        mm3,[ecx+20]
      db $0F,$0F,$E0,$9E       /// pfadd       mm4,mm0
      db $0F,$6F,$CD           /// movq        mm1,mm5
      db $0F,$7F,$10           /// movq        [eax],mm2
      db $0F,$0F,$EF,$B4       /// pfmul       mm5,mm7
      db $0F,$7E,$60,$08       /// movd        [eax+8],mm4
      db $0F,$6F,$D1           /// movq        mm2,mm1
      db $0F,$62,$C9           /// punpckldq   mm1,mm1
      db $0F,$6F,$41,$18       /// movq        mm0,[ecx+24]
      db $0F,$0F,$0A,$B4       /// pfmul       mm1,[edx]
      db $0F,$6A,$D2           /// punpckhdq   mm2,mm2
      db $0F,$0F,$52,$0C,$B4   /// pfmul       mm2,[edx+12]
      db $0F,$0F,$ED,$AE       /// pfacc       mm5,mm5
      db $0F,$6F,$E3           /// movq        mm4,mm3
      db $0F,$62,$DB           /// punpckldq   mm3,mm3
      db $0F,$0F,$5A,$18,$B4   /// pfmul       mm3,[edx+24]
      db $0F,$0F,$D1,$9E       /// pfadd       mm2,mm1
      db $0F,$0F,$E6,$B4       /// pfmul       mm4,mm6
      db $0F,$6F,$C8           /// movq        mm1,mm0
      db $0F,$0F,$D3,$9E       /// pfadd       mm2,mm3
      db $0F,$6E,$59,$20       /// movd        mm3,[ecx+32]
      db $0F,$0F,$E5,$9E       /// pfadd       mm4,mm5
      db $0F,$0F,$C7,$B4       /// pfmul       mm0,mm7
      db $0F,$7F,$50,$0C       /// movq        [eax+12],mm2
      db $0F,$6F,$D1           /// movq        mm2,mm1
      db $0F,$7E,$60,$14       /// movd        [eax+20],mm4
      db $0F,$62,$C9           /// punpckldq   mm1,mm1
      db $0F,$0F,$0A,$B4       /// pfmul       mm1,[edx]
      db $0F,$6A,$D2           /// punpckhdq   mm2,mm2
      db $0F,$0F,$52,$0C,$B4   /// pfmul       mm2,[edx+12]
      db $0F,$0F,$C0,$AE       /// pfacc       mm0,mm0
      db $0F,$0F,$F3,$B4       /// pfmul       mm6,mm3
      db $0F,$62,$DB           /// punpckldq   mm3,mm3
      db $0F,$0F,$5A,$18,$B4   /// pfmul       mm3,[edx+24]
      db $0F,$0F,$D1,$9E       /// pfadd       mm2,mm1
      db $0F,$0F,$F0,$9E       /// pfadd       mm6,mm0
      db $0F,$0F,$D3,$9E       /// pfadd       mm2,mm3

      db $0F,$7E,$70,$20       /// movd        [eax+32],mm6
      db $0F,$7F,$50,$18       /// movq        [eax+24],mm2
      db $0F,$0E               /// femms
    end;
  end
  else {$ENDIF} begin
    result.X.X := m1.X.X * m2.X.X + m1.X.Y * m2.Y.X + m1.X.Z * m2.Z.X;
    result.X.Y := m1.X.X * m2.X.Y + m1.X.Y * m2.Y.Y + m1.X.Z * m2.Z.Y;
    result.X.Z := m1.X.X * m2.X.Z + m1.X.Y * m2.Y.Z + m1.X.Z * m2.Z.Z;
    result.Y.X := m1.Y.X * m2.X.X + m1.Y.Y * m2.Y.X + m1.Y.Z * m2.Z.X;
    result.Y.Y := m1.Y.X * m2.X.Y + m1.Y.Y * m2.Y.Y + m1.Y.Z * m2.Z.Y;
    result.Y.Z := m1.Y.X * m2.X.Z + m1.Y.Y * m2.Y.Z + m1.Y.Z * m2.Z.Z;
    result.Z.X := m1.Z.X * m2.X.X + m1.Z.Y * m2.Y.X + m1.Z.Z * m2.Z.X;
    result.Z.Y := m1.Z.X * m2.X.Y + m1.Z.Y * m2.Y.Y + m1.Z.Z * m2.Z.Y;
    result.Z.Z := m1.Z.X * m2.X.Z + m1.Z.Y * m2.Y.Z + m1.Z.Z * m2.Z.Z;
  end;
end;

// MatrixMultiply (4x4, func)
//
function MatrixMultiply(const m1, m2: TMatrix): TMatrix;
begin
{$IFNDEF GEOMETRY_NO_ASM}
  if vSIMD = 1 then
  begin
    asm
      xchg eax, ecx
      db $0F,$6F,$01           /// movq        mm0,[ecx]
      db $0F,$6F,$49,$08       /// movq        mm1,[ecx+8]
      db $0F,$6F,$22           /// movq        mm4,[edx]
      db $0F,$6A,$D0           /// punpckhdq   mm2,mm0
      db $0F,$6F,$6A,$10       /// movq        mm5,[edx+16]
      db $0F,$6A,$D9           /// punpckhdq   mm3,mm1
      db $0F,$6F,$72,$20       /// movq        mm6,[edx+32]
      db $0F,$62,$C0           /// punpckldq   mm0,mm0
      db $0F,$62,$C9           /// punpckldq   mm1,mm1
      db $0F,$0F,$E0,$B4       /// pfmul       mm4,mm0
      db $0F,$6A,$D2           /// punpckhdq   mm2,mm2
      db $0F,$0F,$42,$08,$B4   /// pfmul       mm0, [edx+8]
      db $0F,$6F,$7A,$30       /// movq        mm7,[edx+48]
      db $0F,$0F,$EA,$B4       /// pfmul       mm5,mm2
      db $0F,$6A,$DB           /// punpckhdq   mm3,mm3
      db $0F,$0F,$52,$18,$B4   /// pfmul       mm2,[edx+24]
      db $0F,$0F,$F1,$B4       /// pfmul       mm6,mm1
      db $0F,$0F,$EC,$9E       /// pfadd       mm5,mm4
      db $0F,$0F,$4A,$28,$B4   /// pfmul       mm1,[edx+40]
      db $0F,$0F,$D0,$9E       /// pfadd       mm2,mm0
      db $0F,$0F,$FB,$B4       /// pfmul       mm7,mm3
      db $0F,$0F,$F5,$9E       /// pfadd       mm6,mm5
      db $0F,$0F,$5A,$38,$B4   /// pfmul       mm3,[edx+56]
      db $0F,$0F,$D1,$9E       /// pfadd       mm2,mm1
      db $0F,$0F,$FE,$9E       /// pfadd       mm7,mm6
      db $0F,$6F,$41,$10       /// movq        mm0,[ecx+16]
      db $0F,$0F,$DA,$9E       /// pfadd       mm3,mm2
      db $0F,$6F,$49,$18       /// movq        mm1,[ecx+24]
      db $0F,$7F,$38           /// movq        [eax],mm7
      db $0F,$6F,$22           /// movq        mm4,[edx]
      db $0F,$7F,$58,$08       /// movq        [eax+8],mm3

      db $0F,$6A,$D0           /// punpckhdq   mm2,mm0
      db $0F,$6F,$6A,$10       /// movq        mm5,[edx+16]
      db $0F,$6A,$D9           /// punpckhdq   mm3,mm1
      db $0F,$6F,$72,$20       /// movq        mm6,[edx+32]
      db $0F,$62,$C0           /// punpckldq   mm0,mm0
      db $0F,$62,$C9           /// punpckldq   mm1,mm1
      db $0F,$0F,$E0,$B4       /// pfmul       mm4,mm0
      db $0F,$6A,$D2           /// punpckhdq   mm2,mm2
      db $0F,$0F,$42,$08,$B4   /// pfmul       mm0,[edx+8]
      db $0F,$6F,$7A,$30       /// movq        mm7,[edx+48]
      db $0F,$0F,$EA,$B4       /// pfmul       mm5,mm2
      db $0F,$6A,$DB           /// punpckhdq   mm3,mm3
      db $0F,$0F,$52,$18,$B4   /// pfmul       mm2,[edx+24]
      db $0F,$0F,$F1,$B4       /// pfmul       mm6,mm1
      db $0F,$0F,$EC,$9E       /// pfadd       mm5,mm4
      db $0F,$0F,$4A,$28,$B4   /// pfmul       mm1,[edx+40]
      db $0F,$0F,$D0,$9E       /// pfadd       mm2,mm0
      db $0F,$0F,$FB,$B4       /// pfmul       mm7,mm3
      db $0F,$0F,$F5,$9E       /// pfadd       mm6,mm5
      db $0F,$0F,$5A,$38,$B4   /// pfmul       mm3,[edx+56]
      db $0F,$0F,$D1,$9E       /// pfadd       mm2,mm1
      db $0F,$0F,$FE,$9E       /// pfadd       mm7,mm6
      db $0F,$6F,$41,$20       /// movq        mm0,[ecx+32]
      db $0F,$0F,$DA,$9E       /// pfadd       mm3,mm2
      db $0F,$6F,$49,$28       /// movq        mm1,[ecx+40]
      db $0F,$7F,$78,$10       /// movq        [eax+16],mm7
      db $0F,$6F,$22           /// movq        mm4,[edx]
      db $0F,$7F,$58,$18       /// movq        [eax+24],mm3

      db $0F,$6A,$D0           /// punpckhdq   mm2,mm0
      db $0F,$6F,$6A,$10       /// movq        mm5,[edx+16]
      db $0F,$6A,$D9           /// punpckhdq   mm3,mm1
      db $0F,$6F,$72,$20       /// movq        mm6,[edx+32]
      db $0F,$62,$C0           /// punpckldq   mm0,mm0
      db $0F,$62,$C9           /// punpckldq   mm1,mm1
      db $0F,$0F,$E0,$B4       /// pfmul       mm4,mm0
      db $0F,$6A,$D2           /// punpckhdq   mm2,mm2
      db $0F,$0F,$42,$08,$B4   /// pfmul       mm0,[edx+8]
      db $0F,$6F,$7A,$30       /// movq        mm7,[edx+48]
      db $0F,$0F,$EA,$B4       /// pfmul       mm5,mm2
      db $0F,$6A,$DB           /// punpckhdq   mm3,mm3
      db $0F,$0F,$52,$18,$B4   /// pfmul       mm2,[edx+24]
      db $0F,$0F,$F1,$B4       /// pfmul       mm6,mm1
      db $0F,$0F,$EC,$9E       /// pfadd       mm5,mm4
      db $0F,$0F,$4A,$28,$B4   /// pfmul       mm1,[edx+40]
      db $0F,$0F,$D0,$9E       /// pfadd       mm2,mm0
      db $0F,$0F,$FB,$B4       /// pfmul       mm7,mm3
      db $0F,$0F,$F5,$9E       /// pfadd       mm6,mm5
      db $0F,$0F,$5A,$38,$B4   /// pfmul       mm3,[edx+56]
      db $0F,$0F,$D1,$9E       /// pfadd       mm2,mm1
      db $0F,$0F,$FE,$9E       /// pfadd       mm7,mm6
      db $0F,$6F,$41,$30       /// movq        mm0,[ecx+48]
      db $0F,$0F,$DA,$9E       /// pfadd       mm3,mm2
      db $0F,$6F,$49,$38       /// movq        mm1,[ecx+56]
      db $0F,$7F,$78,$20       /// movq        [eax+32],mm7
      db $0F,$6F,$22           /// movq        mm4,[edx]
      db $0F,$7F,$58,$28       /// movq        [eax+40],mm3

      db $0F,$6A,$D0           /// punpckhdq   mm2,mm0
      db $0F,$6F,$6A,$10       /// movq        mm5,[edx+16]
      db $0F,$6A,$D9           /// punpckhdq   mm3,mm1
      db $0F,$6F,$72,$20       /// movq        mm6,[edx+32]
      db $0F,$62,$C0           /// punpckldq   mm0,mm0
      db $0F,$62,$C9           /// punpckldq   mm1,mm1
      db $0F,$0F,$E0,$B4       /// pfmul       mm4,mm0
      db $0F,$6A,$D2           /// punpckhdq   mm2,mm2
      db $0F,$0F,$42,$08,$B4   /// pfmul       mm0,[edx+8]
      db $0F,$6F,$7A,$30       /// movq        mm7,[edx+48]
      db $0F,$0F,$EA,$B4       /// pfmul       mm5,mm2
      db $0F,$6A,$DB           /// punpckhdq   mm3,mm3
      db $0F,$0F,$52,$18,$B4   /// pfmul       mm2,[edx+24]
      db $0F,$0F,$F1,$B4       /// pfmul       mm6,mm1
      db $0F,$0F,$EC,$9E       /// pfadd       mm5,mm4
      db $0F,$0F,$4A,$28,$B4   /// pfmul       mm1,[edx+40]
      db $0F,$0F,$D0,$9E       /// pfadd       mm2,mm0
      db $0F,$0F,$FB,$B4       /// pfmul       mm7,mm3
      db $0F,$0F,$F5,$9E       /// pfadd       mm6,mm5
      db $0F,$0F,$5A,$38,$B4   /// pfmul       mm3,[edx+56]
      db $0F,$0F,$D1,$9E       /// pfadd       mm2,mm1
      db $0F,$0F,$FE,$9E       /// pfadd       mm7,mm6
      db $0F,$0F,$DA,$9E       /// pfadd       mm3,mm2
      db $0F,$7F,$78,$30       /// movq        [eax+48],mm7
      db $0F,$7F,$58,$38       /// movq        [eax+56],mm3
      db $0F,$0E               /// femms
    end;
  end
  else {$ENDIF} begin
    result.X.X := m1.X.X * m2.X.X + m1.X.Y * m2.Y.X + m1.X.Z * m2.Z.X + m1.X.W * m2.W.X;
    result.X.Y := m1.X.X * m2.X.Y + m1.X.Y * m2.Y.Y + m1.X.Z * m2.Z.Y + m1.X.W * m2.W.Y;
    result.X.Z := m1.X.X * m2.X.Z + m1.X.Y * m2.Y.Z + m1.X.Z * m2.Z.Z + m1.X.W * m2.W.Z;
    result.X.W := m1.X.X * m2.X.W + m1.X.Y * m2.Y.W + m1.X.Z * m2.Z.W + m1.X.W * m2.W.W;
    result.Y.X := m1.Y.X * m2.X.X + m1.Y.Y * m2.Y.X + m1.Y.Z * m2.Z.X + m1.Y.W * m2.W.X;
    result.Y.Y := m1.Y.X * m2.X.Y + m1.Y.Y * m2.Y.Y + m1.Y.Z * m2.Z.Y + m1.Y.W * m2.W.Y;
    result.Y.Z := m1.Y.X * m2.X.Z + m1.Y.Y * m2.Y.Z + m1.Y.Z * m2.Z.Z + m1.Y.W * m2.W.Z;
    result.Y.W := m1.Y.X * m2.X.W + m1.Y.Y * m2.Y.W + m1.Y.Z * m2.Z.W + m1.Y.W * m2.W.W;
    result.Z.X := m1.Z.X * m2.X.X + m1.Z.Y * m2.Y.X + m1.Z.Z * m2.Z.X + m1.Z.W * m2.W.X;
    result.Z.Y := m1.Z.X * m2.X.Y + m1.Z.Y * m2.Y.Y + m1.Z.Z * m2.Z.Y + m1.Z.W * m2.W.Y;
    result.Z.Z := m1.Z.X * m2.X.Z + m1.Z.Y * m2.Y.Z + m1.Z.Z * m2.Z.Z + m1.Z.W * m2.W.Z;
    result.Z.W := m1.Z.X * m2.X.W + m1.Z.Y * m2.Y.W + m1.Z.Z * m2.Z.W + m1.Z.W * m2.W.W;
    result.W.X := m1.W.X * m2.X.X + m1.W.Y * m2.Y.X + m1.W.Z * m2.Z.X + m1.W.W * m2.W.X;
    result.W.Y := m1.W.X * m2.X.Y + m1.W.Y * m2.Y.Y + m1.W.Z * m2.Z.Y + m1.W.W * m2.W.Y;
    result.W.Z := m1.W.X * m2.X.Z + m1.W.Y * m2.Y.Z + m1.W.Z * m2.Z.Z + m1.W.W * m2.W.Z;
    result.W.W := m1.W.X * m2.X.W + m1.W.Y * m2.Y.W + m1.W.Z * m2.Z.W + m1.W.W * m2.W.W;
  end;
end;

// MatrixMultiply (4x4, proc)
//
procedure MatrixMultiply(const m1, m2: TMatrix; var MResult: TMatrix);
begin
  MResult := MatrixMultiply(m1, m2);
end;

// VectorTransform
//
function VectorTransform(const V: TVector; const M: TMatrix): TVector;
begin
{$IFNDEF GEOMETRY_NO_ASM}
  if vSIMD = 1 then
  begin
    asm
      db $0F,$6F,$00           /// movq        mm0,[eax]
      db $0F,$6F,$48,$08       /// movq        mm1,[eax+8]
      db $0F,$6F,$22           /// movq        mm4,[edx]
      db $0F,$6A,$D0           /// punpckhdq   mm2,mm0
      db $0F,$6F,$6A,$10       /// movq        mm5,[edx+16]
      db $0F,$62,$C0           /// punpckldq   mm0,mm0
      db $0F,$6F,$72,$20       /// movq        mm6,[edx+32]
      db $0F,$0F,$E0,$B4       /// pfmul       mm4,mm0
      db $0F,$6F,$7A,$30       /// movq        mm7,[edx+48]
      db $0F,$6A,$D2           /// punpckhdq   mm2,mm2
      db $0F,$6A,$D9           /// punpckhdq   mm3,mm1
      db $0F,$0F,$EA,$B4       /// pfmul       mm5,mm2
      db $0F,$62,$C9           /// punpckldq   mm1,mm1
      db $0F,$0F,$42,$08,$B4   /// pfmul       mm0,[edx+8]
      db $0F,$6A,$DB           /// punpckhdq   mm3,mm3
      db $0F,$0F,$52,$18,$B4   /// pfmul       mm2,[edx+24]
      db $0F,$0F,$F1,$B4       /// pfmul       mm6,mm1
      db $0F,$0F,$EC,$9E       /// pfadd       mm5,mm4
      db $0F,$0F,$4A,$28,$B4   /// pfmul       mm1,[edx+40]
      db $0F,$0F,$D0,$9E       /// pfadd       mm2,mm0
      db $0F,$0F,$FB,$B4       /// pfmul       mm7,mm3
      db $0F,$0F,$F5,$9E       /// pfadd       mm6,mm5
      db $0F,$0F,$5A,$38,$B4   /// pfmul       mm3,[edx+56]
      db $0F,$0F,$D1,$9E       /// pfadd       mm2,mm1
      db $0F,$0F,$FE,$9E       /// pfadd       mm7,mm6
      db $0F,$0F,$DA,$9E       /// pfadd       mm3,mm2

      db $0F,$7F,$39           /// movq        [ecx],mm7
      db $0F,$7F,$59,$08       /// movq        [ecx+8],mm3
      db $0F,$0E               /// femms
    end
  end
  else {$ENDIF} begin
    result.V[X] := V.V[X] * M.X.X + V.V[Y] * M.Y.X + V.V[Z] * M.Z.X + V.V[W] * M.W.X;
    result.V[Y] := V.V[X] * M.X.Y + V.V[Y] * M.Y.Y + V.V[Z] * M.Z.Y + V.V[W] * M.W.Y;
    result.V[Z] := V.V[X] * M.X.Z + V.V[Y] * M.Y.Z + V.V[Z] * M.Z.Z + V.V[W] * M.W.Z;
    result.V[W] := V.V[X] * M.X.W + V.V[Y] * M.Y.W + V.V[Z] * M.Z.W + V.V[W] * M.W.W;
  end;
end;

// VectorTransform
//
function VectorTransform(const V: TVector; const M: TAffineMatrix): TVector;
begin
  result.V[X] := V.V[X] * M.X.X + V.V[Y] * M.Y.X + V.V[Z] * M.Z.X;
  result.V[Y] := V.V[X] * M.X.Y + V.V[Y] * M.Y.Y + V.V[Z] * M.Z.Y;
  result.V[Z] := V.V[X] * M.X.Z + V.V[Y] * M.Y.Z + V.V[Z] * M.Z.Z;
  result.V[W] := V.V[W];
end;

// VectorTransform
//
function VectorTransform(const V: TAffineVector; const M: TMatrix): TAffineVector;
begin
  result.V[X] := V.V[X] * M.X.X + V.V[Y] * M.Y.X + V.V[Z] * M.Z.X + M.W.X;
  result.V[Y] := V.V[X] * M.X.Y + V.V[Y] * M.Y.Y + V.V[Z] * M.Z.Y + M.W.Y;
  result.V[Z] := V.V[X] * M.X.Z + V.V[Y] * M.Y.Z + V.V[Z] * M.Z.Z + M.W.Z;
end;

// VectorTransform
//
function VectorTransform(const V: TAffineVector; const M: TAffineMatrix): TAffineVector;
begin
{$IFNDEF GEOMETRY_NO_ASM}
  if vSIMD = 1 then
  begin
    asm
      db $0F,$6F,$00           /// movq        mm0,[eax]
      db $0F,$6E,$48,$08       /// movd        mm1,[eax+8]
      db $0F,$6E,$62,$08       /// movd        mm4,[edx+8]
      db $0F,$6F,$D8           /// movq        mm3,mm0
      db $0F,$6E,$52,$20       /// movd        mm2,[edx+32]
      db $0F,$62,$C0           /// punpckldq   mm0,mm0
      db $0F,$62,$62,$14       /// punpckldq   mm4,[edx+20]
      db $0F,$0F,$02,$B4       /// pfmul       mm0,[edx]
      db $0F,$6A,$DB           /// punpckhdq   mm3,mm3
      db $0F,$0F,$D1,$B4       /// pfmul       mm2,mm1
      db $0F,$62,$C9           /// punpckldq   mm1,mm1
      db $0F,$0F,$20,$B4       /// pfmul       mm4,[eax]
      db $0F,$0F,$5A,$0C,$B4   /// pfmul       mm3,[edx+12]
      db $0F,$0F,$4A,$18,$B4   /// pfmul       mm1,[edx+24]
      db $0F,$0F,$E4,$AE       /// pfacc       mm4,mm4
      db $0F,$0F,$D8,$9E       /// pfadd       mm3,mm0
      db $0F,$0F,$E2,$9E       /// pfadd       mm4,mm2
      db $0F,$0F,$D9,$9E       /// pfadd       mm3,mm1

      db $0F,$7E,$61,$08       /// movd        [ecx+8],mm4
      db $0F,$7F,$19           /// movq        [ecx],mm3
      db $0F,$0E               /// femms
    end;
  end
  else {$ENDIF} begin
    result.V[X] := V.V[X] * M.X.X + V.V[Y] * M.Y.X + V.V[Z] * M.Z.X;
    result.V[Y] := V.V[X] * M.X.Y + V.V[Y] * M.Y.Y + V.V[Z] * M.Z.Y;
    result.V[Z] := V.V[X] * M.X.Z + V.V[Y] * M.Y.Z + V.V[Z] * M.Z.Z;
  end;
end;

// MatrixDeterminant (affine)
//
function MatrixDeterminant(const M: TAffineMatrix): single;
begin
  result := M.X.X * (M.Y.Y * M.Z.Z - M.Z.Y * M.Y.Z) - M.X.Y * (M.Y.X * M.Z.Z - M.Z.X * M.Y.Z) +
    M.X.Z * (M.Y.X * M.Z.Y - M.Z.X * M.Y.Y);
end;

// MatrixDetInternal
//
function MatrixDetInternal(const a1, a2, a3, b1, b2, b3, c1, c2, c3: single): single;
// internal version for the determinant of a 3x3 matrix
begin
  result := a1 * (b2 * c3 - b3 * c2) - b1 * (a2 * c3 - a3 * c2) + c1 * (a2 * b3 - a3 * b2);
end;

// MatrixDeterminant (hmg)
//
function MatrixDeterminant(const M: TMatrix): single;
begin
  result := M.X.X * MatrixDetInternal(M.Y.Y, M.Z.Y, M.W.Y, M.Y.Z, M.Z.Z, M.W.Z, M.Y.W, M.Z.W, M.W.W)
    - M.X.Y * MatrixDetInternal(M.Y.X, M.Z.X, M.W.X, M.Y.Z, M.Z.Z, M.W.Z, M.Y.W, M.Z.W, M.W.W) +
    M.X.Z * MatrixDetInternal(M.Y.X, M.Z.X, M.W.X, M.Y.Y, M.Z.Y, M.W.Y, M.Y.W, M.Z.W, M.W.W) - M.X.W
    * MatrixDetInternal(M.Y.X, M.Z.X, M.W.X, M.Y.Y, M.Z.Y, M.W.Y, M.Y.Z, M.Z.Z, M.W.Z);
end;

// AdjointMatrix
//
procedure AdjointMatrix(var M: TMatrix);
var
  a1, a2, a3, a4, b1, b2, b3, b4, c1, c2, c3, c4, d1, d2, d3, d4: single;
begin
  a1 := M.X.X;
  b1 := M.X.Y;
  c1 := M.X.Z;
  d1 := M.X.W;
  a2 := M.Y.X;
  b2 := M.Y.Y;
  c2 := M.Y.Z;
  d2 := M.Y.W;
  a3 := M.Z.X;
  b3 := M.Z.Y;
  c3 := M.Z.Z;
  d3 := M.Z.W;
  a4 := M.W.X;
  b4 := M.W.Y;
  c4 := M.W.Z;
  d4 := M.W.W;

  // row column labeling reversed since we transpose rows & columns
  M.X.X := MatrixDetInternal(b2, b3, b4, c2, c3, c4, d2, d3, d4);
  M.Y.X := -MatrixDetInternal(a2, a3, a4, c2, c3, c4, d2, d3, d4);
  M.Z.X := MatrixDetInternal(a2, a3, a4, b2, b3, b4, d2, d3, d4);
  M.W.X := -MatrixDetInternal(a2, a3, a4, b2, b3, b4, c2, c3, c4);

  M.X.Y := -MatrixDetInternal(b1, b3, b4, c1, c3, c4, d1, d3, d4);
  M.Y.Y := MatrixDetInternal(a1, a3, a4, c1, c3, c4, d1, d3, d4);
  M.Z.Y := -MatrixDetInternal(a1, a3, a4, b1, b3, b4, d1, d3, d4);
  M.W.Y := MatrixDetInternal(a1, a3, a4, b1, b3, b4, c1, c3, c4);

  M.X.Z := MatrixDetInternal(b1, b2, b4, c1, c2, c4, d1, d2, d4);
  M.Y.Z := -MatrixDetInternal(a1, a2, a4, c1, c2, c4, d1, d2, d4);
  M.Z.Z := MatrixDetInternal(a1, a2, a4, b1, b2, b4, d1, d2, d4);
  M.W.Z := -MatrixDetInternal(a1, a2, a4, b1, b2, b4, c1, c2, c4);

  M.X.W := -MatrixDetInternal(b1, b2, b3, c1, c2, c3, d1, d2, d3);
  M.Y.W := MatrixDetInternal(a1, a2, a3, c1, c2, c3, d1, d2, d3);
  M.Z.W := -MatrixDetInternal(a1, a2, a3, b1, b2, b3, d1, d2, d3);
  M.W.W := MatrixDetInternal(a1, a2, a3, b1, b2, b3, c1, c2, c3);
end;

// AdjointMatrix (affine)
//
procedure AdjointMatrix(var M: TAffineMatrix);
var
  a1, a2, a3, b1, b2, b3, c1, c2, c3: single;
begin
  a1    := M.X.X;
  a2    := M.X.Y;
  a3    := M.X.Z;
  b1    := M.Y.X;
  b2    := M.Y.Y;
  b3    := M.Y.Z;
  c1    := M.Z.X;
  c2    := M.Z.Y;
  c3    := M.Z.Z;
  M.X.X := (b2 * c3 - c2 * b3);
  M.Y.X := -(b1 * c3 - c1 * b3);
  M.Z.X := (b1 * c2 - c1 * b2);

  M.X.Y := -(a2 * c3 - c2 * a3);
  M.Y.Y := (a1 * c3 - c1 * a3);
  M.Z.Y := -(a1 * c2 - c1 * a2);

  M.X.Z := (a2 * b3 - b2 * a3);
  M.Y.Z := -(a1 * b3 - b1 * a3);
  M.Z.Z := (a1 * b2 - b1 * a2);
end;

// ScaleMatrix (affine)
//
procedure ScaleMatrix(var M: TAffineMatrix; const factor: single);
var
  i: Integer;
begin
  for i := 0 to 2 do
  begin
    M.V[i].V[0] := M.V[i].V[0] * factor;
    M.V[i].V[1] := M.V[i].V[1] * factor;
    M.V[i].V[2] := M.V[i].V[2] * factor;
  end;
end;

// ScaleMatrix (hmg)
//
procedure ScaleMatrix(var M: TMatrix; const factor: single);
var
  i: Integer;
begin
  for i := 0 to 3 do
  begin
    M.V[i].V[0] := M.V[i].V[0] * factor;
    M.V[i].V[1] := M.V[i].V[1] * factor;
    M.V[i].V[2] := M.V[i].V[2] * factor;
    M.V[i].V[3] := M.V[i].V[3] * factor;
  end;
end;

// TranslateMatrix (affine vec)
//
procedure TranslateMatrix(var M: TMatrix; const V: TAffineVector);
begin
  M.V[3].V[0] := M.V[3].V[0] + V.V[0];
  M.V[3].V[1] := M.V[3].V[1] + V.V[1];
  M.V[3].V[2] := M.V[3].V[2] + V.V[2];
end;

// TranslateMatrix
//
procedure TranslateMatrix(var M: TMatrix; const V: TVector);
begin
  M.V[3].V[0] := M.V[3].V[0] + V.V[0];
  M.V[3].V[1] := M.V[3].V[1] + V.V[1];
  M.V[3].V[2] := M.V[3].V[2] + V.V[2];
end;

// NormalizeMatrix
//
procedure NormalizeMatrix(var M: TMatrix);
begin
  M.V[0].V[3] := 0;
  NormalizeVector(M.V[0]);
  M.V[1].V[3] := 0;
  NormalizeVector(M.V[1]);
  M.V[2] := VectorCrossProduct(M.V[0], M.V[1]);
  M.V[0] := VectorCrossProduct(M.V[1], M.V[2]);
  M.V[3] := WHmgVector;
end;

// TransposeMatrix
//
procedure TransposeMatrix(var M: TAffineMatrix);
var
  f: single;
begin
  f           := M.V[0].V[1];
  M.V[0].V[1] := M.V[1].V[0];
  M.V[1].V[0] := f;
  f           := M.V[0].V[2];
  M.V[0].V[2] := M.V[2].V[0];
  M.V[2].V[0] := f;
  f           := M.V[1].V[2];
  M.V[1].V[2] := M.V[2].V[1];
  M.V[2].V[1] := f;
end;

// TransposeMatrix
//
procedure TransposeMatrix(var M: TMatrix);
var
  f: single;
begin
  f           := M.V[0].V[1];
  M.V[0].V[1] := M.V[1].V[0];
  M.V[1].V[0] := f;
  f           := M.V[0].V[2];
  M.V[0].V[2] := M.V[2].V[0];
  M.V[2].V[0] := f;
  f           := M.V[0].V[3];
  M.V[0].V[3] := M.V[3].V[0];
  M.V[3].V[0] := f;
  f           := M.V[1].V[2];
  M.V[1].V[2] := M.V[2].V[1];
  M.V[2].V[1] := f;
  f           := M.V[1].V[3];
  M.V[1].V[3] := M.V[3].V[1];
  M.V[3].V[1] := f;
  f           := M.V[2].V[3];
  M.V[2].V[3] := M.V[3].V[2];
  M.V[3].V[2] := f;
end;

// InvertMatrix
//
procedure InvertMatrix(var M: TMatrix);
var
  det: single;
begin
  det := MatrixDeterminant(M);
  if Abs(det) < EPSILON then
    M := IdentityHmgMatrix
  else
  begin
    AdjointMatrix(M);
    ScaleMatrix(M, 1 / det);
  end;
end;

// MatrixInvert
//
function MatrixInvert(const M: TMatrix): TMatrix;
begin
  result := M;
  InvertMatrix(result);
end;

// InvertMatrix (affine)
//
procedure InvertMatrix(var M: TAffineMatrix);
var
  det: single;
begin
  det := MatrixDeterminant(M);
  if Abs(det) < EPSILON then
    M := IdentityMatrix
  else
  begin
    AdjointMatrix(M);
    ScaleMatrix(M, 1 / det);
  end;
end;

// MatrixInvert (affine)
//
function MatrixInvert(const M: TAffineMatrix): TAffineMatrix;
begin
  result := M;
  InvertMatrix(result);
end;

// transpose_scale_m33
//
procedure Transpose_Scale_M33(const src: TMatrix; var dest: TMatrix; var scale: single);
// EAX src
// EDX dest
// ECX scale
begin
{$IFNDEF GEOMETRY_NO_ASM}
  asm
    // dest[0][0]:=scale*src[0][0];
    fld   dword ptr [ecx]
    fld   st(0)
    fmul  dword ptr [eax]
    fstp  dword ptr [edx]
    // dest[1][0]:=scale*src[0][1];
    fld   st(0)
    fmul  dword ptr [eax+4]
    fstp  dword ptr [edx+16]
    // dest[2][0]:=scale*src[0][2];
    fmul  dword ptr [eax+8]
    fstp  dword ptr [edx+32]

    // dest[0][1]:=scale*src[1][0];
    fld   dword ptr [ecx]
    fld   st(0)
    fmul  dword ptr [eax+16]
    fstp  dword ptr [edx+4]
    // dest[1][1]:=scale*src[1][1];
    fld   st(0)
    fmul  dword ptr [eax+20]
    fstp  dword ptr [edx+20]
    // dest[2][1]:=scale*src[1][2];
    fmul  dword ptr [eax+24]
    fstp  dword ptr [edx+36]

    // dest[0][2]:=scale*src[2][0];
    fld   dword ptr [ecx]
    fld   st(0)
    fmul  dword ptr [eax+32]
    fstp  dword ptr [edx+8]
    // dest[1][2]:=scale*src[2][1];
    fld   st(0)
    fmul  dword ptr [eax+36]
    fstp  dword ptr [edx+24]
    // dest[2][2]:=scale*src[2][2];
    fmul  dword ptr [eax+40]
    fstp  dword ptr [edx+40]
  end;
{$ELSE}
  dest.V[0].V[0] := scale * src.V[0].V[0];
  dest.V[1].V[0] := scale * src.V[0].V[1];
  dest.V[2].V[0] := scale * src.V[0].V[2];
  dest.V[0].V[1] := scale * src.V[1].V[0];
  dest.V[1].V[1] := scale * src.V[1].V[1];
  dest.V[2].V[1] := scale * src.V[1].V[2];
  dest.V[0].V[2] := scale * src.V[2].V[0];
  dest.V[1].V[2] := scale * src.V[2].V[1];
  dest.V[2].V[2] := scale * src.V[2].V[2];
{$ENDIF}
end;

// AnglePreservingMatrixInvert
//
function AnglePreservingMatrixInvert(const mat: TMatrix): TMatrix;
var
  scale: single;
begin
  scale := VectorNorm(mat.V[0]);

  // Is the submatrix A singular?
  if Abs(scale) < EPSILON then
  begin
    // Matrix M has no inverse
    result := IdentityHmgMatrix;
    Exit;
  end
  else
  begin
    // Calculate the inverse of the square of the isotropic scale factor
    scale := 1.0 / scale;
  end;

  // Fill in last row while CPU is busy with the division
  result.V[0].V[3] := 0.0;
  result.V[1].V[3] := 0.0;
  result.V[2].V[3] := 0.0;
  result.V[3].V[3] := 1.0;

  // Transpose and scale the 3 by 3 upper-left submatrix
  Transpose_Scale_M33(mat, result, scale);

  // Calculate -(transpose(A) / s*s) C
  result.V[3].V[0] := -(result.V[0].V[0] * mat.V[3].V[0] + result.V[1].V[0] * mat.V[3].V[1] +
    result.V[2].V[0] * mat.V[3].V[2]);
  result.V[3].V[1] := -(result.V[0].V[1] * mat.V[3].V[0] + result.V[1].V[1] * mat.V[3].V[1] +
    result.V[2].V[1] * mat.V[3].V[2]);
  result.V[3].V[2] := -(result.V[0].V[2] * mat.V[3].V[0] + result.V[1].V[2] * mat.V[3].V[1] +
    result.V[2].V[2] * mat.V[3].V[2]);
end;

// MatrixDecompose
//
function MatrixDecompose(const M: TMatrix; var Tran: TTransformations): Boolean;
var
  i, J                 : Integer;
  LocMat, pmat, invpmat: TMatrix;
  prhs, psol           : TVector;
  row0, row1, row2     : TAffineVector;
  f                    : single;
begin
  result := False;
  LocMat := M;
  // normalize the matrix
  if LocMat.W.W = 0 then
    Exit;
  for i                := 0 to 3 do
    for J              := 0 to 3 do
      LocMat.V[i].V[J] := LocMat.V[i].V[J] / LocMat.W.W;

  // pmat is used to solve for perspective, but it also provides
  // an easy way to test for singularity of the upper 3x3 component.

  pmat             := LocMat;
  for i            := 0 to 2 do
    pmat.V[i].V[W] := 0;
  pmat.W.W         := 1;

  if MatrixDeterminant(pmat) = 0 then
    Exit;

  // First, isolate perspective.  This is the messiest.
  if (LocMat.X.W <> 0) or (LocMat.Y.W <> 0) or (LocMat.Z.W <> 0) then
  begin
    // prhs is the right hand side of the equation.
    prhs.V[X] := LocMat.X.W;
    prhs.V[Y] := LocMat.Y.W;
    prhs.V[Z] := LocMat.Z.W;
    prhs.V[W] := LocMat.W.W;

    // Solve the equation by inverting pmat and multiplying
    // prhs by the inverse.  (This is the easiest way, not
    // necessarily the best.)

    invpmat := pmat;
    InvertMatrix(invpmat);
    TransposeMatrix(invpmat);
    psol := VectorTransform(prhs, invpmat);

    // stuff the answer away
    Tran[ttPerspectiveX] := psol.V[X];
    Tran[ttPerspectiveY] := psol.V[Y];
    Tran[ttPerspectiveZ] := psol.V[Z];
    Tran[ttPerspectiveW] := psol.V[W];

    // clear the perspective partition
    LocMat.X.W := 0;
    LocMat.Y.W := 0;
    LocMat.Z.W := 0;
    LocMat.W.W := 1;
  end
  else
  begin
    // no perspective
    Tran[ttPerspectiveX] := 0;
    Tran[ttPerspectiveY] := 0;
    Tran[ttPerspectiveZ] := 0;
    Tran[ttPerspectiveW] := 0;
  end;

  // next take care of translation (easy)
  for i := 0 to 2 do
  begin
    Tran[TTransType(Ord(ttTranslateX) + i)] := LocMat.V[W].V[i];
    LocMat.V[W].V[i]                        := 0;
  end;

  // now get scale and shear
  SetVector(row0, LocMat.V[0]);
  SetVector(row1, LocMat.V[1]);
  SetVector(row2, LocMat.V[2]);

  // compute X scale factor and normalize first row
  Tran[ttScaleX] := VectorNorm(row0);
  VectorScale(row0, RSqrt(Tran[ttScaleX]));

  // compute XY shear factor and make 2nd row orthogonal to 1st
  Tran[ttShearXY] := VectorDotProduct(row0, row1);
  f               := -Tran[ttShearXY];
  CombineVector(row1, row0, f);

  // now, compute Y scale and normalize 2nd row
  Tran[ttScaleY] := VectorNorm(row1);
  VectorScale(row1, RSqrt(Tran[ttScaleY]));
  Tran[ttShearXY] := Tran[ttShearXY] / Tran[ttScaleY];

  // compute XZ and YZ shears, orthogonalize 3rd row
  Tran[ttShearXZ] := VectorDotProduct(row0, row2);
  f               := -Tran[ttShearXZ];
  CombineVector(row2, row0, f);
  Tran[ttShearYZ] := VectorDotProduct(row1, row2);
  f               := -Tran[ttShearYZ];
  CombineVector(row2, row1, f);

  // next, get Z scale and normalize 3rd row
  Tran[ttScaleZ] := VectorNorm(row2);
  VectorScale(row2, RSqrt(Tran[ttScaleZ]));
  Tran[ttShearXZ] := Tran[ttShearXZ] / Tran[ttScaleZ];
  Tran[ttShearYZ] := Tran[ttShearYZ] / Tran[ttScaleZ];

  // At this point, the matrix (in rows[]) is orthonormal.
  // Check for a coordinate system flip.  If the determinant
  // is -1, then negate the matrix and the scaling factors.
  if VectorDotProduct(row0, VectorCrossProduct(row1, row2)) < 0 then
  begin
    for i                                 := 0 to 2 do
      Tran[TTransType(Ord(ttScaleX) + i)] := -Tran[TTransType(Ord(ttScaleX) + i)];
    NegateVector(row0);
    NegateVector(row1);
    NegateVector(row2);
  end;

  // now, get the rotations out, as described in the gem
  Tran[ttRotateY] := VectorGeometry.ArcSin(-row0.V[Z]);
  if Cos(Tran[ttRotateY]) <> 0 then
  begin
    Tran[ttRotateX] := VectorGeometry.ArcTan2(row1.V[Z], row2.V[Z]);
    Tran[ttRotateZ] := VectorGeometry.ArcTan2(row0.V[Y], row0.V[X]);
  end
  else
  begin
    Tran[ttRotateX] := VectorGeometry.ArcTan2(row1.V[X], row1.V[Y]);
    Tran[ttRotateZ] := 0;
  end;
  // All done!
  result := True;
end;

function CreateLookAtMatrix(const eye, center, normUp: TVector): TMatrix;
var
  XAxis, YAxis, ZAxis, negEye: TVector;
begin
  ZAxis := VectorSubtract(center, eye);
  NormalizeVector(ZAxis);
  XAxis := VectorCrossProduct(ZAxis, normUp);
  NormalizeVector(XAxis);
  YAxis       := VectorCrossProduct(XAxis, ZAxis);
  result.V[0] := XAxis;
  result.V[1] := YAxis;
  result.V[2] := ZAxis;
  NegateVector(result.V[2]);
  result.V[3] := NullHmgPoint;
  TransposeMatrix(result);
  negEye := eye;
  NegateVector(negEye);
  negEye.V[3] := 1;
  negEye      := VectorTransform(negEye, result);
  result.V[3] := negEye;
end;

function CreateMatrixFromFrustum(Left, Right, Bottom, Top, ZNear, ZFar: single): TMatrix;
begin
  result.V[0].V[0] := 2 * ZNear / (Right - Left);
  result.V[0].V[1] := 0;
  result.V[0].V[2] := 0;
  result.V[0].V[3] := 0;

  result.V[1].V[0] := 0;
  result.V[1].V[1] := 2 * ZNear / (Top - Bottom);
  result.V[1].V[2] := 0;
  result.V[1].V[3] := 0;

  result.V[2].V[0] := (Right + Left) / (Right - Left);
  result.V[2].V[1] := (Top + Bottom) / (Top - Bottom);
  result.V[2].V[2] := -(ZFar + ZNear) / (ZFar - ZNear);
  result.V[2].V[3] := -1;

  result.V[3].V[0] := 0;
  result.V[3].V[1] := 0;
  result.V[3].V[2] := -2 * ZFar * ZNear / (ZFar - ZNear);
  result.V[3].V[3] := 0;
end;

function CreatePerspectiveMatrix(FOV, Aspect, ZNear, ZFar: single): TMatrix;
var
  X, Y: single;
begin
  FOV    := MinFloat(179.9, MaxFloat(0, FOV));
  Y      := ZNear * VectorGeometry.Tan(VectorGeometry.DegToRad(FOV) * 0.5);
  X      := Y * Aspect;
  result := CreateMatrixFromFrustum(-X, X, -Y, Y, ZNear, ZFar);
end;

function CreateOrthoMatrix(Left, Right, Bottom, Top, ZNear, ZFar: single): TMatrix;
begin
  result.V[0].V[0] := 2 / (Right - Left);
  result.V[0].V[1] := 0;
  result.V[0].V[2] := 0;
  result.V[0].V[3] := 0;

  result.V[1].V[0] := 0;
  result.V[1].V[1] := 2 / (Top - Bottom);
  result.V[1].V[2] := 0;
  result.V[1].V[3] := 0;

  result.V[2].V[0] := 0;
  result.V[2].V[1] := 0;
  result.V[2].V[2] := -2 / (ZFar - ZNear);
  result.V[2].V[3] := 0;

  result.V[3].V[0] := (Left + Right) / (Left - Right);
  result.V[3].V[1] := (Bottom + Top) / (Bottom - Top);
  result.V[3].V[2] := (ZNear + ZFar) / (ZNear - ZFar);
  result.V[3].V[3] := 1;
end;

function CreatePickMatrix(X, Y, deltax, deltay: single; const viewport: TVector4i): TMatrix;
begin
  if (deltax <= 0) or (deltay <= 0) then
  begin
    result := IdentityHmgMatrix;
    Exit;
  end;
  // Translate and scale the picked region to the entire window
  result := CreateTranslationMatrix(AffineVectorMake((viewport.V[2] - 2 * (X - viewport.V[0])) /
    deltax, (viewport.V[3] - 2 * (Y - viewport.V[1])) / deltay, 0.0));
  result.V[0].V[0] := viewport.V[2] / deltax;
  result.V[1].V[1] := viewport.V[3] / deltay;
end;

function Project(objectVector: TVector; const ViewProjMatrix: TMatrix; const viewport: TVector4i;
  out WindowVector: TVector): Boolean;
begin
  result            := False;
  objectVector.V[3] := 1.0;
  WindowVector      := VectorTransform(objectVector, ViewProjMatrix);
  if WindowVector.V[3] = 0.0 then
    Exit;
  WindowVector.V[0] := WindowVector.V[0] / WindowVector.V[3];
  WindowVector.V[1] := WindowVector.V[1] / WindowVector.V[3];
  WindowVector.V[2] := WindowVector.V[2] / WindowVector.V[3];
  // Map x, y and z to range 0-1
  WindowVector.V[0] := WindowVector.V[0] * 0.5 + 0.5;
  WindowVector.V[1] := WindowVector.V[1] * 0.5 + 0.5;
  WindowVector.V[2] := WindowVector.V[2] * 0.5 + 0.5;

  // Map x,y to viewport
  WindowVector.V[0] := WindowVector.V[0] * viewport.V[2] + viewport.V[0];
  WindowVector.V[1] := WindowVector.V[1] * viewport.V[3] + viewport.V[1];
  result            := True;
end;

function UnProject(WindowVector: TVector; ViewProjMatrix: TMatrix; const viewport: TVector4i;
  out objectVector: TVector): Boolean;
begin
  result := False;
  InvertMatrix(ViewProjMatrix);
  WindowVector.V[3] := 1.0;
  // Map x and y from window coordinates
  WindowVector.V[0] := (WindowVector.V[0] - viewport.V[0]) / viewport.V[2];
  WindowVector.V[1] := (WindowVector.V[1] - viewport.V[1]) / viewport.V[3];
  // Map to range -1 to 1
  WindowVector.V[0] := WindowVector.V[0] * 2 - 1;
  WindowVector.V[1] := WindowVector.V[1] * 2 - 1;
  WindowVector.V[2] := WindowVector.V[2] * 2 - 1;
  objectVector      := VectorTransform(WindowVector, ViewProjMatrix);
  if objectVector.V[3] = 0.0 then
    Exit;
  objectVector.V[0] := objectVector.V[0] / objectVector.V[3];
  objectVector.V[1] := objectVector.V[1] / objectVector.V[3];
  objectVector.V[2] := objectVector.V[2] / objectVector.V[3];
  result            := True;
end;

// CalcPlaneNormal (func, affine)
//
function CalcPlaneNormal(const p1, p2, p3: TAffineVector): TAffineVector;
var
  V1, V2: TAffineVector;
begin
  VectorSubtract(p2, p1, V1);
  VectorSubtract(p3, p1, V2);
  VectorCrossProduct(V1, V2, result);
  NormalizeVector(result);
end;

// CalcPlaneNormal (proc, affine)
//
procedure CalcPlaneNormal(const p1, p2, p3: TAffineVector; var vr: TAffineVector);
var
  V1, V2: TAffineVector;
begin
  VectorSubtract(p2, p1, V1);
  VectorSubtract(p3, p1, V2);
  VectorCrossProduct(V1, V2, vr);
  NormalizeVector(vr);
end;

// CalcPlaneNormal (proc, hmg)
//
procedure CalcPlaneNormal(const p1, p2, p3: TVector; var vr: TAffineVector); overload;
var
  V1, V2: TVector;
begin
  VectorSubtract(p2, p1, V1);
  VectorSubtract(p3, p1, V2);
  VectorCrossProduct(V1, V2, vr);
  NormalizeVector(vr);
end;

// PlaneMake (point + normal, affine)
//
function PlaneMake(const point, normal: TAffineVector): THmgPlane;
begin
  PAffineVector(@result)^ := normal;
  result.V[3]             := -VectorDotProduct(point, normal);
end;

// PlaneMake (point + normal, hmg)
//
function PlaneMake(const point, normal: TVector): THmgPlane;
begin
  PAffineVector(@result)^ := PAffineVector(@normal)^;
  result.V[3]             := -VectorDotProduct(PAffineVector(@point)^, PAffineVector(@normal)^);
end;

// PlaneMake (3 points, affine)
//
function PlaneMake(const p1, p2, p3: TAffineVector): THmgPlane;
begin
  CalcPlaneNormal(p1, p2, p3, PAffineVector(@result)^);
  result.V[3] := -VectorDotProduct(p1, PAffineVector(@result)^);
end;

// PlaneMake (3 points, hmg)
//
function PlaneMake(const p1, p2, p3: TVector): THmgPlane;
begin
  CalcPlaneNormal(p1, p2, p3, PAffineVector(@result)^);
  result.V[3] := -VectorDotProduct(p1, PAffineVector(@result)^);
end;

// SetPlane
//
procedure SetPlane(var dest: TDoubleHmgPlane; const src: THmgPlane);
begin
  dest.V[0] := src.V[0];
  dest.V[1] := src.V[1];
  dest.V[2] := src.V[2];
  dest.V[3] := src.V[3];
end;

// NormalizePlane
//
procedure NormalizePlane(var plane: THmgPlane);
var
  n: single;
begin
  n := RSqrt(plane.V[0] * plane.V[0] + plane.V[1] * plane.V[1] + plane.V[2] * plane.V[2]);
  ScaleVector(plane, n);
end;

// PlaneEvaluatePoint (affine)
//
function PlaneEvaluatePoint(const plane: THmgPlane; const point: TAffineVector): single;
// EAX contains address of plane
// EDX contains address of point
// result is stored in ST(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [EAX]
  FMUL DWORD PTR [EDX]
  FLD DWORD PTR [EAX + 4]
  FMUL DWORD PTR [EDX + 4]
  FADDP
  FLD DWORD PTR [EAX + 8]
  FMUL DWORD PTR [EDX + 8]
  FADDP
  FLD DWORD PTR [EAX + 12]
  FADDP
  {$ELSE}
begin
  result := plane.V[0] * point.V[0] + plane.V[1] * point.V[1] + plane.V[2] * point.V[2] +
    plane.V[3];
{$ENDIF}
end;

// PlaneEvaluatePoint (hmg)
//
function PlaneEvaluatePoint(const plane: THmgPlane; const point: TVector): single;
// EAX contains address of plane
// EDX contains address of point
// result is stored in ST(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD DWORD PTR [EAX]
  FMUL DWORD PTR [EDX]
  FLD DWORD PTR [EAX + 4]
  FMUL DWORD PTR [EDX + 4]
  FADDP
  FLD DWORD PTR [EAX + 8]
  FMUL DWORD PTR [EDX + 8]
  FADDP
  FLD DWORD PTR [EAX + 12]
  FADDP
  {$ELSE}
begin
  result := plane.V[0] * point.V[0] + plane.V[1] * point.V[1] + plane.V[2] * point.V[2] +
    plane.V[3];
{$ENDIF}
end;

// PointIsInHalfSpace
//
function PointIsInHalfSpace(const point, planePoint, planeNormal: TVector): Boolean;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld   dword ptr [eax]         // 27
  fsub  dword ptr [edx]
  fmul  dword ptr [ecx]
  fld   dword ptr [eax+4]
  fsub  dword ptr [edx+4]
  fmul  dword ptr [ecx+4]
  faddp
  fld   dword ptr [eax+8]
  fsub  dword ptr [edx+8]
  fmul  dword ptr [ecx+8]
  faddp
  ftst
  fstsw ax
  sahf
  setnbe al
  ffree st(0)
  {$ELSE}
begin
  result := (PointPlaneDistance(point, planePoint, planeNormal) > 0); // 44
{$ENDIF}
end;

// PointIsInHalfSpace
//
function PointIsInHalfSpace(const point, planePoint, planeNormal: TAffineVector): Boolean;
begin
  result := (PointPlaneDistance(point, planePoint, planeNormal) > 0);
end;

// PointIsInHalfSpace
//
function PointIsInHalfSpace(const point: TAffineVector; plane: THmgPlane): Boolean;
begin
  result := (PointPlaneDistance(point, plane) > 0);
end;

// PointPlaneDistance
//
function PointPlaneDistance(const point, planePoint, planeNormal: TVector): single;
// EAX contains address of point
// EDX contains address of planepoint
// ECX contains address of planeNormal
// result in St(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld   dword ptr [eax]
  fsub  dword ptr [edx]
  fmul  dword ptr [ecx]

  fld   dword ptr [eax+$4]
  fsub  dword ptr [edx+$4]
  fmul  dword ptr [ecx+$4]
  faddp

  fld   dword ptr [eax+$8]
  fsub  dword ptr [edx+$8]
  fmul  dword ptr [ecx+$8]
  faddp
  {$ELSE}
begin
  result := (point.V[0] - planePoint.V[0]) * planeNormal.V[0] + (point.V[1] - planePoint.V[1]) *
    planeNormal.V[1] + (point.V[2] - planePoint.V[2]) * planeNormal.V[2];
{$ENDIF}
end;

// PointPlaneDistance
//
function PointPlaneDistance(const point, planePoint, planeNormal: TAffineVector): single;
// EAX contains address of point
// EDX contains address of planepoint
// ECX contains address of planeNormal
// result in St(0)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld   dword ptr [eax]
  fsub  dword ptr [edx]
  fmul  dword ptr [ecx]

  fld   dword ptr [eax+$4]
  fsub  dword ptr [edx+$4]
  fmul  dword ptr [ecx+$4]
  faddp

  fld   dword ptr [eax+$8]
  fsub  dword ptr [edx+$8]
  fmul  dword ptr [ecx+$8]
  faddp
  {$ELSE}
begin
  result := (point.V[0] - planePoint.V[0]) * planeNormal.V[0] + (point.V[1] - planePoint.V[1]) *
    planeNormal.V[1] + (point.V[2] - planePoint.V[2]) * planeNormal.V[2];
{$ENDIF}
end;

// PointPlaneDistance
//
function PointPlaneDistance(const point: TAffineVector; plane: THmgPlane): single;
begin
  result := PlaneEvaluatePoint(plane, point);
end;

// PointPlaneOrthoProjection
//
function PointPlaneOrthoProjection(const point: TAffineVector; const plane: THmgPlane;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;
var
  h     : single;
  normal: TAffineVector;
begin
  result := False;

  h := PointPlaneDistance(point, plane);

  if (not bothface) and (h < 0) then
    Exit;

  normal := Vector3fMake(plane);
  inter  := VectorAdd(point, VectorScale(normal, -h));
  result := True;
end;

// PointPlaneProjection
//
function PointPlaneProjection(const point, direction: TAffineVector; const plane: THmgPlane;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;
var
  h, dot: single;
  normal: TAffineVector;
begin
  result := False;

  normal := Vector3fMake(plane);
  dot    := VectorDotProduct(VectorNormalize(direction), normal);

  if (not bothface) and (dot > 0) then
    Exit;

  if Abs(dot) >= 0.000000001 then
  begin
    h      := PointPlaneDistance(point, plane);
    inter  := VectorAdd(point, VectorScale(direction, -h / dot));
    result := True;
  end;
end;

// SegmentPlaneIntersection
//
function SegmentPlaneIntersection(const ptA, ptB: TAffineVector; const plane: THmgPlane;
  var inter: TAffineVector): Boolean;
var
  hA, hB, dot      : single;
  normal, direction: TVector3f;
begin
  result := False;
  hA     := PointPlaneDistance(ptA, plane);
  hB     := PointPlaneDistance(ptB, plane);
  if hA * hB <= 0 then
  begin
    normal    := Vector3fMake(plane);
    direction := VectorNormalize(VectorSubtract(ptB, ptA));
    dot       := VectorDotProduct(direction, normal);
    if Abs(dot) >= 0.000000001 then
    begin
      inter  := VectorAdd(ptA, VectorScale(direction, -hA / dot));
      result := True;
    end;
  end;
end;

// PointTriangleOrthoProjection
//
function PointTriangleOrthoProjection(const point, ptA, ptB, ptC: TAffineVector;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;
var
  plane: THmgPlane;
begin
  result := False;

  plane := PlaneMake(ptA, ptB, ptC);
  if not IsLineIntersectTriangle(point, Vector3fMake(plane), ptA, ptB, ptC) then
    Exit;

  result := PointPlaneOrthoProjection(point, plane, inter, bothface);
end;

// PointTriangleProjection
//
function PointTriangleProjection(const point, direction, ptA, ptB, ptC: TAffineVector;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;
var
  plane: THmgPlane;
begin
  result := False;

  if not IsLineIntersectTriangle(point, direction, ptA, ptB, ptC) then
    Exit;

  plane  := PlaneMake(ptA, ptB, ptC);
  result := PointPlaneProjection(point, direction, plane, inter, bothface);
end;

// IsLineIntersectTriangle
//
function IsLineIntersectTriangle(const point, direction, ptA, ptB, ptC: TAffineVector): Boolean;
var
  PA, PB, PC               : TAffineVector;
  crossAB, crossBC, crossCA: TAffineVector;
begin
  result := False;

  PA := VectorSubtract(ptA, point);
  PB := VectorSubtract(ptB, point);
  PC := VectorSubtract(ptC, point);

  crossAB := VectorCrossProduct(PA, PB);
  crossBC := VectorCrossProduct(PB, PC);

  if VectorDotProduct(crossAB, direction) > 0 then
  begin
    if VectorDotProduct(crossBC, direction) > 0 then
    begin
      crossCA := VectorCrossProduct(PC, PA);
      if VectorDotProduct(crossCA, direction) > 0 then
        result := True;
    end;
  end
  else if VectorDotProduct(crossBC, direction) < 0 then
  begin
    crossCA := VectorCrossProduct(PC, PA);
    if VectorDotProduct(crossCA, direction) < 0 then
      result := True;
  end
end;

// PointQuadOrthoProjection
//
function PointQuadOrthoProjection(const point, ptA, ptB, ptC, ptD: TAffineVector;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;
var
  plane: THmgPlane;
begin
  result := False;

  plane := PlaneMake(ptA, ptB, ptC);
  if not IsLineIntersectQuad(point, Vector3fMake(plane), ptA, ptB, ptC, ptD) then
    Exit;

  result := PointPlaneOrthoProjection(point, plane, inter, bothface);
end;

// PointQuadProjection
//
function PointQuadProjection(const point, direction, ptA, ptB, ptC, ptD: TAffineVector;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;
var
  plane: THmgPlane;
begin
  result := False;

  if not IsLineIntersectQuad(point, direction, ptA, ptB, ptC, ptD) then
    Exit;

  plane  := PlaneMake(ptA, ptB, ptC);
  result := PointPlaneProjection(point, direction, plane, inter, bothface);
end;

// IsLineIntersectQuad
//
function IsLineIntersectQuad(const point, direction, ptA, ptB, ptC, ptD: TAffineVector): Boolean;
var
  PA, PB, PC, PD                    : TAffineVector;
  crossAB, crossBC, crossCD, crossDA: TAffineVector;
begin
  result := False;

  PA := VectorSubtract(ptA, point);
  PB := VectorSubtract(ptB, point);
  PC := VectorSubtract(ptC, point);
  PD := VectorSubtract(ptD, point);

  crossAB := VectorCrossProduct(PA, PB);
  crossBC := VectorCrossProduct(PB, PC);

  if VectorDotProduct(crossAB, direction) > 0 then
  begin
    if VectorDotProduct(crossBC, direction) > 0 then
    begin
      crossCD := VectorCrossProduct(PC, PD);
      if VectorDotProduct(crossCD, direction) > 0 then
      begin
        crossDA := VectorCrossProduct(PD, PA);
        if VectorDotProduct(crossDA, direction) > 0 then
          result := True;
      end;
    end;
  end
  else if VectorDotProduct(crossBC, direction) < 0 then
  begin
    crossCD := VectorCrossProduct(PC, PD);
    if VectorDotProduct(crossCD, direction) < 0 then
    begin
      crossDA := VectorCrossProduct(PD, PA);
      if VectorDotProduct(crossDA, direction) < 0 then
        result := True;
    end;
  end
end;

// PointDiskOrthoProjection
//
function PointDiskOrthoProjection(const point, center, up: TAffineVector; const radius: single;
  var inter: TAffineVector; bothface: Boolean = True): Boolean;
begin
  if PointPlaneOrthoProjection(point, PlaneMake(center, up), inter, bothface) then
    result := (VectorDistance2(inter, center) <= radius * radius)
  else
    result := False;
end;

// PointDiskProjection
//
function PointDiskProjection(const point, direction, center, up: TAffineVector;
  const radius: single; var inter: TAffineVector; bothface: Boolean = True): Boolean;
begin
  if PointPlaneProjection(point, direction, PlaneMake(center, up), inter, bothface) then
    result := VectorDistance2(inter, center) <= radius * radius
  else
    result := False;
end;

// PointLineClosestPoint
//
function PointLineClosestPoint(const point, linePoint, lineDirection: TAffineVector): TAffineVector;
var
  W        : TAffineVector;
  c1, c2, b: single;
begin
  W := VectorSubtract(point, linePoint);

  c1 := VectorDotProduct(W, lineDirection);
  c2 := VectorDotProduct(lineDirection, lineDirection);
  b  := c1 / c2;

  VectorAdd(linePoint, VectorScale(lineDirection, b), result);
end;

// PointLineDistance
//
function PointLineDistance(const point, linePoint, lineDirection: TAffineVector): single;
var
  PB: TAffineVector;
begin
  PB     := PointLineClosestPoint(point, linePoint, lineDirection);
  result := VectorDistance(point, PB);
end;

// PointSegmentClosestPoint
//
function PointSegmentClosestPoint(const point, segmentStart, segmentStop: TVector): TVector;
var
  W, lineDirection: TVector;
  c1, c2, b       : single;
begin
  lineDirection := VectorSubtract(segmentStop, segmentStart);
  W             := VectorSubtract(point, segmentStart);

  c1 := VectorDotProduct(W, lineDirection);
  c2 := VectorDotProduct(lineDirection, lineDirection);
  b  := ClampValue(c1 / c2, 0, 1);

  VectorAdd(segmentStart, VectorScale(lineDirection, b), result);
end;

// PointSegmentClosestPoint
//
function PointSegmentClosestPoint(const point, segmentStart, segmentStop: TAffineVector)
  : TAffineVector;
var
  W, lineDirection: TAffineVector;
  c1, c2, b       : single;
begin
  lineDirection := VectorSubtract(segmentStop, segmentStart);
  W             := VectorSubtract(point, segmentStart);

  c1 := VectorDotProduct(W, lineDirection);
  c2 := VectorDotProduct(lineDirection, lineDirection);
  b  := ClampValue(c1 / c2, 0, 1);

  VectorAdd(segmentStart, VectorScale(lineDirection, b), result);
end;

// PointSegmentDistance
//
function PointSegmentDistance(const point, segmentStart, segmentStop: TAffineVector): single;
var
  PB: TAffineVector;
begin
  PB     := PointSegmentClosestPoint(point, segmentStart, segmentStop);
  result := VectorDistance(point, PB);
end;

// http://geometryalgorithms.com/Archive/algorithm_0104/algorithm_0104B.htm
// SegmentSegmentClosestPoint
//
procedure SegmentSegmentClosestPoint(const S0Start, S0Stop, S1Start, S1Stop: TAffineVector;
  var Segment0Closest, Segment1Closest: TAffineVector);
const
  cSMALL_NUM = 0.000000001;
var
  u, V, W                                           : TAffineVector;
  a, b, c, smalld, e, largeD, sc, sn, sD, tc, tN, tD: single;
begin
  VectorSubtract(S0Stop, S0Start, u);
  VectorSubtract(S1Stop, S1Start, V);
  VectorSubtract(S0Start, S1Start, W);

  a      := VectorDotProduct(u, u);
  b      := VectorDotProduct(u, V);
  c      := VectorDotProduct(V, V);
  smalld := VectorDotProduct(u, W);
  e      := VectorDotProduct(V, W);
  largeD := a * c - b * b;

  sD := largeD;
  tD := largeD;

  if largeD < cSMALL_NUM then
  begin
    sn := 0.0;
    sD := 1.0;
    tN := e;
    tD := c;
  end
  else
  begin
    sn := (b * e - c * smalld);
    tN := (a * e - b * smalld);
    if (sn < 0.0) then
    begin
      sn := 0.0;
      tN := e;
      tD := c;
    end
    else if (sn > sD) then
    begin
      sn := sD;
      tN := e + b;
      tD := c;
    end;
  end;

  if (tN < 0.0) then
  begin
    tN := 0.0;
    // recompute sc for this edge
    if (-smalld < 0.0) then
      sn := 0.0
    else if (-smalld > a) then
      sn := sD
    else
    begin
      sn := -smalld;
      sD := a;
    end;
  end
  else if (tN > tD) then
  begin
    tN := tD;
    // recompute sc for this edge
    if ((-smalld + b) < 0.0) then
      sn := 0
    else if ((-smalld + b) > a) then
      sn := sD
    else
    begin
      sn := (-smalld + b);
      sD := a;
    end;
  end;

  // finally do the division to get sc and tc
  // sc := (abs(sN) < SMALL_NUM ? 0.0 : sN / sD);
  if Abs(sn) < cSMALL_NUM then
    sc := 0
  else
    sc := sn / sD;

  // tc := (abs(tN) < SMALL_NUM ? 0.0 : tN / tD);
  if Abs(tN) < cSMALL_NUM then
    tc := 0
  else
    tc := tN / tD;

  // get the difference of the two closest points
  // Vector   dP = w + (sc * u) - (tc * v);  // = S0(sc) - S1(tc)

  Segment0Closest := VectorAdd(S0Start, VectorScale(u, sc));
  Segment1Closest := VectorAdd(S1Start, VectorScale(V, tc));
end;

// SegmentSegmentDistance
//
function SegmentSegmentDistance(const S0Start, S0Stop, S1Start, S1Stop: TAffineVector): single;
var
  Pb0, PB1: TAffineVector;
begin
  SegmentSegmentClosestPoint(S0Start, S0Stop, S1Start, S1Stop, Pb0, PB1);
  result := VectorDistance(Pb0, PB1);
end;

// LineLineDistance
//
function LineLineDistance(const linePt0, lineDir0, linePt1, lineDir1: TAffineVector): single;
const
  cBIAS = 0.000000001;
var
  det: single;
begin
{$IFNDEF GEOMETRY_NO_ASM}
  asm
    fld   dword ptr [ecx]
    fsub  dword ptr [eax]
    fld   dword ptr [edx+$4]
    fmul  dword ptr [ebp+$40]    // +8
    fld   dword ptr [ebp+$3c]    // +4
    fmul  dword ptr [edx+$8]
    fsubp
    fmulp
    fld   dword ptr [eax+$4]
    fsub  dword ptr [ecx+$4]
    fld   dword ptr [edx]
    fmul  dword ptr [ebp+$40]    // +8
    fld   dword ptr [ebp+$38]    // +0
    fmul  dword ptr [edx+$8]
    fsubp
    fmulp
    faddp
    fld   dword ptr [ecx+$8]
    fsub  dword ptr [eax+$8]
    fld   dword ptr [edx]
    fmul  dword ptr [ebp+$3c]    // +4
    fld   dword ptr [ebp+$38]    // +0
    fmul  dword ptr [edx+$4]
    fsubp
    fmulp
    faddp
    fabs
    fstp  det
  end;
{$ELSE}
  det := Abs((linePt1.V[0] - linePt0.V[0]) * (lineDir0.V[1] * lineDir1.V[2] - lineDir1.V[1] *
    lineDir0.V[2]) - (linePt1.V[1] - linePt0.V[1]) * (lineDir0.V[0] * lineDir1.V[2] - lineDir1.V[0]
    * lineDir0.V[2]) + (linePt1.V[2] - linePt0.V[2]) * (lineDir0.V[0] * lineDir1.V[1] -
    lineDir1.V[0] * lineDir0.V[1]));
{$ENDIF}
  if det < cBIAS then
    result := PointLineDistance(linePt0, linePt1, lineDir1)
  else
    result := det / VectorLength(VectorCrossProduct(lineDir0, lineDir1));
end;

// QuaternionMake
//
function QuaternionMake(const Imag: array of single; Real: single): TQuaternion;
// EAX contains address of Imag
// ECX contains address to result vector
// EDX contains highest index of Imag
// Real part is passed on the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  PUSH EDI
  PUSH ESI
  MOV EDI, ECX
  MOV ESI, EAX
  MOV ECX, EDX
  INC ECX
  REP MOVSD
  MOV EAX, [Real]
  MOV [EDI], EAX
  POP ESI
  POP EDI
  {$ELSE}
var
  n: Integer;
begin
  n := Length(Imag);
  if n >= 1 then
    result.ImagPart.V[0] := Imag[0];
  if n >= 2 then
    result.ImagPart.V[1] := Imag[1];
  if n >= 3 then
    result.ImagPart.V[2] := Imag[2];
  result.RealPart        := Real;
{$ENDIF}
end;

// QuaternionConjugate
//
function QuaternionConjugate(const Q: TQuaternion): TQuaternion;
begin
  result.ImagPart.V[0] := -Q.ImagPart.V[0];
  result.ImagPart.V[1] := -Q.ImagPart.V[1];
  result.ImagPart.V[2] := -Q.ImagPart.V[2];
  result.RealPart      := Q.RealPart;
end;

// QuaternionMagnitude
//
function QuaternionMagnitude(const Q: TQuaternion): single;
begin
  result := Sqrt(VectorNorm(Q.ImagPart) + Sqr(Q.RealPart));
end;

// NormalizeQuaternion
//
procedure NormalizeQuaternion(var Q: TQuaternion);
var
  M, f: single;
begin
  M := QuaternionMagnitude(Q);
  if M > EPSILON2 then
  begin
    f := 1 / M;
    ScaleVector(Q.ImagPart, f);
    Q.RealPart := Q.RealPart * f;
  end
  else
    Q := IdentityQuaternion;
end;

// QuaternionFromPoints
//
function QuaternionFromPoints(const V1, V2: TAffineVector): TQuaternion;
begin
  result.ImagPart := VectorCrossProduct(V1, V2);
  result.RealPart := Sqrt((VectorDotProduct(V1, V2) + 1) / 2);
end;

// QuaternionFromMatrix
//
function QuaternionFromMatrix(const mat: TMatrix): TQuaternion;
// the matrix must be a rotation matrix!
var
  traceMat, S, invS: double;
begin
  traceMat := 1 + mat.V[0].V[0] + mat.V[1].V[1] + mat.V[2].V[2];
  if traceMat > EPSILON2 then
  begin
    S                    := Sqrt(traceMat) * 2;
    invS                 := 1 / S;
    result.ImagPart.V[0] := (mat.V[1].V[2] - mat.V[2].V[1]) * invS;
    result.ImagPart.V[1] := (mat.V[2].V[0] - mat.V[0].V[2]) * invS;
    result.ImagPart.V[2] := (mat.V[0].V[1] - mat.V[1].V[0]) * invS;
    result.RealPart      := 0.25 * S;
  end
  else if (mat.V[0].V[0] > mat.V[1].V[1]) and (mat.V[0].V[0] > mat.V[2].V[2]) then
  begin // Row 0:
    S    := Sqrt(MaxFloat(EPSILON2, cOne + mat.V[0].V[0] - mat.V[1].V[1] - mat.V[2].V[2])) * 2;
    invS := 1 / S;
    result.ImagPart.V[0] := 0.25 * S;
    result.ImagPart.V[1] := (mat.V[0].V[1] + mat.V[1].V[0]) * invS;
    result.ImagPart.V[2] := (mat.V[2].V[0] + mat.V[0].V[2]) * invS;
    result.RealPart      := (mat.V[1].V[2] - mat.V[2].V[1]) * invS;
  end
  else if (mat.V[1].V[1] > mat.V[2].V[2]) then
  begin // Row 1:
    S    := Sqrt(MaxFloat(EPSILON2, cOne + mat.V[1].V[1] - mat.V[0].V[0] - mat.V[2].V[2])) * 2;
    invS := 1 / S;
    result.ImagPart.V[0] := (mat.V[0].V[1] + mat.V[1].V[0]) * invS;
    result.ImagPart.V[1] := 0.25 * S;
    result.ImagPart.V[2] := (mat.V[1].V[2] + mat.V[2].V[1]) * invS;
    result.RealPart      := (mat.V[2].V[0] - mat.V[0].V[2]) * invS;
  end
  else
  begin // Row 2:
    S    := Sqrt(MaxFloat(EPSILON2, cOne + mat.V[2].V[2] - mat.V[0].V[0] - mat.V[1].V[1])) * 2;
    invS := 1 / S;
    result.ImagPart.V[0] := (mat.V[2].V[0] + mat.V[0].V[2]) * invS;
    result.ImagPart.V[1] := (mat.V[1].V[2] + mat.V[2].V[1]) * invS;
    result.ImagPart.V[2] := 0.25 * S;
    result.RealPart      := (mat.V[0].V[1] - mat.V[1].V[0]) * invS;
  end;
  NormalizeQuaternion(result);
end;

// QuaternionMultiply
//
function QuaternionMultiply(const qL, qR: TQuaternion): TQuaternion;
var
  Temp: TQuaternion;
begin
  Temp.RealPart := qL.RealPart * qR.RealPart - qL.ImagPart.V[X] * qR.ImagPart.V[X] - qL.ImagPart.V
    [Y] * qR.ImagPart.V[Y] - qL.ImagPart.V[Z] * qR.ImagPart.V[Z];
  Temp.ImagPart.V[X] := qL.RealPart * qR.ImagPart.V[X] + qL.ImagPart.V[X] * qR.RealPart +
    qL.ImagPart.V[Y] * qR.ImagPart.V[Z] - qL.ImagPart.V[Z] * qR.ImagPart.V[Y];
  Temp.ImagPart.V[Y] := qL.RealPart * qR.ImagPart.V[Y] + qL.ImagPart.V[Y] * qR.RealPart +
    qL.ImagPart.V[Z] * qR.ImagPart.V[X] - qL.ImagPart.V[X] * qR.ImagPart.V[Z];
  Temp.ImagPart.V[Z] := qL.RealPart * qR.ImagPart.V[Z] + qL.ImagPart.V[Z] * qR.RealPart +
    qL.ImagPart.V[X] * qR.ImagPart.V[Y] - qL.ImagPart.V[Y] * qR.ImagPart.V[X];
  result := Temp;
end;

// QuaternionToMatrix
//
function QuaternionToMatrix(quat: TQuaternion): TMatrix;
var
  W, X, Y, Z, xx, xy, xz, xw, yy, yz, yw, zz, zw: single;
begin
  NormalizeQuaternion(quat);
  W                := quat.RealPart;
  X                := quat.ImagPart.V[0];
  Y                := quat.ImagPart.V[1];
  Z                := quat.ImagPart.V[2];
  xx               := X * X;
  xy               := X * Y;
  xz               := X * Z;
  xw               := X * W;
  yy               := Y * Y;
  yz               := Y * Z;
  yw               := Y * W;
  zz               := Z * Z;
  zw               := Z * W;
  result.V[0].V[0] := 1 - 2 * (yy + zz);
  result.V[1].V[0] := 2 * (xy - zw);
  result.V[2].V[0] := 2 * (xz + yw);
  result.V[3].V[0] := 0;
  result.V[0].V[1] := 2 * (xy + zw);
  result.V[1].V[1] := 1 - 2 * (xx + zz);
  result.V[2].V[1] := 2 * (yz - xw);
  result.V[3].V[1] := 0;
  result.V[0].V[2] := 2 * (xz - yw);
  result.V[1].V[2] := 2 * (yz + xw);
  result.V[2].V[2] := 1 - 2 * (xx + yy);
  result.V[3].V[2] := 0;
  result.V[0].V[3] := 0;
  result.V[1].V[3] := 0;
  result.V[2].V[3] := 0;
  result.V[3].V[3] := 1;
end;

// QuaternionToAffineMatrix
//
function QuaternionToAffineMatrix(quat: TQuaternion): TAffineMatrix;
var
  W, X, Y, Z, xx, xy, xz, xw, yy, yz, yw, zz, zw: single;
begin
  NormalizeQuaternion(quat);
  W                := quat.RealPart;
  X                := quat.ImagPart.V[0];
  Y                := quat.ImagPart.V[1];
  Z                := quat.ImagPart.V[2];
  xx               := X * X;
  xy               := X * Y;
  xz               := X * Z;
  xw               := X * W;
  yy               := Y * Y;
  yz               := Y * Z;
  yw               := Y * W;
  zz               := Z * Z;
  zw               := Z * W;
  result.V[0].V[0] := 1 - 2 * (yy + zz);
  result.V[1].V[0] := 2 * (xy - zw);
  result.V[2].V[0] := 2 * (xz + yw);
  result.V[0].V[1] := 2 * (xy + zw);
  result.V[1].V[1] := 1 - 2 * (xx + zz);
  result.V[2].V[1] := 2 * (yz - xw);
  result.V[0].V[2] := 2 * (xz - yw);
  result.V[1].V[2] := 2 * (yz + xw);
  result.V[2].V[2] := 1 - 2 * (xx + yy);
end;

// QuaternionFromAngleAxis
//
function QuaternionFromAngleAxis(const angle: single; const axis: TAffineVector): TQuaternion;
var
  f, S, c: single;
begin
  VectorGeometry.SinCos(VectorGeometry.DegToRad(angle * cOneDotFive), S, c);
  result.RealPart      := c;
  f                    := S / VectorLength(axis);
  result.ImagPart.V[0] := axis.V[0] * f;
  result.ImagPart.V[1] := axis.V[1] * f;
  result.ImagPart.V[2] := axis.V[2] * f;
end;

// QuaternionFromRollPitchYaw
//
function QuaternionFromRollPitchYaw(const r, p, Y: single): TQuaternion;
var
  qp, qy: TQuaternion;
begin
  result := QuaternionFromAngleAxis(r, ZVector);
  qp     := QuaternionFromAngleAxis(p, XVector);
  qy     := QuaternionFromAngleAxis(Y, YVector);

  result := QuaternionMultiply(qp, result);
  result := QuaternionMultiply(qy, result);
end;

// QuaternionFromEuler
//
function QuaternionFromEuler(const X, Y, Z: single; eulerOrder: TEulerOrder): TQuaternion;
// input angles in degrees
var
  gimbalLock  : Boolean;
  quat1, quat2: TQuaternion;

  function EulerToQuat(const X, Y, Z: single; eulerOrder: TEulerOrder): TQuaternion;
  const
    cOrder: array [Low(TEulerOrder) .. High(TEulerOrder)] of array [1 .. 3] of byte = ((1, 2, 3),
      (1, 3, 2), (2, 1, 3),             // eulXYZ, eulXZY, eulYXZ,
      (3, 1, 2), (2, 3, 1), (3, 2, 1)); // eulYZX, eulZXY, eulZYX
  var
    Q: array [1 .. 3] of TQuaternion;
  begin
    Q[cOrder[eulerOrder][1]] := QuaternionFromAngleAxis(X, XVector);
    Q[cOrder[eulerOrder][2]] := QuaternionFromAngleAxis(Y, YVector);
    Q[cOrder[eulerOrder][3]] := QuaternionFromAngleAxis(Z, ZVector);
    result                   := QuaternionMultiply(Q[2], Q[3]);
    result                   := QuaternionMultiply(Q[1], result);
  end;

const
  SMALL_ANGLE = 0.001;
begin
  NormalizeDegAngle(X);
  NormalizeDegAngle(Y);
  NormalizeDegAngle(Z);
  case eulerOrder of
    eulXYZ, eulZYX:
      gimbalLock := Abs(Abs(Y) - 90.0) <= EPSILON2; // cos(Y) = 0;
    eulYXZ, eulZXY:
      gimbalLock := Abs(Abs(X) - 90.0) <= EPSILON2; // cos(X) = 0;
    eulXZY, eulYZX:
      gimbalLock := Abs(Abs(Z) - 90.0) <= EPSILON2; // cos(Z) = 0;
  else
    Assert(False);
    gimbalLock := False;
  end;
  if gimbalLock then
  begin
    case eulerOrder of
      eulXYZ, eulZYX:
        quat1 := EulerToQuat(X, Y - SMALL_ANGLE, Z, eulerOrder);
      eulYXZ, eulZXY:
        quat1 := EulerToQuat(X - SMALL_ANGLE, Y, Z, eulerOrder);
      eulXZY, eulYZX:
        quat1 := EulerToQuat(X, Y, Z - SMALL_ANGLE, eulerOrder);
    end;
    case eulerOrder of
      eulXYZ, eulZYX:
        quat2 := EulerToQuat(X, Y + SMALL_ANGLE, Z, eulerOrder);
      eulYXZ, eulZXY:
        quat2 := EulerToQuat(X + SMALL_ANGLE, Y, Z, eulerOrder);
      eulXZY, eulYZX:
        quat2 := EulerToQuat(X, Y, Z + SMALL_ANGLE, eulerOrder);
    end;
    result := QuaternionSlerp(quat1, quat2, 0.5);
  end
  else
  begin
    result := EulerToQuat(X, Y, Z, eulerOrder);
  end;
end;

// QuaternionToPoints
//
procedure QuaternionToPoints(const Q: TQuaternion; var ArcFrom, ArcTo: TAffineVector);
var
  S, invS: single;
begin
  S := Q.ImagPart.V[X] * Q.ImagPart.V[X] + Q.ImagPart.V[Y] * Q.ImagPart.V[Y];
  if S = 0 then
    SetAffineVector(ArcFrom, 0, 1, 0)
  else
  begin
    invS := RSqrt(S);
    SetAffineVector(ArcFrom, -Q.ImagPart.V[Y] * invS, Q.ImagPart.V[X] * invS, 0);
  end;
  ArcTo.V[X] := Q.RealPart * ArcFrom.V[X] - Q.ImagPart.V[Z] * ArcFrom.V[Y];
  ArcTo.V[Y] := Q.RealPart * ArcFrom.V[Y] + Q.ImagPart.V[Z] * ArcFrom.V[X];
  ArcTo.V[Z] := Q.ImagPart.V[X] * ArcFrom.V[Y] - Q.ImagPart.V[Y] * ArcFrom.V[X];
  if Q.RealPart < 0 then
    SetAffineVector(ArcFrom, -ArcFrom.V[X], -ArcFrom.V[Y], 0);
end;

// LnXP1
//
function LnXP1(X: Extended): Extended;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLDLN2
  MOV     AX, WORD PTR X+8  // exponent
  FLD     X
  CMP     AX, $3FFD         // .4225
  JB      @@1
  FLD1
  FADD
  FYL2X
  JMP     @@2
@@1:
  FYL2XP1
@@2:
  FWAIT
  {$ELSE}
begin
  result := Math.LnXP1(X);
{$ENDIF}
end;

// Log10
//
function Log10(X: Extended): Extended;
// Log.10(X):=Log.2(X) * Log.10(2)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLDLG2     { Log base ten of 2 }
  FLD     X
  FYL2X
  {$ELSE}
begin
  result := Math.Log10(X);
{$ENDIF}
end;

// Log2
//
function Log2(X: Extended): Extended;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD1
  FLD     X
  FYL2X
  {$ELSE}
begin
  result := Math.Log2(X);
{$ENDIF}
end;

// Log2
//
function Log2(X: single): single;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD1
  FLD     X
  FYL2X
  {$ELSE}
begin
{$HINTS OFF}
  result := Math.Log2(X);
{$HINTS ON}
{$ENDIF}
end;

// LogN
//
function LogN(Base, X: Extended): Extended;
// Log.N(X):=Log.2(X) / Log.2(N)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD1
  FLD     X
  FYL2X
  FLD1
  FLD     Base
  FYL2X
  FDIV
  {$ELSE}
begin
  result := Math.LogN(Base, X);
{$ENDIF}
end;

// IntPower
//
function IntPower(Base: Extended; Exponent: Integer): Extended;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  mov     ecx, eax
  cdq
  fld1                      { Result:=1 }
  xor     eax, edx
  sub     eax, edx          { eax:=Abs(Exponent) }
  jz      @@3
  fld     Base
  jmp     @@2
@@1:    fmul    ST, ST            { X:=Base * Base }
@@2:    shr     eax,1
  jnc     @@1
  fmul    ST(1),ST          { Result:=Result * X }
  jnz     @@1
  fstp    st                { pop X from FPU stack }
  cmp     ecx, 0
  jge     @@3
  fld1
  fdivrp                    { Result:=1 / Result }
@@3:
  {$ELSE}
begin
  result := Math.IntPower(Base, Exponent);
{$ENDIF}
end;

// Power
//
function Power(const Base, Exponent: single): single;
begin
{$HINTS OFF}
  if Exponent = cZero then
    result := cOne
  else if (Base = cZero) and (Exponent > cZero) then
    result := cZero
  else if RoundInt(Exponent) = Exponent then
    result := Power(Base, Integer(Round(Exponent)))
  else
    result := Exp(Exponent * Ln(Base));
{$HINTS ON}
end;

// Power (int exponent)
//
function Power(Base: single; Exponent: Integer): single;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  mov     ecx, eax
  cdq
  fld1                      { Result:=1 }
  xor     eax, edx
  sub     eax, edx          { eax:=Abs(Exponent) }
  jz      @@3
  fld     Base
  jmp     @@2
@@1:    fmul    ST, ST            { X:=Base * Base }
@@2:    shr     eax,1
  jnc     @@1
  fmul    ST(1),ST          { Result:=Result * X }
  jnz     @@1
  fstp    st                { pop X from FPU stack }
  cmp     ecx, 0
  jge     @@3
  fld1
  fdivrp                    { Result:=1 / Result }
@@3:
  {$ELSE}
begin
{$HINTS OFF}
  result := Math.Power(Base, Exponent);
{$HINTS ON}
{$ENDIF}
end;

function Power(Base: single; Exponent: Int64): single;
begin
{$HINTS OFF}
  result := Math.Power(Base, Exponent);
{$HINTS ON}
end;

// DegToRad (extended)
//
function DegToRad(const Degrees: Extended): Extended;
begin
  result := Degrees * (PI / 180);
end;

// DegToRad (single)
//
function DegToRad(const Degrees: single): single;
// Result:=Degrees * cPIdiv180;
// don't laugh, Delphi's compiler manages to make a nightmare of this one
// with pushs, pops, etc. in its default compile... (this one is twice faster !)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EBP+8]
  FMUL cPIdiv180
  {$ELSE}
begin
  result := Degrees * cPIdiv180;
{$ENDIF}
end;

// RadToDeg (extended)
//
function RadToDeg(const Radians: Extended): Extended;
begin
  result := Radians * (180 / PI);
end;

// RadToDeg (single)
//
function RadToDeg(const Radians: single): single;
// Result:=Radians * c180divPI;
// don't laugh, Delphi's compiler manages to make a nightmare of this one
// with pushs, pops, etc. in its default compile... (this one is twice faster !)
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EBP+8]
  FMUL c180divPI
  {$ELSE}
begin
  result := Radians * c180divPI;
{$ENDIF}
end;

// NormalizeAngle
//
function NormalizeAngle(angle: single): single;
begin
  result := angle - Int(angle * cInv2PI) * c2PI;
  if result > PI then
    result := result - 2 * PI
  else if result < -PI then
    result := result + 2 * PI;
end;

// NormalizeDegAngle
//
function NormalizeDegAngle(angle: single): single;
begin
  result := angle - Int(angle * cInv360) * c360;
  if result > c180 then
    result := result - c360
  else if result < -c180 then
    result := result + c360;
end;

{$IFDEF GLS_PLATFORM_HAS_EXTENDED}

// SinCos (Extended)
//
procedure SinCos(const Theta: Extended; out Sin, Cos: Extended);
// EAX contains address of Sin
// EDX contains address of Cos
// Theta is passed over the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  Theta
  FSinCos
  FSTP TBYTE PTR [EDX]    // cosine
  FSTP TBYTE PTR [EAX]    // sine
  {$ELSE}
begin
  Math.SinCos(Theta, Sin, Cos);
{$ENDIF}
end;
{$ENDIF GLS_PLATFORM_HAS_EXTENDED}

// SinCos (Double)
//
procedure SinCos(const Theta: double; out Sin, Cos: double);
// EAX contains address of Sin
// EDX contains address of Cos
// Theta is passed over the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  Theta
  FSinCos
  FSTP QWORD PTR [EDX]    // cosine
  FSTP QWORD PTR [EAX]    // sine
  {$ELSE}
var
  S, c: Extended;
begin
  Math.SinCos(Theta, S, c);
{$HINTS OFF}
  Sin := S;
  Cos := c;
{$HINTS ON}
{$ENDIF}
end;

// SinCos (Single)
//
procedure SinCos(const Theta: single; out Sin, Cos: single);
// EAX contains address of Sin
// EDX contains address of Cos
// Theta is passed over the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  Theta
  FSinCos
  FSTP DWORD PTR [EDX]    // cosine
  FSTP DWORD PTR [EAX]    // sine
  {$ELSE}
var
  S, c: Extended;
begin
  Math.SinCos(Theta, S, c);
{$HINTS OFF}
  Sin := S;
  Cos := c;
{$HINTS ON}
{$ENDIF}
end;

{$IFDEF GLS_PLATFORM_HAS_EXTENDED}

// SinCos (Extended w radius)
//
procedure SinCos(const Theta, radius: double; out Sin, Cos: Extended);
// EAX contains address of Sin
// EDX contains address of Cos
// Theta is passed over the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  theta
  FSinCos
  FMUL radius
  FSTP TBYTE PTR [EDX]    // cosine
  FMUL radius
  FSTP TBYTE PTR [EAX]    // sine
  {$ELSE}
var
  S, c: Extended;
begin
  Math.SinCos(Theta, S, c);
  Sin := S * radius;
  Cos := c * radius;
{$ENDIF}
end;
{$ENDIF GLS_PLATFORM_HAS_EXTENDED}

// SinCos (Double w radius)
//
procedure SinCos(const Theta, radius: double; out Sin, Cos: double);
// EAX contains address of Sin
// EDX contains address of Cos
// Theta is passed over the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  theta
  FSinCos
  FMUL radius
  FSTP QWORD PTR [EDX]    // cosine
  FMUL radius
  FSTP QWORD PTR [EAX]    // sine
  {$ELSE}
var
  S, c: Extended;
begin
  Math.SinCos(Theta, S, c);
  Sin := S * radius;
  Cos := c * radius;
{$ENDIF}
end;

// SinCos (Single w radius)
//
procedure SinCos(const Theta, radius: single; out Sin, Cos: single);
// EAX contains address of Sin
// EDX contains address of Cos
// Theta is passed over the stack
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  theta
  FSinCos
  FMUL radius
  FSTP DWORD PTR [EDX]    // cosine
  FMUL radius
  FSTP DWORD PTR [EAX]    // sine
  {$ELSE}
var
  S, c: Extended;
begin
  Math.SinCos(Theta, S, c);
  Sin := S * radius;
  Cos := c * radius;
{$ENDIF}
end;

// PrepareSinCosCache
//
procedure PrepareSinCosCache(var S, c: array of single; startAngle, stopAngle: single);
var
  i             : Integer;
  d, alpha, beta: single;
begin
  Assert((High(S) = High(c)) and (Low(S) = Low(c)));
  stopAngle := stopAngle + 1E-5;
  if High(S) > Low(S) then
    d := cPIdiv180 * (stopAngle - startAngle) / (High(S) - Low(S))
  else
    d := 0;

  if High(S) - Low(S) < 1000 then
  begin
    // Fast computation (approx 5.5x)
    alpha := 2 * Sqr(Sin(d * 0.5));
    beta  := Sin(d);
    VectorGeometry.SinCos(startAngle * cPIdiv180, S[Low(S)], c[Low(S)]);
    for i := Low(S) to High(S) - 1 do
    begin
      // Make use of the incremental formulae:
      // cos (theta+delta) = cos(theta) - [alpha*cos(theta) + beta*sin(theta)]
      // sin (theta+delta) = sin(theta) - [alpha*sin(theta) - beta*cos(theta)]
      c[i + 1] := c[i] - alpha * c[i] - beta * S[i];
      S[i + 1] := S[i] - alpha * S[i] + beta * c[i];
    end;
  end
  else
  begin
    // Slower, but maintains precision when steps are small
    startAngle := startAngle * cPIdiv180;
    for i      := Low(S) to High(S) do
      VectorGeometry.SinCos((i - Low(S)) * d + startAngle, S[i], c[i]);
  end;
end;

// ArcCos (Extended)
//
function ArcCos(const X: Extended): Extended;
begin
  result := VectorGeometry.ArcTan2(Sqrt(1 - Sqr(X)), X);
end;

// ArcCos (Single)
//
function ArcCos(const X: single): single;
// Result:=ArcTan2(Sqrt(c1 - X * X), X);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD   X
  FMUL  ST, ST
  FSUBR cOne
  FSQRT
  FLD   X
  FPATAN
  {$ELSE}
begin
{$HINTS OFF}
{$IFDEF FPC}
  if Abs(X) > 1.0 then
    result := Math.ArcCos(Sign(X))
  else
{$ENDIF}
    result := Math.ArcCos(X);
{$HINTS ON}
{$ENDIF}
end;

// ArcSin (Extended)
//
function ArcSin(const X: Extended): Extended;
begin
  result := VectorGeometry.ArcTan2(X, Sqrt(1 - Sqr(X)))
end;

// ArcSin (Single)
//
function ArcSin(const X: single): single;
// Result:=ArcTan2(X, Sqrt(1 - X * X))
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD   X
  FLD   ST
  FMUL  ST, ST
  FSUBR cOne
  FSQRT
  FPATAN
  {$ELSE}
begin
{$HINTS OFF}
  result := Math.ArcSin(X);
{$HINTS ON}
{$ENDIF}
end;

// ArcTan2 (Extended)
//
function ArcTan2(const Y, X: Extended): Extended;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  Y
  FLD  X
  FPATAN
  {$ELSE}
begin
  result := Math.ArcTan2(Y, X);
{$ENDIF}
end;

// ArcTan2 (Single)
//
function ArcTan2(const Y, X: single): single;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  Y
  FLD  X
  FPATAN
  {$ELSE}
begin
{$HINTS OFF}
  result := Math.ArcTan2(Y, X);
{$HINTS ON}
{$ENDIF}
end;

// FastArcTan2
//
function FastArcTan2(Y, X: single): single;
// accuracy of about 0.07 rads
const
  cEpsilon: single = 1E-10;
var
  abs_y: single;
begin
  abs_y := Abs(Y) + cEpsilon; // prevent 0/0 condition
  if Y < 0 then
  begin
    if X >= 0 then
      result := cPIdiv4 * (X - abs_y) / (X + abs_y) - cPIdiv4
    else
      result := cPIdiv4 * (X + abs_y) / (abs_y - X) - c3PIdiv4;
  end
  else
  begin
    if X >= 0 then
      result := cPIdiv4 - cPIdiv4 * (X - abs_y) / (X + abs_y)
    else
      result := c3PIdiv4 - cPIdiv4 * (X + abs_y) / (abs_y - X);
  end;
end;

// Tan (Extended)
//
function Tan(const X: Extended): Extended;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  X
  FPTAN
  FSTP ST(0)      // FPTAN pushes 1.0 after result
  {$ELSE}
begin
  result := Math.Tan(X);
{$ENDIF}
end;

// Tan (Single)
//
function Tan(const X: single): single;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  X
  FPTAN
  FSTP ST(0)      // FPTAN pushes 1.0 after result
  {$ELSE}
begin
{$HINTS OFF}
  result := Math.Tan(X);
{$HINTS ON}
{$ENDIF}
end;

// CoTan (Extended)
//
function CoTan(const X: Extended): Extended;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  X
  FPTAN
  FDIVRP
  {$ELSE}
begin
  result := Math.CoTan(X);
{$ENDIF}
end;

// CoTan (Single)
//
function CoTan(const X: single): single;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  X
  FPTAN
  FDIVRP
  {$ELSE}
begin
{$HINTS OFF}
  result := Math.CoTan(X);
{$HINTS ON}
{$ENDIF}
end;

// Sinh
//
function Sinh(const X: single): single;
{$IFDEF GEOMETRY_NO_ASM}
begin
  result := 0.5 * (Exp(X) - Exp(-X));
{$ELSE}
asm
  fld   x
  call  RegisterBasedExp
  fld   x
  fchs
  call  RegisterBasedExp
  fsub
  fmul  cOneDotFive
  {$ENDIF}
end;

// Sinh
//
function Sinh(const X: double): double;
{$IFDEF GEOMETRY_NO_ASM}
begin
  result := 0.5 * (Exp(X) - Exp(-X));
{$ELSE}
asm
  fld   x
  call  RegisterBasedExp
  fld   x
  fchs
  call  RegisterBasedExp
  fsub
  fmul  cOneDotFive
  {$ENDIF}
end;

// Cosh
//
function Cosh(const X: single): single;
{$IFDEF GEOMETRY_NO_ASM}
begin
  result := 0.5 * (Exp(X) + Exp(-X));
{$ELSE}
asm
  fld   x
  call  RegisterBasedExp
  fld   x
  fchs
  call  RegisterBasedExp
  fadd
  fmul  cOneDotFive
  {$ENDIF}
end;

// Cosh
//
function Cosh(const X: double): double;
{$IFDEF GEOMETRY_NO_ASM}
begin
  result := 0.5 * (Exp(X) + Exp(-X));
{$ELSE}
asm
  fld   x
  call  RegisterBasedExp
  fld   x
  fchs
  call  RegisterBasedExp
  fadd
  fmul  cOneDotFive
  {$ENDIF}
end;

// RSqrt
//
function RSqrt(V: single): single;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU
@@3DNow:
  lea eax, [ebp+8]
  db $0F,$6E,$00           /// movd mm0, [eax]
  db $0F,$0F,$C8,$97       /// pfrsqrt  mm1, mm0

  db $0F,$6F,$D1           /// movq     mm2, mm1
  db $0F,$0F,$C9,$B4       /// pfmul    mm1, mm1
  db $0F,$0F,$C8,$A7       /// pfrsqit1 mm1, mm0
  db $0F,$0F,$CA,$B6       /// pfrcpit2 mm1, mm2

  db $0F,$7E,$08           /// movd [eax], mm1
  db $0F,$0E               /// femms
  fld dword ptr [eax]
  jmp @@End

@@FPU:
  fld v
  fsqrt
  fld1
  fdivr
@@End:
  {$ELSE}
begin
  result := 1 / Sqrt(V);
{$ENDIF}
end;

// ISqrt
//
function ISqrt(i: Integer): Integer;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  push     eax
  test     vSIMD, 1
  jz @@FPU
@@3DNow:
  db $0F,$6E,$04,$24       /// movd     mm0, [esp]
  db $0F,$0F,$C8,$0D       /// pi2fd    mm1, mm0
  db $0F,$0F,$D1,$97       /// pfrsqrt  mm2, mm1
  db $0F,$0F,$DA,$96       /// pfrcp    mm3, mm2
  db $0F,$0F,$E3,$1D       /// pf2id    mm4, mm3
  db $0F,$7E,$24,$24       /// movd     [esp], mm4
  db $0F,$0E               /// femms
  pop      eax
  ret
@@FPU:
  fild     dword ptr [esp]
  fsqrt
  fistp    dword ptr [esp]
  pop      eax
  {$ELSE}
begin
{$HINTS OFF}
  result := Round(Sqrt(i));
{$HINTS ON}
{$ENDIF}
end;

// ILength
//
function ILength(X, Y: Integer): Integer;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  push     edx
  push     eax
  fild     dword ptr [esp]
  fmul     ST(0), ST(0)
  fild     dword ptr [esp+4]
  fmul     ST(0), ST(0)
  faddp
  fsqrt
  fistp    dword ptr [esp+4]
  pop      edx
  pop      eax
  {$ELSE}
begin
{$HINTS OFF}
  result := Round(Sqrt(X * X + Y * Y));
{$HINTS ON}
{$ENDIF}
end;

// ILength
//
function ILength(X, Y, Z: Integer): Integer;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  push     ecx
  push     edx
  push     eax
  fild     dword ptr [esp]
  fmul     ST(0), ST(0)
  fild     dword ptr [esp+4]
  fmul     ST(0), ST(0)
  faddp
  fild     dword ptr [esp+8]
  fmul     ST(0), ST(0)
  faddp
  fsqrt
  fistp    dword ptr [esp+8]
  pop      ecx
  pop      edx
  pop      eax
  {$ELSE}
begin
{$HINTS OFF}
  result := Round(Sqrt(X * X + Y * Y + Z * Z));
{$HINTS ON}
{$ENDIF}
end;

// RLength
//
function RLength(X, Y: single): single;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  fld  x
  fmul x
  fld  y
  fmul y
  fadd
  fsqrt
  fld1
  fdivr
  {$ELSE}
begin
  result := 1 / Sqrt(X * X + Y * Y);
{$ENDIF}
end;

// RegisterBasedExp
//
{$IFNDEF GEOMETRY_NO_ASM}

procedure RegisterBasedExp;
asm   // Exp(x) = 2^(x.log2(e))
  fldl2e
  fmul
  fld      st(0)
  frndint
  fsub     st(1), st
  fxch     st(1)
  f2xm1
  fld1
  fadd
  fscale
  fstp     st(1)
end;
{$ENDIF}

// RandomPointOnSphere
//
procedure RandomPointOnSphere(var p: TAffineVector);
var
  T, W: single;
begin
  p.V[2] := 2 * Random - 1;
  T      := 2 * PI * Random;
  W      := Sqrt(1 - p.V[2] * p.V[2]);
  VectorGeometry.SinCos(T, W, p.V[1], p.V[0]);
end;

// RoundInt (single)
//
function RoundInt(V: single): single;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD     v
  FRNDINT
  {$ELSE}
begin
{$HINTS OFF}
  result := Int(V + cOneDotFive);
{$HINTS ON}
{$ENDIF}
end;

// RoundInt (extended)
//
function RoundInt(V: Extended): Extended;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD     v
  FRNDINT
  {$ELSE}
begin
  result := Int(V + 0.5);
{$ENDIF}
end;

{$IFNDEF GEOMETRY_NO_ASM}

// Trunc64 (extended)
//
function Trunc64(V: Extended): Int64;
asm
  SUB     ESP,12
  FSTCW   [ESP]
  FLDCW   cwChop
  FLD     v
  FISTP   qword ptr [ESP+4]
  FLDCW   [ESP]
  POP     ECX
  POP     EAX
  POP     EDX
end;

// Trunc (single)
//
function Trunc(V: single): Integer;
asm
  SUB     ESP,8
  FSTCW   [ESP]
  FLDCW   cwChop
  FLD     v
  FISTP   dword ptr [ESP+4]
  FLDCW   [ESP]
  POP     ECX
  POP     EAX
end;

// Int (Extended)
//
function Int(V: Extended): Extended;
asm
  SUB     ESP,4
  FSTCW   [ESP]
  FLDCW   cwChop
  FLD     v
  FRNDINT
  FLDCW   [ESP]
  ADD     ESP,4
end;

// Int (Single)
//
function Int(V: single): single;
asm
  SUB     ESP,4
  FSTCW   [ESP]
  FLDCW   cwChop
  FLD     v
  FRNDINT
  FLDCW   [ESP]
  ADD     ESP,4
end;

// Frac (Extended)
//
function Frac(V: Extended): Extended;
asm
  SUB     ESP,4
  FSTCW   [ESP]
  FLDCW   cwChop
  FLD     v
  FLD     ST
  FRNDINT
  FSUB
  FLDCW   [ESP]
  ADD     ESP,4
end;

// Frac (Extended)
//
function Frac(V: single): single;
asm
  SUB     ESP,4
  FSTCW   [ESP]
  FLDCW   cwChop
  FLD     v
  FLD     ST
  FRNDINT
  FSUB
  FLDCW   [ESP]
  ADD     ESP,4
end;

// Round64 (Single);
//
function Round64(V: single): Int64;
asm
  SUB     ESP,8
  FLD     v
  FISTP   qword ptr [ESP]
  POP     EAX
  POP     EDX
end;

// Round64 (Extended);
//
function Round64(V: Extended): Int64;
asm
  FLD      v
  FISTP    qword ptr [v]           // use v as storage to place the result
  MOV      EAX, dword ptr [v]
  MOV      EDX, dword ptr [v+4]
end;

// Round (Single);
//
function Round(V: single): Integer;
asm
  FLD     v
  FISTP   DWORD PTR [v]     // use v as storage to place the result
  MOV     EAX, [v]
end;

{$ELSE}

function Trunc(X: Extended): Int64;
begin
  result := System.Trunc(X);
end;

function Round(X: Extended): Int64;
begin
  result := System.Round(X);
end;

function Frac(X: Extended): Extended;
begin
  result := System.Frac(X);
end;

{$ENDIF}

// Ceil64 (Extended)
//
function Ceil64(V: Extended): Int64; overload;
begin
  if Frac(V) > 0 then
    result := Trunc(V) + 1
  else
    result := Trunc(V);
end;

// Ceil (Single)
//
function Ceil(V: single): Integer; overload;
begin
{$HINTS OFF}
  if Frac(V) > 0 then
    result := Trunc(V) + 1
  else
    result := Trunc(V);
{$HINTS ON}
end;

// Floor64 (Extended)
//
function Floor64(V: Extended): Int64; overload;
begin
  if V < 0 then
    result := Trunc(V) - 1
  else
    result := Trunc(V);
end;

// Floor (Single)
//
function Floor(V: single): Integer; overload;
begin
{$HINTS OFF}
  if V < 0 then
    result := Trunc(V) - 1
  else
    result := Trunc(V);
{$HINTS ON}
end;

// Sign
//
function Sign(X: single): Integer;
begin
  if X < 0 then
    result := -1
  else if X > 0 then
    result := 1
  else
    result := 0;
end;

// SignStrict
//
function SignStrict(X: single): Integer;
begin
  if X < 0 then
    result := -1
  else
    result := 1
end;

// ScaleAndRound
//
function ScaleAndRound(i: Integer; var S: single): Integer;
{$IFNDEF GEOMETRY_NO_ASM}
asm
  push  eax
  fild  dword ptr [esp]
  fmul  dword ptr [edx]
  fistp dword ptr [esp]
  pop   eax
  {$ELSE}
begin
{$HINTS OFF}
  result := Round(i * S);
{$HINTS ON}
{$ENDIF}
end;

// IsInRange (single)
//
function IsInRange(const X, a, b: single): Boolean;
begin
  if a < b then
    result := (a <= X) and (X <= b)
  else
    result := (b <= X) and (X <= a);
end;

// IsInRange (double)
//
function IsInRange(const X, a, b: double): Boolean;
begin
  if a < b then
    result := (a <= X) and (X <= b)
  else
    result := (b <= X) and (X <= a);
end;

// IsInCube (affine)
//
function IsInCube(const p, d: TAffineVector): Boolean; overload;
begin
  result := ((p.V[0] >= -d.V[0]) and (p.V[0] <= d.V[0])) and
    ((p.V[1] >= -d.V[1]) and (p.V[1] <= d.V[1])) and ((p.V[2] >= -d.V[2]) and (p.V[2] <= d.V[2]));
end;

// IsInCube (hmg)
//
function IsInCube(const p, d: TVector): Boolean; overload;
begin
  result := ((p.V[0] >= -d.V[0]) and (p.V[0] <= d.V[0])) and
    ((p.V[1] >= -d.V[1]) and (p.V[1] <= d.V[1])) and ((p.V[2] >= -d.V[2]) and (p.V[2] <= d.V[2]));
end;

// MinFloat (single)
//
function MinFloat(values: PSingleArray; nbItems: Integer): single;
var
  i, k: Integer;
begin
  if nbItems > 0 then
  begin
    k     := 0;
    for i := 1 to nbItems - 1 do
      if values^[i] < values^[k] then
        k  := i;
    result := values^[k];
  end
  else
    result := 0;
end;

// MinFloat (double)
//
function MinFloat(values: PDoubleArray; nbItems: Integer): double;
var
  i, k: Integer;
begin
  if nbItems > 0 then
  begin
    k     := 0;
    for i := 1 to nbItems - 1 do
      if values^[i] < values^[k] then
        k  := i;
    result := values^[k];
  end
  else
    result := 0;
end;

// MinFloat (extended)
//
function MinFloat(values: PExtendedArray; nbItems: Integer): Extended;
var
  i, k: Integer;
begin
  if nbItems > 0 then
  begin
    k     := 0;
    for i := 1 to nbItems - 1 do
      if values^[i] < values^[k] then
        k  := i;
    result := values^[k];
  end
  else
    result := 0;
end;

// MinFloat (array)
//
function MinFloat(const V: array of single): single;
var
  i: Integer;
begin
  if Length(V) > 0 then
  begin
    result := V[0];
    for i  := 1 to High(V) do
      if V[i] < result then
        result := V[i];
  end
  else
    result := 0;
end;

// MinFloat (single 2)
//
function MinFloat(const V1, V2: single): single;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 < V2 then
    result := V1
  else
    result := V2;
{$ELSE}
asm
  fld     v1
  fld     v2
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DB,$C1                 /// fcmovnb st(0), st(1)
  ffree   st(1)
  {$ENDIF}
end;

// MinFloat (double 2)
//
function MinFloat(const V1, V2: double): double;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 < V2 then
    result := V1
  else
    result := V2;
{$ELSE}
asm
  fld     v1
  fld     v2
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DB,$C1                 /// fcmovnb st(0), st(1)
  ffree   st(1)
  {$ENDIF}
end;

{$IFDEF GLS_PLATFORM_HAS_EXTENDED}

// MinFloat (extended 2)
//
function MinFloat(const V1, V2: Extended): Extended;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 < V2 then
    result := V1
  else
    result := V2;
{$ELSE}
asm
  fld     v1
  fld     v2
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DB,$C1                 /// fcmovnb st(0), st(1)
  ffree   st(1)
  {$ENDIF}
end;
{$ENDIF GLS_PLATFORM_HAS_EXTENDED}

// MinFloat
//
function MinFloat(const V1, V2, V3: single): single;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 <= V2 then
    if V1 <= V3 then
      result := V1
    else if V3 <= V2 then
      result := V3
    else
      result := V2
  else if V2 <= V3 then
    result := V2
  else if V3 <= V1 then
    result := V3
  else
    result := V1;
{$ELSE}
asm
  fld     v1
  fld     v2
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DB,$C1                 /// fcmovnb st(0), st(1)
  ffree   st(1)
  fld     v3
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DB,$C1                 /// fcmovnb st(0), st(1)
  ffree   st(1)
  {$ENDIF}
end;

// MinFloat (double)
//
function MinFloat(const V1, V2, V3: double): double;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 <= V2 then
    if V1 <= V3 then
      result := V1
    else if V3 <= V2 then
      result := V3
    else
      result := V2
  else if V2 <= V3 then
    result := V2
  else if V3 <= V1 then
    result := V3
  else
    result := V1;
{$ELSE}
asm
  fld     v1
  fld     v2
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DB,$C1                 /// fcmovnb st(0), st(1)
  ffree   st(1)
  fld     v3
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DB,$C1                 /// fcmovnb st(0), st(1)
  ffree   st(1)
  {$ENDIF}
end;

{$IFDEF GLS_PLATFORM_HAS_EXTENDED}

// MinFloat
//
function MinFloat(const V1, V2, V3: Extended): Extended;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 <= V2 then
    if V1 <= V3 then
      result := V1
    else if V3 <= V2 then
      result := V3
    else
      result := V2
  else if V2 <= V3 then
    result := V2
  else if V3 <= V1 then
    result := V3
  else
    result := V1;
{$ELSE}
asm
  fld     v1
  fld     v2
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DB,$C1                 /// fcmovnb st(0), st(1)
  ffree   st(1)
  fld     v3
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DB,$C1                 /// fcmovnb st(0), st(1)
  ffree   st(1)
  {$ENDIF}
end;
{$ENDIF GLS_PLATFORM_HAS_EXTENDED}

// MaxFloat (single)
//
function MaxFloat(values: PSingleArray; nbItems: Integer): single; overload;
var
  i, k: Integer;
begin
  if nbItems > 0 then
  begin
    k     := 0;
    for i := 1 to nbItems - 1 do
      if values^[i] > values^[k] then
        k  := i;
    result := values^[k];
  end
  else
    result := 0;
end;

// MaxFloat (double)
//
function MaxFloat(values: PDoubleArray; nbItems: Integer): double; overload;
var
  i, k: Integer;
begin
  if nbItems > 0 then
  begin
    k     := 0;
    for i := 1 to nbItems - 1 do
      if values^[i] > values^[k] then
        k  := i;
    result := values^[k];
  end
  else
    result := 0;
end;

// MaxFloat (extended)
//
function MaxFloat(values: PExtendedArray; nbItems: Integer): Extended; overload;
var
  i, k: Integer;
begin
  if nbItems > 0 then
  begin
    k     := 0;
    for i := 1 to nbItems - 1 do
      if values^[i] > values^[k] then
        k  := i;
    result := values^[k];
  end
  else
    result := 0;
end;

// MaxFloat
//
function MaxFloat(const V: array of single): single;
var
  i: Integer;
begin
  if Length(V) > 0 then
  begin
    result := V[0];
    for i  := 1 to High(V) do
      if V[i] > result then
        result := V[i];
  end
  else
    result := 0;
end;

// MaxFloat
//
function MaxFloat(const V1, V2: single): single;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 > V2 then
    result := V1
  else
    result := V2;
{$ELSE}
asm
  fld     v1
  fld     v2
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DA,$C1                 /// fcmovb  st(0), st(1)
  ffree   st(1)
  {$ENDIF}
end;

// MaxFloat
//
function MaxFloat(const V1, V2: double): double;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 > V2 then
    result := V1
  else
    result := V2;
{$ELSE}
asm
  fld     v1
  fld     v2
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DA,$C1                 /// fcmovb  st(0), st(1)
  ffree   st(1)
  {$ENDIF}
end;

{$IFDEF GLS_PLATFORM_HAS_EXTENDED}

// MaxFloat
//
function MaxFloat(const V1, V2: Extended): Extended;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 > V2 then
    result := V1
  else
    result := V2;
{$ELSE}
asm
  fld     v1
  fld     v2
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DA,$C1                 /// fcmovb  st(0), st(1)
  ffree   st(1)
  {$ENDIF}
end;
{$ENDIF GLS_PLATFORM_HAS_EXTENDED}

// MaxFloat
//
function MaxFloat(const V1, V2, V3: single): single;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 >= V2 then
    if V1 >= V3 then
      result := V1
    else if V3 >= V2 then
      result := V3
    else
      result := V2
  else if V2 >= V3 then
    result := V2
  else if V3 >= V1 then
    result := V3
  else
    result := V1;
{$ELSE}
asm
  fld     v1
  fld     v2
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DA,$C1                 /// fcmovb  st(0), st(1)
  ffree   st(1)
  fld     v3
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DA,$C1                 /// fcmovb  st(0), st(1)
  ffree   st(1)
  {$ENDIF}
end;

// MaxFloat
//
function MaxFloat(const V1, V2, V3: double): double;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 >= V2 then
    if V1 >= V3 then
      result := V1
    else if V3 >= V2 then
      result := V3
    else
      result := V2
  else if V2 >= V3 then
    result := V2
  else if V3 >= V1 then
    result := V3
  else
    result := V1;
{$ELSE}
asm
  fld     v1
  fld     v2
  fld     v3
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DA,$C1                 /// fcmovb  st(0), st(1)
  db $DB,$F2                 /// fcomi   st(0), st(2)
  db $DA,$C2                 /// fcmovb  st(0), st(2)
  ffree   st(2)
  ffree   st(1)
  {$ENDIF}
end;

{$IFDEF GLS_PLATFORM_HAS_EXTENDED}

// MaxFloat
//
function MaxFloat(const V1, V2, V3: Extended): Extended;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 >= V2 then
    if V1 >= V3 then
      result := V1
    else if V3 >= V2 then
      result := V3
    else
      result := V2
  else if V2 >= V3 then
    result := V2
  else if V3 >= V1 then
    result := V3
  else
    result := V1;
{$ELSE}
asm
  fld     v1
  fld     v2
  fld     v3
  db $DB,$F1                 /// fcomi   st(0), st(1)
  db $DA,$C1                 /// fcmovb  st(0), st(1)
  db $DB,$F2                 /// fcomi   st(0), st(2)
  db $DA,$C2                 /// fcmovb  st(0), st(2)
  ffree   st(2)
  ffree   st(1)
  {$ENDIF}
end;
{$ENDIF GLS_PLATFORM_HAS_EXTENDED}

// MinInteger (2 int)
//
function MinInteger(const V1, V2: Integer): Integer;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 < V2 then
    result := V1
  else
    result := V2;
{$ELSE}
asm
  cmp   eax, edx
  db $0F,$4F,$C2             /// cmovg eax, edx
  {$ENDIF}
end;

// MinInteger (2 card)
//
function MinInteger(const V1, V2: Cardinal): Cardinal;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 < V2 then
    result := V1
  else
    result := V2;
{$ELSE}
asm
  cmp   eax, edx
  db $0F,$47,$C2             /// cmova eax, edx
  {$ENDIF}
end;

// MinInteger
//
function MinInteger(const V1, V2, V3: Integer): Integer;
begin
  if V1 <= V2 then
    if V1 <= V3 then
      result := V1
    else if V3 <= V2 then
      result := V3
    else
      result := V2
  else if V2 <= V3 then
    result := V2
  else if V3 <= V1 then
    result := V3
  else
    result := V1;
end;

// MinInteger
//
function MinInteger(const V1, V2, V3: Cardinal): Cardinal;
begin
  if V1 <= V2 then
    if V1 <= V3 then
      result := V1
    else if V3 <= V2 then
      result := V3
    else
      result := V2
  else if V2 <= V3 then
    result := V2
  else if V3 <= V1 then
    result := V3
  else
    result := V1;
end;

// MaxInteger (2 int)
//
function MaxInteger(const V1, V2: Integer): Integer;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 > V2 then
    result := V1
  else
    result := V2;
{$ELSE}
asm
  cmp   eax, edx
  db $0F,$4C,$C2             /// cmovl eax, edx
  {$ENDIF}
end;

// MaxInteger (2 card)
//
function MaxInteger(const V1, V2: Cardinal): Cardinal;
{$IFDEF GEOMETRY_NO_ASM}
begin
  if V1 > V2 then
    result := V1
  else
    result := V2;
{$ELSE}
asm
  cmp   eax, edx
  db $0F,$42,$C2             /// cmovb eax, edx
  {$ENDIF}
end;

// MaxInteger
//
function MaxInteger(const V1, V2, V3: Integer): Integer;
begin
  if V1 >= V2 then
    if V1 >= V3 then
      result := V1
    else if V3 >= V2 then
      result := V3
    else
      result := V2
  else if V2 >= V3 then
    result := V2
  else if V3 >= V1 then
    result := V3
  else
    result := V1;
end;

// MaxInteger
//
function MaxInteger(const V1, V2, V3: Cardinal): Cardinal;
begin
  if V1 >= V2 then
    if V1 >= V3 then
      result := V1
    else if V3 >= V2 then
      result := V3
    else
      result := V2
  else if V2 >= V3 then
    result := V2
  else if V3 >= V1 then
    result := V3
  else
    result := V1;
end;

function ClampInteger(const value, min, max: Integer): Integer;
begin
  result := MinInteger(MaxInteger(value, min), max);
end;

function ClampInteger(const value, min, max: Cardinal): Cardinal;
begin
  result := MinInteger(MaxInteger(value, min), max);
end;

// TriangleArea
//
function TriangleArea(const p1, p2, p3: TAffineVector): single;
begin
  result := 0.5 * VectorLength(VectorCrossProduct(VectorSubtract(p2, p1), VectorSubtract(p3, p1)));
end;

// PolygonArea
//
function PolygonArea(const p: PAffineVectorArray; nSides: Integer): single;
var
  r         : TAffineVector;
  i         : Integer;
  p1, p2, p3: PAffineVector;
begin
  result := 0;
  if nSides > 2 then
  begin
    RstVector(r);
    p1    := @p[0];
    p2    := @p[1];
    for i := 2 to nSides - 1 do
    begin
      p3 := @p[i];
      AddVector(r, VectorCrossProduct(VectorSubtract(p2^, p1^), VectorSubtract(p3^, p1^)));
      p2 := p3;
    end;
    result := VectorLength(r) * 0.5;
  end;
end;

// TriangleSignedArea
//
function TriangleSignedArea(const p1, p2, p3: TAffineVector): single;
begin
  result := 0.5 * ((p2.V[0] - p1.V[0]) * (p3.V[1] - p1.V[1]) - (p3.V[0] - p1.V[0]) *
    (p2.V[1] - p1.V[1]));
end;

// PolygonSignedArea
//
function PolygonSignedArea(const p: PAffineVectorArray; nSides: Integer): single;
var
  i         : Integer;
  p1, p2, p3: PAffineVector;
begin
  result := 0;
  if nSides > 2 then
  begin
    p1    := @(p^[0]);
    p2    := @(p^[1]);
    for i := 2 to nSides - 1 do
    begin
      p3     := @(p^[i]);
      result := result + (p2^.V[0] - p1^.V[0]) * (p3^.V[1] - p1^.V[1]) - (p3^.V[0] - p1^.V[0]) *
        (p2^.V[1] - p1^.V[1]);
      p2 := p3;
    end;
    result := result * 0.5;
  end;
end;

// ScaleFloatArray (raw)
//
procedure ScaleFloatArray(values: PSingleArray; nb: Integer; var factor: single);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU

  push  edx
  shr   edx, 2
  or    edx, edx
  jz    @@FPU

  db $0F,$6E,$39           /// movd        mm7, [ecx]
  db $0F,$62,$FF           /// punpckldq   mm7, mm7

@@3DNowLoop:
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$00           /// movq  mm0, [eax]
  db $0F,$6F,$48,$08       /// movq  mm1, [eax+8]
  db $0F,$0F,$C7,$B4       /// pfmul mm0, mm7
  db $0F,$0F,$CF,$B4       /// pfmul mm1, mm7
  db $0F,$7F,$00           /// movq  [eax], mm0
  db $0F,$7F,$48,$08       /// movq  [eax+8], mm1

  add   eax, 16
  dec   edx
  jnz   @@3DNowLoop

  pop   edx
  and   edx, 3
  db $0F,$0E               /// femms

@@FPU:
  push  edx
  shr   edx, 1
  or    edx, edx
  jz    @@FPULone

@@FPULoop:
  fld   dword ptr [eax]
  fmul  dword ptr [ecx]
  fstp  dword ptr [eax]
  fld   dword ptr [eax+4]
  fmul  dword ptr [ecx]
  fstp  dword ptr [eax+4]

  add   eax, 8
  dec   edx
  jnz   @@FPULoop

@@FPULone:
  pop   edx
  test  edx, 1
  jz    @@End

  fld   dword ptr [eax]
  fmul  dword ptr [ecx]
  fstp  dword ptr [eax]

@@End:
  {$ELSE}
var
  i: Integer;
begin
  for i        := 0 to nb - 1 do
    values^[i] := values^[i] * factor;
{$ENDIF}
end;

// ScaleFloatArray (array)
//
procedure ScaleFloatArray(var values: TSingleArray; factor: single);
begin
  if Length(values) > 0 then
    ScaleFloatArray(@values[0], Length(values), factor);
end;

// OffsetFloatArray (raw)
//
procedure OffsetFloatArray(values: PSingleArray; nb: Integer; var delta: single);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test vSIMD, 1
  jz @@FPU

  push  edx
  shr   edx, 2
  or    edx, edx
  jz    @@FPU

  db $0F,$6E,$39           /// movd  mm7, [ecx]
  db $0F,$62,$FF           /// punpckldq   mm7, mm7

@@3DNowLoop:
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$00           /// movq  mm0, [eax]
  db $0F,$6F,$48,$08       /// movq  mm1, [eax+8]
  db $0F,$0F,$C7,$9E       /// pfadd mm0, mm7
  db $0F,$0F,$CF,$9E       /// pfadd mm1, mm7
  db $0F,$7F,$00           /// movq  [eax], mm0
  db $0F,$7F,$48,$08       /// movq  [eax+8], mm1

  add   eax, 16
  dec   edx
  jnz   @@3DNowLoop

  pop   edx
  and   edx, 3
  db $0F,$0E               /// femms

@@FPU:
  push  edx
  shr   edx, 1
  or    edx, edx
  jz    @@FPULone

@@FPULoop:
  fld   dword ptr [eax]
  fadd  dword ptr [ecx]
  fstp  dword ptr [eax]
  fld   dword ptr [eax+4]
  fadd  dword ptr [ecx]
  fstp  dword ptr [eax+4]

  add   eax, 8
  dec   edx
  jnz   @@FPULoop

@@FPULone:
  pop   edx
  test  edx, 1
  jz    @@End

  fld   dword ptr [eax]
  fadd  dword ptr [ecx]
  fstp  dword ptr [eax]

@@End:
  {$ELSE}
var
  i: Integer;
begin
  for i        := 0 to nb - 1 do
    values^[i] := values^[i] + delta;
{$ENDIF}
end;

// ScaleFloatArray (array)
//
procedure OffsetFloatArray(var values: array of single; delta: single);
begin
  if Length(values) > 0 then
    ScaleFloatArray(@values[0], Length(values), delta);
end;

// OffsetFloatArray (raw, raw)
//
procedure OffsetFloatArray(valuesDest, valuesDelta: PSingleArray; nb: Integer);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  test  ecx, ecx
  jz    @@End

@@FPULoop:
  dec   ecx
  fld   dword ptr [eax+ecx*4]
  fadd  dword ptr [edx+ecx*4]
  fstp  dword ptr [eax+ecx*4]
  jnz   @@FPULoop

@@End:
  {$ELSE}
var
  i: Integer;
begin
  for i            := 0 to nb - 1 do
    valuesDest^[i] := valuesDest^[i] + valuesDelta^[i];
{$ENDIF}
end;

// MaxXYZComponent
//
function MaxXYZComponent(const V: TVector): single; overload;
begin
  result := MaxFloat(V.V[0], V.V[1], V.V[2]);
end;

// MaxXYZComponent
//
function MaxXYZComponent(const V: TAffineVector): single; overload;
begin
  result := MaxFloat(V.V[0], V.V[1], V.V[2]);
end;

// MinXYZComponent
//
function MinXYZComponent(const V: TVector): single; overload;
begin
  if V.V[0] <= V.V[1] then
    if V.V[0] <= V.V[2] then
      result := V.V[0]
    else if V.V[2] <= V.V[1] then
      result := V.V[2]
    else
      result := V.V[1]
  else if V.V[1] <= V.V[2] then
    result := V.V[1]
  else if V.V[2] <= V.V[0] then
    result := V.V[2]
  else
    result := V.V[0];
end;

// MinXYZComponent
//
function MinXYZComponent(const V: TAffineVector): single; overload;
begin
  result := MinFloat(V.V[0], V.V[1], V.V[2]);
end;

// MaxAbsXYZComponent
//
function MaxAbsXYZComponent(V: TVector): single;
begin
  AbsVector(V);
  result := MaxXYZComponent(V);
end;

// MinAbsXYZComponent
//
function MinAbsXYZComponent(V: TVector): single;
begin
  AbsVector(V);
  result := MinXYZComponent(V);
end;

// MaxVector (hmg)
//
procedure MaxVector(var V: TVector; const V1: TVector);
begin
  if V1.V[0] > V.V[0] then
    V.V[0] := V1.V[0];
  if V1.V[1] > V.V[1] then
    V.V[1] := V1.V[1];
  if V1.V[2] > V.V[2] then
    V.V[2] := V1.V[2];
  if V1.V[3] > V.V[3] then
    V.V[3] := V1.V[3];
end;

// MaxVector (affine)
//
procedure MaxVector(var V: TAffineVector; const V1: TAffineVector); overload;
begin
  if V1.V[0] > V.V[0] then
    V.V[0] := V1.V[0];
  if V1.V[1] > V.V[1] then
    V.V[1] := V1.V[1];
  if V1.V[2] > V.V[2] then
    V.V[2] := V1.V[2];
end;

// MinVector (hmg)
//
procedure MinVector(var V: TVector; const V1: TVector);
begin
  if V1.V[0] < V.V[0] then
    V.V[0] := V1.V[0];
  if V1.V[1] < V.V[1] then
    V.V[1] := V1.V[1];
  if V1.V[2] < V.V[2] then
    V.V[2] := V1.V[2];
  if V1.V[3] < V.V[3] then
    V.V[3] := V1.V[3];
end;

// MinVector (affine)
//
procedure MinVector(var V: TAffineVector; const V1: TAffineVector);
begin
  if V1.V[0] < V.V[0] then
    V.V[0] := V1.V[0];
  if V1.V[1] < V.V[1] then
    V.V[1] := V1.V[1];
  if V1.V[2] < V.V[2] then
    V.V[2] := V1.V[2];
end;

// SortArrayAscending (extended)
//
procedure SortArrayAscending(var a: array of Extended);
var
  i, J, M: Integer;
  buf    : Extended;
begin
  for i := Low(a) to High(a) - 1 do
  begin
    M     := i;
    for J := i + 1 to High(a) do
      if a[J] < a[M] then
        M := J;
    if M <> i then
    begin
      buf  := a[M];
      a[M] := a[i];
      a[i] := buf;
    end;
  end;
end;

// ClampValue (min-max)
//
function ClampValue(const aValue, aMin, aMax: single): single;
// begin
{$IFNDEF GEOMETRY_NO_ASM}
asm   // 118
  fld   aValue
  fcom  aMin
  fstsw ax
  sahf
  jb    @@ReturnMin
@@CompMax:
  fcom  aMax
  fstsw ax
  sahf
  jnbe  @@ReturnMax
  pop   ebp
  ret   $0C

@@ReturnMax:
  fld   aMax
  jmp @@End
@@ReturnMin:
  fld   aMin
@@End:
  ffree st(1)
end;
{$ELSE}
begin // 134
  if aValue < aMin then
    result := aMin
  else if aValue > aMax then
    result := aMax
  else
    result := aValue;
end;
{$ENDIF}

// ClampValue (min-)
//
function ClampValue(const aValue, aMin: single): single;
begin
  if aValue < aMin then
    result := aMin
  else
    result := aValue;
end;

// MakeAffineDblVector
//
function MakeAffineDblVector(var V: array of double): TAffineDblVector;
begin
  result.V[0] := V[0];
  result.V[1] := V[1];
  result.V[2] := V[2];
end;

// MakeDblVector
//
function MakeDblVector(var V: array of double): THomogeneousDblVector;
// creates a vector from given values
// EAX contains address of V
// ECX contains address to result vector
// EDX contains highest index of V
{$IFNDEF GEOMETRY_NO_ASM}
asm
  PUSH EDI
  PUSH ESI
  MOV EDI, ECX
  MOV ESI, EAX
  MOV ECX, 8
  REP MOVSD
  POP ESI
  POP EDI
  {$ELSE}
begin
  result.V[0] := V[0];
  result.V[1] := V[1];
  result.V[2] := V[2];
  result.V[3] := V[3];
{$ENDIF}
end;

// PointInPolygon
//
function PointInPolygon(var xp, yp: array of single; X, Y: single): Boolean;
// The code below is from Wm. Randolph Franklin <wrf@ecse.rpi.edu>
// with some minor modifications for speed.  It returns 1 for strictly
// interior points, 0 for strictly exterior, and 0 or 1 for points on
// the boundary.
var
  i, J: Integer;
begin
  result := False;
  if High(xp) = High(yp) then
  begin
    J     := High(xp);
    for i := 0 to High(xp) do
    begin
      if ((((yp[i] <= Y) and (Y < yp[J])) or ((yp[J] <= Y) and (Y < yp[i]))) and
        (X < (xp[J] - xp[i]) * (Y - yp[i]) / (yp[J] - yp[i]) + xp[i])) then
        result := not result;
      J        := i;
    end;
  end;
end;

// IsPointInPolygon
//
function IsPointInPolygon(Polygon: array of TPoint; p: TPoint): Boolean;
var
  a     : array of TPoint;
  n, i  : Integer;
  inside: Boolean;
begin
  n := High(Polygon) + 1;
  SetLength(a, n + 2);
  a[0]     := p;
  for i    := 1 to n do
    a[i]   := Polygon[i - 1];
  a[n + 1] := a[0];
  inside   := True;

  for i := 1 to n do
  begin
    if (a[0].Y > a[i].Y) xor (a[0].Y <= a[i + 1].Y) then
      Continue;
    if (a[0].X - a[i].X) < ((a[0].Y - a[i].Y) * (a[i + 1].X - a[i].X) / (a[i + 1].Y - a[i].Y)) then
      inside := not inside;
  end;
  inside := not inside;

  result := inside;
end;

// DivMod
//
procedure DivMod(Dividend: Integer; Divisor: Word; var result, Remainder: Word);
{$IFNDEF GEOMETRY_NO_ASM}
asm
  push  ebx
  mov   ebx, edx
  mov   edx, eax
  shr   edx, 16
  div   bx
  mov   ebx, remainder
  mov   [ecx], ax
  mov   [ebx], dx
  pop   ebx
  {$ELSE}
begin
  result    := Dividend div Divisor;
  Remainder := Dividend mod Divisor;
{$ENDIF}
end;

// ConvertRotation
//
function ConvertRotation(const Angles: TAffineVector): TVector;

{ Rotation of the Angle t about the axis (X, Y, Z) is given by:

  | X^2 + (1-X^2) Cos(t),    XY(1-Cos(t))  +  Z Sin(t), XZ(1-Cos(t))-Y Sin(t) |
  M = | XY(1-Cos(t))-Z Sin(t), Y^2 + (1-Y^2) Cos(t),      YZ(1-Cos(t)) + X Sin(t) |
  | XZ(1-Cos(t)) + Y Sin(t), YZ(1-Cos(t))-X Sin(t),   Z^2 + (1-Z^2) Cos(t)    |

  Rotation about the three axes (Angles a1, a2, a3) can be represented as
  the product of the individual rotation matrices:

  | 1  0       0       | | Cos(a2) 0 -Sin(a2) | |  Cos(a3) Sin(a3) 0 |
  | 0  Cos(a1) Sin(a1) | * | 0       1  0       | * | -Sin(a3) Cos(a3) 0 |
  | 0 -Sin(a1) Cos(a1) | | Sin(a2) 0  Cos(a2) | |  0       0       1 |
  Mx                       My                     Mz

  We now want to solve for X, Y, Z, and t given 9 equations in 4 unknowns.
  Using the diagonal elements of the two matrices, we get:

  X^2 + (1-X^2) Cos(t) = M[0][0]
  Y^2 + (1-Y^2) Cos(t) = M[1][1]
  Z^2 + (1-Z^2) Cos(t) = M[2][2]

  Adding the three equations, we get:

  X^2  +  Y^2  +  Z^2 - (M[0][0]  +  M[1][1]  +  M[2][2]) =
  - (3 - X^2 - Y^2 - Z^2) Cos(t)

  Since (X^2  +  Y^2  +  Z^2) = 1, we can rewrite as:

  Cos(t) = (1 - (M[0][0]  +  M[1][1]  +  M[2][2])) / 2

  Solving for t, we get:

  t = Acos(((M[0][0]  +  M[1][1]  +  M[2][2]) - 1) / 2)

  We can substitute t into the equations for X^2, Y^2, and Z^2 above
  to get the values for X, Y, and Z.  To find the proper signs we note
  that:

  2 X Sin(t) = M[1][2] - M[2][1]
  2 Y Sin(t) = M[2][0] - M[0][2]
  2 Z Sin(t) = M[0][1] - M[1][0]
}
var
  Axis1, Axis2                 : TVector3f;
  M, m1, m2                    : TMatrix;
  cost, cost1, sint, s1, s2, s3: single;
  i                            : Integer;
begin
  // see if we are only rotating about a single Axis
  if Abs(Angles.V[X]) < EPSILON then
  begin
    if Abs(Angles.V[Y]) < EPSILON then
    begin
      SetVector(result, 0, 0, 1, Angles.V[Z]);
      Exit;
    end
    else if Abs(Angles.V[Z]) < EPSILON then
    begin
      SetVector(result, 0, 1, 0, Angles.V[Y]);
      Exit;
    end
  end
  else if (Abs(Angles.V[Y]) < EPSILON) and (Abs(Angles.V[Z]) < EPSILON) then
  begin
    SetVector(result, 1, 0, 0, Angles.V[X]);
    Exit;
  end;

  // make the rotation matrix
  Axis1 := XVector;
  M     := CreateRotationMatrix(Axis1, Angles.V[X]);

  Axis2 := YVector;
  m2    := CreateRotationMatrix(Axis2, Angles.V[Y]);
  m1    := MatrixMultiply(M, m2);

  Axis2 := ZVector;
  m2    := CreateRotationMatrix(Axis2, Angles.V[Z]);
  M     := MatrixMultiply(m1, m2);

  cost := ((M.X.X + M.Y.Y + M.Z.Z) - 1) / 2;
  if cost < -1 then
    cost := -1
  else if cost > 1 - EPSILON then
  begin
    // Bad Angle - this would cause a crash
    SetVector(result, XHmgVector);
    Exit;
  end;

  cost1 := 1 - cost;
  SetVector(result, Sqrt((M.X.X - cost) / cost1), Sqrt((M.Y.Y - cost) / cost1),
    Sqrt((M.Z.Z - cost) / cost1), VectorGeometry.ArcCos(cost));

  sint := 2 * Sqrt(1 - cost * cost); // This is actually 2 Sin(t)

  // Determine the proper signs
  for i := 0 to 7 do
  begin
    if (i and 1) > 1 then
      s1 := -1
    else
      s1 := 1;
    if (i and 2) > 1 then
      s2 := -1
    else
      s2 := 1;
    if (i and 4) > 1 then
      s3 := -1
    else
      s3 := 1;
    if (Abs(s1 * result.V[X] * sint - M.Y.Z + M.Z.Y) < EPSILON2) and
      (Abs(s2 * result.V[Y] * sint - M.Z.X + M.X.Z) < EPSILON2) and
      (Abs(s3 * result.V[Z] * sint - M.X.Y + M.Y.X) < EPSILON2) then
    begin
      // We found the right combination of signs
      result.V[X] := result.V[X] * s1;
      result.V[Y] := result.V[Y] * s2;
      result.V[Z] := result.V[Z] * s3;
      Exit;
    end;
  end;
end;

// QuaternionSlerp
//
function QuaternionSlerp(const QStart, QEnd: TQuaternion; Spin: Integer; T: single): TQuaternion;
var
  beta,           // complementary interp parameter
  Theta,          // Angle between A and B
  sint, cost,     // sine, cosine of theta
  phi  : single;  // theta plus spins
  bflip: Boolean; // use negativ t?
begin
  // cosine theta
  cost := VectorAngleCosine(QStart.ImagPart, QEnd.ImagPart);

  // if QEnd is on opposite hemisphere from QStart, use -QEnd instead
  if cost < 0 then
  begin
    cost  := -cost;
    bflip := True;
  end
  else
    bflip := False;

  // if QEnd is (within precision limits) the same as QStart,
  // just linear interpolate between QStart and QEnd.
  // Can't do spins, since we don't know what direction to spin.

  if (1 - cost) < EPSILON then
    beta := 1 - T
  else
  begin
    // normal case
    Theta := VectorGeometry.ArcCos(cost);
    phi   := Theta + Spin * PI;
    sint  := Sin(Theta);
    beta  := Sin(Theta - T * phi) / sint;
    T     := Sin(T * phi) / sint;
  end;

  if bflip then
    T := -T;

  // interpolate
  result.ImagPart.V[X] := beta * QStart.ImagPart.V[X] + T * QEnd.ImagPart.V[X];
  result.ImagPart.V[Y] := beta * QStart.ImagPart.V[Y] + T * QEnd.ImagPart.V[Y];
  result.ImagPart.V[Z] := beta * QStart.ImagPart.V[Z] + T * QEnd.ImagPart.V[Z];
  result.RealPart      := beta * QStart.RealPart + T * QEnd.RealPart;
end;

// QuaternionSlerp
//
function QuaternionSlerp(const source, dest: TQuaternion; const T: single): TQuaternion;
var
  to1                                : array [0 .. 4] of single;
  omega, cosom, sinom, scale0, scale1: Extended;
  // t goes from 0 to 1
  // absolute rotations
begin
  // calc cosine
  cosom := source.ImagPart.V[0] * dest.ImagPart.V[0] + source.ImagPart.V[1] * dest.ImagPart.V[1] +
    source.ImagPart.V[2] * dest.ImagPart.V[2] + source.RealPart * dest.RealPart;
  // adjust signs (if necessary)
  if cosom < 0 then
  begin
    cosom  := -cosom;
    to1[0] := -dest.ImagPart.V[0];
    to1[1] := -dest.ImagPart.V[1];
    to1[2] := -dest.ImagPart.V[2];
    to1[3] := -dest.RealPart;
  end
  else
  begin
    to1[0] := dest.ImagPart.V[0];
    to1[1] := dest.ImagPart.V[1];
    to1[2] := dest.ImagPart.V[2];
    to1[3] := dest.RealPart;
  end;
  // calculate coefficients
  if ((1.0 - cosom) > EPSILON2) then
  begin // standard case (slerp)
    omega  := VectorGeometry.ArcCos(cosom);
    sinom  := 1 / Sin(omega);
    scale0 := Sin((1.0 - T) * omega) * sinom;
    scale1 := Sin(T * omega) * sinom;
  end
  else
  begin // "from" and "to" quaternions are very close
    // ... so we can do a linear interpolation
    scale0 := 1.0 - T;
    scale1 := T;
  end;
  // calculate final values
  result.ImagPart.V[0] := scale0 * source.ImagPart.V[0] + scale1 * to1[0];
  result.ImagPart.V[1] := scale0 * source.ImagPart.V[1] + scale1 * to1[1];
  result.ImagPart.V[2] := scale0 * source.ImagPart.V[2] + scale1 * to1[2];
  result.RealPart      := scale0 * source.RealPart + scale1 * to1[3];
  NormalizeQuaternion(result);
end;

// VectorDblToFlt
//
function VectorDblToFlt(const V: THomogeneousDblVector): THomogeneousVector;
// converts a vector containing double sized values into a vector with single sized values
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  QWORD PTR [EAX]
  FSTP DWORD PTR [EDX]
  FLD  QWORD PTR [EAX + 8]
  FSTP DWORD PTR [EDX + 4]
  FLD  QWORD PTR [EAX + 16]
  FSTP DWORD PTR [EDX + 8]
  FLD  QWORD PTR [EAX + 24]
  FSTP DWORD PTR [EDX + 12]
  {$ELSE}
begin
{$HINTS OFF}
  result.V[0] := V.V[0];
  result.V[1] := V.V[1];
  result.V[2] := V.V[2];
  result.V[3] := V.V[3];
{$HINTS ON}
{$ENDIF}
end;

// VectorAffineDblToFlt
//
function VectorAffineDblToFlt(const V: TAffineDblVector): TAffineVector;
// converts a vector containing double sized values into a vector with single sized values
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  QWORD PTR [EAX]
  FSTP DWORD PTR [EDX]
  FLD  QWORD PTR [EAX + 8]
  FSTP DWORD PTR [EDX + 4]
  FLD  QWORD PTR [EAX + 16]
  FSTP DWORD PTR [EDX + 8]
  {$ELSE}
begin
{$HINTS OFF}
  result.V[0] := V.V[0];
  result.V[1] := V.V[1];
  result.V[2] := V.V[2];
{$HINTS ON}
{$ENDIF}
end;

// VectorAffineFltToDbl
//
function VectorAffineFltToDbl(const V: TAffineVector): TAffineDblVector;
// converts a vector containing single sized values into a vector with double sized values
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSTP QWORD PTR [EDX]
  FLD  DWORD PTR [EAX + 4]
  FSTP QWORD PTR [EDX + 8]
  FLD  DWORD PTR [EAX + 8]
  FSTP QWORD PTR [EDX + 16]
  {$ELSE}
begin
  result.V[0] := V.V[0];
  result.V[1] := V.V[1];
  result.V[2] := V.V[2];
{$ENDIF}
end;

// VectorFltToDbl
//
function VectorFltToDbl(const V: TVector): THomogeneousDblVector;
// converts a vector containing single sized values into a vector with double sized values
{$IFNDEF GEOMETRY_NO_ASM}
asm
  FLD  DWORD PTR [EAX]
  FSTP QWORD PTR [EDX]
  FLD  DWORD PTR [EAX + 4]
  FSTP QWORD PTR [EDX + 8]
  FLD  DWORD PTR [EAX + 8]
  FSTP QWORD PTR [EDX + 16]
  FLD  DWORD PTR [EAX + 12]
  FSTP QWORD PTR [EDX + 24]
  {$ELSE}
begin
  result.V[0] := V.V[0];
  result.V[1] := V.V[1];
  result.V[2] := V.V[2];
  result.V[3] := V.V[3];
{$ENDIF}
end;

// ----------------- coordinate system manipulation functions -----------------------------------------------------------

// Turn (Y axis)
//
function Turn(const Matrix: TMatrix; angle: single): TMatrix;
begin
  result := MatrixMultiply(Matrix, CreateRotationMatrix(AffineVectorMake(Matrix.V[1].V[0],
    Matrix.V[1].V[1], Matrix.V[1].V[2]), angle));
end;

// Turn (direction)
//
function Turn(const Matrix: TMatrix; const MasterUp: TAffineVector; angle: single): TMatrix;
begin
  result := MatrixMultiply(Matrix, CreateRotationMatrix(MasterUp, angle));
end;

// Pitch (X axis)
//
function Pitch(const Matrix: TMatrix; angle: single): TMatrix;
begin
  result := MatrixMultiply(Matrix, CreateRotationMatrix(AffineVectorMake(Matrix.V[0].V[0],
    Matrix.V[0].V[1], Matrix.V[0].V[2]), angle));
end;

// Pitch (direction)
//
function Pitch(const Matrix: TMatrix; const MasterRight: TAffineVector; angle: single)
  : TMatrix; overload;
begin
  result := MatrixMultiply(Matrix, CreateRotationMatrix(MasterRight, angle));
end;

// Roll (Z axis)
//
function Roll(const Matrix: TMatrix; angle: single): TMatrix;
begin
  result := MatrixMultiply(Matrix, CreateRotationMatrix(AffineVectorMake(Matrix.V[2].V[0],
    Matrix.V[2].V[1], Matrix.V[2].V[2]), angle));
end;

// Roll (direction)
//
function Roll(const Matrix: TMatrix; const MasterDirection: TAffineVector; angle: single)
  : TMatrix; overload;
begin
  result := MatrixMultiply(Matrix, CreateRotationMatrix(MasterDirection, angle));
end;

// RayCastPlaneIntersect (plane defined by point+normal)
//
function RayCastPlaneIntersect(const rayStart, rayVector: TVector;
  const planePoint, planeNormal: TVector; intersectPoint: PVector = nil): Boolean;
var
  sp  : TVector;
  T, d: single;
begin
  d      := VectorDotProduct(rayVector, planeNormal);
  result := ((d > EPSILON2) or (d < -EPSILON2));
  if result and Assigned(intersectPoint) then
  begin
    VectorSubtract(planePoint, rayStart, sp);
    d := 1 / d; // will keep one FPU unit busy during dot product calculation
    T := VectorDotProduct(sp, planeNormal) * d;
    if T > 0 then
      VectorCombine(rayStart, rayVector, T, intersectPoint^)
    else
      result := False;
  end;
end;

// RayCastPlaneXZIntersect
//
function RayCastPlaneXZIntersect(const rayStart, rayVector: TVector; const planeY: single;
  intersectPoint: PVector = nil): Boolean;
var
  T: single;
begin
  if rayVector.V[1] = 0 then
    result := False
  else
  begin
    T := (rayStart.V[1] - planeY) / rayVector.V[1];
    if T < 0 then
    begin
      if Assigned(intersectPoint) then
        VectorCombine(rayStart, rayVector, T, intersectPoint^);
      result := True;
    end
    else
      result := False;
  end;
end;

// RayCastTriangleIntersect
//
function RayCastTriangleIntersect(const rayStart, rayVector: TVector;
  const p1, p2, p3: TAffineVector; intersectPoint: PVector = nil;
  intersectNormal: PVector = nil): Boolean;
var
  pvec                : TAffineVector;
  V1, V2, qvec, tvec  : TVector;
  T, u, V, det, invDet: single;
begin
  VectorSubtract(p2, p1, V1);
  VectorSubtract(p3, p1, V2);
  VectorCrossProduct(rayVector, V2, pvec);
  det := VectorDotProduct(V1, pvec);
  if ((det < EPSILON2) and (det > -EPSILON2)) then
  begin // vector is parallel to triangle's plane
    result := False;
    Exit;
  end;
  invDet := cOne / det;
  VectorSubtract(rayStart, p1, tvec);
  u := VectorDotProduct(tvec, pvec) * invDet;
  if (u < 0) or (u > 1) then
    result := False
  else
  begin
    qvec   := VectorCrossProduct(tvec, V1);
    V      := VectorDotProduct(rayVector, qvec) * invDet;
    result := (V >= 0) and (u + V <= 1);
    if result then
    begin
      T := VectorDotProduct(V2, qvec) * invDet;
      if T > 0 then
      begin
        if intersectPoint <> nil then
          VectorCombine(rayStart, rayVector, T, intersectPoint^);
        if intersectNormal <> nil then
          VectorCrossProduct(V1, V2, intersectNormal^);
      end
      else
        result := False;
    end;
  end;
end;

// RayCastMinDistToPoint
//
function RayCastMinDistToPoint(const rayStart, rayVector: TVector; const point: TVector): single;
var
  proj: single;
begin
  proj := PointProject(point, rayStart, rayVector);
  if proj <= 0 then
    proj := 0; // rays don't go backward!
  result := VectorDistance(point, VectorCombine(rayStart, rayVector, 1, proj));
end;

// RayCastIntersectsSphere
//
function RayCastIntersectsSphere(const rayStart, rayVector: TVector; const sphereCenter: TVector;
  const SphereRadius: single): Boolean;
var
  proj: single;
begin
  proj := PointProject(sphereCenter, rayStart, rayVector);
  if proj <= 0 then
    proj := 0; // rays don't go backward!
  result := (VectorDistance2(sphereCenter, VectorCombine(rayStart, rayVector, 1, proj)) <=
    Sqr(SphereRadius));
end;

// RayCastSphereIntersect
//
function RayCastSphereIntersect(const rayStart, rayVector: TVector; const sphereCenter: TVector;
  const SphereRadius: single; var i1, i2: TVector): Integer;
var
  proj, d2 : single;
  id2      : Integer;
  projPoint: TVector;
begin
  proj := PointProject(sphereCenter, rayStart, rayVector);
  VectorCombine(rayStart, rayVector, proj, projPoint);
  d2  := SphereRadius * SphereRadius - VectorDistance2(sphereCenter, projPoint);
  id2 := PInteger(@d2)^;
  if id2 >= 0 then
  begin
    if id2 = 0 then
    begin
      if PInteger(@proj)^ > 0 then
      begin
        VectorCombine(rayStart, rayVector, proj, i1);
        result := 1;
        Exit;
      end;
    end
    else if id2 > 0 then
    begin
      d2 := Sqrt(d2);
      if proj >= d2 then
      begin
        VectorCombine(rayStart, rayVector, proj - d2, i1);
        VectorCombine(rayStart, rayVector, proj + d2, i2);
        result := 2;
        Exit;
      end
      else if proj + d2 >= 0 then
      begin
        VectorCombine(rayStart, rayVector, proj + d2, i1);
        result := 1;
        Exit;
      end;
    end;
  end;
  result := 0;
end;

// RayCastBoxIntersect
//
function RayCastBoxIntersect(const rayStart, rayVector, aMinExtent, aMaxExtent: TAffineVector;
  intersectPoint: PAffineVector = nil): Boolean;
var
  i, planeInd           : Integer;
  ResAFV, MaxDist, plane: TAffineVector;
  isMiddle              : array [0 .. 2] of Boolean;
begin
  // Find plane.
  result := True;
  for i  := 0 to 2 do
    if rayStart.V[i] < aMinExtent.V[i] then
    begin
      plane.V[i]  := aMinExtent.V[i];
      isMiddle[i] := False;
      result      := False;
    end
    else if rayStart.V[i] > aMaxExtent.V[i] then
    begin
      plane.V[i]  := aMaxExtent.V[i];
      isMiddle[i] := False;
      result      := False;
    end
    else
    begin
      isMiddle[i] := True;
    end;
  if result then
  begin
    // rayStart inside box.
    if intersectPoint <> nil then
      intersectPoint^ := rayStart;
  end
  else
  begin
    // Distance to plane.
    planeInd := 0;
    for i    := 0 to 2 do
      if isMiddle[i] or (rayVector.V[i] = 0) then
        MaxDist.V[i] := -1
      else
      begin
        MaxDist.V[i] := (plane.V[i] - rayStart.V[i]) / rayVector.V[i];
        if MaxDist.V[i] > 0 then
        begin
          if MaxDist.V[planeInd] < MaxDist.V[i] then
            planeInd := i;
          result     := True;
        end;
      end;
    // Inside box ?
    if result then
    begin
      for i := 0 to 2 do
        if planeInd = i then
          ResAFV.V[i] := plane.V[i]
        else
        begin
          ResAFV.V[i] := rayStart.V[i] + MaxDist.V[planeInd] * rayVector.V[i];
          result      := (ResAFV.V[i] >= aMinExtent.V[i]) and (ResAFV.V[i] <= aMaxExtent.V[i]);
          if not result then
            Exit;
        end;
      if intersectPoint <> nil then
        intersectPoint^ := ResAFV;
    end;
  end;
end;

// SphereVisibleRadius
//
function SphereVisibleRadius(distance, radius: single): single;
var
  d2, r2, ir, tr: single;
begin
  d2 := distance * distance;
  r2 := radius * radius;
  ir := Sqrt(d2 - r2);
  tr := (d2 + r2 - Sqr(ir)) / (2 * ir);

  result := Sqrt(r2 + Sqr(tr));
end;

// IntersectLinePlane
//
function IntersectLinePlane(const point, direction: TVector; const plane: THmgPlane;
  intersectPoint: PVector = nil): Integer;
var
  a, b: Extended;
  T   : single;
begin
  a := VectorDotProduct(plane, direction); // direction projected to plane normal
  b := PlaneEvaluatePoint(plane, point);   // distance to plane
  if a = 0 then
  begin // direction is parallel to plane
    if b = 0 then
      result := -1 // line is inside plane
    else
      result := 0; // line is outside plane
  end
  else
  begin
    if Assigned(intersectPoint) then
    begin
      T               := -b / a; // parameter of intersection
      intersectPoint^ := point;
      // calculate intersection = p + t*d
      CombineVector(intersectPoint^, direction, T);
    end;
    result := 1;
  end;
end;

// TriangleBoxIntersect
//
function IntersectTriangleBox(const p1, p2, p3, aMinExtent, aMaxExtent: TAffineVector): Boolean;
var
  RayDir, iPoint                         : TAffineVector;
  BoxDiagPt, BoxDiagPt2, BoxDiagDir, iPnt: TVector;
begin
  // Triangle edge (p2, p1) - Box intersection
  VectorSubtract(p2, p1, RayDir);
  result := RayCastBoxIntersect(p1, RayDir, aMinExtent, aMaxExtent, @iPoint);
  if result then
    result := VectorNorm(VectorSubtract(p1, iPoint)) < VectorNorm(VectorSubtract(p1, p2));
  if result then
    Exit;

  // Triangle edge (p3, p1) - Box intersection
  VectorSubtract(p3, p1, RayDir);
  result := RayCastBoxIntersect(p1, RayDir, aMinExtent, aMaxExtent, @iPoint);
  if result then
    result := VectorNorm(VectorSubtract(p1, iPoint)) < VectorNorm(VectorSubtract(p1, p3));
  if result then
    Exit;

  // Triangle edge (p2, p3) - Box intersection
  VectorSubtract(p2, p3, RayDir);
  result := RayCastBoxIntersect(p3, RayDir, aMinExtent, aMaxExtent, @iPoint);
  if result then
    result := VectorNorm(VectorSubtract(p3, iPoint)) < VectorNorm(VectorSubtract(p3, p2));
  if result then
    Exit;

  // Triangle - Box diagonal 1 intersection
  BoxDiagPt := VectorMake(aMinExtent);
  VectorSubtract(aMaxExtent, aMinExtent, BoxDiagDir);
  result := RayCastTriangleIntersect(BoxDiagPt, BoxDiagDir, p1, p2, p3, @iPnt);
  if result then
    result := VectorNorm(VectorSubtract(BoxDiagPt, iPnt)) <
      VectorNorm(VectorSubtract(aMaxExtent, aMinExtent));
  if result then
    Exit;

  // Triangle - Box diagonal 2 intersection
  BoxDiagPt  := VectorMake(aMinExtent.V[0], aMinExtent.V[1], aMaxExtent.V[2]);
  BoxDiagPt2 := VectorMake(aMaxExtent.V[0], aMaxExtent.V[1], aMinExtent.V[2]);
  VectorSubtract(BoxDiagPt2, BoxDiagPt, BoxDiagDir);
  result := RayCastTriangleIntersect(BoxDiagPt, BoxDiagDir, p1, p2, p3, @iPnt);
  if result then
    result := VectorNorm(VectorSubtract(BoxDiagPt, iPnt)) <
      VectorNorm(VectorSubtract(BoxDiagPt, BoxDiagPt2));
  if result then
    Exit;

  // Triangle - Box diagonal 3 intersection
  BoxDiagPt  := VectorMake(aMinExtent.V[0], aMaxExtent.V[1], aMinExtent.V[2]);
  BoxDiagPt2 := VectorMake(aMaxExtent.V[0], aMinExtent.V[1], aMaxExtent.V[2]);
  VectorSubtract(BoxDiagPt, BoxDiagPt, BoxDiagDir);
  result := RayCastTriangleIntersect(BoxDiagPt, BoxDiagDir, p1, p2, p3, @iPnt);
  if result then
    result := VectorLength(VectorSubtract(BoxDiagPt, iPnt)) <
      VectorLength(VectorSubtract(BoxDiagPt, BoxDiagPt));
  if result then
    Exit;

  // Triangle - Box diagonal 4 intersection
  BoxDiagPt  := VectorMake(aMaxExtent.V[0], aMinExtent.V[1], aMinExtent.V[2]);
  BoxDiagPt2 := VectorMake(aMinExtent.V[0], aMaxExtent.V[1], aMaxExtent.V[2]);
  VectorSubtract(BoxDiagPt, BoxDiagPt, BoxDiagDir);
  result := RayCastTriangleIntersect(BoxDiagPt, BoxDiagDir, p1, p2, p3, @iPnt);
  if result then
    result := VectorLength(VectorSubtract(BoxDiagPt, iPnt)) <
      VectorLength(VectorSubtract(BoxDiagPt, BoxDiagPt));
end;

// IntersectSphereBox
//
function IntersectSphereBox(const SpherePos: TVector; const SphereRadius: single;
  const BoxMatrix: TMatrix; // Up Direction and Right must be normalized!
  // Use CubDepht, CubeHeight and CubeWidth
  // for scale TGLCube.
  const BoxScale: TAffineVector; intersectPoint: PAffineVector = nil; normal: PAffineVector = nil;
  depth: PSingle = nil): Boolean;

  function dDOTByColumn(const V: TAffineVector; const M: TMatrix; const aColumn: Integer): single;
  begin
    result := V.V[0] * M.V[0].V[aColumn] + V.V[1] * M.V[1].V[aColumn] + V.V[2] * M.V[2].V[aColumn];
  end;

  function dDotByRow(const V: TAffineVector; const M: TMatrix; const aRow: Integer): single;
  begin
    // Equal with: Result := VectorDotProduct(v, AffineVectorMake(m[aRow]));
    result := V.V[0] * M.V[aRow].V[0] + V.V[1] * M.V[aRow].V[1] + V.V[2] * M.V[aRow].V[2];
  end;

  function dDotMatrByColumn(const V: TAffineVector; const M: TMatrix): TAffineVector;
  begin
    result.V[0] := dDOTByColumn(V, M, 0);
    result.V[1] := dDOTByColumn(V, M, 1);
    result.V[2] := dDOTByColumn(V, M, 2);
  end;

  function dDotMatrByRow(const V: TAffineVector; const M: TMatrix): TAffineVector;
  begin
    result.V[0] := dDotByRow(V, M, 0);
    result.V[1] := dDotByRow(V, M, 1);
    result.V[2] := dDotByRow(V, M, 2);
  end;

var
  tmp, l, T, p, Q, r               : TAffineVector;
  FaceDistance, MinDistance, Depth1: single;
  mini, i                          : Integer;
  isSphereCenterInsideBox          : Boolean;
begin
  // this is easy. get the sphere center `p' relative to the box, and then clip
  // that to the boundary of the box (call that point `q'). if q is on the
  // boundary of the box and |p-q| is <= sphere radius, they touch.
  // if q is inside the box, the sphere is inside the box, so set a contact
  // normal to push the sphere to the closest box face.

  p.V[0] := SpherePos.V[0] - BoxMatrix.V[3].V[0];
  p.V[1] := SpherePos.V[1] - BoxMatrix.V[3].V[1];
  p.V[2] := SpherePos.V[2] - BoxMatrix.V[3].V[2];

  isSphereCenterInsideBox := True;
  for i                   := 0 to 2 do
  begin
    l.V[i] := 0.5 * BoxScale.V[i];
    T.V[i] := dDotByRow(p, BoxMatrix, i);
    if T.V[i] < -l.V[i] then
    begin
      T.V[i]                  := -l.V[i];
      isSphereCenterInsideBox := False;
    end
    else if T.V[i] > l.V[i] then
    begin
      T.V[i]                  := l.V[i];
      isSphereCenterInsideBox := False;
    end;
  end;

  if isSphereCenterInsideBox then
  begin

    MinDistance := l.V[0] - Abs(T.V[0]);
    mini        := 0;
    for i       := 1 to 2 do
    begin
      FaceDistance := l.V[i] - Abs(T.V[i]);
      if FaceDistance < MinDistance then
      begin
        MinDistance := FaceDistance;
        mini        := i;
      end;
    end;

    if intersectPoint <> nil then
      intersectPoint^ := AffineVectorMake(SpherePos);

    if normal <> nil then
    begin
      tmp := NullVector;
      if T.V[mini] > 0 then
        tmp.V[mini] := 1
      else
        tmp.V[mini] := -1;
      normal^       := dDotMatrByRow(tmp, BoxMatrix);
    end;

    if depth <> nil then
      depth^ := MinDistance + SphereRadius;

    result := True;
  end
  else
  begin
    Q      := dDotMatrByColumn(T, BoxMatrix);
    r      := VectorSubtract(p, Q);
    Depth1 := SphereRadius - VectorLength(r);
    if Depth1 < 0 then
    begin
      result := False;
    end
    else
    begin
      if intersectPoint <> nil then
        intersectPoint^ := VectorAdd(Q, AffineVectorMake(BoxMatrix.V[3]));
      if normal <> nil then
      begin
        normal^ := VectorNormalize(r);
      end;
      if depth <> nil then
        depth^ := Depth1;
      result   := True;
    end;
  end;
end;

// ExtractFrustumFromModelViewProjection
//
function ExtractFrustumFromModelViewProjection(const modelViewProj: TMatrix): TFrustum;
begin
  with result do
  begin
    // extract left plane
    pLeft.V[0] := modelViewProj.V[0].V[3] + modelViewProj.V[0].V[0];
    pLeft.V[1] := modelViewProj.V[1].V[3] + modelViewProj.V[1].V[0];
    pLeft.V[2] := modelViewProj.V[2].V[3] + modelViewProj.V[2].V[0];
    pLeft.V[3] := modelViewProj.V[3].V[3] + modelViewProj.V[3].V[0];
    NormalizePlane(pLeft);
    // extract top plane
    pTop.V[0] := modelViewProj.V[0].V[3] - modelViewProj.V[0].V[1];
    pTop.V[1] := modelViewProj.V[1].V[3] - modelViewProj.V[1].V[1];
    pTop.V[2] := modelViewProj.V[2].V[3] - modelViewProj.V[2].V[1];
    pTop.V[3] := modelViewProj.V[3].V[3] - modelViewProj.V[3].V[1];
    NormalizePlane(pTop);
    // extract right plane
    pRight.V[0] := modelViewProj.V[0].V[3] - modelViewProj.V[0].V[0];
    pRight.V[1] := modelViewProj.V[1].V[3] - modelViewProj.V[1].V[0];
    pRight.V[2] := modelViewProj.V[2].V[3] - modelViewProj.V[2].V[0];
    pRight.V[3] := modelViewProj.V[3].V[3] - modelViewProj.V[3].V[0];
    NormalizePlane(pRight);
    // extract bottom plane
    pBottom.V[0] := modelViewProj.V[0].V[3] + modelViewProj.V[0].V[1];
    pBottom.V[1] := modelViewProj.V[1].V[3] + modelViewProj.V[1].V[1];
    pBottom.V[2] := modelViewProj.V[2].V[3] + modelViewProj.V[2].V[1];
    pBottom.V[3] := modelViewProj.V[3].V[3] + modelViewProj.V[3].V[1];
    NormalizePlane(pBottom);
    // extract far plane
    pFar.V[0] := modelViewProj.V[0].V[3] - modelViewProj.V[0].V[2];
    pFar.V[1] := modelViewProj.V[1].V[3] - modelViewProj.V[1].V[2];
    pFar.V[2] := modelViewProj.V[2].V[3] - modelViewProj.V[2].V[2];
    pFar.V[3] := modelViewProj.V[3].V[3] - modelViewProj.V[3].V[2];
    NormalizePlane(pFar);
    // extract near plane
    pNear.V[0] := modelViewProj.V[0].V[3] + modelViewProj.V[0].V[2];
    pNear.V[1] := modelViewProj.V[1].V[3] + modelViewProj.V[1].V[2];
    pNear.V[2] := modelViewProj.V[2].V[3] + modelViewProj.V[2].V[2];
    pNear.V[3] := modelViewProj.V[3].V[3] + modelViewProj.V[3].V[2];
    NormalizePlane(pNear);
  end;
end;

// IsVolumeClipped
//
function IsVolumeClipped(const objPos: TAffineVector; const objRadius: single;
  const Frustum: TFrustum): Boolean;
var
  negRadius: single;
begin
  negRadius := -objRadius;
  result    := (PlaneEvaluatePoint(Frustum.pLeft, objPos) < negRadius) or
    (PlaneEvaluatePoint(Frustum.pTop, objPos) < negRadius) or
    (PlaneEvaluatePoint(Frustum.pRight, objPos) < negRadius) or
    (PlaneEvaluatePoint(Frustum.pBottom, objPos) < negRadius) or
    (PlaneEvaluatePoint(Frustum.pNear, objPos) < negRadius) or
    (PlaneEvaluatePoint(Frustum.pFar, objPos) < negRadius);
end;

// IsVolumeClipped
//
function IsVolumeClipped(const objPos: TVector; const objRadius: single;
  const Frustum: TFrustum): Boolean;
begin
  result := IsVolumeClipped(PAffineVector(@objPos)^, objRadius, Frustum);
end;

// IsVolumeClipped
//
function IsVolumeClipped(const min, max: TAffineVector; const Frustum: TFrustum): Boolean;
begin
  // change box to sphere
  result := IsVolumeClipped(VectorScale(VectorAdd(min, max), 0.5),
    VectorDistance(min, max) * 0.5, Frustum);
end;

// MakeParallelProjectionMatrix
//
function MakeParallelProjectionMatrix(const plane: THmgPlane; const dir: TVector): TMatrix;
// Based on material from a course by William D. Shoaff (www.cs.fit.edu)
var
  dot, invDot: single;
begin
  dot := plane.V[0] * dir.V[0] + plane.V[1] * dir.V[1] + plane.V[2] * dir.V[2];
  if Abs(dot) < 1E-5 then
  begin
    result := IdentityHmgMatrix;
    Exit;
  end;
  invDot := 1 / dot;

  result.V[0].V[0] := (plane.V[1] * dir.V[1] + plane.V[2] * dir.V[2]) * invDot;
  result.V[1].V[0] := (-plane.V[1] * dir.V[0]) * invDot;
  result.V[2].V[0] := (-plane.V[2] * dir.V[0]) * invDot;
  result.V[3].V[0] := (-plane.V[3] * dir.V[0]) * invDot;

  result.V[0].V[1] := (-plane.V[0] * dir.V[1]) * invDot;
  result.V[1].V[1] := (plane.V[0] * dir.V[0] + plane.V[2] * dir.V[2]) * invDot;
  result.V[2].V[1] := (-plane.V[2] * dir.V[1]) * invDot;
  result.V[3].V[1] := (-plane.V[3] * dir.V[1]) * invDot;

  result.V[0].V[2] := (-plane.V[0] * dir.V[2]) * invDot;
  result.V[1].V[2] := (-plane.V[1] * dir.V[2]) * invDot;
  result.V[2].V[2] := (plane.V[0] * dir.V[0] + plane.V[1] * dir.V[1]) * invDot;
  result.V[3].V[2] := (-plane.V[3] * dir.V[2]) * invDot;

  result.V[0].V[3] := 0;
  result.V[1].V[3] := 0;
  result.V[2].V[3] := 0;
  result.V[3].V[3] := 1;
end;

// MakeShadowMatrix
//
function MakeShadowMatrix(const planePoint, planeNormal, lightPos: TVector): TMatrix;
var
  planeNormal3, dot: single;
begin
  // Find the last coefficient by back substitutions
  planeNormal3 := -(planeNormal.V[0] * planePoint.V[0] + planeNormal.V[1] * planePoint.V[1] +
    planeNormal.V[2] * planePoint.V[2]);
  // Dot product of plane and light position
  dot := planeNormal.V[0] * lightPos.V[0] + planeNormal.V[1] * lightPos.V[1] + planeNormal.V[2] *
    lightPos.V[2] + planeNormal3 * lightPos.V[3];
  // Now do the projection
  // First column
  result.V[0].V[0] := dot - lightPos.V[0] * planeNormal.V[0];
  result.V[1].V[0] := -lightPos.V[0] * planeNormal.V[1];
  result.V[2].V[0] := -lightPos.V[0] * planeNormal.V[2];
  result.V[3].V[0] := -lightPos.V[0] * planeNormal3;
  // Second column
  result.V[0].V[1] := -lightPos.V[1] * planeNormal.V[0];
  result.V[1].V[1] := dot - lightPos.V[1] * planeNormal.V[1];
  result.V[2].V[1] := -lightPos.V[1] * planeNormal.V[2];
  result.V[3].V[1] := -lightPos.V[1] * planeNormal3;
  // Third Column
  result.V[0].V[2] := -lightPos.V[2] * planeNormal.V[0];
  result.V[1].V[2] := -lightPos.V[2] * planeNormal.V[1];
  result.V[2].V[2] := dot - lightPos.V[2] * planeNormal.V[2];
  result.V[3].V[2] := -lightPos.V[2] * planeNormal3;
  // Fourth Column
  result.V[0].V[3] := -lightPos.V[3] * planeNormal.V[0];
  result.V[1].V[3] := -lightPos.V[3] * planeNormal.V[1];
  result.V[2].V[3] := -lightPos.V[3] * planeNormal.V[2];
  result.V[3].V[3] := dot - lightPos.V[3] * planeNormal3;
end;

// MakeReflectionMatrix
//
function MakeReflectionMatrix(const planePoint, planeNormal: TAffineVector): TMatrix;
var
  pv2: single;
begin
  // Precalcs
  pv2 := 2 * VectorDotProduct(planePoint, planeNormal);
  // 1st column
  result.V[0].V[0] := 1 - 2 * Sqr(planeNormal.V[0]);
  result.V[0].V[1] := -2 * planeNormal.V[0] * planeNormal.V[1];
  result.V[0].V[2] := -2 * planeNormal.V[0] * planeNormal.V[2];
  result.V[0].V[3] := 0;
  // 2nd column
  result.V[1].V[0] := -2 * planeNormal.V[1] * planeNormal.V[0];
  result.V[1].V[1] := 1 - 2 * Sqr(planeNormal.V[1]);
  result.V[1].V[2] := -2 * planeNormal.V[1] * planeNormal.V[2];
  result.V[1].V[3] := 0;
  // 3rd column
  result.V[2].V[0] := -2 * planeNormal.V[2] * planeNormal.V[0];
  result.V[2].V[1] := -2 * planeNormal.V[2] * planeNormal.V[1];
  result.V[2].V[2] := 1 - 2 * Sqr(planeNormal.V[2]);
  result.V[2].V[3] := 0;
  // 4th column
  result.V[3].V[0] := pv2 * planeNormal.V[0];
  result.V[3].V[1] := pv2 * planeNormal.V[1];
  result.V[3].V[2] := pv2 * planeNormal.V[2];
  result.V[3].V[3] := 1;
end;

// PackRotationMatrix
//
function PackRotationMatrix(const mat: TMatrix): TPackedRotationMatrix;
var
  Q: TQuaternion;
const
  cFact: single = 32767;
begin
  Q := QuaternionFromMatrix(mat);
  NormalizeQuaternion(Q);
{$HINTS OFF}
  if Q.RealPart < 0 then
  begin
    result[0] := Round(-Q.ImagPart.V[0] * cFact);
    result[1] := Round(-Q.ImagPart.V[1] * cFact);
    result[2] := Round(-Q.ImagPart.V[2] * cFact);
  end
  else
  begin
    result[0] := Round(Q.ImagPart.V[0] * cFact);
    result[1] := Round(Q.ImagPart.V[1] * cFact);
    result[2] := Round(Q.ImagPart.V[2] * cFact);
  end;
{$HINTS ON}
end;

// UnPackRotationMatrix
//
function UnPackRotationMatrix(const packedMatrix: TPackedRotationMatrix): TMatrix;
var
  Q: TQuaternion;
const
  cFact: single = 1 / 32767;
begin
  Q.ImagPart.V[0] := packedMatrix[0] * cFact;
  Q.ImagPart.V[1] := packedMatrix[1] * cFact;
  Q.ImagPart.V[2] := packedMatrix[2] * cFact;
  Q.RealPart      := 1 - VectorNorm(Q.ImagPart);
  if Q.RealPart < 0 then
    Q.RealPart := 0
  else
    Q.RealPart := Sqrt(Q.RealPart);
  result       := QuaternionToMatrix(Q);
end;

// BarycentricCoordinates
//
function BarycentricCoordinates(const V1, V2, V3, p: TAffineVector; var u, V: single): Boolean;
var
  a1, a2       : Integer;
  n, e1, e2, pt: TAffineVector;
begin
  // calculate edges
  VectorSubtract(V1, V3, e1);
  VectorSubtract(V2, V3, e2);

  // calculate p relative to v3
  VectorSubtract(p, V3, pt);

  // find the dominant axis
  n := VectorCrossProduct(e1, e2);
  AbsVector(n);
  a1 := 0;
  if n.V[1] > n.V[a1] then
    a1 := 1;
  if n.V[2] > n.V[a1] then
    a1 := 2;

  // use dominant axis for projection
  case a1 of
    0:
      begin
        a1 := 1;
        a2 := 2;
      end;
    1:
      begin
        a1 := 0;
        a2 := 2;
      end;
  else // 2:
    a1 := 0;
    a2 := 1;
  end;

  // solve for u and v
  u := (pt.V[a2] * e2.V[a1] - pt.V[a1] * e2.V[a2]) / (e1.V[a2] * e2.V[a1] - e1.V[a1] * e2.V[a2]);
  V := (pt.V[a2] * e1.V[a1] - pt.V[a1] * e1.V[a2]) / (e2.V[a2] * e1.V[a1] - e2.V[a1] * e1.V[a2]);

  result := (u >= 0) and (V >= 0) and (u + V <= 1);
end;

{ ***************************************************************************** }

// VectorMake functions
// 2x
function Vector2fMake(const X, Y: single): TVector2f;
begin
  result.V[0] := X;
  result.V[1] := Y;
end;

function Vector2iMake(const X, Y: longint): TVector2i;
begin
  result.V[0] := X;
  result.V[1] := Y;
end;

function Vector2sMake(const X, Y: smallint): TVector2s;
begin
  result.V[0] := X;
  result.V[1] := Y;
end;

function Vector2dMake(const X, Y: double): TVector2d;
begin
  result.V[0] := X;
  result.V[1] := Y;
end;

function Vector2bMake(const X, Y: byte): TVector2b;
begin
  result.V[0] := X;
  result.V[1] := Y;
end;

// **************

function Vector2fMake(const Vector: TVector3f): TVector2f;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
end;

function Vector2iMake(const Vector: TVector3i): TVector2i;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
end;

function Vector2sMake(const Vector: TVector3s): TVector2s;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
end;

function Vector2dMake(const Vector: TVector3d): TVector2d;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
end;

function Vector2bMake(const Vector: TVector3b): TVector2b;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
end;

// **********

function Vector2fMake(const Vector: TVector4f): TVector2f;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
end;

function Vector2iMake(const Vector: TVector4i): TVector2i;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
end;

function Vector2sMake(const Vector: TVector4s): TVector2s;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
end;

function Vector2dMake(const Vector: TVector4d): TVector2d;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
end;

function Vector2bMake(const Vector: TVector4b): TVector2b;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
end;

{ ***************************************************************************** }

// 3x
function Vector3fMake(const X, Y, Z: single): TVector3f;
begin
  result.V[0] := X;
  result.V[1] := Y;
  result.V[2] := Z;
end;

function Vector3iMake(const X, Y, Z: longint): TVector3i;
begin
  result.V[0] := X;
  result.V[1] := Y;
  result.V[2] := Z;
end;

function Vector3sMake(const X, Y, Z: smallint): TVector3s;
begin
  result.V[0] := X;
  result.V[1] := Y;
  result.V[2] := Z;
end;

function Vector3dMake(const X, Y, Z: double): TVector3d;
begin
  result.V[0] := X;
  result.V[1] := Y;
  result.V[2] := Z;
end;

function Vector3bMake(const X, Y, Z: byte): TVector3b;
begin
  result.V[0] := X;
  result.V[1] := Y;
  result.V[2] := Z;
end;

// *******

function Vector3fMake(const Vector: TVector2f; const Z: single): TVector3f;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Z;
end;

function Vector3iMake(const Vector: TVector2i; const Z: longint): TVector3i;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Z;
end;

function Vector3sMake(const Vector: TVector2s; const Z: smallint): TVector3s;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Z;
end;

function Vector3dMake(const Vector: TVector2d; const Z: double): TVector3d;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Z;
end;

function Vector3bMake(const Vector: TVector2b; const Z: byte): TVector3b;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Z;
end;

// *******

function Vector3fMake(const Vector: TVector4f): TVector3f;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Vector.V[2];
end;

function Vector3iMake(const Vector: TVector4i): TVector3i;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Vector.V[2];
end;

function Vector3sMake(const Vector: TVector4s): TVector3s;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Vector.V[2];
end;

function Vector3dMake(const Vector: TVector4d): TVector3d;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Vector.V[2];
end;

function Vector3bMake(const Vector: TVector4b): TVector3b;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Vector.V[2];
end;

{ ***************************************************************************** }

// 4x
function Vector4fMake(const X, Y, Z, W: single): TVector4f;
begin
  result.V[0] := X;
  result.V[1] := Y;
  result.V[2] := Z;
  result.V[3] := W;
end;

function Vector4iMake(const X, Y, Z, W: longint): TVector4i;
begin
  result.V[0] := X;
  result.V[1] := Y;
  result.V[2] := Z;
  result.V[3] := W;
end;

function Vector4sMake(const X, Y, Z, W: smallint): TVector4s;
begin
  result.V[0] := X;
  result.V[1] := Y;
  result.V[2] := Z;
  result.V[3] := W;
end;

function Vector4dMake(const X, Y, Z, W: double): TVector4d;
begin
  result.V[0] := X;
  result.V[1] := Y;
  result.V[2] := Z;
  result.V[3] := W;
end;

function Vector4bMake(const X, Y, Z, W: byte): TVector4b;
begin
  result.V[0] := X;
  result.V[1] := Y;
  result.V[2] := Z;
  result.V[3] := W;
end;

// ********

function Vector4fMake(const Vector: TVector3f; const W: single): TVector4f;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Vector.V[2];
  result.V[3] := W;
end;

function Vector4iMake(const Vector: TVector3i; const W: longint): TVector4i;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Vector.V[2];
  result.V[3] := W;
end;

function Vector4sMake(const Vector: TVector3s; const W: smallint): TVector4s;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Vector.V[2];
  result.V[3] := W;
end;

function Vector4dMake(const Vector: TVector3d; const W: double): TVector4d;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Vector.V[2];
  result.V[3] := W;
end;

function Vector4bMake(const Vector: TVector3b; const W: byte): TVector4b;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Vector.V[2];
  result.V[3] := W;
end;

// *******

function Vector4fMake(const Vector: TVector2f; const Z: single; const W: single): TVector4f;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Z;
  result.V[3] := W;
end;

function Vector4iMake(const Vector: TVector2i; const Z: longint; const W: longint): TVector4i;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Z;
  result.V[3] := W;
end;

function Vector4sMake(const Vector: TVector2s; const Z: smallint; const W: smallint): TVector4s;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Z;
  result.V[3] := W;
end;

function Vector4dMake(const Vector: TVector2d; const Z: double; const W: double): TVector4d;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Z;
  result.V[3] := W;
end;

function Vector4bMake(const Vector: TVector2b; const Z: byte; const W: byte): TVector4b;
begin
  result.V[0] := Vector.V[0];
  result.V[1] := Vector.V[1];
  result.V[2] := Z;
  result.V[3] := W;
end;

{ ***************************************************************************** }

// 2
function VectorEquals(const Vector1, Vector2: TVector2f): Boolean;
begin
  result := (Vector1.V[0] = Vector2.V[0]) and (Vector1.V[1] = Vector2.V[1]);
end;

function VectorEquals(const Vector1, Vector2: TVector2i): Boolean;
begin
  result := (Vector1.V[0] = Vector2.V[0]) and (Vector1.V[1] = Vector2.V[1]);
end;

function VectorEquals(const V1, V2: TVector2d): Boolean;
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]);
end;

function VectorEquals(const V1, V2: TVector2s): Boolean;
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]);
end;

function VectorEquals(const V1, V2: TVector2b): Boolean;
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]);
end;

{ ***************************************************************************** }

// 3
function VectorEquals(const V1, V2: TVector3i): Boolean;
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]) and (V1.V[2] = V2.V[2]);
end;

function VectorEquals(const V1, V2: TVector3d): Boolean;
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]) and (V1.V[2] = V2.V[2]);
end;

function VectorEquals(const V1, V2: TVector3s): Boolean;
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]) and (V1.V[2] = V2.V[2]);
end;

function VectorEquals(const V1, V2: TVector3b): Boolean;
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]) and (V1.V[2] = V2.V[2]);
end;

{ ***************************************************************************** }

// 4
function VectorEquals(const V1, V2: TVector4i): Boolean;
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]) and (V1.V[2] = V2.V[2]) and
    (V1.V[3] = V2.V[3]);
end;

function VectorEquals(const V1, V2: TVector4d): Boolean;
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]) and (V1.V[2] = V2.V[2]) and
    (V1.V[3] = V2.V[3]);
end;

function VectorEquals(const V1, V2: TVector4s): Boolean;
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]) and (V1.V[2] = V2.V[2]) and
    (V1.V[3] = V2.V[3]);
end;

function VectorEquals(const V1, V2: TVector4b): Boolean;
begin
  result := (V1.V[0] = V2.V[0]) and (V1.V[1] = V2.V[1]) and (V1.V[2] = V2.V[2]) and
    (V1.V[3] = V2.V[3]);
end;

{ ***************************************************************************** }

// 3x3f
function MatrixEquals(const Matrix1, Matrix2: TMatrix3f): Boolean;
begin
  result := VectorEquals(Matrix1.V[0], Matrix2.V[0]) and VectorEquals(Matrix1.V[1], Matrix2.V[1])
    and VectorEquals(Matrix1.V[2], Matrix2.V[2]);
end;

// 3x3i
function MatrixEquals(const Matrix1, Matrix2: TMatrix3i): Boolean;
begin
  result := VectorEquals(Matrix1.V[0], Matrix2.V[0]) and VectorEquals(Matrix1.V[1], Matrix2.V[1])
    and VectorEquals(Matrix1.V[2], Matrix2.V[2]);
end;

// 3x3d
function MatrixEquals(const Matrix1, Matrix2: TMatrix3d): Boolean;
begin
  result := VectorEquals(Matrix1.V[0], Matrix2.V[0]) and VectorEquals(Matrix1.V[1], Matrix2.V[1])
    and VectorEquals(Matrix1.V[2], Matrix2.V[2]);
end;

// 3x3s
function MatrixEquals(const Matrix1, Matrix2: TMatrix3s): Boolean;
begin
  result := VectorEquals(Matrix1.V[0], Matrix2.V[0]) and VectorEquals(Matrix1.V[1], Matrix2.V[1])
    and VectorEquals(Matrix1.V[2], Matrix2.V[2]);
end;

// 3x3b
function MatrixEquals(const Matrix1, Matrix2: TMatrix3b): Boolean;
begin
  result := VectorEquals(Matrix1.V[0], Matrix2.V[0]) and VectorEquals(Matrix1.V[1], Matrix2.V[1])
    and VectorEquals(Matrix1.V[2], Matrix2.V[2]);
end;

{ ***************************************************************************** }

// 4x4f
function MatrixEquals(const Matrix1, Matrix2: TMatrix4f): Boolean;
begin
  result := VectorEquals(Matrix1.V[0], Matrix2.V[0]) and VectorEquals(Matrix1.V[1], Matrix2.V[1])
    and VectorEquals(Matrix1.V[2], Matrix2.V[2]) and VectorEquals(Matrix1.V[3], Matrix2.V[3]);
end;

// 4x4i
function MatrixEquals(const Matrix1, Matrix2: TMatrix4i): Boolean;
begin
  result := VectorEquals(Matrix1.V[0], Matrix2.V[0]) and VectorEquals(Matrix1.V[1], Matrix2.V[1])
    and VectorEquals(Matrix1.V[2], Matrix2.V[2]) and VectorEquals(Matrix1.V[3], Matrix2.V[3]);
end;

// 4x4d
function MatrixEquals(const Matrix1, Matrix2: TMatrix4d): Boolean;
begin
  result := VectorEquals(Matrix1.V[0], Matrix2.V[0]) and VectorEquals(Matrix1.V[1], Matrix2.V[1])
    and VectorEquals(Matrix1.V[2], Matrix2.V[2]) and VectorEquals(Matrix1.V[3], Matrix2.V[3]);
end;

// 4x4s
function MatrixEquals(const Matrix1, Matrix2: TMatrix4s): Boolean;
begin
  result := VectorEquals(Matrix1.V[0], Matrix2.V[0]) and VectorEquals(Matrix1.V[1], Matrix2.V[1])
    and VectorEquals(Matrix1.V[2], Matrix2.V[2]) and VectorEquals(Matrix1.V[3], Matrix2.V[3]);
end;

// 4x4b
function MatrixEquals(const Matrix1, Matrix2: TMatrix4b): Boolean;
begin
  result := VectorEquals(Matrix1.V[0], Matrix2.V[0]) and VectorEquals(Matrix1.V[1], Matrix2.V[1])
    and VectorEquals(Matrix1.V[2], Matrix2.V[2]) and VectorEquals(Matrix1.V[3], Matrix2.V[3]);
end;

{ ***************************************************************************** }

// Vector comparison functions:
// 3f
function VectorMoreThen(const SourceVector, ComparedVector: TVector3f): Boolean; overload;
begin
  result := (SourceVector.V[0] > ComparedVector.V[0]) and (SourceVector.V[1] > ComparedVector.V[1])
    and (SourceVector.V[2] > ComparedVector.V[2]);
end;

function VectorMoreEqualThen(const SourceVector, ComparedVector: TVector3f): Boolean; overload;
begin
  result := (SourceVector.V[0] >= ComparedVector.V[0]) and (SourceVector.V[1] >= ComparedVector.V[1]
    ) and (SourceVector.V[2] >= ComparedVector.V[2]);
end;

function VectorLessThen(const SourceVector, ComparedVector: TVector3f): Boolean; overload;
begin
  result := (SourceVector.V[0] < ComparedVector.V[0]) and (SourceVector.V[1] < ComparedVector.V[1])
    and (SourceVector.V[2] < ComparedVector.V[2]);
end;

function VectorLessEqualThen(const SourceVector, ComparedVector: TVector3f): Boolean; overload;
begin
  result := (SourceVector.V[0] <= ComparedVector.V[0]) and (SourceVector.V[1] <= ComparedVector.V[1]
    ) and (SourceVector.V[2] <= ComparedVector.V[2]);
end;

// 4f
function VectorMoreThen(const SourceVector, ComparedVector: TVector4f): Boolean; overload;
begin
  result := (SourceVector.V[0] > ComparedVector.V[0]) and (SourceVector.V[1] > ComparedVector.V[1])
    and (SourceVector.V[2] > ComparedVector.V[2]) and (SourceVector.V[3] > ComparedVector.V[3]);
end;

function VectorMoreEqualThen(const SourceVector, ComparedVector: TVector4f): Boolean; overload;
begin
  result := (SourceVector.V[0] >= ComparedVector.V[0]) and (SourceVector.V[1] >= ComparedVector.V[1]
    ) and (SourceVector.V[2] >= ComparedVector.V[2]) and (SourceVector.V[3] >= ComparedVector.V[3]);
end;

function VectorLessThen(const SourceVector, ComparedVector: TVector4f): Boolean; overload;
begin
  result := (SourceVector.V[0] < ComparedVector.V[0]) and (SourceVector.V[1] < ComparedVector.V[1])
    and (SourceVector.V[2] < ComparedVector.V[2]) and (SourceVector.V[3] < ComparedVector.V[3]);
end;

function VectorLessEqualThen(const SourceVector, ComparedVector: TVector4f): Boolean; overload;
begin
  result := (SourceVector.V[0] <= ComparedVector.V[0]) and (SourceVector.V[1] <= ComparedVector.V[1]
    ) and (SourceVector.V[2] <= ComparedVector.V[2]) and (SourceVector.V[3] <= ComparedVector.V[3]);
end;

// 3i
// Vector comparison functions:
function VectorMoreThen(const SourceVector, ComparedVector: TVector3i): Boolean; overload;
begin
  result := (SourceVector.V[0] > ComparedVector.V[0]) and (SourceVector.V[1] > ComparedVector.V[1])
    and (SourceVector.V[2] > ComparedVector.V[2]);
end;

function VectorMoreEqualThen(const SourceVector, ComparedVector: TVector3i): Boolean; overload;
begin
  result := (SourceVector.V[0] >= ComparedVector.V[0]) and (SourceVector.V[1] >= ComparedVector.V[1]
    ) and (SourceVector.V[2] >= ComparedVector.V[2]);
end;

function VectorLessThen(const SourceVector, ComparedVector: TVector3i): Boolean; overload;
begin
  result := (SourceVector.V[0] < ComparedVector.V[0]) and (SourceVector.V[1] < ComparedVector.V[1])
    and (SourceVector.V[2] < ComparedVector.V[2]);
end;

function VectorLessEqualThen(const SourceVector, ComparedVector: TVector3i): Boolean; overload;
begin
  result := (SourceVector.V[0] <= ComparedVector.V[0]) and (SourceVector.V[1] <= ComparedVector.V[1]
    ) and (SourceVector.V[2] <= ComparedVector.V[2]);
end;

// 4i
function VectorMoreThen(const SourceVector, ComparedVector: TVector4i): Boolean; overload;
begin
  result := (SourceVector.V[0] > ComparedVector.V[0]) and (SourceVector.V[1] > ComparedVector.V[1])
    and (SourceVector.V[2] > ComparedVector.V[2]) and (SourceVector.V[3] > ComparedVector.V[3]);
end;

function VectorMoreEqualThen(const SourceVector, ComparedVector: TVector4i): Boolean; overload;
begin
  result := (SourceVector.V[0] >= ComparedVector.V[0]) and (SourceVector.V[1] >= ComparedVector.V[1]
    ) and (SourceVector.V[2] >= ComparedVector.V[2]) and (SourceVector.V[3] >= ComparedVector.V[3]);
end;

function VectorLessThen(const SourceVector, ComparedVector: TVector4i): Boolean; overload;
begin
  result := (SourceVector.V[0] < ComparedVector.V[0]) and (SourceVector.V[1] < ComparedVector.V[1])
    and (SourceVector.V[2] < ComparedVector.V[2]) and (SourceVector.V[3] < ComparedVector.V[3]);
end;

function VectorLessEqualThen(const SourceVector, ComparedVector: TVector4i): Boolean; overload;
begin
  result := (SourceVector.V[0] <= ComparedVector.V[0]) and (SourceVector.V[1] <= ComparedVector.V[1]
    ) and (SourceVector.V[2] <= ComparedVector.V[2]) and (SourceVector.V[3] <= ComparedVector.V[3]);
end;

// 3s
// Vector comparison functions:
function VectorMoreThen(const SourceVector, ComparedVector: TVector3s): Boolean; overload;
begin
  result := (SourceVector.V[0] > ComparedVector.V[0]) and (SourceVector.V[1] > ComparedVector.V[1])
    and (SourceVector.V[2] > ComparedVector.V[2]);
end;

function VectorMoreEqualThen(const SourceVector, ComparedVector: TVector3s): Boolean; overload;
begin
  result := (SourceVector.V[0] >= ComparedVector.V[0]) and (SourceVector.V[1] >= ComparedVector.V[1]
    ) and (SourceVector.V[2] >= ComparedVector.V[2]);
end;

function VectorLessThen(const SourceVector, ComparedVector: TVector3s): Boolean; overload;
begin
  result := (SourceVector.V[0] < ComparedVector.V[0]) and (SourceVector.V[1] < ComparedVector.V[1])
    and (SourceVector.V[2] < ComparedVector.V[2]);
end;

function VectorLessEqualThen(const SourceVector, ComparedVector: TVector3s): Boolean; overload;
begin
  result := (SourceVector.V[0] <= ComparedVector.V[0]) and (SourceVector.V[1] <= ComparedVector.V[1]
    ) and (SourceVector.V[2] <= ComparedVector.V[2]);
end;

// 4s
function VectorMoreThen(const SourceVector, ComparedVector: TVector4s): Boolean; overload;
begin
  result := (SourceVector.V[0] > ComparedVector.V[0]) and (SourceVector.V[1] > ComparedVector.V[1])
    and (SourceVector.V[2] > ComparedVector.V[2]) and (SourceVector.V[3] > ComparedVector.V[3]);
end;

function VectorMoreEqualThen(const SourceVector, ComparedVector: TVector4s): Boolean; overload;
begin
  result := (SourceVector.V[0] >= ComparedVector.V[0]) and (SourceVector.V[1] >= ComparedVector.V[1]
    ) and (SourceVector.V[2] >= ComparedVector.V[2]) and (SourceVector.V[3] >= ComparedVector.V[3]);
end;

function VectorLessThen(const SourceVector, ComparedVector: TVector4s): Boolean; overload;
begin
  result := (SourceVector.V[0] < ComparedVector.V[0]) and (SourceVector.V[1] < ComparedVector.V[1])
    and (SourceVector.V[2] < ComparedVector.V[2]) and (SourceVector.V[3] < ComparedVector.V[3]);
end;

function VectorLessEqualThen(const SourceVector, ComparedVector: TVector4s): Boolean; overload;
begin
  result := (SourceVector.V[0] <= ComparedVector.V[0]) and (SourceVector.V[1] <= ComparedVector.V[1]
    ) and (SourceVector.V[2] <= ComparedVector.V[2]) and (SourceVector.V[3] <= ComparedVector.V[3]);
end;

// ComparedNumber
// 3f
function VectorMoreThen(const SourceVector: TVector3f; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] > ComparedNumber) and (SourceVector.V[1] > ComparedNumber) and
    (SourceVector.V[2] > ComparedNumber);
end;

function VectorMoreEqualThen(const SourceVector: TVector3f; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] >= ComparedNumber) and (SourceVector.V[1] >= ComparedNumber) and
    (SourceVector.V[2] >= ComparedNumber);
end;

function VectorLessThen(const SourceVector: TVector3f; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] < ComparedNumber) and (SourceVector.V[1] < ComparedNumber) and
    (SourceVector.V[2] < ComparedNumber);
end;

function VectorLessEqualThen(const SourceVector: TVector3f; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] <= ComparedNumber) and (SourceVector.V[1] <= ComparedNumber) and
    (SourceVector.V[2] <= ComparedNumber);
end;

// 4f
function VectorMoreThen(const SourceVector: TVector4f; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] > ComparedNumber) and (SourceVector.V[1] > ComparedNumber) and
    (SourceVector.V[2] > ComparedNumber) and (SourceVector.V[3] > ComparedNumber);
end;

function VectorMoreEqualThen(const SourceVector: TVector4f; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] >= ComparedNumber) and (SourceVector.V[1] >= ComparedNumber) and
    (SourceVector.V[2] >= ComparedNumber) and (SourceVector.V[3] >= ComparedNumber);
end;

function VectorLessThen(const SourceVector: TVector4f; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] < ComparedNumber) and (SourceVector.V[1] < ComparedNumber) and
    (SourceVector.V[2] < ComparedNumber) and (SourceVector.V[3] < ComparedNumber);
end;

function VectorLessEqualThen(const SourceVector: TVector4f; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] <= ComparedNumber) and (SourceVector.V[1] <= ComparedNumber) and
    (SourceVector.V[2] <= ComparedNumber) and (SourceVector.V[3] <= ComparedNumber);
end;

// 3i
function VectorMoreThen(const SourceVector: TVector3i; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] > ComparedNumber) and (SourceVector.V[1] > ComparedNumber) and
    (SourceVector.V[2] > ComparedNumber);
end;

function VectorMoreEqualThen(const SourceVector: TVector3i; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] >= ComparedNumber) and (SourceVector.V[1] >= ComparedNumber) and
    (SourceVector.V[2] >= ComparedNumber);
end;

function VectorLessThen(const SourceVector: TVector3i; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] < ComparedNumber) and (SourceVector.V[1] < ComparedNumber) and
    (SourceVector.V[2] < ComparedNumber);
end;

function VectorLessEqualThen(const SourceVector: TVector3i; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] <= ComparedNumber) and (SourceVector.V[1] <= ComparedNumber) and
    (SourceVector.V[2] <= ComparedNumber);
end;

// 4i
function VectorMoreThen(const SourceVector: TVector4i; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] > ComparedNumber) and (SourceVector.V[1] > ComparedNumber) and
    (SourceVector.V[2] > ComparedNumber) and (SourceVector.V[3] > ComparedNumber);
end;

function VectorMoreEqualThen(const SourceVector: TVector4i; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] >= ComparedNumber) and (SourceVector.V[1] >= ComparedNumber) and
    (SourceVector.V[2] >= ComparedNumber) and (SourceVector.V[3] >= ComparedNumber);
end;

function VectorLessThen(const SourceVector: TVector4i; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] < ComparedNumber) and (SourceVector.V[1] < ComparedNumber) and
    (SourceVector.V[2] < ComparedNumber) and (SourceVector.V[3] < ComparedNumber);
end;

function VectorLessEqualThen(const SourceVector: TVector4i; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] <= ComparedNumber) and (SourceVector.V[1] <= ComparedNumber) and
    (SourceVector.V[2] <= ComparedNumber) and (SourceVector.V[3] <= ComparedNumber);
end;

// 3s
function VectorMoreThen(const SourceVector: TVector3s; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] > ComparedNumber) and (SourceVector.V[1] > ComparedNumber) and
    (SourceVector.V[2] > ComparedNumber);
end;

function VectorMoreEqualThen(const SourceVector: TVector3s; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] >= ComparedNumber) and (SourceVector.V[1] >= ComparedNumber) and
    (SourceVector.V[2] >= ComparedNumber);
end;

function VectorLessThen(const SourceVector: TVector3s; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] < ComparedNumber) and (SourceVector.V[1] < ComparedNumber) and
    (SourceVector.V[2] < ComparedNumber);
end;

function VectorLessEqualThen(const SourceVector: TVector3s; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] <= ComparedNumber) and (SourceVector.V[1] <= ComparedNumber) and
    (SourceVector.V[2] <= ComparedNumber);
end;

// 4s
function VectorMoreThen(const SourceVector: TVector4s; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] > ComparedNumber) and (SourceVector.V[1] > ComparedNumber) and
    (SourceVector.V[2] > ComparedNumber) and (SourceVector.V[3] > ComparedNumber);
end;

function VectorMoreEqualThen(const SourceVector: TVector4s; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] >= ComparedNumber) and (SourceVector.V[1] >= ComparedNumber) and
    (SourceVector.V[2] >= ComparedNumber) and (SourceVector.V[3] >= ComparedNumber);
end;

function VectorLessThen(const SourceVector: TVector4s; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] < ComparedNumber) and (SourceVector.V[1] < ComparedNumber) and
    (SourceVector.V[2] < ComparedNumber) and (SourceVector.V[3] < ComparedNumber);
end;

function VectorLessEqualThen(const SourceVector: TVector4s; const ComparedNumber: single)
  : Boolean; overload;
begin
  result := (SourceVector.V[0] <= ComparedNumber) and (SourceVector.V[1] <= ComparedNumber) and
    (SourceVector.V[2] <= ComparedNumber) and (SourceVector.V[3] <= ComparedNumber);
end;

{ : Determine if 2 rectanges intersect. }
function RectanglesIntersect(const ACenterOfRect1, ACenterOfRect2, ASizeOfRect1,
  ASizeOfRect2: TVector2f): Boolean;
begin
  result := (Abs(ACenterOfRect1.V[0] - ACenterOfRect2.V[0]) < (ASizeOfRect1.V[0] + ASizeOfRect2.V[0]
    ) / 2) and (Abs(ACenterOfRect1.V[1] - ACenterOfRect2.V[1]) <
    (ASizeOfRect1.V[1] + ASizeOfRect2.V[1]) / 2);
end;

{ : Determine if BigRect completely contains SmallRect. }
function RectangleContains(const ACenterOfBigRect1, ACenterOfSmallRect2, ASizeOfBigRect1,
  ASizeOfSmallRect2: TVector2f; const AEps: single = 0.0): Boolean;
begin
  result := (Abs(ACenterOfBigRect1.V[0] - ACenterOfSmallRect2.V[0]) + ASizeOfSmallRect2.V[0] / 2 -
    ASizeOfBigRect1.V[0] / 2 < AEps) and (Abs(ACenterOfBigRect1.V[1] - ACenterOfSmallRect2.V[1]) +
    ASizeOfSmallRect2.V[1] / 2 - ASizeOfBigRect1.V[1] / 2 < AEps);
end;

function GetSafeTurnAngle(const AOriginalPosition, AOriginalUpVector, ATargetPosition,
  AMoveAroundTargetCenter: TVector): TVector2f;
var
  pitchangle0, pitchangle1, turnangle0, turnangle1, pitchangledif, turnangledif, dx0, dy0, dz0, dx1,
    dy1, dz1: double;
  Sign      : ShortInt;
begin
  // determine relative positions to determine the lines which form the angles
  // distances from initial camera pos to target object
  dx0 := AOriginalPosition.V[0] - AMoveAroundTargetCenter.V[0];
  dy0 := AOriginalPosition.V[1] - AMoveAroundTargetCenter.V[1];
  dz0 := AOriginalPosition.V[2] - AMoveAroundTargetCenter.V[2];

  // distances from final camera pos to target object
  dx1 := ATargetPosition.V[0] - AMoveAroundTargetCenter.V[0];
  dy1 := ATargetPosition.V[1] - AMoveAroundTargetCenter.V[1];
  dz1 := ATargetPosition.V[2] - AMoveAroundTargetCenter.V[2];

  // just to make sure we don't get division by 0 exceptions
  if dx0 = 0 then
    dx0 := 0.001;
  if dy0 = 0 then
    dy0 := 0.001;
  if dz0 = 0 then
    dz0 := 0.001;
  if dx1 = 0 then
    dx1 := 0.001;
  if dy1 = 0 then
    dy1 := 0.001;
  if dz1 = 0 then
    dz1 := 0.001;

  // determine "pitch" and "turn" angles for the initial and  final camera position
  // the formulas differ depending on the camera.Up vector
  // I tested all quadrants for all possible integer FJoblist.Camera.Up directions
  if Abs(AOriginalUpVector.V[2]) = 1 then // Z=1/-1
  begin
    Sign        := Round(AOriginalUpVector.V[2] / Abs(AOriginalUpVector.V[2]));
    pitchangle0 := arctan(dz0 / Sqrt(Sqr(dx0) + Sqr(dy0)));
    pitchangle1 := arctan(dz1 / Sqrt(Sqr(dx1) + Sqr(dy1)));
    turnangle0  := arctan(dy0 / dx0);
    if (dx0 < 0) and (dy0 < 0) then
      turnangle0 := -(PI - turnangle0)
    else if (dx0 < 0) and (dy0 > 0) then
      turnangle0 := -(PI - turnangle0);
    turnangle1   := arctan(dy1 / dx1);
    if (dx1 < 0) and (dy1 < 0) then
      turnangle1 := -(PI - turnangle1)
    else if (dx1 < 0) and (dy1 > 0) then
      turnangle1 := -(PI - turnangle1);
  end
  else if Abs(AOriginalUpVector.V[1]) = 1 then // Y=1/-1
  begin
    Sign        := Round(AOriginalUpVector.V[1] / Abs(AOriginalUpVector.V[1]));
    pitchangle0 := arctan(dy0 / Sqrt(Sqr(dx0) + Sqr(dz0)));
    pitchangle1 := arctan(dy1 / Sqrt(Sqr(dx1) + Sqr(dz1)));
    turnangle0  := -arctan(dz0 / dx0);
    if (dx0 < 0) and (dz0 < 0) then
      turnangle0 := -(PI - turnangle0)
    else if (dx0 < 0) and (dz0 > 0) then
      turnangle0 := -(PI - turnangle0);
    turnangle1   := -arctan(dz1 / dx1);
    if (dx1 < 0) and (dz1 < 0) then
      turnangle1 := -(PI - turnangle1)
    else if (dx1 < 0) and (dz1 > 0) then
      turnangle1 := -(PI - turnangle1);
  end
  else if Abs(AOriginalUpVector.V[0]) = 1 then // X=1/-1
  begin
    Sign        := Round(AOriginalUpVector.V[0] / Abs(AOriginalUpVector.V[0]));
    pitchangle0 := arctan(dx0 / Sqrt(Sqr(dz0) + Sqr(dy0)));
    pitchangle1 := arctan(dx1 / Sqrt(Sqr(dz1) + Sqr(dy1)));
    turnangle0  := arctan(dz0 / dy0);
    if (dz0 > 0) and (dy0 > 0) then
      turnangle0 := -(PI - turnangle0)
    else if (dz0 < 0) and (dy0 > 0) then
      turnangle0 := -(PI - turnangle0);
    turnangle1   := arctan(dz1 / dy1);
    if (dz1 > 0) and (dy1 > 0) then
      turnangle1 := -(PI - turnangle1)
    else if (dz1 < 0) and (dy1 > 0) then
      turnangle1 := -(PI - turnangle1);
  end
  else
  begin
    Raise Exception.Create('The Camera.Up vector may contain only -1, 0 or 1');
  end;

  // determine pitch and turn angle differences
  pitchangledif := Sign * (pitchangle1 - pitchangle0);
  turnangledif  := Sign * (turnangle1 - turnangle0);

  if Abs(turnangledif) > PI then
    turnangledif := -Abs(turnangledif) / turnangledif * (2 * PI - Abs(turnangledif));

  // Determine rotation speeds
  result.V[0] := VectorGeometry.RadToDeg(-pitchangledif);
  result.V[1] := VectorGeometry.RadToDeg(turnangledif);
end;

function GetSafeTurnAngle(const AOriginalPosition, AOriginalUpVector, ATargetPosition,
  AMoveAroundTargetCenter: TAffineVector): TVector2f;
var
  pitchangle0, pitchangle1, turnangle0, turnangle1, pitchangledif, turnangledif, dx0, dy0, dz0, dx1,
    dy1, dz1: double;
  Sign      : ShortInt;
begin
  // determine relative positions to determine the lines which form the angles
  // distances from initial camera pos to target object
  dx0 := AOriginalPosition.V[0] - AMoveAroundTargetCenter.V[0];
  dy0 := AOriginalPosition.V[1] - AMoveAroundTargetCenter.V[1];
  dz0 := AOriginalPosition.V[2] - AMoveAroundTargetCenter.V[2];

  // distances from final camera pos to target object
  dx1 := ATargetPosition.V[0] - AMoveAroundTargetCenter.V[0];
  dy1 := ATargetPosition.V[1] - AMoveAroundTargetCenter.V[1];
  dz1 := ATargetPosition.V[2] - AMoveAroundTargetCenter.V[2];

  // just to make sure we don't get division by 0 exceptions
  if dx0 = 0 then
    dx0 := 0.001;
  if dy0 = 0 then
    dy0 := 0.001;
  if dz0 = 0 then
    dz0 := 0.001;
  if dx1 = 0 then
    dx1 := 0.001;
  if dy1 = 0 then
    dy1 := 0.001;
  if dz1 = 0 then
    dz1 := 0.001;

  // determine "pitch" and "turn" angles for the initial and  final camera position
  // the formulas differ depending on the camera.Up vector
  // I tested all quadrants for all possible integer FJoblist.Camera.Up directions
  if Abs(AOriginalUpVector.V[2]) = 1 then // Z=1/-1
  begin
    Sign        := Round(AOriginalUpVector.V[2] / Abs(AOriginalUpVector.V[2]));
    pitchangle0 := arctan(dz0 / Sqrt(Sqr(dx0) + Sqr(dy0)));
    pitchangle1 := arctan(dz1 / Sqrt(Sqr(dx1) + Sqr(dy1)));
    turnangle0  := arctan(dy0 / dx0);
    if (dx0 < 0) and (dy0 < 0) then
      turnangle0 := -(PI - turnangle0)
    else if (dx0 < 0) and (dy0 > 0) then
      turnangle0 := -(PI - turnangle0);
    turnangle1   := arctan(dy1 / dx1);
    if (dx1 < 0) and (dy1 < 0) then
      turnangle1 := -(PI - turnangle1)
    else if (dx1 < 0) and (dy1 > 0) then
      turnangle1 := -(PI - turnangle1);
  end
  else if Abs(AOriginalUpVector.V[1]) = 1 then // Y=1/-1
  begin
    Sign        := Round(AOriginalUpVector.V[1] / Abs(AOriginalUpVector.V[1]));
    pitchangle0 := arctan(dy0 / Sqrt(Sqr(dx0) + Sqr(dz0)));
    pitchangle1 := arctan(dy1 / Sqrt(Sqr(dx1) + Sqr(dz1)));
    turnangle0  := -arctan(dz0 / dx0);
    if (dx0 < 0) and (dz0 < 0) then
      turnangle0 := -(PI - turnangle0)
    else if (dx0 < 0) and (dz0 > 0) then
      turnangle0 := -(PI - turnangle0);
    turnangle1   := -arctan(dz1 / dx1);
    if (dx1 < 0) and (dz1 < 0) then
      turnangle1 := -(PI - turnangle1)
    else if (dx1 < 0) and (dz1 > 0) then
      turnangle1 := -(PI - turnangle1);
  end
  else if Abs(AOriginalUpVector.V[0]) = 1 then // X=1/-1
  begin
    Sign        := Round(AOriginalUpVector.V[0] / Abs(AOriginalUpVector.V[0]));
    pitchangle0 := arctan(dx0 / Sqrt(Sqr(dz0) + Sqr(dy0)));
    pitchangle1 := arctan(dx1 / Sqrt(Sqr(dz1) + Sqr(dy1)));
    turnangle0  := arctan(dz0 / dy0);
    if (dz0 > 0) and (dy0 > 0) then
      turnangle0 := -(PI - turnangle0)
    else if (dz0 < 0) and (dy0 > 0) then
      turnangle0 := -(PI - turnangle0);
    turnangle1   := arctan(dz1 / dy1);
    if (dz1 > 0) and (dy1 > 0) then
      turnangle1 := -(PI - turnangle1)
    else if (dz1 < 0) and (dy1 > 0) then
      turnangle1 := -(PI - turnangle1);
  end
  else
  begin
    Raise Exception.Create('The Camera.Up vector may contain only -1, 0 or 1');
  end;

  // determine pitch and turn angle differences
  pitchangledif := Sign * (pitchangle1 - pitchangle0);
  turnangledif  := Sign * (turnangle1 - turnangle0);

  if Abs(turnangledif) > PI then
    turnangledif := -Abs(turnangledif) / turnangledif * (2 * PI - Abs(turnangledif));

  // Determine rotation speeds
  result.V[0] := VectorGeometry.RadToDeg(-pitchangledif);
  result.V[1] := VectorGeometry.RadToDeg(turnangledif);
end;

function MoveObjectAround(const AMovingObjectPosition, AMovingObjectUp, ATargetPosition: TVector;
  pitchDelta, turnDelta: single): TVector;
var
  originalT2C, normalT2C, normalCameraRight: TVector;
  pitchNow, dist                           : single;
begin
  // normalT2C points away from the direction the camera is looking
  originalT2C := VectorSubtract(AMovingObjectPosition, ATargetPosition);
  SetVector(normalT2C, originalT2C);
  dist := VectorLength(normalT2C);
  NormalizeVector(normalT2C);
  // normalRight points to the camera's right
  // the camera is pitching around this axis.
  normalCameraRight := VectorCrossProduct(AMovingObjectUp, normalT2C);
  if VectorLength(normalCameraRight) < 0.001 then
    SetVector(normalCameraRight, XVector) // arbitrary vector
  else
    NormalizeVector(normalCameraRight);
  // calculate the current pitch.
  // 0 is looking down and PI is looking up
  pitchNow := VectorGeometry.ArcCos(VectorDotProduct(AMovingObjectUp, normalT2C));
  pitchNow := ClampValue(pitchNow + VectorGeometry.DegToRad(pitchDelta), 0 + 0.025, PI - 0.025);
  // create a new vector pointing up and then rotate it down
  // into the new position
  SetVector(normalT2C, AMovingObjectUp);
  RotateVector(normalT2C, normalCameraRight, -pitchNow);
  RotateVector(normalT2C, AMovingObjectUp, -VectorGeometry.DegToRad(turnDelta));
  ScaleVector(normalT2C, dist);
  result := VectorAdd(AMovingObjectPosition, VectorSubtract(normalT2C, originalT2C));
end;

{ : Calcualtes Angle between 2 Vectors: (A-CenterPoint) and (B-CenterPoint). In radians. }
function AngleBetweenVectors(const a, b, ACenterPoint: TVector): single;
begin
  result := VectorGeometry.ArcCos(VectorAngleCosine(VectorNormalize(VectorSubtract(a, ACenterPoint)
    ), VectorNormalize(VectorSubtract(b, ACenterPoint))));
end;

{ : Calcualtes Angle between 2 Vectors: (A-CenterPoint) and (B-CenterPoint). In radians. }
function AngleBetweenVectors(const a, b, ACenterPoint: TAffineVector): single;
begin
  result := VectorGeometry.ArcCos(VectorAngleCosine(VectorNormalize(VectorSubtract(a, ACenterPoint)
    ), VectorNormalize(VectorSubtract(b, ACenterPoint))));
end;

{ : AOriginalPosition - Object initial position.
  ACenter - some point, from which is should be distanced.

  ADistance + AFromCenterSpot - distance, which object should keep from ACenter
  or
  ADistance + not AFromCenterSpot - distance, which object should shift from his current position away from center.
}
function ShiftObjectFromCenter(const AOriginalPosition: TVector; const ACenter: TVector;
  const ADistance: single; const AFromCenterSpot: Boolean): TVector;
var
  lDirection: TVector;
begin
  lDirection := VectorNormalize(VectorSubtract(AOriginalPosition, ACenter));
  if AFromCenterSpot then
    result := VectorAdd(ACenter, VectorScale(lDirection, ADistance))
  else
    result := VectorAdd(AOriginalPosition, VectorScale(lDirection, ADistance))
end;

{ : AOriginalPosition - Object initial position.
  ACenter - some point, from which is should be distanced.

  ADistance + AFromCenterSpot - distance, which object should keep from ACenter
  or
  ADistance + not AFromCenterSpot - distance, which object should shift from his current position away from center.
}
function ShiftObjectFromCenter(const AOriginalPosition: TAffineVector; const ACenter: TAffineVector;
  const ADistance: single; const AFromCenterSpot: Boolean): TAffineVector;
var
  lDirection: TAffineVector;
begin
  lDirection := VectorNormalize(VectorSubtract(AOriginalPosition, ACenter));
  if AFromCenterSpot then
    result := VectorAdd(ACenter, VectorScale(lDirection, ADistance))
  else
    result := VectorAdd(AOriginalPosition, VectorScale(lDirection, ADistance))
end;

// --------------------------------------------------------------
// --------------------------------------------------------------
// --------------------------------------------------------------
initialization

// --------------------------------------------------------------
// --------------------------------------------------------------
// --------------------------------------------------------------

{$IFNDEF GEOMETRY_NO_ASM}
try
  // detect 3DNow! capable CPU (adapted from AMD's "3DNow! Porting Guide")
  asm
    pusha
    mov  eax, $80000000
    db $0F,$A2               /// cpuid
    cmp  eax, $80000000
    jbe @@No3DNow
    mov  eax, $80000001
    db $0F,$A2               /// cpuid
    test edx, $80000000
    jz @@No3DNow
    mov vSIMD, 1
  @@No3DNow:
    popa
  end;
except
  // trap for old/exotics CPUs
  vSIMD := 0;
end;
{$ELSE}
  vSIMD := 0;
{$ENDIF}

end.
